/*
 * This file is part of the Scalasca software (https://www.scalasca.org)
 *
 * Copyright (c) 2022,
 * Forschungszentrum Juelich GmbH, Germany
 *
 * This software may be modified and distributed under the terms of
 * a BSD-style license.  See the COPYING file in the package base
 * directory for details.
 */


#ifndef UTILS_PROFILE_H
#define UTILS_PROFILE_H

/**
 * @file
 * @ingroup UTILS_Profile_module
 * @brief   Wrapper module for manual Score-P instrumentation.
 */

/**
 * @defgroup UTILS_Profile_module Manual Score-P instrumentation wrapper.
 *
 * This module provides an interface to wrap manual Score-P instrumentation.
 * It provides a number of macros that either map to corresponding macros of
 * the Score-P instrumentation API, or expand to nothing in case Score-P
 * user instrumentation is not enabled.
 *
 * @{
 */

/**
 * @def UTILS_PROFILE_FUNC(name)
 * Marks the start of a function body to be profiled.  The function region is
 * automatically closed when the function body is left.
 *
 * @param name Function name
 */

/**
 * @def UTILS_PROFILE_BLOCK(name)
 * Marks the begin of a code block to be profiled until the current scope is
 * left.
 *
 * @param name Code region name
 */

/**
 * @def UTILS_PROFILE_BEGIN(name)
 * Marks the start of a user-defined region to be profiled.  The end of the
 * region needs to be marked by UTILS_PROFILE_END(name), with the begin and
 * end calls correctly nested.
 *
 * @attention This call performs a lookup by name, i.e., use it sparingly
 *            and only for coarse-grained regions!
 *
 * @param name Code region name
 */

/**
 * @def UTILS_PROFILE_END(name)
 * Marks the end of a user-defined region to be profiles.  The begin of the
 * region needs to be marked by UTILS_PROFILE_BEGIN(name), with the begin and
 * end calls correctly nested.
 *
 * @attention This call performs a lookup by name, i.e., use it sparingly
 *            and only for coarse-grained regions!
 *
 * @param name Code region name
 */

/**
 * @def UTILS_PROFILE_ONLY(...)
 * Marks code, especially variables, which are only accessed in profiling mode.
 *
 * @note Do not use a `;` after this statement.  But ... needs to be a valid C
 *       statement including `;`.
 */

/* *INDENT-OFF* */
#ifdef SCOREP_USER_ENABLE

#include <scorep/SCOREP_User.h>

#define UTILS_PROFILE_FUNC( name ) \
    SCOREP_USER_REGION( name, SCOREP_USER_REGION_TYPE_FUNCTION )

#define UTILS_PROFILE_BLOCK( name ) \
    SCOREP_USER_REGION( name, SCOREP_USER_REGION_TYPE_COMMON )

#define UTILS_PROFILE_BEGIN( name ) \
    SCOREP_USER_REGION_BY_NAME_BEGIN( name, SCOREP_USER_REGION_TYPE_COMMON )

#define UTILS_PROFILE_END( name ) \
    SCOREP_USER_REGION_BY_NAME_END( name )

#define UTILS_PROFILE_ONLY( ... ) __VA_ARGS__

#else    /* !SCOREP_USER_ENABLE */

#define UTILS_PROFILE_FUNC( name )
#define UTILS_PROFILE_BLOCK( name )
#define UTILS_PROFILE_BEGIN( name )
#define UTILS_PROFILE_END( name )
#define UTILS_PROFILE_ONLY( ... )

#endif    /* !SCOREP_USER_ENABLE */
/* *INDENT-ON* */

/** @} */


#endif    /* UTILS_PROFILE_H */
