/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2023                                                     **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_mpi
 *  @brief   Declaration of the MpiHandleProvider interface and MPI handle
 *           management helper functions.
 *
 *  This header file provides the declaration of the interface class template
 *  MpiHandleProvider, as well as helper functions to create/query various
 *  types of MPI handles from PEARL definition objects.
 **/
/*-------------------------------------------------------------------------*/


#ifndef PEARL_MPIHELPERS_H
#define PEARL_MPIHELPERS_H


#include <mpi.h>


namespace pearl
{
// --- Forward declarations -------------------------------------------------

class CommSet;
class Communicator;


/*-------------------------------------------------------------------------*/
/**
 *  @interface MpiHandleProvider
 *  @ingroup   PEARL_mpi
 *  @brief     Interface implemented by definition objects providing MPI
 *             handles.
 *
 *  The MpiHandleProvider class template defines the common interface
 *  implemented by definition objects providing MPI handles.
 *
 *  @tparam handleT
 *      MPI handle type
 **/
/*-------------------------------------------------------------------------*/

template< typename handleT >
class MpiHandleProvider
{
    public:
        // --- Public methods ------------------------------

        /// @name Query functions
        /// @{

        /// @brief Get MPI handle.
        ///
        /// Returns the MPI handle associated with the definition object.
        ///
        /// @returns
        ///     MPI handle
        ///
        virtual handleT
        getHandle() const = 0;

        /// @}


    protected:
        // --- Protected methods ---------------------------

        /// @name Constructors & destructor
        /// @{

        /// @brief Destructor.
        ///
        /// Destroys the instance.  Protected to prevent polymorphic deletion
        /// via the MpiHandleProvider interface.
        ///
        ~MpiHandleProvider() = default;

        /// @}
};


// --- Function prototypes --------------------------------------------------

/// @name MPI handle creation
/// @{

/// @brief Create MPI group handle.
///
/// Creates an MPI group handle corresponding to the provided communication
/// set definition object @p commSet, and stores it at the memory location
/// referenced by @p handle.  The given communication set must be of paradigm
/// MPI.
///
/// This call is a local operation.  It may abort if the underlying MPI calls
/// fail (MPI errors are fatal).
///
/// @param[in] commSet
///     Communication set definition
/// @param[out] handle
///     Corresponding MPI group handle
///
/// @pre
///     Communication set paradigm is MPI
/// @throws std::bad_alloc
///     on failure to allocate storage
///
void
createMpiHandle(const CommSet& commSet,
                MPI_Group&     handle);

/// @brief Create MPI communicator handle.
///
/// Creates an MPI communicator handle corresponding to the provided
/// communicator definition object @p comm, and stores it at the memory
/// location referenced by @p handle.  The given communicator must be
/// of paradigm MPI.
///
/// This call is collective over `MPI_COMM_WORLD`, and therefore must be
/// invoked by all processes.  It may abort if the underlying MPI calls
/// fail (MPI errors are fatal).
///
/// @param[in] comm
///     %Communicator definition
/// @param[out] handle
///     Corresponding MPI communicator handle
///
/// @pre
///     Communication set paradigm is MPI
/// @throws std::bad_alloc
///     on failure to allocate storage
///
void
createMpiHandle(const Communicator& comm,
                MPI_Comm&           handle);

/// @}
/// @name MPI handle query
/// @{

/// @brief Query MPI communicator handle.
///
/// Returns the MPI handle associated to the provided communicator definition
/// object @p comm, which must be of paradigm MPI.
///
/// @param comm
///     %Communicator definition object
/// @returns
///     Associated MPI communicator handle
///
/// @throws pearl::FatalError
///     if @p comm is not an MPI communicator definition
///
MPI_Comm
getMpiHandle(const Communicator& comm);

/// @brief Query MPI communicator handle.
///
/// Returns the MPI handle associated to the provided communicator definition
/// object @p comm, which must be of paradigm MPI.
///
/// @param comm
///     Communicator definition object
/// @returns
///     Associated MPI communicator handle
///
/// @throws pearl::FatalError
///     if @p comm is not an MPI communicator definition
///
MPI_Comm
getMpiHandle(const Communicator* comm);

/// @}
}    // namespace pearl


#endif    // !PEARL_MPIHELPERS_H
