/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_mpi
 *  @brief   Declaration of the class RemoteData.
 *
 *  This header file provides the declaration of the class RemoteData.
 **/
/*-------------------------------------------------------------------------*/


#ifndef PEARL_REMOTEDATA_H
#define PEARL_REMOTEDATA_H


#include <cstdint>
#include <map>
#include <vector>

#include <pearl/RemoteEvent.h>
#include <pearl/RemoteEventSet.h>


namespace pearl
{
// --- Forward declarations -------------------------------------------------

class Buffer;
class Communicator;
class GlobalDefs;
class MpiMessage;


/*-------------------------------------------------------------------------*/
/**
 *  @class   RemoteData
 *  @ingroup PEARL_mpi
 *  @brief   Container class for a set of remote events and memory buffers
 *           with associated roles.
 **/
/*-------------------------------------------------------------------------*/

class RemoteData
{
    public:
        // --- Public methods ------------------------------

        /// @name Constructors & destructor
        /// @{

        ~RemoteData();

        /// @}
        /// @name Adding/updating data
        /// @{

        void
        addEvent(const RemoteEvent& event,
                 std::uint32_t      key);

        void
        updateEvent(const RemoteEvent& event,
                    std::uint32_t      key);

        void
        addBuffer(Buffer*       buffer,
                  std::uint32_t key);

        /// @}
        /// @name Retrieving data
        /// @{

        RemoteEvent
        getEvent(std::uint32_t key) const;

        Buffer*
        getBuffer(std::uint32_t key) const;

        /// @}
        /// @name Removing data
        /// @{

        void
        clear();

        /// @}
        /// @name Receiving data
        /// @{

        void
        recv(const GlobalDefs&   defs,
             const Communicator& comm,
             int                 source,
             int                 tag);

        /// @}


    private:
        // --- Type definitions ----------------------------

        /// Container type for role |-@> index mapping
        typedef std::map< std::uint32_t, std::uint32_t > KeyMap;

        /// Container type for remote buffers
        typedef std::vector< Buffer* > BufferList;


        // --- Data members --------------------------------

        /// Current set of events
        RemoteEventSet mEvents;

        /// Mapping table key (role ID) |-@> index for buffers
        KeyMap mBufferKeys;

        /// Current set of buffers
        BufferList mBuffers;


        // --- Private methods -----------------------------

        /// @name Unpacking messages
        /// @{

        void
        unpack(const GlobalDefs& defs,
               MpiMessage&       message);

        /// @}
};
}    // namespace pearl


#endif    // !PEARL_REMOTEDATA_H
