/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2021                                                     **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @brief Backported functionality from post-C++11 `<type_traits>`.
 *
 *  This header file provides a compatibility layer for selected (useful)
 *  capabilities of the C++ standard library header `<type_traits>` that
 *  were introduced in language standard versions newer than C++11.
 *
 *  @note
 *      This file is not intended to be feature-complete, but only contains
 *      functionality used in the code base.
 **/
/*-------------------------------------------------------------------------*/


#ifndef CXXSLB_TYPE_TRAITS_H
#define CXXSLB_TYPE_TRAITS_H


#include <type_traits>


namespace cxxslb
{
/// @name Type modifications
/// @{

#if __cplusplus >= 201402L
    using std::remove_extent_t;
#else    // C++11
    /// @brief Removes one extent from the given array type.
    ///
    /// If @p T is an array of some type @em X, provides the type @em X,
    /// otherwise type is @p T.  Note that if @p T is a multidimensional
    /// array, only the first dimension is removed.
    ///
    /// @tparam T
    ///     Array type
    ///
    /// @isocpp{14,type_traits,N3655}
    ///
    template< class T >
    using remove_extent_t = typename std::remove_extent< T >::type;
#endif    // C++11

/// @}
/// @name Miscellaneous transformations
/// @{

#if __cplusplus >= 201402L
    using std::underlying_type_t;
#else    // C++11
    /// @brief Obtains the underlying integer type for a given enumeration
    ///        type.
    ///
    /// If @p T is a complete enumeration type, provides the underlying
    /// integer type of @p T.
    ///
    /// @tparam T
    ///     Enumeration type
    ///
    /// @isocpp{14,type_traits,N3655}
    ///
    template< class T >
    using underlying_type_t = typename std::underlying_type< T >::type;
#endif    // C++11

/// @}
}    // namespace cxxslb


#endif    // !CXXSLB_TYPE_TRAITS_H
