/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include <pearl/LocalData.h>

#include <algorithm>
#include <memory>
#include <string>

#include <pearl/Error.h>
#include <pearl/MpiMessage.h>

using namespace std;
using namespace pearl;


// --- Constructors & destructor --------------------------------------------

LocalData::~LocalData()
{
    clear();
}


// --- Adding/updating data -------------------------------------------------

void
LocalData::addEvent(const Event& event,
                    uint32_t     key)
{
    mEvents.addEvent(event, key);
}


void
LocalData::updateEvent(const Event& event,
                       uint32_t     key)
{
    mEvents.updateEvent(event, key);
}


void
LocalData::addBuffer(Buffer*  buffer,
                     uint32_t key)
{
    mBufferKeys.insert(KeyMap::value_type(key, mBuffers.size()));
    mBuffers.push_back(buffer);
}


// --- Retrieving data ------------------------------------------------------

Event
LocalData::getEvent(uint32_t key) const
{
    return mEvents.getEvent(key);
}


Buffer*
LocalData::getBuffer(uint32_t key) const
{
    KeyMap::const_iterator it = mBufferKeys.find(key);
    if (it == mBufferKeys.end())
    {
        throw RuntimeError("LocalData::getBuffer(uint32_t) "
                           "-- Unknown key: " + std::to_string(key));
    }

    return mBuffers[it->second];
}


// --- Removing data --------------------------------------------------------

void
LocalData::clear()
{
    mEvents.clear();

    mBufferKeys.clear();
    std::for_each(mBuffers.begin(), mBuffers.end(),
                  std::default_delete< Buffer >());
    mBuffers.clear();
}


// --- Sending data ---------------------------------------------------------

MpiMessage*
LocalData::isend(const Communicator& comm,
                 int                 dest,
                 int                 tag) const
{
    MpiMessage* message = new MpiMessage(comm);
    pack(*message);
    message->isend(dest, tag);

    return message;
}


// --- Packing messages (private) -------------------------------------------

void
LocalData::pack(MpiMessage& message) const
{
    mEvents.pack(message);

    // Pack buffers
    const uint32_t keyCount = mBufferKeys.size();
    message.put_uint32(keyCount);
    for (const auto& mapItem : mBufferKeys)
    {
        message.put_uint32(mapItem.first);
        message.put_uint32(mapItem.second);
    }

    const uint32_t bufferCount = mBuffers.size();
    message.put_uint32(bufferCount);
    for (const auto& bufferPtr : mBuffers)
    {
        message.put_buffer(*bufferPtr);
    }
}
