/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_mpi
 *  @brief   Definition of MPI handle management helper functions.
 *
 *  This file provides the definition of helper functions to create various
 *  types of MPI handles from PEARL definition objects.
 **/
/*-------------------------------------------------------------------------*/


#include <config.h>

#include <pearl/MpiHelpers.h>

#include <cstdint>
#include <string>
#include <typeinfo>
#include <vector>

#include <mpi.h>

#include <UTILS_Error.h>

#include <pearl/CommSet.h>
#include <pearl/Communicator.h>
#include <pearl/Error.h>
#include <pearl/Flags.h>
#include <pearl/GroupingSet.h>
#include <pearl/Paradigm.h>


namespace pearl
{
// --- MPI handle creation --------------------------------------------------

void
createMpiHandle(const CommSet& commSet,
                MPI_Group&     handle)
{
    UTILS_ASSERT(commSet.getParadigm() == Paradigm::MPI);

    // Special case 1: GROUP_EMPTY
    const auto size = commSet.numRanks();
    if (size == 0)
    {
        handle = MPI_GROUP_EMPTY;

        return;
    }

    // Special case 2: GROUP_SELF
    const auto properties = commSet.getProperties();
    if (properties.test(GroupingSet::PROPERTY_SELF))
    {
        MPI_Comm_group(MPI_COMM_SELF, &handle);

        return;
    }

    // Special case 3: GROUP_WORLD
    if (properties.test(GroupingSet::PROPERTY_WORLD))
    {
        MPI_Comm_group(MPI_COMM_WORLD, &handle);

        return;
    }

    // Regular case: arbitrary group with size > 1
    std::vector< int > ranks;
    ranks.reserve(size);
    for (std::uint32_t rank = 0; rank < size; ++rank)
    {
        ranks.push_back(commSet.getGlobalRank(rank));
    }

    MPI_Group worldGroup;
    MPI_Comm_group(MPI_COMM_WORLD, &worldGroup);
    MPI_Group_incl(worldGroup, size, &ranks[0], &handle);
    MPI_Group_free(&worldGroup);
}


void
createMpiHandle(const Communicator& comm,
                MPI_Comm&           handle)
{
    UTILS_ASSERT(comm.getParadigm() == Paradigm::MPI);

    MPI_Group commGroup;
    createMpiHandle(comm.getCommSet(), commGroup);
    MPI_Comm_create(MPI_COMM_WORLD, commGroup, &handle);
    MPI_Group_free(&commGroup);
}


// --- MPI handle query -----------------------------------------------------

MPI_Comm
getMpiHandle(const Communicator& comm)
{
    try
    {
        using MpiCommHandleProvider = MpiHandleProvider< MPI_Comm >;

        auto& mpiComm = dynamic_cast< const MpiCommHandleProvider& >(comm);

        return mpiComm.getHandle();
    }
    catch (std::bad_cast& ex)
    {
        throw FatalError("pearl::getMpiHandle -- "
                         "communicator "
                         + std::to_string(comm.getId())
                         + " is not of paradigm MPI");
    }
}


/// @todo Remove once interfaces have been converted to const references
///
MPI_Comm
getMpiHandle(const Communicator* const comm)
{
    UTILS_ASSERT(comm != nullptr);

    return getMpiHandle(*comm);
}
}    // namespace pearl
