/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  Copyright (c) 2021-2023                                                **
**  RWTH Aachen University, IT Center                                      **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_mpi
 *  @brief   Implementation of the class MpiRmaGats_rep.
 *
 *  This file provides the implementation of the class MpiRmaGats_rep.
 **/
/*-------------------------------------------------------------------------*/


#include <config.h>

#include <pearl/MpiRmaGats_rep.h>

#include <cstdint>
#include <iostream>

#include <pearl/Buffer.h>
#include <pearl/GlobalDefs.h>
#include <pearl/MpiGroup.h>
#include <pearl/RmaWindow.h>

#include "iomanip_detail.h"

using namespace std;
using namespace pearl;
using namespace pearl::detail;


// --- Constructors & destructor --------------------------------------------

MpiRmaGats_rep::MpiRmaGats_rep(timestamp_t      timestamp,
                               const RmaWindow& window,
                               const MpiGroup&  group,
                               bool             sync)
    : Event_rep(timestamp),
      mWindow(window),
      mGroup(group),
      mSync(sync)
{
}


MpiRmaGats_rep::MpiRmaGats_rep(const GlobalDefs& defs,
                               Buffer&           buffer)
    : Event_rep(defs, buffer),
      mWindow(defs.getRmaWindow(buffer.get_uint32())),
      mGroup(dynamic_cast< const MpiGroup& >(defs.getGroupingSet(buffer.get_uint32()))),
      mSync(buffer.get_uint8())
{
}


// --- Event type information -----------------------------------------------

event_t
MpiRmaGats_rep::getType() const
{
    return MPI_RMA_GATS;
}


bool
MpiRmaGats_rep::isOfType(event_t type) const
{
    return (  (type == MPI_RMA_GATS)
           || (type == GROUP_ALL));
}


// --- Access event data ----------------------------------------------------

const RmaWindow&
MpiRmaGats_rep::getWindow() const
{
    return mWindow;
}


const MpiGroup&
MpiRmaGats_rep::getGroup() const
{
    return mGroup;
}


bool
MpiRmaGats_rep::isSync() const
{
    return mSync;
}


// --- Serialize event data (protected) -------------------------------------

void
MpiRmaGats_rep::pack(Buffer& buffer) const
{
    Event_rep::pack(buffer);

    buffer.put_uint32(mWindow.getId());
    buffer.put_uint32(mGroup.getId());
    buffer.put_uint8(mSync);
}


// --- Generate human-readable output of event data (protected) -------------

ostream&
MpiRmaGats_rep::output(ostream& stream) const
{
    Event_rep::output(stream);

    // Print data
    const ios_base::fmtflags fmtflags = stream.setf(ios_base::boolalpha);
    stream << sep
           << "window: " << mWindow
           << sep
           << "group: " << mGroup
           << sep
           << "sync: " << mSync;
    stream.flags(fmtflags);

    return stream;
}
