/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  Copyright (c) 2022-2023                                                **
**  RWTH Aachen University, IT Center                                      **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_mpi
 *  @brief   Implementation of the class MpiRmaLock_rep.
 *
 *  This file provides the implementation of the class MpiRmaLock_rep.
 **/
/*-------------------------------------------------------------------------*/


#include <config.h>

#include <pearl/MpiRmaLock_rep.h>

#include <iostream>

#include <pearl/Buffer.h>
#include <pearl/GlobalDefs.h>
#include <pearl/RmaWindow.h>

#include "iomanip_detail.h"

using namespace std;
using namespace pearl;
using namespace pearl::detail;


// --- Constructors & destructor --------------------------------------------

MpiRmaLock_rep::MpiRmaLock_rep(timestamp_t      timestamp,
                               uint32_t         remote,
                               const RmaWindow& window,
                               bool             exclusive)
    : Event_rep(timestamp),
      mWindow(window),
      mRemote(remote),
      mExclusive(exclusive)
{
}


MpiRmaLock_rep::MpiRmaLock_rep(const GlobalDefs& defs,
                               Buffer&           buffer)
    : Event_rep(defs, buffer),
      mWindow(defs.getRmaWindow(buffer.get_uint32())),
      mRemote(buffer.get_uint32()),
      mExclusive(buffer.get_uint8())
{
}


// --- Event type information -----------------------------------------------

event_t
MpiRmaLock_rep::getType() const
{
    return MPI_RMA_LOCK;
}


bool
MpiRmaLock_rep::isOfType(event_t type) const
{
    return (  (type == MPI_RMA_LOCK)
           || (type == GROUP_ALL));
}


// --- Access event data ----------------------------------------------------

const RmaWindow&
MpiRmaLock_rep::getWindow() const
{
    return mWindow;
}


uint32_t
MpiRmaLock_rep::getRemote() const
{
    return mRemote;
}


bool
MpiRmaLock_rep::isExclusive() const
{
    return mExclusive;
}


// --- Serialize event data (protected) -------------------------------------

void
MpiRmaLock_rep::pack(Buffer& buffer) const
{
    Event_rep::pack(buffer);

    buffer.put_uint32(mWindow.getId());
    buffer.put_uint32(mRemote);
    buffer.put_uint8(mExclusive);
}


// --- Generate human-readable output of event data (protected) -------------

ostream&
MpiRmaLock_rep::output(ostream& stream) const
{
    Event_rep::output(stream);

    // Print data
    const ios_base::fmtflags fmtflags = stream.setf(ios_base::boolalpha);
    stream << sep
           << "remote: " << mRemote
           << sep
           << "window: " << mWindow
           << sep
           << "exclusive: " << mExclusive;

    stream.flags(fmtflags);

    return stream;
}
