/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include <pearl/RemoteData.h>

#include <algorithm>
#include <memory>
#include <string>

#include <pearl/Error.h>
#include <pearl/MpiMessage.h>

using namespace std;
using namespace pearl;


// --- Constructors & destructor --------------------------------------------

RemoteData::~RemoteData()
{
    clear();
}


// --- Adding/updating data -------------------------------------------------

void
RemoteData::addEvent(const RemoteEvent& event,
                     uint32_t           key)
{
    mEvents.addEvent(event, key);
}


void
RemoteData::updateEvent(const RemoteEvent& event,
                        uint32_t           key)
{
    mEvents.updateEvent(event, key);
}


void
RemoteData::addBuffer(Buffer*  buffer,
                      uint32_t key)
{
    mBufferKeys.insert(KeyMap::value_type(key, mBuffers.size()));
    mBuffers.push_back(buffer);
}


// --- Retrieving data ------------------------------------------------------

RemoteEvent
RemoteData::getEvent(uint32_t key) const
{
    return mEvents.getEvent(key);
}


Buffer*
RemoteData::getBuffer(uint32_t key) const
{
    KeyMap::const_iterator it = mBufferKeys.find(key);
    if (it == mBufferKeys.end())
    {
        throw RuntimeError("RemoteData::getBuffer(uint32_t) "
                           "-- Unknown key: " + std::to_string(key));
    }

    return mBuffers[it->second];
}


// --- Removing data --------------------------------------------------------

void
RemoteData::clear()
{
    mEvents.clear();

    mBufferKeys.clear();
    std::for_each(mBuffers.begin(), mBuffers.end(),
                  std::default_delete< Buffer >());
    mBuffers.clear();
}


// --- Receiving data -------------------------------------------------------

void
RemoteData::recv(const GlobalDefs&   defs,
                 const Communicator& comm,
                 int                 source,
                 int                 tag)
{
    MpiMessage message(comm);
    message.recv(source, tag);
    unpack(defs, message);
}


// --- Unpacking messages (private) -----------------------------------------

void
RemoteData::unpack(const GlobalDefs& defs,
                   MpiMessage&       message)
{
    mEvents.unpack(defs, message);

    // Unpack buffers
    const uint32_t keyCount = message.get_uint32();
    for (uint32_t i = 0; i < keyCount; ++i)
    {
        const uint32_t key   = message.get_uint32();
        const uint32_t index = message.get_uint32();

        mBufferKeys.insert(KeyMap::value_type(key, index));
    }

    const uint32_t bufferCount = message.get_uint32();
    for (uint32_t i = 0; i < bufferCount; ++i)
    {
        Buffer* buffer = message.get_buffer();

        mBuffers.push_back(buffer);
    }
}
