/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2016-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#ifndef PEARL_DEFINITIONCONTAINER_H
    #error "Do not include this file!  Use 'DefinitionContainer.h' instead."
#endif


#include <algorithm>
#include <stdexcept>
#include <string>


namespace pearl
{
namespace detail
{
// --- Constructors & destructor --------------------------------------------

template< typename DefClassT >
DefinitionContainer< DefClassT >::DefinitionContainer()
    : mIsSorted(true),
      mIsCompact(true)
{
}


template< typename DefClassT >
DefinitionContainer< DefClassT >::~DefinitionContainer()
{
    std::for_each(mItems.begin(), mItems.end(),
                  std::default_delete< DefClassT >());
}


// --- Query functions ------------------------------------------------------

template< typename DefClassT >
bool
DefinitionContainer< DefClassT >::isEmpty() const
{
    return mItems.empty();
}


template< typename DefClassT >
std::size_t
DefinitionContainer< DefClassT >::numItems() const
{
    return mItems.size();
}


template< typename DefClassT >
DefClassT*
DefinitionContainer< DefClassT >::getItemById(const typename DefClassT::IdType id) const
{
    // Compact IDs: direct lookup
    if (mIsCompact)
    {
        if (id >= mItems.size())
        {
            throw std::invalid_argument("pearl::detail::DefinitionContainer<T>::"
                                        "getItemById(T::IdType) -- "
                                        "invalid ID: " + std::to_string(id));
        }

        return mItems[id];
    }

    // Sorted IDs: binary search
    if (mIsSorted)
    {
        typename ContainerType::const_iterator it =
            std::lower_bound(mItems.begin(), mItems.end(), id,
                             [](const DefClassT* const lhs,
                                const typename DefClassT::IdType rhs)
                             {
                                 return (lhs->getId() < rhs);
                             });
        if (  (it == mItems.end())
           || ((*it)->getId() != id))
        {
            throw std::invalid_argument("pearl::detail::DefinitionContainer<T>::"
                                        "getItemById(T::IdType) -- "
                                        "invalid ID: " + std::to_string(id));
        }

        return *it;
    }

    // Fallback: linear search
    typename ContainerType::const_iterator it =
        std::find_if(mItems.begin(), mItems.end(),
                     [id](const DefClassT* const lhs)
                     {
                         return (lhs->getId() == id);
                     });
    if (it == mItems.end())
    {
        throw std::invalid_argument("pearl::detail::DefinitionContainer<T>::"
                                    "getItemById(T::IdType) -- "
                                    "invalid ID: " + std::to_string(id));
    }

    return *it;
}


template< typename DefClassT >
DefClassT*
DefinitionContainer< DefClassT >::getItemByIndex(const std::size_t index) const
{
    if (index >= mItems.size())
    {
        throw std::out_of_range("pearl::detail::DefinitionContainer<T>::"
                                "getItemByIndex(size_t) -- "
                                "index out of range: " + std::to_string(index));
    }

    return mItems[index];
}


// --- Insertion functions --------------------------------------------------

template< typename DefClassT >
void
DefinitionContainer< DefClassT >::addItem(std::unique_ptr< DefClassT > item)
{
    if (!item)
    {
        throw std::invalid_argument("pearl::detail::DefinitionContainer<T>::"
                                    "addItem(unique_ptr<T>) -- "
                                    "invalid argument: `nullptr`");
    }

    const typename DefClassT::IdType nextCompactId = mItems.size();
    const DefClassT* const           lastItem      = mItems.empty() ? nullptr : mItems.back();

    mItems.push_back(item.get());

    if (mIsCompact && (item->getId() != nextCompactId))
    {
        mIsCompact = false;
    }
    if (  !mIsCompact && mIsSorted
       && lastItem && (item->getId() < lastItem->getId()))
    {
        mIsSorted = false;
    }

    // Since we've successfully stored the raw pointer in `mItems`,
    // resource leaks can no longer occur.  Thus it's safe to ignore the
    // return value here:  NOLINTNEXTLINE(bugprone-unused-return-value)
    item.release();
}


// --- Miscellaneous functions ----------------------------------------------

template< typename DefClassT >
void
DefinitionContainer< DefClassT >::optimizeLookups()
{
    if (!mIsSorted)
    {
        std::sort(mItems.begin(), mItems.end(),
                  [](const DefClassT* const lhs,
                     const DefClassT* const rhs)
                  {
                      return (lhs->getId() < rhs->getId());
                  });
        mIsSorted = true;
    }

    // As a second step, one could check whether the IDs are compact.
    // However, such a check would have linear complexity, i.e., one would
    // need a sufficiently large number of lookups to amortize this cost
    // via the faster lookups.  For typical global definition data such as
    // the one generated by Score-P, this is unlikely to pay off.  Thus,
    // this additional check is intentionally not carried out.
}
}    // namespace pearl::detail
}    // namespace pearl
