/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_base
 *  @brief   Declaration of the class MutableGlobalDefs.
 *
 *  This header file provides the declaration of the class MutableGlobalDefs.
 **/
/*-------------------------------------------------------------------------*/


#ifndef PEARL_MUTABLEGLOBALDEFS_H
#define PEARL_MUTABLEGLOBALDEFS_H


#include <cstdint>
#include <memory>
#include <vector>

#include <pearl/GlobalDefs.h>
#include <pearl/Paradigm.h>
#include <pearl/pearl_types.h>


namespace pearl
{
// --- Forward declarations -------------------------------------------------

class CommLocationSet;


namespace detail
{
/*-------------------------------------------------------------------------*/
/**
 *  @class   MutableGlobalDefs
 *  @ingroup PEARL_base
 **/
/*-------------------------------------------------------------------------*/

class MutableGlobalDefs
    : public pearl::GlobalDefs
{
    public:
        // --- Public methods ------------------------------

        /// @name Constructors & destructor
        /// @{

        /// @brief Default constructor.
        ///
        /// Constructs a new, empty instance.
        ///
        /// @exception_strong
        ///
        /// @throws std::bad_alloc
        ///     on failure to allocate storage space
        ///
        MutableGlobalDefs();

        /// @}
        /// @name Insertion functions for global definitions
        /// @{

        /// @brief Add %CallingContext definition.
        ///
        /// Adds the CallingContext definition object owned by @p item to the
        /// global definitions container, with transfer of ownership from the
        /// caller to the container.
        ///
        /// @param item
        ///     Smart pointer to the CallingContext definition object to be
        ///     added
        ///
        /// @exception_safety
        ///     If no reallocations happen, the container remains unchanged in
        ///     case of exceptions (strong guarantee).  Otherwise the container
        ///     is guaranteed to end in a valid state (basic guarantee).
        ///
        /// @throws std::invalid_argument
        ///     on @p item not owning an object
        /// @throws std::bad_alloc
        ///     on failure to reallocate storage space during expansion
        ///
        void
        addCallingContext(std::unique_ptr< CallingContext > item);

        /// @brief Add %Callpath definition.
        ///
        /// Adds the Callpath definition object owned by @p item to the global
        /// definitions container, with transfer of ownership from the caller
        /// to the container.  Note that callpath definitions must be added
        /// with consecutive identifiers, i.e., the identifier of @p item has
        /// to match the value returned by numCallpaths() before the call.
        ///
        /// @param item
        ///     Smart pointer to the Callpath definition object to be added
        ///
        /// @exception_safety
        ///     If no reallocations happen, the container remains unchanged in
        ///     case of exceptions (strong guarantee).  Otherwise the container
        ///     is guaranteed to end in a valid state (basic guarantee).
        ///
        /// @throws std::invalid_argument
        ///     on @p item not owning an object or non-consecutive IDs
        /// @throws std::bad_alloc
        ///     on failure to reallocate storage space during expansion
        ///
        void
        addCallpath(std::unique_ptr< Callpath > item);

        /// @brief Add %Callsite definition.
        ///
        /// Adds the Callsite definition object owned by @p item to the global
        /// definitions container, with transfer of ownership from the caller
        /// to the container.
        ///
        /// @param item
        ///     Smart pointer to the Callsite definition object to be added
        ///
        /// @exception_safety
        ///     If no reallocations happen, the container remains unchanged in
        ///     case of exceptions (strong guarantee).  Otherwise the container
        ///     is guaranteed to end in a valid state (basic guarantee).
        ///
        /// @throws std::invalid_argument
        ///     on @p item not owning an object
        /// @throws std::bad_alloc
        ///     on failure to reallocate storage space during expansion
        ///
        void
        addCallsite(std::unique_ptr< Callsite > item);

        /// @brief Add Cartesian topology dimension definition.
        ///
        /// Adds the Cartesian grid topology dimension definition object owned
        /// by @p item to the global definitions container, with transfer of
        /// ownership from the caller to the container.
        ///
        /// @param item
        ///     Smart pointer to the Cartesian grid topology dimension
        ///     definition object to be added
        ///
        /// @exception_safety
        ///     If no reallocations happen, the container remains unchanged in
        ///     case of exceptions (strong guarantee).  Otherwise the container
        ///     is guaranteed to end in a valid state (basic guarantee).
        ///
        /// @throws std::invalid_argument
        ///     on @p item not owning an object
        /// @throws std::bad_alloc
        ///     on failure to reallocate storage space during expansion
        ///
        void
        addCartDimension(std::unique_ptr< CartDimension > item);

        /// @brief Add Cartesian topology definition.
        ///
        /// Adds the Cartesian grid topology definition object owned by @p item
        /// to the global definitions container, with transfer of ownership
        /// from the caller to the container.
        ///
        /// @param item
        ///     Smart pointer to the Cartesian grid topology definition object
        ///     to be added
        ///
        /// @exception_safety
        ///     If no reallocations happen, the container remains unchanged in
        ///     case of exceptions (strong guarantee).  Otherwise the container
        ///     is guaranteed to end in a valid state (basic guarantee).
        ///
        /// @throws std::invalid_argument
        ///     on @p item not owning an object
        /// @throws std::bad_alloc
        ///     on failure to reallocate storage space during expansion
        ///
        void
        addCartTopology(std::unique_ptr< CartTopology > item);

        /// @brief Add %Communicator definition.
        ///
        /// Adds the Communicator definition object owned by @p item to the
        /// global definitions container, with transfer of ownership from the
        /// caller to the container.
        ///
        /// @param item
        ///     Smart pointer to the Communicator definition object to be added
        ///
        /// @exception_safety
        ///     If no reallocations happen, the container remains unchanged in
        ///     case of exceptions (strong guarantee).  Otherwise the container
        ///     is guaranteed to end in a valid state (basic guarantee).
        ///
        /// @throws std::invalid_argument
        ///     on @p item not owning an object
        /// @throws std::bad_alloc
        ///     on failure to reallocate storage space during expansion
        ///
        void
        addCommunicator(std::unique_ptr< Communicator > item);

        /// @brief Add %GroupingSet definition.
        ///
        /// Adds the GroupingSet definition object owned by @p item to the
        /// global definitions container, with transfer of ownership from the
        /// caller to the container.
        ///
        /// Note that there can be at most one communication location set
        /// definition (see class CommLocationSet) for each paradigm.  This is
        /// ensured by the validate() member function.
        ///
        /// @param item
        ///     Smart pointer to the GroupingSet definition object to be added
        ///
        /// @exception_safety
        ///     If no reallocations happen, the container remains unchanged in
        ///     case of exceptions (strong guarantee).  Otherwise the container
        ///     is guaranteed to end in a valid state (basic guarantee).
        ///
        /// @throws std::invalid_argument
        ///     on @p item not owning an object
        /// @throws std::bad_alloc
        ///     on failure to reallocate storage space during expansion
        ///
        void
        addGroupingSet(std::unique_ptr< GroupingSet > item);

        /// @brief Add %Location definition.
        ///
        /// Adds the Location definition object owned by @p item to the global
        /// definitions container, with transfer of ownership from the caller
        /// to the container.
        ///
        /// @param item
        ///     Smart pointer to the Location definition object to be added
        ///
        /// @exception_safety
        ///     If no reallocations happen, the container remains unchanged in
        ///     case of exceptions (strong guarantee).  Otherwise the container
        ///     is guaranteed to end in a valid state (basic guarantee).
        ///
        /// @throws std::invalid_argument
        ///     on @p item not owning an object
        /// @throws std::bad_alloc
        ///     on failure to reallocate storage space during expansion
        ///
        void
        addLocation(std::unique_ptr< Location > item);

        /// @brief Add %LocationGroup definition.
        ///
        /// Adds the LocationGroup definition object owned by @p item to the
        /// global definitions container, with transfer of ownership from the
        /// caller to the container.
        ///
        /// @param item
        ///     Smart pointer to the LocationGroup definition object to be
        //      added
        ///
        /// @exception_safety
        ///     If no reallocations happen, the container remains unchanged in
        ///     case of exceptions (strong guarantee).  Otherwise the container
        ///     is guaranteed to end in a valid state (basic guarantee).
        ///
        /// @throws std::invalid_argument
        ///     on @p item not owning an object
        /// @throws std::bad_alloc
        ///     on failure to reallocate storage space during expansion
        ///
        void
        addLocationGroup(std::unique_ptr< LocationGroup > item);

        /// @brief Add %Metric definition.
        ///
        /// Adds the Metric definition object owned by @p item to the global
        /// definitions container, with transfer of ownership from the caller
        /// to the container.
        ///
        /// @param item
        ///     Smart pointer to the Metric definition object to be added
        ///
        /// @exception_safety
        ///     If no reallocations happen, the container remains unchanged in
        ///     case of exceptions (strong guarantee).  Otherwise the container
        ///     is guaranteed to end in a valid state (basic guarantee).
        ///
        /// @throws std::invalid_argument
        ///     on @p item not owning an object
        /// @throws std::bad_alloc
        ///     on failure to reallocate storage space during expansion
        ///
        void
        addMetric(std::unique_ptr< Metric > item);

        /// @brief Add %Region definition.
        ///
        /// Adds the Region definition object owned by @p item to the global
        /// definitions container, with transfer of ownership from the caller
        /// to the container.
        ///
        /// @param item
        ///     Smart pointer to the Region definition object to be added
        ///
        /// @exception_safety
        ///     If no reallocations happen, the container remains unchanged in
        ///     case of exceptions (strong guarantee).  Otherwise the container
        ///     is guaranteed to end in a valid state (basic guarantee).
        ///
        /// @throws std::invalid_argument
        ///     on @p item not owning an object
        /// @throws std::bad_alloc
        ///     on failure to reallocate storage space during expansion
        ///
        void
        addRegion(std::unique_ptr< Region > item);

        /// @brief Add %RmaWindow definition.
        ///
        /// Adds the RmaWindow definition object owned by @p item to the global
        /// definitions container, with transfer of ownership from the caller
        /// to the container.
        ///
        /// @param item
        ///     Smart pointer to the RmaWindow definition object to be added
        ///
        /// @exception_safety
        ///     If no reallocations happen, the container remains unchanged in
        ///     case of exceptions (strong guarantee).  Otherwise the container
        ///     is guaranteed to end in a valid state (basic guarantee).
        ///
        /// @throws std::invalid_argument
        ///     on @p item not owning an object
        /// @throws std::bad_alloc
        ///     on failure to reallocate storage space during expansion
        ///
        void
        addRmaWindow(std::unique_ptr< RmaWindow > item);

        /// @brief Add %SourceLocation definition.
        ///
        /// Adds the SourceLocation definition object owned by @p item to the
        /// global definitions container, with transfer of ownership from the
        /// caller to the container.
        ///
        /// @param item
        ///     Smart pointer to the SourceLocation definition object to be
        //      added
        ///
        /// @exception_safety
        ///     If no reallocations happen, the container remains unchanged in
        ///     case of exceptions (strong guarantee).  Otherwise the container
        ///     is guaranteed to end in a valid state (basic guarantee).
        ///
        /// @throws std::invalid_argument
        ///     on @p item not owning an object
        /// @throws std::bad_alloc
        ///     on failure to reallocate storage space during expansion
        ///
        void
        addSourceLocation(std::unique_ptr< SourceLocation > item);

        /// @brief Add %String definition.
        ///
        /// Adds the String definition object owned by @p item to the global
        /// definitions container, with transfer of ownership from the caller
        /// to the container.
        ///
        /// @param item
        ///     Smart pointer to the String definition object to be added
        ///
        /// @exception_safety
        ///     If no reallocations happen, the container remains unchanged in
        ///     case of exceptions (strong guarantee).  Otherwise the container
        ///     is guaranteed to end in a valid state (basic guarantee).
        ///
        /// @throws std::invalid_argument
        ///     on @p item not owning an object
        /// @throws std::bad_alloc
        ///     on failure to reallocate storage space during expansion
        ///
        void
        addString(std::unique_ptr< String > item);

        /// @brief Add %SystemNode definition.
        ///
        /// Adds the SystemNode definition object owned by @p item to the global
        /// definitions container, with transfer of ownership from the caller
        /// to the container.
        ///
        /// @param item
        ///     Smart pointer to the SystemNode definition object to be added
        ///
        /// @exception_safety
        ///     If no reallocations happen, the container remains unchanged in
        ///     case of exceptions (strong guarantee).  Otherwise the container
        ///     is guaranteed to end in a valid state (basic guarantee).
        ///
        /// @throws std::invalid_argument
        ///     on @p item not owning an object
        /// @throws std::bad_alloc
        ///     on failure to reallocate storage space during expansion
        ///
        void
        addSystemNode(std::unique_ptr< SystemNode > item);

        /// @}
        /// @name Definition amendment functions
        /// @{

        /// @brief Add rank coordinate to Cartesian topology definition.
        ///
        /// Adds the given @p coordinate to the list of Cartesian coordinates
        /// for a particular @p rank in the Cartesian grid topology definition
        /// with the identifier @p cartTopoId.
        ///
        /// @param cartTopoId
        ///     Global Cartesian grid topology identifier
        /// @param rank
        ///     Local rank in associated communicator
        /// @param coordinate
        ///     Cartesian coordinate to be added
        ///
        /// @exception_safety
        ///     If no reallocations happen, the container remains unchanged in
        ///     case of exceptions (strong guarantee).  Otherwise the container
        ///     is guaranteed to end in a valid state (basic guarantee).
        ///
        /// @throws RuntimeError
        ///     on failure to find the Cartesian topology with the given
        ///         @p cartTopoId<br/>
        ///     on communicator &ne; Communicator::UNDEFINED and rank out of
        ///         range<br/>
        ///     on invalid @#dimensions<br/>
        ///     on coordinate value out of range
        /// @throws std::bad_alloc
        ///     on failure to allocate storage
        ///
        void
        assignCartCoordinate(CartTopology::IdType  cartTopoId,
                             std::uint32_t         rank,
                             const CartCoordinate& coordinate);

        /// @brief Assign communicator to Cartesian topology definition.
        ///
        /// Assigns the given @p communicator to the Cartesian grid topology
        /// with the global identifier @p cartTopoId.
        ///
        /// @param cartTopoId
        ///     Global Cartesian grid topology identifier
        /// @param communicator
        ///     New associated communicator
        ///
        /// @exception_strong
        ///
        /// @throws RuntimeError
        ///     on failure to find the Cartesian topology with the given
        ///         @p cartTopoId
        ///
        void
        assignCartCommunicator(CartTopology::IdType cartTopoId,
                               const Communicator&  communicator);

        /// @}
        /// @name Query functions for communication location set definitions
        /// @{

        /// @brief Communication location set lookup by paradigm.
        ///
        /// Returns a constant reference to the communication location set
        /// definition object with the given @p paradigm.
        ///
        /// @param paradigm
        ///     %Paradigm of the communication location set definition object
        ///     to be retrieved
        /// @returns
        ///     Constant reference to the corresponding communication location
        ///     set object
        ///
        /// @exception_strong
        ///
        /// @throws pearl::RuntimeError
        ///     on failure to find a communication location set definition
        ///     object with the given @p paradigm
        ///
        const CommLocationSet&
        getCommLocationSet(Paradigm paradigm) const;

        /// @}
        /// @name Internal methods
        /// @{

        void
        validate();

        void
        setup();

        /// @}
        /// @name Set internal regions
        /// @{

        void
        setPausingRegion(Region* region);

        void
        setFlushingRegion(Region* region);

        void
        setTaskRootRegion(Region* region);

        void
        setThreadRootRegion(Region* region);

        /// @}
        /// @name Call tree handling
        /// @{

        CallTree*
        getCallTree() const;

        void
        setCallTree(CallTree* callTree);

        /// @}
        /// @name Global time offset handling
        /// @{

        timestamp_t
        getGlobalOffset() const;

        void
        setGlobalOffset(timestamp_t offset);

        /// @brief Set the experiment's realtime offset.
        ///
        /// Sets the realtime offset of the experiment's trace data in
        /// nanoseconds since 1970-01-01T00:00 UTC, or `UINT64_MAX` if this
        /// information is unavailable.
        ///
        /// @param offset
        ///     Realtime trace data offset
        void
        setRealtimeOffset(std::uint64_t offset);

        /// @}


    private:
        // --- Type definitions ----------------------------

        /// Symbolic names for definition types
        enum DefinitionType
        {
            UNDEFINED,    ///< Only used as initial value

            CALLINGCONTEXT,
            CALLPATH,
            CALLSITE,
            CARTDIMENSION,
            CARTTOPOLOGY,
            COMMUNICATOR,
            GROUPINGSET,
            LOCATION,
            LOCATIONGROUP,
            METRIC,
            REGION,
            RMAWINDOW,
            SOURCELOCATION,
            STRING,
            SYSTEMNODE
        };


        /// Container type for communication location set definitions
        typedef std::vector< CommLocationSet* > CommLocationSetContainer;


        // --- Data members --------------------------------

        /// Communication location set definitions
        CommLocationSetContainer mCommLocationSets;

        /// Global time offset subtracted during trace preprocessing
        timestamp_t mGlobalOffset;

        /// Stores the type of the last definition object added
        DefinitionType mLastDefType;


        // --- Private methods -----------------------------

        /// @name Definition type tracking
        /// @{

        void
        trackType(DefinitionType currDefType);

        /// @}
};
}    // namespace pearl::detail
}    // namespace pearl


#endif    // !PEARL_MUTABLEGLOBALDEFS_H
