/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2022-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_base
 *  @brief   Implementation of PEARL-internal debug functionality.
 *
 *  This file provides the implementation of PEARL-internal functions to
 *  initialize and finalize debug output redirection to per-process files.
 **/
/*-------------------------------------------------------------------------*/


#include <config.h>

#include "pearl_debug.h"

#include <cstdio>
#include <cstdlib>
#include <memory>
#include <string>

#include <UTILS_Debug.h>
#include <UTILS_Error.h>

#include "Threading.h"
#include "pearl_ipc.h"


// --- Local variables & helper functions ----------------------------------

namespace
{
/// @brief Custom deleter for the debug log stream.
///
/// This function object implements a custom deleter for the debug log
/// stream to be used in conjunction with `std::unique_ptr` for automatic
/// resource management.
///
struct closeDebugLog
{
    /// @brief Reset debug output redirection and close log file.
    ///
    /// Disables the debug output redirection to process-local files and
    /// afterwards closes the log file referenced by the given file pointer
    /// @p fp.
    ///
    /// @param fp
    ///     Debug log file pointer
    ///
    void
    operator()(std::FILE* fp) const
    {
        UTILS_Debug_SetLogStream(nullptr);
        std::fclose(fp);
    }
};


/// Smart pointer managing the debug log file pointer.
std::unique_ptr< std::FILE, closeDebugLog > logStream;
}    // unnamed namespace


namespace pearl
{
namespace detail
{
// --- Debugging functions  -------------------------------------------------

void
debugInit()
{
    const char* const logPrefix = getenv("SCALASCA_DEBUG_LOG");
    if (logPrefix != nullptr)
    {
        std::string logName(logPrefix);
        logName += '.';
        logName += std::to_string(ipcGetRank(ipcCommWorld));

        logStream.reset(fopen(logName.c_str(), "w"));
        if (!logStream)
        {
            UTILS_ERROR_POSIX("Failed to open debug log file '%s'", logName.c_str());
            std::exit(EXIT_FAILURE);
        }

        UTILS_Debug_SetLogStream(logStream.get());
    }
}


void
debugFinalize()
{
    static Mutex mutex;

    mutex.lock();
    logStream.reset();
    mutex.unlock();
}
}    // namespace pearl::detail
}    // namespace pearl
