/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @ingroup PEARL_ipc
 *  @brief   Implementation of inter-process communication functions for
 *           single-process codes.
 *
 *  This file provides a "dummy" implementation of the PEARL-internal
 *  inter-process communication layer for single-process codes.  See
 *  pearl_ipc.h for a detailed description of the API.
 **/
/*-------------------------------------------------------------------------*/


#include <config.h>

#include "pearl_ipc.h"

#include <climits>
#include <cstddef>
#include <cstdlib>
#include <cstring>

#include <UTILS_Error.h>

using namespace std;
using namespace pearl::detail;


// --- Local helper functions -----------------------------------------------

namespace
{
/// @brief Map PEARL IPC datatype to element size in bytes
///
/// Returns the size of a single element of the given PEARL IPC @a datatype
/// in bytes.
///
/// @param ipcDatatype
///     PEARL IPC datatype
/// @returns
///     Size of a single element in bytes
///
inline size_t
getElementSize(const IpcDatatype ipcDatatype)
{
    static const size_t elementSizes[] = {
        sizeof(int8_t),
        sizeof(uint8_t),
        sizeof(int16_t),
        sizeof(uint16_t),
        sizeof(int32_t),
        sizeof(uint32_t),
        sizeof(int64_t),
        sizeof(uint64_t),
        sizeof(float),
        sizeof(double)
    };

    return elementSizes[ipcDatatype];
}
}    // unnamed namespace


// --- Type definitions -----------------------------------------------------

struct pearl::detail::IpcComm
{
};


// --- Global variables -----------------------------------------------------

IpcComm* pearl::detail::ipcCommWorld = nullptr;


// --- IPC layer: Environmental management ----------------------------------

void
pearl::detail::ipcInit()
{
    UTILS_BUG_ON(ipcCommWorld != nullptr,
                 "PEARL IPC layer already initialized");

    ipcCommWorld = new IpcComm;
}


void
pearl::detail::ipcFinalize()
{
    UTILS_BUG_ON(ipcCommWorld == nullptr,
                 "PEARL IPC layer not initialized");

    delete ipcCommWorld;
    ipcCommWorld = nullptr;
}


void
pearl::detail::ipcAbort()
{
    std::abort();
}


// --- IPC layer: Communication contexts ------------------------------------

uint32_t
pearl::detail::ipcGetSize(const IpcComm* const comm)
{
    UTILS_ASSERT(comm != nullptr);

    return 1;
}


uint32_t
pearl::detail::ipcGetRank(const IpcComm* const comm)
{
    UTILS_ASSERT(comm != nullptr);

    return 0;
}


void
pearl::detail::ipcCommSplit(const IpcComm* const comm,
                            const uint32_t       color,
                            const uint32_t       key,
                            IpcComm** const      newcomm)
{
    UTILS_ASSERT(comm != nullptr);
    UTILS_ASSERT(color <= INT_MAX);
    UTILS_ASSERT(key <= INT_MAX);
    UTILS_ASSERT(newcomm != nullptr);

    *newcomm = new IpcComm;
}


void
pearl::detail::ipcCommFree(IpcComm** const comm)
{
    UTILS_ASSERT((comm != nullptr) && (*comm != ipcCommWorld));

    delete *comm;
    *comm = nullptr;
}


// --- IPC layer: Collective communication ----------------------------------

void
pearl::detail::ipcBarrier(const IpcComm* const comm)
{
    UTILS_ASSERT(comm != nullptr);

    // Single process => no synchronization necessary
}


void
pearl::detail::ipcBroadcast(void* const          buffer,
                            const uint32_t       count,
                            const IpcDatatype    datatype,
                            const uint32_t       root,
                            const IpcComm* const comm)
{
    UTILS_ASSERT(comm != nullptr);
    UTILS_ASSERT((buffer != nullptr) || (count == 0));
    UTILS_ASSERT(count <= INT_MAX);
    UTILS_ASSERT(root == 0);

    // Single process => data stays in buffer on root
}


void
pearl::detail::ipcGather(const void* const    sendBuffer,
                         void* const          recvBuffer,
                         const uint32_t       count,
                         const IpcDatatype    datatype,
                         const uint32_t       root,
                         const IpcComm* const comm)
{
    UTILS_ASSERT(comm != nullptr);
    UTILS_ASSERT((sendBuffer != nullptr) || (count == 0));
    UTILS_ASSERT((recvBuffer != nullptr) || (count == 0));
    UTILS_ASSERT((sendBuffer != recvBuffer) || (count == 0));
    UTILS_ASSERT(count <= INT_MAX);
    UTILS_ASSERT(root == 0);

    // Single process => copy data from sendBuffer to recvBuffer
    const size_t numBytes = count * getElementSize(datatype);
    memcpy(recvBuffer, sendBuffer, numBytes);
}


void
pearl::detail::ipcGatherv(const void* const     sendBuffer,
                          const uint32_t        sendCount,
                          void* const           recvBuffer,
                          const uint32_t* const recvCounts,
                          const IpcDatatype     datatype,
                          const uint32_t        root,
                          const IpcComm* const  comm)
{
    UTILS_ASSERT(comm != nullptr);
    UTILS_ASSERT((sendBuffer != nullptr) || (sendCount == 0));
    UTILS_ASSERT((recvBuffer != nullptr) || (sendCount == 0));
    UTILS_ASSERT((sendBuffer != recvBuffer) || (sendCount == 0));
    UTILS_ASSERT((recvCounts != nullptr) && (recvCounts[0] == sendCount));
    UTILS_ASSERT(root == 0);

    // Single process => copy data from sendBuffer to recvBuffer
    const size_t numBytes = sendCount * getElementSize(datatype);
    memcpy(recvBuffer, sendBuffer, numBytes);
}


void
pearl::detail::ipcScatter(const void* const    sendBuffer,
                          void* const          recvBuffer,
                          const uint32_t       count,
                          const IpcDatatype    datatype,
                          const uint32_t       root,
                          const IpcComm* const comm)
{
    UTILS_ASSERT(comm != nullptr);
    UTILS_ASSERT((sendBuffer != nullptr) || (count == 0));
    UTILS_ASSERT((recvBuffer != nullptr) || (count == 0));
    UTILS_ASSERT((sendBuffer != recvBuffer) || (count == 0));
    UTILS_ASSERT(count <= INT_MAX);
    UTILS_ASSERT(root == 0);

    // Single process => copy data from sendBuffer to recvBuffer
    const size_t numBytes = count * getElementSize(datatype);
    memcpy(recvBuffer, sendBuffer, numBytes);
}


void
pearl::detail::ipcScatterv(const void* const     sendBuffer,
                           const uint32_t* const sendCounts,
                           void* const           recvBuffer,
                           const uint32_t        recvCount,
                           const IpcDatatype     datatype,
                           const uint32_t        root,
                           const IpcComm* const  comm)
{
    UTILS_ASSERT(comm != nullptr);
    UTILS_ASSERT((sendBuffer != nullptr) || (recvCount == 0));
    UTILS_ASSERT((sendCounts != nullptr) && (sendCounts[0] == recvCount));
    UTILS_ASSERT((recvBuffer != nullptr) || (recvCount == 0));
    UTILS_ASSERT((sendBuffer != recvBuffer) || (recvCount == 0));
    UTILS_ASSERT(root == 0);

    // Single process => copy data from sendBuffer to recvBuffer
    const size_t numBytes = recvCount * getElementSize(datatype);
    memcpy(recvBuffer, sendBuffer, numBytes);
}
