/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2017-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#ifndef PEARL_MEMORYCHUNKTREE_H
    #error "Do not include this file!  Use 'MemoryChunkTree.h' instead."
#endif


#include <UTILS_Error.h>

#include "MemoryChunk.h"


namespace pearl
{
namespace detail
{
inline MemoryChunk*
insert(MemoryChunk*       root,
       MemoryChunk* const chunk,
       const std::size_t  chunkSize)
{
    UTILS_ASSERT(chunk != nullptr);

    // Special case: Empty tree => chunk is new root
    if (root == nullptr)
    {
        return chunk;
    }

    // Make closest leaf node the new (temporary) root
    root = splay(root, chunk, chunkSize);

    // Chunk is in left subtree
    //   => split on left link of root and join subtrees using chunk
    if (chunk->isLessThan(root, chunkSize))
    {
        chunk->mLeft  = root->mLeft;
        chunk->mRight = root;
        root->mLeft   = nullptr;
    }

    // Chunk is in right subtree
    //   => split on right link of root and join subtrees using chunk
    else
    {
        chunk->mLeft  = root;
        chunk->mRight = root->mRight;
        root->mRight  = nullptr;
    }

    // Chunk is new root
    return chunk;
}


inline MemoryChunk*
remove(MemoryChunk*      root,
       const std::size_t chunkSize)
{
    UTILS_ASSERT(root != nullptr);

    // No left subtree => root of right subtree is new root
    if (root->mLeft == nullptr)
    {
        root = root->mRight;
    }

    // Otherwise => join left and right subtrees
    else
    {
        MemoryChunk* const left  = root->mLeft;
        MemoryChunk* const right = root->mRight;

        // Make "largest" node in left subtree the new root (with no right
        // child) and add right subtree of previous root as its right subtree
        root         = splay(left, root, chunkSize);
        root->mRight = right;
    }

    return root;
}


inline MemoryChunk*
splay(MemoryChunk*      root,
      void* const       ptr,
      const std::size_t chunkSize)
{
    UTILS_ASSERT(root != nullptr);

    // ptr is in left subtree
    if (  root->isGreaterThan(ptr, chunkSize)
       && (root->mLeft != nullptr))
    {
        // Case 2: zig-zig (left/left)
        if (  root->mLeft->isGreaterThan(ptr, chunkSize)
           && (root->mLeft->mLeft != nullptr))
        {
            // Recursively make chunk containing ptr (or the closest leaf)
            // the root of left-left
            root->mLeft->mLeft = splay(root->mLeft->mLeft, ptr, chunkSize);

            // Two single rotations
            //   => first rotation brings left to root,
            //      second rotation (outside the `if`) left-left to root
            root = rotateRight(root);
        }

        // Case 3: zig-zag (left/right)
        else if (  root->mLeft->isLessThan(ptr, chunkSize)
                && (root->mLeft->mRight != nullptr))
        {
            // Recursively make chunk containing ptr (or the closest leaf)
            // the root of left-right
            root->mLeft->mRight = splay(root->mLeft->mRight, ptr, chunkSize);

            // Double rotation
            //   => first rotation brings left-right to left,
            //      second rotation (outside the `if`) left-right to root
            root->mLeft = rotateLeft(root->mLeft);
        }

        // Case 1: zig (left)
        // Terminating single rotation => brings left to root
        root = rotateRight(root);
    }

    // ptr is in right subtree
    else if (  root->isLessThan(ptr, chunkSize)
            && (root->mRight != nullptr))
    {
        // Case 2: zig-zig (right/right)
        if (  root->mRight->isLessThan(ptr, chunkSize)
           && (root->mRight->mRight != nullptr))
        {
            // Recursively make chunk containing ptr (or the closest leaf)
            // the root of right-right
            root->mRight->mRight = splay(root->mRight->mRight, ptr, chunkSize);

            // Two single rotations
            //   => first rotation brings right to root,
            //      second rotation (outside the `if`) right-right to root
            root = rotateLeft(root);
        }

        // Case 3: zig-zag (right/left)
        else if (  root->mRight->isGreaterThan(ptr, chunkSize)
                && (root->mRight->mLeft != nullptr))
        {
            // Recursively make chunk containing ptr (or the closest leaf)
            // the root of right-left
            root->mRight->mLeft = splay(root->mRight->mLeft, ptr, chunkSize);

            // Double rotation
            //   => first rotation brings right-left to right,
            //      second rotation (outside the `if`) right-left to root
            root->mRight = rotateRight(root->mRight);
        }

        // Case 1: zig (right)
        // Terminating single rotation => brings right to root
        root = rotateLeft(root);
    }

    // root now contains ptr or is closest leaf if ptr is not in tree
    return root;
}


inline MemoryChunk*
rotateLeft(MemoryChunk* const root)
{
    UTILS_ASSERT(root != nullptr);
    UTILS_ASSERT(root->mRight != nullptr);

    MemoryChunk* const temp = root->mRight;

    root->mRight = temp->mLeft;
    temp->mLeft  = root;

    return temp;
}


inline MemoryChunk*
rotateRight(MemoryChunk* const root)
{
    UTILS_ASSERT(root != nullptr);
    UTILS_ASSERT(root->mLeft != nullptr);

    MemoryChunk* const temp = root->mLeft;

    root->mLeft  = temp->mRight;
    temp->mRight = root;

    return temp;
}
}    // namespace pearl::detail
}    // namespace pearl
