/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "MpiDatatypes.h"

#include "WaitstateStats.h"

using namespace scout;


// --- Global variables -----------------------------------------------------

/// MPI datatype handle for TopMostSevere struct
MPI_Datatype scout::TOPSEVERE;

/// MPI datatype handle for array with TOP_SEVERE_INSTANCES entries of
/// type TopMostSevere
MPI_Datatype scout::TOPSEVEREARRAY;

/// MPI datatype handle for TopMostSevereMaxSum struct
MPI_Datatype scout::TOPSEVERECOLL;


// --- Related functions ----------------------------------------------------

/**
 *  @brief Creates all SCOUT-specific MPI datatypes.
 *
 *  This function creates all SCOUT-specific MPI derived datatypes. It can
 *  only be called after MPI has been successfully initialized.
 **/
void
scout::InitDatatypes()
{
    // Define MPI datatype for TopMostSevere struct
    TopMostSevere topMost[2];    // We need 2 entries to account for padding
    int           counts[4];
    MPI_Aint      displs[4];
    MPI_Datatype  types[4];
    MPI_Datatype  tmpType;

    counts[0] = 3;
    types[0]  = MPI_DOUBLE;
    counts[1] = 1;
    types[1]  = SCALASCA_MPI_UINT32_T;
    counts[2] = 1;
    types[2]  = MPI_INT;
    MPI_Get_address(&topMost[0].idletime, &displs[0]);
    MPI_Get_address(&topMost[0].cnode, &displs[1]);
    MPI_Get_address(&topMost[0].rank, &displs[2]);
    MPI_Get_address(&topMost[1], &displs[3]);
    displs[3] -= displs[0];
    displs[2] -= displs[0];
    displs[1] -= displs[0];
    displs[0]  = 0;
    MPI_Type_create_struct(3, counts, displs, types, &tmpType);
    MPI_Type_create_resized(tmpType, 0, displs[3], &TOPSEVERE);
    MPI_Type_free(&tmpType);
    MPI_Type_commit(&TOPSEVERE);

    // Define MPI datatype for array of TopMostSevere struct entries
    MPI_Type_contiguous(TOP_SEVERE_INSTANCES, TOPSEVERE, &TOPSEVEREARRAY);
    MPI_Type_commit(&TOPSEVEREARRAY);

    // Define MPI datatype for TopMostSevereMaxSum struct
    TopMostSevereMaxSum topMostCollMaxSum;

    counts[0] = 3;
    types[0]  = MPI_DOUBLE;
    counts[1] = 1;
    types[1]  = SCALASCA_MPI_UINT32_T;
    counts[2] = 1;
    types[2]  = MPI_INT;
    counts[3] = 1;
    types[3]  = MPI_DOUBLE;
    MPI_Get_address(&topMostCollMaxSum.idletime, &displs[0]);
    MPI_Get_address(&topMostCollMaxSum.cnode, &displs[1]);
    MPI_Get_address(&topMostCollMaxSum.rank, &displs[2]);
    MPI_Get_address(&topMostCollMaxSum.idletime_max, &displs[3]);
    displs[3] -= displs[0];
    displs[2] -= displs[0];
    displs[1] -= displs[0];
    displs[0]  = 0;
    MPI_Type_create_struct(4, counts, displs, types, &TOPSEVERECOLL);
    MPI_Type_commit(&TOPSEVERECOLL);
}


/**
 *  @brief Releases all SCOUT-specific MPI datatypes.
 *
 *  This function releases all SCOUT-specific MPI derived datatypes created
 *  by InitDatatypes(). It must be called before MPI is finalized.
 **/
void
scout::FreeDatatypes()
{
    // Release MPI datatypes
    MPI_Type_free(&TOPSEVERECOLL);
    MPI_Type_free(&TOPSEVEREARRAY);
    MPI_Type_free(&TOPSEVERE);
}
