/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2012-2021                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 * @file
 * @brief Declaration of the class MpiNonBlockingBarrier.
 *
 * This header file provides the declaration of the class
 * MpiNonBlockingBarrier, which provides an implementation of a
 * non-blocking barrier using MPI point-to-point communication.
 */
/*-------------------------------------------------------------------------*/


#ifndef SCOUT_MPINONBLOCKINGBARRIER_H
#define SCOUT_MPINONBLOCKINGBARRIER_H


#include <vector>

#include <mpi.h>


namespace scout
{
/*-------------------------------------------------------------------------*/
/**
 * @class MpiNonBlockingBarrier
 * @brief Implementation of a non-blocking barrier.
 *
 * The MpiNonBlockingBarrier class provides an implementation of a
 * non-blocking barrier based on MPI point-to-point communication using
 * the dissemination barrier algorithm.
 */
/*-------------------------------------------------------------------------*/

class MpiNonBlockingBarrier
{
    public:
        // --- Public methods ------------------------------

        /// @name Construction & destruction
        /// @{

        /// @brief Constructs a non-blocking barrier over a given communicator.
        ///
        /// This constructor creates an object which represents a non-blocking
        /// barrier over a given MPI communicator.
        ///
        /// @param comm
        ///     The communicator over which the barrier shall be created.
        ///
        explicit
        MpiNonBlockingBarrier(MPI_Comm comm);

        /// @}
        /// @name Blocking intialization & finalization
        /// @{

        /// @brief Initialize the barrier
        ///
        /// This method has to be called once before using the non-blocking
        /// barrier.  Calling one of the other methods first may lead to
        /// undefined behavior.  This method requires synchronization!
        ///
        void
        init();

        /// @brief Finalize the barrier
        ///
        /// This method has to be called once after using the non-blocking
        /// barrier.  Calling one of the other methods afterwards may lead
        /// to undefined behavior.  This method requires synchronization!
        ///
        void
        finalize();

        /// @}
        /// @name Non-blocking execution of a barrier
        /// @{

        /// @brief Starts the barrier
        ///
        /// This method starts the barrier.  Until every participating process
        /// has called this method, test() will not return `true` and wait()
        /// will not terminate.
        ///
        void
        start();

        /// @brief Tests whether the barrier is done
        ///
        /// This method returns `false` until every process has called start().
        /// Afterwards the barrier is complete and test() will return `true`.
        ///
        /// @returns
        ///     `true` if the barrier is completed, `false` otherwise
        ///
        bool
        test();

        /// @brief Waits until completion of the barrier
        ///
        /// Waits until every participating process at least called the barrier
        /// start() method.  Communication between other processes however might
        /// not be finished upon returning.
        ///
        void
        wait();

        /// @}


    private:
        // --- Data members --------------------------------

        /// Current step number
        int mStep;

        /// Total number of steps
        int mNumSteps;

        /// Communicator size
        int mSize;

        /// Rank within communicator
        int mRank;

        /// Communicator
        MPI_Comm mCommunicator;

        /// Requests array
        std::vector< MPI_Request > mRequests;


        // --- Private methods -----------------------------

        /// @brief Perform a step of the dissemination barrier
        ///
        /// Executes a single step of the dissemination barrier algorithm.
        ///
        void
        performStep();
};
}    // namespace scout


#endif    // !SCOUT_MPINONBLOCKINGBARRIER_H
