/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "MpiOperators.h"

#include <algorithm>
#include <cstring>

#include <UTILS_Error.h>

#include "MpiDatatypes.h"
#include "WaitstateStats.h"

using namespace std;
using namespace scout;


// --- Global variables -----------------------------------------------------

/// MPI reduction operator handle for most-severe instance reduction
MPI_Op scout::MAX_TOPSEVEREARRAY;

/// MPI reduction operator handle for collective wait-state information
/// reduction
MPI_Op scout::MAXSUM_TOPSEVERECOLL;


// --- Local function prototypes --------------------------------------------

extern "C" void
top_severe(TopMostSevere* in,
           TopMostSevere* inout,
           int*           len,
           MPI_Datatype*  type);

extern "C" void
top_severe_coll_maxsum(TopMostSevereMaxSum* in,
                       TopMostSevereMaxSum* inout,
                       int*                 len,
                       MPI_Datatype*        type);


// --- Related functions ----------------------------------------------------

/**
 *  @brief Creates all SCOUT-specific MPI reduction operators.
 *
 *  This function creates all SCOUT-specific MPI reduction operators. It can
 *  only be called after MPI has been successfully initialized.
 **/
void
scout::InitOperators()
{
    // Define MPI reduction operators
    MPI_Op_create((MPI_User_function*)top_severe, true, &MAX_TOPSEVEREARRAY);
    MPI_Op_create((MPI_User_function*)top_severe_coll_maxsum, true,
                  &MAXSUM_TOPSEVERECOLL);
}


/**
 *  @brief Releases all SCOUT-specific MPI reduction operators.
 *
 *  This function releases all SCOUT-specific MPI reduction operators created
 *  by InitOperators(). It must be called before MPI is finalized.
 **/
void
scout::FreeOperators()
{
    // Release MPI reduction operators
    MPI_Op_free(&MAX_TOPSEVEREARRAY);
    MPI_Op_free(&MAXSUM_TOPSEVERECOLL);
}


// --- Local functions ------------------------------------------------------

/// @brief Merges most-severe waitstate instance tracking vectors.
///
/// Element-wise merge of local most-severe waitstate instance tracking
/// vectors stored in the given arrays @p inoutvec and @p invec of length
/// @p len.  Each element is a sub-array of length TOP_SEVERE_INSTANCES,
/// with the top-most severe instances sorted by their severity.
///
/// @param invec
///     Input array
/// @param inoutvec
///     Input/output array
/// @param len
///     Pointer to number of array elements
/// @param datatype
///     MPI datatype
///
/// @pre
///     datatype == scout::TOPSEVEREARRAY
/// @pre
///     Sub-arrays must be sorted
///
void
top_severe(TopMostSevere* invec,
           TopMostSevere* inoutvec,
           int*           len,
           MPI_Datatype*  datatype)
{
    UTILS_ASSERT(*datatype == TOPSEVEREARRAY);

    auto* inBegin = invec;
    auto* ioBegin = inoutvec;

    for (int i = 0; i < *len; ++i)
    {
        auto* inEnd = inBegin + TOP_SEVERE_INSTANCES;
        auto* ioEnd = ioBegin + TOP_SEVERE_INSTANCES;

        for (auto* inIt = inBegin; inIt != inEnd; ++inIt)
        {
            // Search for an existing entry with identical callpath ID
            auto* it = std::find_if(ioBegin, ioEnd,
                                    [inIt](const TopMostSevere& entry)
                                    {
                                        return (entry.cnode == inIt->cnode);
                                    });

            // Found => Update entry and "bubble" it to the correct position
            if (it != ioEnd)
            {
                *it = std::max(*it, *inIt);

                auto* prev = std::prev(it);
                while (  (it != ioBegin)
                      && (*prev < *it))
                {
                    std::swap(*prev, *it);

                    it = prev--;
                }

                continue;
            }

            // Not found => Insert item at the correct position
            it = std::find_if(ioBegin, ioEnd,
                              [inIt](const TopMostSevere& entry)
                              {
                                  return (entry < *inIt);
                              });
            if (it == ioEnd)
            {
                // Case 1: Item less than any existing entry => no update
                continue;
            }

            // Case 2: Insert, dropping smallest entry
            std::move_backward(it, std::prev(ioEnd), ioEnd);
            *it = *inIt;
        }

        inBegin = inEnd;
        ioBegin = ioEnd;
    }
}


/// @brief Merges partial collective waitstate information.
///
/// Element-wise merge of the partial collective waitstate information
/// stored in the given arrays @p inoutvec and @p invec of length @p len,
/// using the TopMostSevereMaxSum merge operation.
///
/// @param invec
///     Input array
/// @param inoutvec
///     Input/output array
/// @param len
///     Pointer to number of array elements
/// @param datatype
///     MPI datatype
///
/// @pre
///     datatype == scout::TOPSEVERECOLL
///
void
top_severe_coll_maxsum(TopMostSevereMaxSum* invec,
                       TopMostSevereMaxSum* inoutvec,
                       int*                 len,
                       MPI_Datatype*        datatype)
{
    UTILS_ASSERT(*datatype == TOPSEVERECOLL);

    for (int i = 0; i < *len; ++i)
    {
        inoutvec[i] = merge(inoutvec[i], invec[i]);
    }
}
