/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @brief Declaration of the class Pattern.
 *
 *  This header file provides the declaration of the abstract class Pattern,
 *  which defines the interface that has to be implemented by all performance
 *  properties calculated by SCOUT.
 **/
/*-------------------------------------------------------------------------*/


#ifndef SCOUT_PATTERN_H
#define SCOUT_PATTERN_H


#include <string>

#include <cubew_metric.h>

#include "scout_types.h"


// --- Forward declarations -------------------------------------------------

namespace pearl
{
class LocalTrace;
}    // namespace pearl


namespace scout
{
// --- Forward declarations -------------------------------------------------

struct ReportData;


/*-------------------------------------------------------------------------*/
/**
 *  @class Pattern
 *  @brief Abstract base class for all properties calculated by SCOUT.
 *
 *  The Pattern class serves as an abstract base class for all performance
 *  properties calculated by SCOUT and defines the interface that has to be
 *  implemented by concrete derived classes.  This includes basic metrics
 *  such as @em Time and @em Visits, as well as the inefficiency patterns
 *  (i.e., wait states) occurring at synchronization points, delay costs,
 *  etc.
 **/
/*-------------------------------------------------------------------------*/

class Pattern
{
    public:
        // --- Public methods ------------------------------

        /// @name Constructors & destructor
        /// @{

        /// @brief Destructor.
        ///
        /// Destroys the instance and performs pattern-specific data
        /// management (i.e., cleanup) tasks.
        ///
        virtual
        ~Pattern();

        /// @}
        /// @name Callback registration
        /// @{

        /// @brief Registers the pattern-specific replay callbacks.
        ///
        /// Registers the replay callbacks that need to be triggered for this
        /// pattern during each of the trace analysis replay phases.  Generic
        /// callbacks defined by base classes are registered first, followed
        /// by the pattern-specific callbacks.
        ///
        /// Implicitly also triggers the init() member function to perform
        /// pattern-specific data management (i.e., setup) tasks.
        ///
        /// @param cbmanagers
        ///     Map of CallbackManager objects with which the callbacks should
        ///     be registered
        ///
        virtual void
        registerCallbacks(CallbackManagerMap& cbmanagers);

        /// @}
        /// @name Pattern information access
        /// @{

        /// @brief Get pattern identifier.
        ///
        /// Returns the unique numeric identifier associated with this pattern.
        ///
        /// @returns
        ///     %Pattern identifier
        ///
        virtual long
        getId() const = 0;

        /// @brief Get parent pattern identifier.
        ///
        /// Returns the unique numeric identifier associated with the parent
        /// pattern in the pattern hierarchy.
        ///
        /// @returns
        ///     Parent pattern identifier or PAT_NONE for top-level patterns
        ///
        virtual long
        getParent() const;

        /// @brief Get pattern name.
        ///
        /// Returns the display name of the pattern.
        ///
        /// @returns
        ///     %Pattern display name
        ///
        virtual std::string
        getName() const = 0;

        /// @brief Get the pattern's unique name.
        ///
        /// Returns the @"unique name@" of the pattern which can be used to
        /// uniquely identify a particular pattern.
        ///
        /// @returns
        ///     %Pattern unique name
        ///
        virtual std::string
        getUniqueName() const = 0;

        /// @brief Get pattern description.
        ///
        /// Returns a brief description (approximately one line of text) that
        /// is embedded as pattern info in the generated analysis report file.
        ///
        /// @returns
        ///     %Pattern description
        ///
        virtual std::string
        getDescription() const = 0;

        /// @brief Get the pattern's unit of measurement.
        ///
        /// Returns the unit of measurement for this pattern.  It can be either
        /// "sec" (i.e., seconds), "occ" (i.e., occurrences) or "bytes".
        ///
        /// @returns
        ///     Unit of measurement
        ///
        virtual std::string
        getUnit() const = 0;

        /// @brief Get the pattern's storage mode.
        ///
        /// Returns the storage mode (inclusive vs. exclusive) used for this
        /// pattern.
        ///
        /// @returns
        ///     Storage mode
        ///
        virtual CubeMetricType
        getStorageMode() const = 0;

        /// @brief Get the pattern's storage data type.
        ///
        /// Returns the (Cube) string representation of the data type used by
        /// this pattern.
        ///
        /// @returns
        ///     Data type
        ///
        virtual std::string
        getDatatype() const = 0;

        /// @brief Get pattern documentation URL.
        ///
        /// Returns the URL where a more detailed documentation of the pattern
        /// can be found.  It is composed of a base URL and the unique name.
        ///
        /// @returns
        ///     %Pattern documentation URL
        ///
        std::string
        getDocumentationUrl() const;

        /// @brief Get pattern visibility.
        ///
        /// Returns the external visibility of the pattern.  This allows for
        /// @"internal@" patterns which can define callbacks and store data,
        /// but which are not written to the generated analysis report file.
        ///
        /// @returns
        ///     %Pattern visibility
        ///
        virtual bool
        isVisible() const;

        /// @}
        /// @name Metric value collation and writing
        /// @{

        /// @brief Collate and write metric values to analysis report.
        ///
        /// Generates the severity section of the CUBE report for the pattern.
        /// Therefore, it first collates the values from each process/thread
        /// on a @"master node@" (details depend on the parallel programming
        /// paradigm used) and then writes the gathered data to file.  This
        /// is done incrementally per call-tree node.
        ///
        /// @param data
        ///     Temporary data used for report writing
        /// @param rank
        ///     Global process rank (MPI rank or 0)
        /// @param trace
        ///     Local trace-data object
        ///
        virtual void
        collateSeverities(ReportData&              data,
                          int                      rank,
                          const pearl::LocalTrace& trace) = 0;

        /// @}


    protected:
        // --- Protected methods ---------------------------

        /// @name Pattern data management
        /// @{

        /// @brief Initialize local data.
        ///
        /// This method is implicitly called from registerCallbacks() and can
        /// be overwritten in derived classes to initialize pattern-local data.
        /// The default implementation is empty.
        ///
        virtual void
        init();

        /// @}
};
}    // namespace scout


#endif    // !SCOUT_PATTERN_H
