/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2014-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "ReportData.h"

using namespace std;
using namespace pearl;
using namespace scout;


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @brief Implementation of the ReportData struct.
 *
 *  This file provides the implementation of the member functions defined
 *  for the ReportData struct.
 **/
/*-------------------------------------------------------------------------*/

/// @todo Refactor to avoid resource leaks in case of exceptions.
ReportData::ReportData(const GlobalDefs& defs,
                       int               rank)
    : cb(nullptr),
      mCallpathUsage(nullptr),
      mCallpathUsageSize(0),
      mCallpathOrder(nullptr),
      mRecvCounts(nullptr),
      mRecvDispls(nullptr),
      mEqualThreads(true)
{
    // These variables are implicitly shared!
    static uint8_t*  callpathUsage = nullptr;
    static uint32_t* callpathOrder = nullptr;

    // Initialize collation arrays
    const size_t numCallpaths = defs.numCallpaths();

    mCallpathUsageSize = (numCallpaths + 7) / 8;
    #pragma omp master
    {
        // Allocate callpath usage bit field
        callpathUsage = new uint8_t[mCallpathUsageSize];

        // Allocate preferred callpath order vector
        callpathOrder = new uint32_t[numCallpaths];

        // Allocate and initialize global collation data structures
        // (only relevant on rank 0)
        // FIXME: deal with non-CPU locations / non-process groups
        const uint32_t numRanks   = defs.numLocationGroups();
        const uint32_t numThreads = defs.getLocationGroup(rank).numLocations();
        if (rank == 0)
        {
            // Gatherv receive counts and displacements
            mRecvCounts = new int[numRanks];
            mRecvDispls = new int[numRanks];

            // Determine whether all processes used the same number of threads
            // and initialize gatherv arrays
            int displacement = 0;
            for (uint32_t i = 0; i < numRanks; ++i)
            {
                const LocationGroup& lgroup  = defs.getLocationGroup(i);
                const uint32_t       threads = lgroup.numLocations();

                mEqualThreads = mEqualThreads && (threads == numThreads);

                mRecvCounts[i] = threads;
                mRecvDispls[i] = displacement;

                displacement += threads;
            }
        }
        else
        {
            // Determine whether all processes used the same number of threads
            for (uint32_t i = 0; i < numRanks; ++i)
            {
                const LocationGroup& lgroup  = defs.getLocationGroup(i);
                const uint32_t       threads = lgroup.numLocations();

                mEqualThreads = mEqualThreads && (threads == numThreads);
            }
        }
    }

    // Share process-local arrays across all threads
    #pragma omp barrier
    mCallpathUsage = callpathUsage;
    mCallpathOrder = callpathOrder;
}


ReportData::~ReportData()
{
    #pragma omp master
    {
        delete[] mCallpathUsage;
        delete[] mCallpathOrder;
        delete[] mRecvCounts;
        delete[] mRecvDispls;
    }
}
