/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @brief Declaration of the class template TypedPattern.
 *
 *  This header file provides the declaration of the class template
 *  TypedPattern, which serves as an abstract base class for performance
 *  properties of a particular data type.
 **/
/*-------------------------------------------------------------------------*/


#ifndef SCOUT_TYPEDPATTERN_H
#define SCOUT_TYPEDPATTERN_H


#include <cstdint>
#include <map>

#include "Pattern.h"

#if HAVE(SCALASCA_DEBUG)
    #include <UTILS_Debug.h>
#endif    // HAVE(SCALASCA_DEBUG)


// --- Forward declarations -------------------------------------------------

namespace pearl
{
class Callpath;
}    // namespace pearl


// --- Helper macros --------------------------------------------------------

#if !HAVE(SCALASCA_DEBUG)
    /// @brief Increment metric value
    ///
    /// Increments the metric value for the given @p callpath by @p delta.
    /// In debug mode, also print a corresponding debug message if the
    /// runtime debug level includes @p mask.
    ///
    /// @param mask
    ///     Debug level bit mask
    /// @param callpath
    ///     Pointer to target callpath
    /// @param delta
    ///     Value increment
    ///
    #define INC_METRIC_VALUE(mask, callpath, delta) \
        incrementValue(callpath, delta)

    /// @brief Decrement metric value
    ///
    /// Decrements the metric value for the given @p callpath by @p delta.
    /// In debug mode, also print a corresponding debug message if the
    /// runtime debug level includes @p mask.
    ///
    /// @param mask
    ///     Debug level bit mask
    /// @param callpath
    ///     Pointer to target callpath
    /// @param delta
    ///     Value decrement
    ///
    #define DEC_METRIC_VALUE(mask, callpath, delta) \
        decrementValue(callpath, delta)
#else    // HAVE(SCALASCA_DEBUG)
    #define INC_METRIC_VALUE(mask, callpath, delta) \
        incrementValue(callpath,                    \
                       delta,                       \
                       mask,                        \
                       __FILE__,                    \
                       __LINE__,                    \
                       UTILS_FUNCTION_NAME)

    #define DEC_METRIC_VALUE(mask, callpath, delta) \
        decrementValue(callpath,                    \
                       delta,                       \
                       mask,                        \
                       __FILE__,                    \
                       __LINE__,                    \
                       UTILS_FUNCTION_NAME)
#endif    // HAVE(SCALASCA_DEBUG)


namespace scout
{
/*-------------------------------------------------------------------------*/
/**
 *  @class TypedPattern
 *  @brief Abstract base class template for properties of a particular data
 *         type.
 *
 *  The class template TypedPattern serves as an abstract base class for all
 *  performance properties calculated by SCOUT using a particular data type.
 *  It provides both data storage and collation/writing of metric values.
 *  At present, the only supported data types are `double` and `uint64_t`.
 *
 *  @tparam dataT
 *      Data type of metric values
 **/
/*-------------------------------------------------------------------------*/

template< typename dataT >
class TypedPattern
    : public Pattern
{
    public:
        // --- Public methods ------------------------------

        /// @name Pattern information access
        /// @{

        /// @copydoc scout::Pattern::getDatatype() const
        ///
        std::string
        getDatatype() const override;

        /// @name Metric value collation and writing
        /// @{

        /// @copydoc scout::Pattern::collateSeverities(ReportData&,int,const pearl::LocalTrace&)
        ///
        void
        collateSeverities(ReportData&              data,
                          int                      rank,
                          const pearl::LocalTrace& trace) override;

        /// @}


    protected:
        // --- Protected methods ---------------------------

        /// @name Metric value modification
        /// @{

        /// @brief Increment metric value.
        ///
        /// Increments the metric value for the given @p callpath by @p delta
        /// and returns the value of the updated map element.
        ///
        /// @attention
        ///     Avoid calling this member function directly.  Instead, use the
        ///     @ref INC_METRIC_VALUE helper macro which will select the
        ///     approproate overload depending on whether debugging code is
        ///     enabled or not.
        ///
        /// @param callpath
        ///     Pointer to target callpath
        /// @param delta
        ///     Value increment
        /// @returns
        ///     Value of the updated element
        ///
        dataT
        incrementValue(pearl::Callpath* callpath,
                       dataT            delta);

        /// @brief Increment metric value (debug overload).
        ///
        /// Increments the metric value for the given @p callpath by @p delta
        /// and returns the value of the updated map element.  In addition,
        /// a debug message including @p file, @p line, and @p function is
        /// printed if the runtime debug level includes the given @p mask.
        ///
        /// @attention
        ///     Avoid calling this member function directly.  Instead, use the
        ///     @ref INC_METRIC_VALUE helper macro which will select the
        ///     approproate overload depending on whether debugging code is
        ///     enabled or not.
        ///
        /// @param callpath
        ///     Pointer to target callpath
        /// @param delta
        ///     Value increment
        /// @param mask
        ///     Debug level bit mask
        /// @param file
        ///     Source file name
        /// @param line
        ///     Source file line
        /// @param function
        ///     Function name
        /// @returns
        ///     Value of the updated element
        ///
        dataT
        incrementValue(pearl::Callpath* callpath,
                       dataT            delta,
                       std::uint64_t    mask,
                       const char*      file,
                       std::uint64_t    line,
                       const char*      function);

        /// @brief Decrement metric value.
        ///
        /// Decrements the metric value for the given @p callpath by @p delta
        /// and returns the value of the updated map element.
        ///
        /// @attention
        ///     Avoid calling this member function directly.  Instead, use the
        ///     @ref DEC_METRIC_VALUE helper macro which will select the
        ///     approproate overload depending on whether debugging code is
        ///     enabled or not.
        ///
        /// @param callpath
        ///     Pointer to target callpath
        /// @param delta
        ///     Value decrement
        /// @returns
        ///     Value of the updated element
        ///
        dataT
        decrementValue(pearl::Callpath* callpath,
                       dataT            delta);

        /// @brief Decrement metric value (debug overload).
        ///
        /// Decrements the metric value for the given @p callpath by @p delta
        /// and returns the value of the updated map element.  In addition,
        /// a debug message including @p file, @p line, and @p function is
        /// printed if the runtime debug level includes the given @p mask.
        ///
        /// @attention
        ///     Avoid calling this member function directly.  Instead, use the
        ///     @ref DEC_METRIC_VALUE helper macro which will select the
        ///     approproate overload depending on whether debugging code is
        ///     enabled or not.
        ///
        /// @param callpath
        ///     Pointer to target callpath
        /// @param delta
        ///     Value decrement
        /// @param mask
        ///     Debug level bit mask
        /// @param file
        ///     Source file name
        /// @param line
        ///     Source file line
        /// @param function
        ///     Function name
        /// @returns
        ///     Value of the updated element
        ///
        dataT
        decrementValue(pearl::Callpath* callpath,
                       dataT            delta,
                       std::uint64_t    mask,
                       const char*      file,
                       std::uint64_t    line,
                       const char*      function);

        /// @}


        // --- Data members --------------------------------

        /// Local severity map.  Stores a (sparse) callpath @f$\times@f$
        /// value matrix.
        std::map< pearl::Callpath*, dataT > m_severity;
};
}    // namespace scout


#endif    // !SCOUT_TYPEDPATTERN_H
