/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2014                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#ifndef SCOUT_WAITSTATESTATS_H
    #error "Do not include this file!  Use 'WaitstateStats.h' instead."
#endif


#include <algorithm>
#include <tuple>


namespace scout
{
// --------------------------------------------------------------------------
//
//  TopMostSevere & related functions
//
// --------------------------------------------------------------------------

// --- Constructors & destructor --------------------------------------------

inline
TopMostSevere::TopMostSevere()
    : TopMostSevere(0.0, 0.0, 0.0, 0, 0)
{
}


inline
TopMostSevere::TopMostSevere(const pearl::timestamp_t      idletime_,
                             const pearl::timestamp_t      entertime_,
                             const pearl::timestamp_t      exittime_,
                             const pearl::Callpath::IdType cnode_,
                             const int                     rank_)
    : idletime(idletime_),
      entertime(entertime_),
      exittime(exittime_),
      cnode(cnode_),
      rank(rank_)
{
}


// --- Comparison operators -------------------------------------------------

inline bool
operator<(const TopMostSevere& lhs,
          const TopMostSevere& rhs)
{
    // exittime = entertime + idletime, thus irrelevant for the comparison
    return (std::tie(lhs.idletime, lhs.cnode, lhs.rank, lhs.entertime)
            < std::tie(rhs.idletime, lhs.cnode, lhs.rank, lhs.entertime));
}


// --------------------------------------------------------------------------
//
//  TopMostSevereMaxSum & related functions
//
// --------------------------------------------------------------------------

// --- Constructors & destructor --------------------------------------------

inline
TopMostSevereMaxSum::TopMostSevereMaxSum()
    : idletime_max(0.0)
{
}


inline
TopMostSevereMaxSum::TopMostSevereMaxSum(const pearl::timestamp_t      idletime_,
                                         const pearl::timestamp_t      entertime_,
                                         const pearl::timestamp_t      exittime_,
                                         const pearl::Callpath::IdType cnode_,
                                         const int                     rank_)
    : TopMostSevere(idletime_,
                    entertime_,
                    exittime_,
                    cnode_,
                    rank_),
      idletime_max(idletime_)
{
}


// --- Reduction operations -------------------------------------------------

inline TopMostSevereMaxSum
merge(const TopMostSevereMaxSum& lhs,
      const TopMostSevereMaxSum& rhs)
{
    TopMostSevereMaxSum result(lhs);

    result.idletime += rhs.idletime;
    result.entertime = std::min(result.entertime, rhs.entertime);
    result.exittime  = std::max(result.exittime, rhs.exittime);

    if (std::tie(result.idletime_max, result.rank, result.cnode)
        < std::tie(rhs.idletime_max, rhs.rank, rhs.cnode))
    {
        result.idletime_max = rhs.idletime_max;
        result.cnode        = rhs.cnode;
        result.rank         = rhs.rank;
    }

    return result;
}
}    // namespace scout
