/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2014                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


/*-------------------------------------------------------------------------*/
/**
 *  @file
 *  @brief Declaration of helper classes to track waitstate instance
 *         statistics information.
 *
 *  This header file provides the declaration of various helper classes used
 *  to track and process information about detected waitstate instances.
 */
/*-------------------------------------------------------------------------*/


#ifndef SCOUT_WAITSTATESTATS_H
#define SCOUT_WAITSTATESTATS_H


#include <pearl/Callpath.h>
#include <pearl/pearl_types.h>


// --- Preprocessor defines -------------------------------------------------

/// Maximum number of most-severe instances to be tracked
#define TOP_SEVERE_INSTANCES    5


namespace scout
{
/*-------------------------------------------------------------------------*/
/**
 *  @struct TopMostSevere
 *  @brief  Helper struct used to track information about detected waitstate
 *          instances.
 *
 *  Instances of the TopMostSevere struct store information about a single
 *  waitstate instance detected by the trace analysis.
 */
/*-------------------------------------------------------------------------*/

struct TopMostSevere
{
    // --- Public methods ----------------------------------

    /// @name Constructors & destructor
    /// @{

    /// @brief Default constructor.
    ///
    /// Initializes all member variables to zero.
    ///
    TopMostSevere();

    /// @brief Constructor.
    ///
    /// Initializes all member variables with the given values.
    ///
    /// @param idletime_
    ///     Idle (i.e., waiting) time
    /// @param entertime_
    ///     Begin timestamp
    /// @param exittime_
    ///     End timestamp
    /// @param cnode_
    ///     Callpath identifier
    /// @param rank_
    ///     Process rank
    ///
    TopMostSevere(pearl::timestamp_t      idletime_,
                  pearl::timestamp_t      entertime_,
                  pearl::timestamp_t      exittime_,
                  pearl::Callpath::IdType cnode_,
                  int                     rank_);

    /// @}


    // --- Data members ------------------------------------

    /// Total idle (i.e., waiting) time spent in the waitstate
    pearl::timestamp_t idletime;

    /// Begin timestamp of the waitstate situation
    pearl::timestamp_t entertime;

    /// End timestamp of the waitstate situation
    pearl::timestamp_t exittime;

    /// Callpath ID from the process/thread exhibiting the maximum
    /// idle time
    pearl::Callpath::IdType cnode;

    /// Process rank exhibiting the maximum idle time
    int rank;
};


// --- Related functions ----------------------------------------------------

/// @name Comparison operators
/// @{

/// @brief Less-than comparison operator for TopMostSevere instances.
///
/// Returns whether the operand @p lhs compares "less than" the operand
/// @p rhs, imposing a strict weak ordering considering (in order of
/// significance) idle times, callpath IDs, ranks, and enter time.
///
/// @param lhs
///     Left-hand operand
/// @param rhs
///     Right-hand operand
/// @returns
///     True if @p lhs compares "less than" @p rhs, false otherwise
///
inline bool
operator<(const TopMostSevere& lhs,
          const TopMostSevere& rhs);

/// @}


/*-------------------------------------------------------------------------*/
/**
 *  @struct TopMostSevereMaxSum
 *  @brief  Helper struct used to merge local waitstate instance information
 *          for collective operations.
 *
 *  Instances of the TopMostSevereMaxSum helper struct are used in
 *  reductions to merge local waitstate instance information into global
 *  instance information for collective operations.
 */
/*-------------------------------------------------------------------------*/

struct TopMostSevereMaxSum
    : TopMostSevere
{
    // --- Public methods ----------------------------------

    /// @name Constructors & destructor
    /// @{

    /// @brief Default constructor.
    ///
    /// Initializes all member variables to zero.
    ///
    TopMostSevereMaxSum();

    /// @brief Constructor
    ///
    /// Initializes all member variables with the given values.
    ///
    /// @param idletime_
    ///     Idle (i.e., waiting) time
    /// @param entertime_
    ///     Begin timestamp
    /// @param exittime_
    ///     End timestamp
    /// @param cnode_
    ///     Callpath identifier
    /// @param rank_
    ///     Process rank
    ///
    TopMostSevereMaxSum(pearl::timestamp_t      idletime_,
                        pearl::timestamp_t      entertime_,
                        pearl::timestamp_t      exittime_,
                        pearl::Callpath::IdType cnode_,
                        int                     rank_);

    /// @}


    // --- Data members ------------------------------------

    /// Maximum local idle (i.e., waiting) time in the collective waitstate
    pearl::timestamp_t idletime_max;
};


// --- Related functions ----------------------------------------------------

/// @name Reduction operations
/// @{

/// @brief   Merges partial collective waitstate information.
/// @relates TopMostSevereMaxSum
///
/// Merges the partial collective waitstate information provided in @p lhs
/// and @p rhs.  The returned result includes the sum of both idle times,
/// the minimum begin and maximum end timestamp, as well as the maximum local
/// idle time with the corresponding process rank and callpath ID (with ties
/// broken in this order).  This operation is commutative.
///
/// @param lhs
///     Left-hand operand
/// @param rhs
///     Right-hand operand
/// @returns
///     Merged collective waitstate information
///
TopMostSevereMaxSum
merge(const TopMostSevereMaxSum& lhs,
      const TopMostSevereMaxSum& rhs);

/// @}
}    // namespace scout


// --- Inline methods & related functions -----------------------------------

#include "WaitstateStats-inl.h"


#endif    // !SCOUT_WAITSTATESTATS_H
