/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "Pattern.h"

#include <algorithm>
#include <cassert>
#include <cctype>
#include <functional>
#include <sstream>
#include <stdexcept>

#include "Helper.h"
#include "IndentStream.h"

using namespace std;


// --- Type definitions -----------------------------------------------------

/// @brief Compound type for type-specific pattern information.
///
/// The TypeInfo struct combines all information related to a specific
/// pattern type, such as its type name, its base class, and potential
/// conditional compilation guards.
///
struct Pattern::TypeInfo
{
    /// Pattern type name
    const char* const mType;

    /// Pattern base class name.  Can either be a concrete base class or
    /// a template instantiation, with `@DATATYPE@` to be replaced by the
    /// patterns data type.
    const char* const mBaseClassTempl;

    /// Beginning conditional compilation guard if applicable,
    /// `nullptr` otherwise
    const char* const mGuardBegin;

    /// Closing conditional compilation guard if applicable,
    /// `nullptr` otherwise
    const char* const mGuardEnd;
};


// --- Utility functions ----------------------------------------------------

namespace
{
struct fo_tolower
{
    int
    operator()(int x) const
    {
        return std::tolower(x);
    }
};


string
lowercase(const string& str)
{
    string result(str);

    transform(str.begin(), str.end(), result.begin(), fo_tolower());

    return result;
}


string
wrappedCharConstant(string                  text,
                    const string::size_type width,
                    const string::size_type indent)
{
    string result("\"");

    while (text.length() > width)
    {
        string::size_type pos = text.rfind(' ', width);

        result += text.substr(0, pos);
        if (pos != string::npos)
        {
            result += "\"\n";
            result += string(indent, ' ') + '"';
        }
        text.erase(0, pos);
    }

    return (result + text + '"');
}


bool
stringStartsWith(const string& str,
                 const string& prefix)
{
    return (str.compare(0, prefix.length(), prefix) == 0);
}


string
getCallbackName(const string& group,
                const string& event)
{
    string result = event;
    if (!group.empty())
    {
        result.append("_").append(group);
    }

    return result;
}
}    // unnamed namespace


// --- Constructors & destructor --------------------------------------------

Pattern::Pattern(const string& identifier)
    : mIdentifier(identifier),
      mParent(nullptr),
      mTypeInfo(nullptr),
      mIsHidden(false),
      mProvidesDocs(true)
{
}


// --- Get pattern attributes -----------------------------------------------

string
Pattern::getIdentifier() const
{
    return mIdentifier;
}


string
Pattern::getClassName() const
{
    return mClassName;
}


bool
Pattern::providesCode() const
{
    return !(mIsHidden && mCallbacks.empty());
}


bool
Pattern::providesDocs() const
{
    return mProvidesDocs;
}


// --- Set pattern attributes -----------------------------------------------

void
Pattern::setDisplayName(const string& name)
{
    if (!mDisplayName.empty())
    {
        throw invalid_argument("Duplicate NAME definition.");
    }

    mDisplayName = name;
}


void
Pattern::setClassName(const string& name)
{
    if (!mClassName.empty())
    {
        throw invalid_argument("Duplicate CLASS definition.");
    }

    mClassName = name;
}


void
Pattern::setDocName(const string& name)
{
    if (!mDocName.empty())
    {
        throw invalid_argument("Duplicate DOCNAME definition.");
    }

    mDocName = name;
}


void
Pattern::setParent(Pattern* const parent)
{
    assert(parent != nullptr);

    if (mParent != nullptr)
    {
        throw invalid_argument("Duplicate PARENT definition.");
    }
    if (parent == this)
    {
        throw invalid_argument("Recursive PARENT definition: pattern references itself");
    }

    parent->mChildren.push_back(this);

    mParent = parent;
}


void
Pattern::setType(const string& type)
{
    if (mTypeInfo != nullptr)
    {
        throw invalid_argument("Duplicate TYPE definition.");
    }

    // *INDENT-OFF*
    static const TypeInfo typeInfo[] = {
        { "Generic",
          "TypedPattern< @DATATYPE@ >",
          nullptr,
          nullptr
        },
        { "MPI",
          "TypedPattern< @DATATYPE@ >",
          "#if defined(_MPI)\n",
          "#endif    // _MPI\n",
        },
        { "OMP",
          "TypedPattern< @DATATYPE@ >",
          "#if defined(_OPENMP)\n",
          "#endif    // _OPENMP\n",
        },
        { "PTHREAD",
          "TypedPattern< @DATATYPE@ >",
          "#if defined(_OPENMP)\n",
          "#endif    // _OPENMP\n",
        },
        { "DELAY_MPIDEP",
          "TypedPattern< @DATATYPE@ >",
          "#if defined(_MPI)\n",
          "#endif    // _MPI\n",
        },
        { "DELAY_LT_MPIDEP",
          "LongTermDelayPattern",
          "#if defined(_MPI)\n",
          "#endif    // _MPI\n",
        },
        { "DELAY_OMPDEP",
          "TypedPattern< @DATATYPE@ >",
          "#if defined(_OPENMP)\n",
          "#endif    // _OPENMP\n",
        },
        { "DELAY_LT_OMPDEP",
          "LongTermDelayPattern",
          "#if defined(_OPENMP)\n",
          "#endif    // _OPENMP\n",
        },

        { nullptr,
          nullptr,
          nullptr,
          nullptr
        }
    };
    // *INDENT-ON*

    // Search entry for type name
    for (mTypeInfo = typeInfo;
         mTypeInfo->mType && (type != mTypeInfo->mType);
         ++mTypeInfo)
    {
    }
    if (mTypeInfo == nullptr)
    {
        throw invalid_argument("Unknown pattern type: " + type);
    }
}


void
Pattern::setDelayKey(const string& key)
{
    if (!mDelayKey.empty())
    {
        throw invalid_argument("Duplicate DELAY_KEY definition.");
    }

    mDelayKey = key;
}


void
Pattern::setInfoString(const string& text)
{
    if (!mInfoString.empty())
    {
        throw invalid_argument("Duplicate INFO definition.");
    }

    mInfoString = text;
}


void
Pattern::setDescription(const string& text)
{
    if (!mDescription.empty())
    {
        throw invalid_argument("Duplicate DESCR definition.");
    }

    mDescription = text;
}


void
Pattern::setDiagnosis(const string& text)
{
    if (!mDiagnosis.empty())
    {
        throw invalid_argument("Duplicate DIAGNOSIS definition.");
    }

    mDiagnosis = text;
}


void
Pattern::setUnit(const string& unit)
{
    if (!mUnit.empty())
    {
        throw invalid_argument("Duplicate UNIT definition.");
    }
    if (  (unit != "sec")
       && (unit != "occ")
       && (unit != "bytes"))
    {
        throw invalid_argument("Unknown unit: " + unit);
    }

    mUnit = unit;
}


void
Pattern::setStorageMode(const string& mode)
{
    if (!mStorageMode.empty())
    {
        throw invalid_argument("Duplicate MODE definition.");
    }
    if (  (mode != "inclusive")
       && (mode != "exclusive"))
    {
        throw invalid_argument("Unknown storage mode: " + mode);
    }

    mStorageMode = mode;
}


void
Pattern::setConditionalExpr(const string& expr)
{
    if (!mConditionalExpr.empty())
    {
        throw invalid_argument("Duplicate CONDITION definition.");
    }

    mConditionalExpr = expr;
}


void
Pattern::setPrivateData(const string& code)
{
    if (!mPrivateData.empty())
    {
        throw invalid_argument("Duplicate DATA definition.");
    }

    mPrivateData = code;
}


void
Pattern::setInitCode(const string& code)
{
    if (!mInitCode.empty())
    {
        throw invalid_argument("Duplicate INIT definition.");
    }

    mInitCode = code;
}


void
Pattern::setCleanupCode(const string& code)
{
    if (!mCleanupCode.empty())
    {
        throw invalid_argument("Duplicate CLEANUP definition.");
    }

    mCleanupCode = code;
}


void
Pattern::setStaticInitCode(const string& code)
{
    if (!mStaticInitCode.empty())
    {
        throw invalid_argument("Duplicate STATICINIT definition.");
    }

    mStaticInitCode = code;
}


void
Pattern::hide()
{
    if (mIsHidden)
    {
        throw invalid_argument("Duplicate HIDDEN definition.");
    }

    mIsHidden = true;
}


void
Pattern::suppressDocs()
{
    if (!mProvidesDocs)
    {
        throw invalid_argument("Duplicate NODOCS definition.");
    }

    mProvidesDocs = false;
}


void
Pattern::addCallback(const std::string& group,
                     const string&      event,
                     const string&      code)
{
    const string cbname = getCallbackName(group, event);

    map< string, string >::iterator it = mCallbacks.lower_bound(cbname);
    if (  (it != mCallbacks.end())
       && !(mCallbacks.key_comp()(cbname, it->first)))
    {
        const string phase = group.empty() ? "main" : group;

        throw invalid_argument("Callback already defined: "
                               + event + " (" + phase + ')');
    }

    mCallbacks.insert(it, make_pair(cbname, code));
    mCallbackGroups[group].push_back(event);
}


// --- Finalize pattern definition ------------------------------------------

void
Pattern::finalize()
{
    // Apply defaults
    if (mTypeInfo == nullptr)
    {
        setType("Generic");
    }

    // Check for mandatory definitions
    if (mDisplayName.empty())
    {
        throw runtime_error("Missing NAME definition.");
    }
    if (mClassName.empty())
    {
        throw runtime_error("Missing CLASS definition.");
    }
    if (mInfoString.empty())
    {
        throw runtime_error("Missing INFO definition.");
    }
    if (mUnit.empty())
    {
        throw runtime_error("Missing UNIT definition.");
    }
    if (mStorageMode.empty())
    {
        throw runtime_error("Missing MODE definition.");
    }
    if (  isLongTermDelayPattern()
       && mDelayKey.empty())
    {
        throw runtime_error("Missing DELAY_KEY definition.");
    }

    // Apply fallbacks
    if (mDocName.empty())
    {
        mDocName = mDisplayName;
    }
    if (mDescription.empty())
    {
        mDescription = mInfoString;
    }

    // Resolve (optional) template type in base class name
    const string keyword = "@DATATYPE@";

    mBaseClassName = mTypeInfo->mBaseClassTempl;
    string::size_type pos = mBaseClassName.find(keyword);
    if (pos != string::npos)
    {
        const string datatype = (mUnit == "sec")
                                ? "double"
                                : "uint64_t";

        mBaseClassName.replace(pos, keyword.length(), datatype);
    }
}


// --- Generate code --------------------------------------------------------

void
Pattern::writeImplementation(FILE* const fp,
                             int&        indent) const
{
    assert(mTypeInfo != nullptr);

    static bool isFirst = true;
    if (isFirst)
    {
        isFirst = false;
    }
    else
    {
        IndentStream(fp, indent)
            << '\n'
            << '\n';
    }

    writePreprocGuardBegin(fp, indent);
    writeClassDeclaration(fp, indent);
    writeClassDefinition(fp, indent);
    writePreprocGuardEnd(fp, indent);
}


void
Pattern::writePreprocGuardBegin(FILE* const fp,
                                int&        indent) const
{
    if (mTypeInfo->mGuardBegin != nullptr)
    {
        IndentStream(fp, indent)
            << mTypeInfo->mGuardBegin;
        indent++;
    }
}


void
Pattern::writePreprocGuardEnd(FILE* const fp,
                              int&        indent) const
{
    if (mTypeInfo->mGuardEnd != nullptr)
    {
        indent--;
        IndentStream(fp, indent)
            << mTypeInfo->mGuardEnd;
    }
}


void
Pattern::writeConditionalBegin(FILE* const fp,
                               int&        indent) const
{
    if (!mConditionalExpr.empty())
    {
        IndentStream(fp, indent)
            << "if (" << mConditionalExpr << ")\n"
            << "{\n";
        indent++;
    }
}


void
Pattern::writeConditionalEnd(FILE* const fp,
                             int&        indent) const
{
    if (!mConditionalExpr.empty())
    {
        indent--;
        IndentStream(fp, indent)
            << "}\n";
    }
}


// --- Generate documentation -----------------------------------------------

void
Pattern::writeDocumentation(FILE* const       fp,
                            const bool        isFirst,
                            const PatternMap& patternMap)
{
    processReferences("DESCR", mDescription, patternMap);
    processReferences("DIAGNOSIS", mDiagnosis, patternMap);

    fprintf(fp, "\n");
    if (!isFirst)
    {
        fprintf(fp, "<hr width=\"75%%\" align=\"center\">\n\n");
    }
    fprintf(fp, "<a name=\"%s\"><h3>%s</h3></a>\n",
            lowercase(mIdentifier).c_str(), mDocName.c_str());
    if (mIsHidden)
    {
        fprintf(fp, "(only available after "
                    "<a href=\"#remapping_info\">remapping</a>)\n");
    }
    fprintf(fp, "<dl>\n");

    /* Description */
    fprintf(fp, "<dt><b>Description:</b></dt>\n");
    if (mDescription.empty())
    {
        fprintf(fp, "<dd>No description available.</dd>\n");
    }
    else
    {
        fprintf(fp, "<dd>\n%s\n</dd>\n", mDescription.c_str());
    }

    /* Unit */
    fprintf(fp, "<dt><b>Unit:</b></dt>\n");
    if (mUnit == "sec")
    {
        fprintf(fp, "<dd>Seconds</dd>\n");
    }
    else if (mUnit == "occ")
    {
        fprintf(fp, "<dd>Counts</dd>\n");
    }
    else if (mUnit == "bytes")
    {
        fprintf(fp, "<dd>Bytes</dd>\n");
    }

    /* Diagnosis */
    if (!mDiagnosis.empty())
    {
        fprintf(fp, "<dt><b>Diagnosis:</b></dt>\n");
        fprintf(fp, "<dd>\n%s\n</dd>\n", mDiagnosis.c_str());
    }

    /* Parent */
    string parentName;
    if (mParent == nullptr)
    {
        parentName = "None";
    }
    else
    {
        parentName = "<a href=\"#" + lowercase(mParent->mIdentifier) + "\">"
                     + mParent->mDocName + "</a>";
    }
    fprintf(fp, "<dt><b>Parent metric:</b></dt>\n"
                "<dd>%s</dd>\n", parentName.c_str());

    /* Children */
    fprintf(fp, "<dt><b>Sub-metrics:</b></dt>\n<dd>\n");
    if (mChildren.empty())
    {
        fprintf(fp, "    None");
    }
    else
    {
        for (vector< Pattern* >::const_iterator child = mChildren.begin();
             child != mChildren.end();
             ++child)
        {
            if (child != mChildren.begin())
            {
                fprintf(fp, "<br/>\n");
            }
            fprintf(fp, "    <a href=\"#%s\">%s</a>",
                    lowercase((*child)->mIdentifier).c_str(),
                    (*child)->mDocName.c_str());
        }
    }
    fprintf(fp, "\n</dd>\n");

    fprintf(fp, "</dl>\n");
}


// --- Get pattern attributes (private) -------------------------------------

bool
Pattern::isLongTermDelayPattern() const
{
    assert(mTypeInfo != nullptr);

    return stringStartsWith(mTypeInfo->mType, "DELAY_LT_");
}


// --- Generate code (private) ----------------------------------------------

void
Pattern::writeClassDeclaration(FILE* const fp,
                               int&        indent) const
{
    assert(!mBaseClassName.empty());

    IndentStream(fp, indent)
        << separatorComment(indent, C_STYLE) << '\n'
        << "/**\n"
        << " *  @class " << mClassName << '\n'
        << " *  @brief " << mDocName << " metric.\n"
        << " *\n"
        << " *  Implements the \"" << mDocName << "\" metric.\n"
        << " */\n"
        << separatorComment(indent, C_STYLE) << '\n'
        << '\n';

    IndentStream(fp, indent)
        << "class " << mClassName << '\n'
        << "    : public " << mBaseClassName << '\n'
        << "{\n";
    indent++;

    writeClassDeclPublic(fp, indent);
    writeClassDeclProtected(fp, indent);
    writeClassDeclPrivate(fp, indent);

    indent--;
    IndentStream(fp, indent)
        << "};\n";
}


void
Pattern::writeClassDeclPublic(FILE* const fp,
                              int&        indent) const
{
    IndentStream(fp, indent)
        << "public:\n";
    indent++;

    writeClassDeclCtorDtor(fp, indent);
    writeClassDeclRegistration(fp, indent);
    writeClassDeclInfoAccess(fp, indent);

    indent--;
}


void
Pattern::writeClassDeclProtected(FILE* const fp,
                                 int&        indent) const
{
    assert(!mBaseClassName.empty());

    if (mInitCode.empty())
    {
        return;
    }

    IndentStream(fp, indent)
        << '\n'
        << '\n'
        << "protected:\n";
    indent++;

    IndentStream(fp, indent)
        << "/// @name Pattern data management\n"
        << "/// @{\n"
        << '\n'
        << "/// @brief Initialize local data.\n"
        << "///\n"
        << "/// This method is implicitly called from registerCallbacks() and\n"
        << "/// initializes pattern-local data.\n"
        << "///\n"
        << "void\n"
        << "init() override\n"
        << "{\n";
    indent++;

    IndentStream(fp, indent)
        << mBaseClassName << "::init();\n"
        << '\n'
        << mInitCode << '\n';

    indent--;
    IndentStream(fp, indent)
        << "}\n"
        << '\n'
        << "/// @}\n";

    indent--;
}


void
Pattern::writeClassDeclPrivate(FILE* const fp,
                               int&        indent) const
{
    if (  mPrivateData.empty()
       && mCallbacks.empty())
    {
        return;
    }

    IndentStream(fp, indent)
        << '\n'
        << '\n'
        << "private:\n";
    indent++;

    writeClassDeclPrivateData(fp, indent);

    if (  !mPrivateData.empty()
       && !mCallbacks.empty())
    {
        IndentStream(fp, indent)
            << '\n';
    }

    writeClassDeclCallbacks(fp, indent);

    indent--;
}


void
Pattern::writeClassDeclCtorDtor(FILE* const fp,
                                int&        indent) const
{
    if (  mCleanupCode.empty()
       && !isLongTermDelayPattern())
    {
        return;
    }

    IndentStream(fp, indent)
        << "/// @name Constructors & destructor\n"
        << "/// @{\n"
        << '\n';

    writeClassDeclCtor(fp, indent);
    writeClassDeclDtor(fp, indent);

    IndentStream(fp, indent)
        << "/// @}\n";
}


void
Pattern::writeClassDeclRegistration(FILE* const fp,
                                    int&        indent) const
{
    if (mCallbacks.empty())
    {
        return;
    }

    IndentStream(fp, indent)
        << "/// @name Callback registration\n"
        << "/// @{\n"
        << '\n'
        << "/// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)\n"
        << "///\n"
        << "void\n"
        << "registerCallbacks(CallbackManagerMap& cbmanagers) override;\n"
        << '\n'
        << "/// @}\n";
}


void
Pattern::writeClassDeclInfoAccess(FILE* const fp,
                                  int&        indent) const
{
    IndentStream(fp, indent)
        << "/// @name Pattern information access\n"
        << "/// @{\n"
        << '\n';

    writeClassDeclGetId(fp, indent);
    writeClassDeclGetParent(fp, indent);
    writeClassDeclGetName(fp, indent);
    writeClassDeclGetUniqueName(fp, indent);
    writeClassDeclGetDescription(fp, indent);
    writeClassDeclGetUnit(fp, indent);
    writeClassDeclGetStorageMode(fp, indent);
    writeClassDeclIsVisible(fp, indent);

    IndentStream(fp, indent)
        << "/// @}\n";
}


void
Pattern::writeClassDeclCallbacks(FILE* const fp,
                                 int&        indent) const
{
    if (mCallbacks.empty())
    {
        return;
    }

    for (const auto& group : mCallbackGroups)
    {
        const string phase = group.first.empty() ? "main" : group.first;

        IndentStream(fp, indent)
            << "/// @name Callback functions (" << phase << ")\n"
            << "/// @{\n"
            << '\n';

        for (const auto& callback : group.second)
        {
            // Parameter indent: callback name + "_cb("
            const string cbName = getCallbackName(group.first, callback);
            const string paramIndent(cbName.length() + 4, ' ');

            IndentStream(fp, indent)
                << "/// @brief Callback function for " << callback << " events.\n"
                << "///\n"
                << "/// Callback function triggered by " << callback << " events\n"
                << "/// during the @\"" << phase << "@\" replay pass.\n"
                << "///\n"
                << "/// @param cbmanager\n"
                << "///     Callback manager instance\n"
                << "/// @param user_event\n"
                << "///     User event triggering this callback (if applicable)\n"
                << "/// @param event\n"
                << "///     Currently processed event\n"
                << "/// @param cdata\n"
                << "///     Callback data object\n"
                << "///\n"
                << "void\n"
                << lowercase(cbName) << "_cb("
                << "const CallbackManager& cbmanager,\n"
                << paramIndent
                << "int                    user_event,\n"
                << paramIndent
                << "const Event&           event,\n"
                << paramIndent
                << "CallbackData*          cdata);\n"
                << '\n';
        }

        IndentStream(fp, indent)
            << "/// @}\n";
    }
}


void
Pattern::writeClassDeclCtor(FILE* const fp,
                            int&        indent) const
{
    assert(!mBaseClassName.empty());

    if (!isLongTermDelayPattern())
    {
        return;
    }

    IndentStream(fp, indent)
        << "/// @brief Constructor.\n"
        << "///\n"
        << "/// Creates a new instance.\n"
        << "///\n"
        << mClassName << "()\n";
    indent++;

    // Parameter indent: "  " + base class name + "("
    string paramIndent(mBaseClassName.length() + 3, ' ');
    IndentStream(fp, indent)
        << ": " << mBaseClassName << "(BUFFER_RCOST_" << mDelayKey << ",\n"
        << paramIndent << "KEY_RCOST_" << mDelayKey << ",\n"
        << paramIndent << "DELAY_" << mDelayKey << ")\n";
    indent--;

    IndentStream(fp, indent)
        << "{\n"
        << "}\n"
        << '\n';
}


void
Pattern::writeClassDeclDtor(FILE* const fp,
                            int&        indent) const
{
    if (mCleanupCode.empty())
    {
        return;
    }

    IndentStream(fp, indent)
        << "/// @copydoc scout::Pattern::~Pattern()\n"
        << "///\n"
        << '~' << mClassName << "() override\n"
        << "{\n";
    indent++;

    IndentStream(fp, indent)
        << mCleanupCode << '\n';

    indent--;
    IndentStream(fp, indent)
        << "}\n"
        << '\n';
}


void
Pattern::writeClassDeclGetId(FILE* const fp,
                             int&        indent) const
{
    IndentStream(fp, indent)
        << "/// @copydoc scout::Pattern::getId()\n"
        << "///\n"
        << "long\n"
        << "getId() const override\n"
        << "{\n";
    indent++;

    IndentStream(fp, indent)
        << "return PAT_" << mIdentifier << ";\n";

    indent--;
    IndentStream(fp, indent)
        << "}\n"
        << '\n';
}


void
Pattern::writeClassDeclGetParent(FILE* const fp,
                                 int&        indent) const
{
    IndentStream(fp, indent)
        << "/// @copydoc scout::Pattern::getParent()\n"
        << "///\n"
        << "long\n"
        << "getParent() const override\n"
        << "{\n";
    indent++;

    const string parentId = (mParent == nullptr) ? "NONE" : mParent->mIdentifier;
    IndentStream(fp, indent)
        << "return PAT_" << parentId << ";\n";

    indent--;
    IndentStream(fp, indent)
        << "}\n"
        << '\n';
}


void
Pattern::writeClassDeclGetName(FILE* const fp,
                               int&        indent) const
{
    IndentStream(fp, indent)
        << "/// @copydoc scout::Pattern::getName()\n"
        << "///\n"
        << "string\n"
        << "getName() const override\n"
        << "{\n";
    indent++;

    IndentStream(fp, indent)
        << "return " << wrappedCharConstant(mDisplayName, 52, 7) << ";\n";

    indent--;
    IndentStream(fp, indent)
        << "}\n"
        << '\n';
}


void
Pattern::writeClassDeclGetUniqueName(FILE* const fp,
                                     int&        indent) const
{
    IndentStream(fp, indent)
        << "/// @copydoc scout::Pattern::getUniqueName()\n"
        << "///\n"
        << "string\n"
        << "getUniqueName() const override\n"
        << "{\n";
    indent++;

    IndentStream(fp, indent)
        << "return \"" << lowercase(mIdentifier) << "\";\n";

    indent--;
    IndentStream(fp, indent)
        << "}\n"
        << '\n';
}


void
Pattern::writeClassDeclGetDescription(FILE* const fp,
                                      int&        indent) const
{
    IndentStream(fp, indent)
        << "/// @copydoc scout::Pattern::getDescription()\n"
        << "///\n"
        << "string\n"
        << "getDescription() const override\n"
        << "{\n";
    indent++;

    IndentStream(fp, indent)
        << "return " << wrappedCharConstant(mInfoString, 52, 7) << ";\n";

    indent--;
    IndentStream(fp, indent)
        << "}\n"
        << '\n';
}


void
Pattern::writeClassDeclGetUnit(FILE* const fp,
                               int&        indent) const
{
    IndentStream(fp, indent)
        << "/// @copydoc scout::Pattern::getUnit()\n"
        << "///\n"
        << "string\n"
        << "getUnit() const override\n"
        << "{\n";
    indent++;

    IndentStream(fp, indent)
        << "return \"" << mUnit << "\";\n";

    indent--;
    IndentStream(fp, indent)
        << "}\n"
        << '\n';
}


void
Pattern::writeClassDeclGetStorageMode(FILE* const fp,
                                      int&        indent) const
{
    IndentStream(fp, indent)
        << "/// @copydoc scout::Pattern::getStorageMode()\n"
        << "///\n"
        << "CubeMetricType\n"
        << "getStorageMode() const override\n"
        << "{\n";
    indent++;

    string mode;
    if (mStorageMode == "inclusive")
    {
        mode = "CUBE_METRIC_INCLUSIVE";
    }
    else if (mStorageMode == "exclusive")
    {
        mode = "CUBE_METRIC_EXCLUSIVE";
    }
    IndentStream(fp, indent)
        << "return " << mode << ";\n";

    indent--;
    IndentStream(fp, indent)
        << "}\n"
        << '\n';
}


void
Pattern::writeClassDeclIsVisible(FILE* const fp,
                                 int&        indent) const
{
    if (!mIsHidden)
    {
        return;
    }

    IndentStream(fp, indent)
        << "/// @copydoc scout::Pattern::isVisible()\n"
        << "///\n"
        << "bool\n"
        << "isVisible() const override\n"
        << "{\n";
    indent++;

    IndentStream(fp, indent)
        << "return false;\n";

    indent--;
    IndentStream(fp, indent)
        << "}\n"
        << '\n';
}


void
Pattern::writeClassDeclPrivateData(FILE* const fp,
                                   int&        indent) const
{
    if (mPrivateData.empty())
    {
        return;
    }

    IndentStream(fp, indent)
        << mPrivateData << '\n';
}


void
Pattern::writeClassDefinition(FILE* const fp,
                              int&        indent) const
{
    writeClassDefStaticInit(fp, indent);
    writeClassDefRegistration(fp, indent);
    writeClassDefCallbacks(fp, indent);
}


void
Pattern::writeClassDefStaticInit(FILE* const fp,
                                 int&        indent) const
{
    if (mStaticInitCode.empty())
    {
        return;
    }

    IndentStream(fp, indent)
        << '\n'
        << '\n'
        << separatorComment(indent, CPP_STYLE, "Static class data") << '\n'
        << '\n'
        << mStaticInitCode << '\n';
}


void
Pattern::writeClassDefRegistration(FILE* const fp,
                                   int&        indent) const
{
    assert(!mBaseClassName.empty());

    if (mCallbacks.empty())
    {
        return;
    }

    IndentStream(fp, indent)
        << '\n'
        << '\n'
        << separatorComment(indent, CPP_STYLE, "Callback registration") << '\n'
        << '\n'
        << "void\n"
        << mClassName << "::" << "registerCallbacks(CallbackManagerMap& cbmanagers)\n"
        << "{\n";
    indent++;

    IndentStream(fp, indent)
        << mBaseClassName << "::registerCallbacks(cbmanagers);\n"
        << '\n'
        << "CallbackManagerMap::iterator    it;\n"
        << "CallbackManagerMap::mapped_type cbmanager;\n";

    for (map< string, vector< string > >::const_iterator git = mCallbackGroups.begin();
         git != mCallbackGroups.end();
         ++git)
    {
        const string phase = git->first.empty() ? "main" : git->first;

        IndentStream(fp, indent)
            << '\n'
            << "// Register " << phase << " callbacks\n"
            << "it = cbmanagers.find(\"" << lowercase(git->first) << "\");\n"
            << "UTILS_ASSERT(it != cbmanagers.end());\n"
            << '\n'
            << "cbmanager = it->second;\n";

        for (vector< string >::const_iterator cit = git->second.begin();
             cit != git->second.end();
             ++cit)
        {
            IndentStream(fp, indent)
                << "cbmanager->register_callback(\n"
                << "    " << (*cit) << ",\n"
                << "    PEARL_create_callback(\n"
                << "        this,\n"
                << "        &"
                << mClassName
                << "::"
                << lowercase(getCallbackName(git->first, *cit))
                << "_cb));\n";
        }
    }

    indent--;
    IndentStream(fp, indent)
        << "}\n";
}


void
Pattern::writeClassDefCallbacks(FILE* const fp,
                                int&        indent) const
{
    if (mCallbacks.empty())
    {
        return;
    }

    for (const auto& group : mCallbackGroups)
    {
        const string phase = group.first.empty() ? "main" : group.first;

        IndentStream(fp, indent)
            << '\n'
            << '\n'
            << separatorComment(indent, CPP_STYLE, "Callback functions (" + phase + ")");

        for (const auto& callback : group.second)
        {
            // Parameter indent: class name + "::" + callback name + "_cb("
            const string cbName = getCallbackName(group.first, callback);
            const string paramIndent(mClassName.length() + cbName.length() + 6, ' ');

            IndentStream(fp, indent)
                << '\n'
                << '\n'
                << "void\n"
                << mClassName << "::" << lowercase(cbName) << "_cb("
                << "const CallbackManager& cbmanager,\n"
                << paramIndent
                << "const int              user_event,\n"
                << paramIndent
                << "const Event&           event,\n"
                << paramIndent
                << "CallbackData* const    cdata)\n"
                << "{\n";
            indent++;

            // Heuristic to check whether the data object is used
            auto it = mCallbacks.find(cbName);
            assert(it != mCallbacks.end());
            if (it->second.find("data") != string::npos)
            {
                IndentStream(fp, indent)
                    << "SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);\n"
                    << '\n';
            }

            IndentStream(fp, indent)
                << it->second << '\n';

            indent--;
            IndentStream(fp, indent)
                << "}\n";
        }
    }
}


// --- Generate documentation (private) -------------------------------------

void
Pattern::processReferences(const string&     section,
                           string&           text,
                           const PatternMap& patternMap)
{
    /* Remove trailing spaces */
    string::reverse_iterator it = text.rbegin();
    while ((it != text.rend()) && (*it == ' '))
    {
        ++it;
    }
    text.erase(it.base(), text.end());

    /* Keyword substitution: Cross-references */
    string::size_type spos = text.find("@ref(");
    while (spos != string::npos)
    {
        /* Search for closing brace */
        string::size_type epos = text.find(')', spos);
        if (epos == string::npos)
        {
            throw runtime_error("Error in " + section + " of " + mIdentifier
                                + ": \")\" missing.");
        }

        /* Extract & validate ID */
        string                                  id(text, spos + 5, epos - (spos + 5));
        map< string, Pattern* >::const_iterator pat = patternMap.find(id);
        if (pat == patternMap.end())
        {
            throw runtime_error("Error in " + section + " of " + mIdentifier
                                + ": Unknown pattern name " + id + '.');
        }

        /* Insert anchor */
        text.replace(spos, epos - spos + 1,
                     "<a href=\"#" + lowercase(pat->second->mIdentifier) + "\">"
                     + pat->second->mDocName + "</a>");

        /* Search for new reference */
        spos = text.find("@ref(");
    }

    /* Keyword substitution: Images */
    spos = text.find("@img(");
    while (spos != string::npos)
    {
        /* Search for closing brace */
        string::size_type epos = text.find(')', spos);
        if (epos == string::npos)
        {
            throw runtime_error("Error in " + section + " of " + mIdentifier
                                + ": \")\" missing.");
        }

        /* Extract image name */
        string id(text, spos + 5, epos - (spos + 5));

        /* Insert image reference */
        text.replace(spos, epos - spos + 1,
                     "<br>\n"
                     "<div align=\"center\">\n"
                     "<img src=\""
                     + id
                     + "\" alt=\""
                     + mDisplayName
                     + " Example\">\n"
                       "</div>\n"
                       "<br>\n");

        /* Search for new reference */
        spos = text.find("@img(");
    }
}
