/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#ifndef SCOUT_GENERATOR_PATTERN_H
#define SCOUT_GENERATOR_PATTERN_H


#include <cstdio>
#include <map>
#include <string>
#include <vector>


class Pattern
{
    public:
        // --- Type definitions ----------------------------

        /// Mapping pattern id -@> pattern
        typedef std::map< std::string, Pattern* > PatternMap;


        // --- Public methods ------------------------------

        /// @name Constructors & destructor
        /// @{

        explicit
        Pattern(const std::string& identifier);

        /// @}
        /// @name Get pattern attributes
        /// @{

        std::string
        getIdentifier() const;

        std::string
        getClassName() const;

        bool
        providesCode() const;

        bool
        providesDocs() const;

        /// @}
        /// @name Set pattern attributes
        /// @{

        void
        setDisplayName(const std::string& name);

        void
        setClassName(const std::string& name);

        void
        setDocName(const std::string& name);

        void
        setParent(Pattern* parent);

        void
        setType(const std::string& type);

        void
        setDelayKey(const std::string& key);

        void
        setInfoString(const std::string& text);

        void
        setDescription(const std::string& text);

        void
        setDiagnosis(const std::string& text);

        void
        setUnit(const std::string& unit);

        void
        setStorageMode(const std::string& mode);

        void
        setConditionalExpr(const std::string& expr);

        void
        setPrivateData(const std::string& code);

        void
        setInitCode(const std::string& code);

        void
        setCleanupCode(const std::string& code);

        void
        setStaticInitCode(const std::string& code);

        void
        hide();

        void
        suppressDocs();

        void
        addCallback(const std::string& group,
                    const std::string& event,
                    const std::string& code);

        /// @}
        /// @name Finalize pattern definition
        /// @{

        void
        finalize();

        /// @}
        /// @name Generate code
        /// @{

        void
        writeImplementation(std::FILE* fp,
                            int&       indent) const;

        void
        writePreprocGuardBegin(std::FILE* fp,
                               int&       indent) const;

        void
        writePreprocGuardEnd(std::FILE* fp,
                             int&       indent) const;

        void
        writeConditionalBegin(std::FILE* fp,
                              int&       indent) const;

        void
        writeConditionalEnd(std::FILE* fp,
                            int&       indent) const;

        /// @}
        /// @name Generate documentation
        /// @{

        void
        writeDocumentation(std::FILE*        fp,
                           bool              isFirst,
                           const PatternMap& patternMap);

        /// @}


    private:
        // --- Forward declarations ------------------------

        struct TypeInfo;


        // --- Data members --------------------------------

        /// Symbolic pattern name (identifier)
        std::string mIdentifier;

        /// Pattern display name
        std::string mDisplayName;

        /// Pattern implementation class name
        std::string mClassName;

        /// Pattern implementation base class name
        std::string mBaseClassName;

        /// Pattern name in documentation
        std::string mDocName;

        /// Parent pattern
        Pattern* mParent;

        /// Child patterns
        std::vector< Pattern* > mChildren;

        /// Type specific pattern information
        const TypeInfo* mTypeInfo;

        // Delay event and cost name suffix
        std::string mDelayKey;

        /// Brief pattern description
        std::string mInfoString;

        /// Detailed pattern description (HTML)
        std::string mDescription;

        /// Diagnosis description (HTML)
        std::string mDiagnosis;

        /// Unit of measurement
        std::string mUnit;

        /// Data storage mode (inclusive/exclusive)
        std::string mStorageMode;

        /// Conditional expression controlling pattern creation
        std::string mConditionalExpr;

        /// Code snippet for private static/non-static data (and code)
        std::string mPrivateData;

        /// Code snippet to initialize non-static private data
        std::string mInitCode;

        /// Code snippet to clean up non-static private data
        std::string mCleanupCode;

        /// Code snippet to initialize static private data
        std::string mStaticInitCode;

        /// Flag indicating visibility of pattern in unmapped CUBE file
        bool mIsHidden;

        /// Flag indicating whether documentation should be generated
        bool mProvidesDocs;

        /// Mapping callback group -@> events
        std::map< std::string, std::vector< std::string > > mCallbackGroups;

        /// Mapping event -@> callback code
        std::map< std::string, std::string > mCallbacks;


        // --- Private methods -----------------------------

        /// @name Get pattern attributes
        /// @{

        bool
        isLongTermDelayPattern() const;

        /// @}
        /// @name Generate code
        /// @{

        void
        writeClassDeclaration(std::FILE* fp,
                              int&       indent) const;

        void
        writeClassDeclPublic(std::FILE* fp,
                             int&       indent) const;

        void
        writeClassDeclProtected(std::FILE* fp,
                                int&       indent) const;

        void
        writeClassDeclPrivate(std::FILE* fp,
                              int&       indent) const;

        void
        writeClassDeclCtorDtor(std::FILE* fp,
                               int&       indent) const;

        void
        writeClassDeclRegistration(std::FILE* fp,
                                   int&       indent) const;

        void
        writeClassDeclInfoAccess(std::FILE* fp,
                                 int&       indent) const;

        void
        writeClassDeclCallbacks(std::FILE* fp,
                                int&       indent) const;

        void
        writeClassDeclCtor(std::FILE* fp,
                           int&       indent) const;

        void
        writeClassDeclDtor(std::FILE* fp,
                           int&       indent) const;

        void
        writeClassDeclGetId(std::FILE* fp,
                            int&       indent) const;

        void
        writeClassDeclGetParent(std::FILE* fp,
                                int&       indent) const;

        void
        writeClassDeclGetName(std::FILE* fp,
                              int&       indent) const;

        void
        writeClassDeclGetUniqueName(std::FILE* fp,
                                    int&       indent) const;

        void
        writeClassDeclGetDescription(std::FILE* fp,
                                     int&       indent) const;

        void
        writeClassDeclGetUnit(std::FILE* fp,
                              int&       indent) const;

        void
        writeClassDeclGetStorageMode(std::FILE* fp,
                                     int&       indent) const;

        void
        writeClassDeclIsVisible(std::FILE* fp,
                                int&       indent) const;

        void
        writeClassDeclPrivateData(std::FILE* fp,
                                  int&       indent) const;

        void
        writeClassDefinition(std::FILE* fp,
                             int&       indent) const;

        void
        writeClassDefStaticInit(std::FILE* fp,
                                int&       indent) const;

        void
        writeClassDefRegistration(std::FILE* fp,
                                  int&       indent) const;

        void
        writeClassDefCallbacks(std::FILE* fp,
                               int&       indent) const;

        /// @}
        /// @name Generate documentation
        /// @{

        void
        processReferences(const std::string& section,
                          std::string&       text,
                          const PatternMap&  patternMap);

        /// @}
};


#endif    // !SCOUT_GENERATOR_PATTERN_H
