%{
/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2024                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2014                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  Copyright (c) 2021-2023                                                **
**  RWTH Aachen University, IT Center                                      **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include <algorithm>
#include <cctype>
#include <cstdio>
#include <cstdlib>
#include <exception>
#include <functional>
#include <map>
#include <memory>
#include <sstream>
#include <string>
#include <vector>

#include "Helper.h"
#include "IndentStream.h"
#include "Pattern.h"

using namespace std;


#define YYSTYPE std::string


/*--- Constants -----------------------------------------------------------*/

const char* copyright =
    "/****************************************************************************\n"
    "**  SCALASCA    http://www.scalasca.org/                                   **\n"
    "*****************************************************************************\n"
    "**  Copyright (c) 1998-2024                                                **\n"
    "**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **\n"
    "**                                                                         **\n"
    "**  Copyright (c) 2009-2014                                                **\n"
    "**  German Research School for Simulation Sciences GmbH,                   **\n"
    "**  Laboratory for Parallel Programming                                    **\n"
    "**                                                                         **\n"
    "**  Copyright (c) 2019-2023                                                **\n"
    "**  RWTH Aachen University, IT Center                                      **\n"
    "**                                                                         **\n"
    "**  This software may be modified and distributed under the terms of       **\n"
    "**  a BSD-style license.  See the COPYING file in the package base         **\n"
    "**  directory for details.                                                 **\n"
    "****************************************************************************/\n"
    "\n\n";


/*--- Global variables ----------------------------------------------------*/

extern FILE* yyin;

long lineno = 1;
long depth  = 0;

string inpFilename;
string incFilename;
int    incLine;
int    codeLine;

string prolog;
string prefix("Patterns_gen");

map< string,Pattern* > id2pattern;    // pattern id |-> pattern
vector< Pattern* >     pattern;       // pattern list
Pattern*               current;       // current pattern
string                 callbackgroup; // current callback group


/*--- Function prototypes -------------------------------------------------*/

int
main(int    argc,
     char** argv);

void
write_header();

void
write_impl();

void
write_html();

string
uppercase(const string& str);

string
preprocessText(const string& input,
               int&          lineNumber);

void
yyerror(const string& message);

extern int
yylex();

extern void
include_file(const string& filename);


%}


%token STRING TEXT
%token CALLBACKS CLASS CLEANUP CONDITION DATA DELAY_KEY DESCR DOCNAME
%token DIAGNOSIS HIDDEN INCLUDE INFO INIT MODE NAME NODOCS PARENT PATTERN
%token PROLOG STATICINIT TYPE UNIT


%%

File            :   Body
                ;


Body            :   BodyItem
                |   Body BodyItem
                ;


BodyItem        :   Include
                |   Pattern
                |   Prolog
                ;


Include         :   INCLUDE String
                    {
                        include_file($2);
                    }
                ;


Prolog          :   PROLOG CodeBlock
                    {
                        if (!prolog.empty())
                        {
                            prolog += "\n\n";
                        }

                        prolog += $2;
                    }
                ;


Pattern         :   PATTERN String
                    {
                        if (id2pattern.find($2) != id2pattern.end())
                        {
                            yyerror("Pattern already defined: " + $2);
                        }

                        current = new Pattern($2);
                        pattern.push_back(current);
                        id2pattern.insert(make_pair($2, current));
                    }
                    '=' '[' PatternDef ']'
                    {
                        try
                        {
                            current->finalize();
                        }
                        catch (exception& ex)
                        {
                            yyerror(string("Incomplete pattern definition: ")
                                    + ex.what());
                        }
                    }
                ;


PatternDef      :   PatternDef DefItem
                |   DefItem
                ;


DefItem         :   Name
                |   Classname
                |   Docname
                |   Parent
                |   Type
                |   DelayKey
                |   Hidden
                |   NoDocs
                |   Info
                |   Description
                |   Diagnosis
                |   Unit
                |   Mode
                |   Condition
                |   Init
                |   StaticInit
                |   Cleanup
                |   Data
                |   Callbacks
                ;


Name            :   NAME '=' String
                    {
                        try
                        {
                            current->setDisplayName($3);
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


Classname       :   CLASS '=' String
                    {
                        try
                        {
                            current->setClassName($3);
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


Docname         :   DOCNAME '=' String
                    {
                        try
                        {
                            current->setDocName($3);
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


Parent          :   PARENT '=' String
                    {
                        map< string, Pattern* >::iterator it = id2pattern.find($3);
                        if (it == id2pattern.end())
                        {
                            yyerror("Unknown pattern: " + $3);
                        }

                        try
                        {
                            current->setParent(it->second);
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


Type            :   TYPE '=' String
                    {
                        try
                        {
                            current->setType($3);
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


DelayKey        :   DELAY_KEY '=' String
                    {
                        try
                        {
                            current->setDelayKey($3);
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


Hidden          :   HIDDEN
                    {
                        try
                        {
                            current->hide();
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


NoDocs          :   NODOCS
                    {
                        try
                        {
                            current->suppressDocs();
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


Info            :   INFO '=' String
                    {
                        try
                        {
                            current->setInfoString($3);
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


Description     :   DESCR '=' TextBlock
                    {
                        try
                        {
                            current->setDescription($3);
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


Diagnosis       :   DIAGNOSIS '=' TextBlock
                    {
                        try
                        {
                            current->setDiagnosis($3);
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


Unit            :   UNIT '=' String
                    {
                        try
                        {
                            current->setUnit($3);
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


Mode            :   MODE '=' String
                    {
                        try
                        {
                            current->setStorageMode($3);
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


Condition       :   CONDITION '=' String
                    {
                        try
                        {
                            current->setConditionalExpr($3);
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


Init            :   INIT '=' CodeBlock
                    {
                        try
                        {
                            current->setInitCode($3);
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


StaticInit      :   STATICINIT '=' CodeBlock
                    {
                        try
                        {
                            current->setStaticInitCode($3);
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


Cleanup         :   CLEANUP '=' CodeBlock
                    {
                        try
                        {
                            current->setCleanupCode($3);
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


Data            :   DATA '=' CodeBlock
                    {
                        try
                        {
                            current->setPrivateData($3);
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


Callbacks       :   CALLBACKS '(' String ')'
                    {
                        callbackgroup = $3;
                    }
                    '=' '[' CbList ']'
                |   CALLBACKS
                    {
                        callbackgroup = "";
                    }
                    '=' '[' CbList ']'
                ;


CbList          :   CbList CbItem
                |   CbItem
                ;


CbItem          :   String '=' CodeBlock
                    {
                        try
                        {
                            current->addCallback(callbackgroup, $1, $3);
                        }
                        catch (exception& ex)
                        {
                            yyerror(ex.what());
                        }
                    }
                ;


String          :   '"' STRING '"'
                    {
                        $$ = $2;
                    }
                ;


CodeBlock       :   '{'
                    {
                        codeLine = lineno;
                    }
                    Text '}'
                    {
                        string codeText = preprocessText($3, codeLine);

                        // Account for the extra blank line added after the
                        // line directive
                        codeLine--;
                        if (codeLine <= 0)
                        {
                            yyerror("Code block in first line unsupported!");
                        }

                        stringstream lineDirective;
                        lineDirective << "#line "
                                      << codeLine
                                      << " \""
                                      << (incFilename.empty()
                                         ? inpFilename
                                         : incFilename)
                                      << "\"\n"
                                      << '\n';

                        $$ = lineDirective.str() + codeText;
                    }
                ;


TextBlock       :   '{' Text '}'
                    {
                        int dummy = 0;
                        $$ = preprocessText($2, dummy);
                    }
                ;


Text            :   Text TEXT
                    {
                        $$ = $1 + $2;
                    }
                |   TEXT
                    {
                        $$ = $1;
                    }
                ;


%%


// --------------------------------------------------------------------------
//
// gen_pattern <description file>
//
// --------------------------------------------------------------------------

int
main(int    argc,
     char** argv)
{
    /* Check command line arguments */
    if (argc != 2)
    {
        fprintf(stderr, "Usage: gen_patterns <description file>\n");
        exit(EXIT_FAILURE);
    }

    /* Store input filename */
    inpFilename = argv[1];

    /* Open input file */
    yyin = fopen(inpFilename.c_str(), "r");
    if (!yyin)
    {
        fprintf(stderr, "Could not open file \"%s\".\n", inpFilename.c_str());
        exit(EXIT_FAILURE);
    }

    /* Parse input file */
    yyparse();

    /* Close input file */
    fclose(yyin);

    /* Write output */
    write_header();
    write_impl();
    write_html();

    /* Clean up */
    std::for_each(pattern.begin(), pattern.end(),
                  std::default_delete< Pattern >());

    return EXIT_SUCCESS;
}


void
write_header()
{
    /* Open header file */
    string filename = prefix + ".h";
    FILE* fp = fopen(filename.c_str(), "w");
    if (!fp)
    {
        fprintf(stderr, "Could not open output file \"%s\".\n", filename.c_str());
        exit(EXIT_FAILURE);
    }

    /* Temporary strings */
    string prefix_upper = uppercase(prefix);

    /* Write copyright notice */
    fprintf(fp, "%s", copyright);

    /* Open include guard */
    fprintf(fp, "#ifndef SCOUT_%s_H\n", prefix_upper.c_str());
    fprintf(fp, "#define SCOUT_%s_H\n", prefix_upper.c_str());
    fprintf(fp, "\n\n");

    /* Open namespace */
    fprintf(fp, "namespace scout\n"
                "{\n");

    /* Determine format string for pattern constants */
    string::size_type                  length = 0;
    vector< Pattern* >::const_iterator it     = pattern.begin();
    while (it != pattern.end())
    {
        length = max(length, (*it)->getIdentifier().length());
        ++it;
    }
    ostringstream fmt;
    fmt << "const long PAT_%-" << length << "s = %d;\n";

    /* Write pattern constants */
    int num = -1;
    fprintf(fp, "%s\n\n", separatorComment(0, CPP_STYLE, "Constants").c_str());
    fprintf(fp, fmt.str().c_str(), "NONE", num++);
    it = pattern.begin();
    while (it != pattern.end())
    {
        fprintf(fp, fmt.str().c_str(), (*it)->getIdentifier().c_str(), num++);
        ++it;
    }
    fprintf(fp, "\n\n");

    /* Write forward declarations */
    fprintf(fp, "%s\n"
                "\n"
                "class AnalyzeTask;\n"
                "\n\n",
                separatorComment(0, CPP_STYLE, "Forward declarations").c_str());

    /* Write function prototypes */
    fprintf(fp, "%s\n"
                "\n"
                "void\n"
                "create_patterns(AnalyzeTask* analyzer);\n",
                separatorComment(0, CPP_STYLE, "Pattern registration").c_str());

    /* Close namespace */
    fprintf(fp, "}    // namespace scout\n"
                "\n\n");

    /* Close include guard */
    fprintf(fp, "#endif    // !SCOUT_%s_H\n", prefix_upper.c_str());

    /* Close file */
    fclose(fp);
}


void
write_impl()
{
    /* Open implementation file */
    string filename = prefix + ".cpp";
    FILE* fp = fopen(filename.c_str(), "w");
    if (!fp)
    {
        fprintf(stderr, "Could not open output file \"%s\".\n", filename.c_str());
        exit(EXIT_FAILURE);
    }

    /* Write copyright notice */
    int indent = 0;
    IndentStream(fp, indent)
        << copyright;

    /* Write common includes */
    IndentStream(fp, indent)
        << "#include <config.h>\n"
        << '\n'
        << "#include \"Patterns_gen.h\"\n"
        << '\n'
        << "#include <map>\n"
        << "#include <string>\n"
        << '\n'
        << "#include <UTILS_Error.h>\n"
        << '\n'
        << "#include <cubew_cube.h>\n"
        << '\n'
        << "#include <pearl/Callback.h>\n"
        << "#include <pearl/CallbackManager.h>\n"
        << "#include <pearl/Event.h>\n"
        << '\n'
        << "#include \"AnalyzeTask.h\"\n"
        << "#include \"CbData.h\"\n"
        << "#include \"TypedPattern.h\"\n"
        << "#include \"scout_types.h\"\n"
        << '\n'
        << "#if defined(_MPI)\n"
        << "    #include \"LongTermDelayPattern.h\"\n"
        << "#endif    // _MPI\n"
        << '\n';

    /* Write "using" directives */
    IndentStream(fp, indent)
        << "using namespace std;\n"
        << "using namespace pearl;\n"
        << '\n'
        << '\n';

    /* Write prolog */
    IndentStream(fp, indent)
        << prolog << '\n'
        << '\n'
        << '\n';

    /* Open 'scout::metric' namespace */
    IndentStream(fp, indent)
        << "namespace scout\n"
        << "{\n"
        << "namespace metric\n"
        << "{\n";

    /* Write class implementation */
    for (vector< Pattern* >::const_iterator it = pattern.begin();
         it != pattern.end();
         ++it)
    {
        if ((*it)->providesCode())
        {
            (*it)->writeImplementation(fp, indent);
        }
    }

    /* Close 'scout::metric' namespace */
    IndentStream(fp, indent)
        << "}    // namespace scout::metric\n"
        << '\n'
        << '\n';

    /* Write function implementation */
    IndentStream(fp, indent)
        << separatorComment(indent, CPP_STYLE, "Pattern registration") << '\n'
        << '\n'
        << "void\n"
        << "create_patterns(AnalyzeTask* const analyzer)\n"
        << "{\n";
    indent++;

    for (vector< Pattern* >::const_iterator it = pattern.begin();
         it != pattern.end();
         ++it)
    {
        if ((*it)->providesCode())
        {
            (*it)->writePreprocGuardBegin(fp, indent);
            (*it)->writeConditionalBegin(fp, indent);

            IndentStream(fp, indent)
                << "analyzer->addPattern(new metric::" << (*it)->getClassName() << "());\n";

            (*it)->writeConditionalEnd(fp, indent);
            (*it)->writePreprocGuardEnd(fp, indent);
        }
    }

    indent--;
    IndentStream(fp, indent)
        << "}\n";

    /* Close 'scout' namespace */
    IndentStream(fp, indent)
        << "}    // namespace scout\n";

    /* Close file */
    fclose(fp);
}


void
write_html()
{
    /* Open documentation file */
    string filename = prefix + ".html";
    FILE* fp = fopen(filename.c_str(), "w");
    if (!fp)
    {
        fprintf(stderr, "Could not open output file \"%s\".\n", filename.c_str());
        exit(EXIT_FAILURE);
    }

    /* Write file header */
    fprintf(fp, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n"
                "<html>\n"
                "<head>\n"
                "<title>Performance properties</title>\n"
                "</head>\n"
                "<body>\n");

    /* Write page header */
    fprintf(fp, "<h2>Performance properties</h2>\n");

    /* Write class documentation */
    try
    {
        bool isFirst = true;
        for (vector< Pattern* >::const_iterator it = pattern.begin();
             it != pattern.end();
             ++it)
        {
            if ((*it)->providesDocs())
            {
                (*it)->writeDocumentation(fp, isFirst, id2pattern);
                isFirst = false;
            }
        }
    }
    catch (exception& ex)
    {
        fprintf(stderr, "%s\n", ex.what());
        exit(EXIT_FAILURE);
    }

    /* Close file */
    fclose(fp);
}


void
yyerror(const string& message)
{
    if (!incFilename.empty())
    {
        fprintf(stderr, "In included file \"%s\":\n   ", incFilename.c_str());
    }

    fprintf(stderr, "Line %ld: %s\n", lineno, message.c_str());
    exit(EXIT_FAILURE);
}


struct fo_toupper
{
    int
    operator()(int x) const
    {
        return std::toupper(x);
    }
};


string
uppercase(const string& str)
{
    string result(str);

    transform(str.begin(), str.end(), result.begin(), fo_toupper());

    return result;
}


string
preprocessText(const string& input,
               int&          lineNumber)
{
    const string::size_type inputLength = input.length();

    string            result;
    string::size_type baseIndent = string::npos;
    string::size_type beginPos   = 0;
    do
    {
        // Determine character index right after end of line
        string::size_type endPos = input.find('\n', beginPos);
        if (endPos == string::npos)
        {
            endPos = inputLength;
        }
        else
        {
            endPos++;
        }

        const string::size_type lineLength = endPos - beginPos;
        string                  line(input, beginPos, lineLength);

        // Try to determine base indentation (implicit assumption: first
        // non-blank line of block is properly indented)
        if (baseIndent == string::npos)
        {
            baseIndent = line.find_first_not_of(" \t\n\r\f\v");

            // Drop leading blank lines; requires adjustment of line number
            // for '#line' directive
            if (baseIndent == string::npos)
            {
                lineNumber++;
                beginPos += lineLength;
                continue;
            }
        }

        // Strip off leading indentation, keeping lines with less spaces intact
        string::size_type textIndent = line.find_first_not_of(" \t");
        line.erase(0, min(textIndent, baseIndent));
        result += line;

        beginPos += lineLength;
    }
    while (beginPos < inputLength);

    // Trim trailing whitespaces
    beginPos = result.find_last_not_of(" \t\n\r\f\v");
    if (beginPos != string::npos)
    {
        beginPos++;
        if (beginPos < result.length())
        {
            result.erase(beginPos);
        }
    }

    return result;
}
