/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2023                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2014                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#ifndef SCOUT_SCOUT_TYPES_H
#define SCOUT_SCOUT_TYPES_H


#include <cstdint>
#include <map>
#include <stack>

#include <pearl/Callpath.h>
#include <pearl/Event.h>
#include <pearl/Location.h>
#include <pearl/pearl_types.h>

#include "TimeRank.h"
#include "fixed_priority_map.h"


/*-------------------------------------------------------------------------*/
/**
 * @file  scout_types.h
 * @brief Declaration of SCOUT-specific datatypes.
 *
 * This header file provides the declaration of some SCOUT-specific
 * datatypes.
 */
/*-------------------------------------------------------------------------*/

// --- Forward declarations -------------------------------------------------

namespace pearl
{
class CallbackManager;
}    // namespace pearl


// --- Type definitions -----------------------------------------------------

namespace scout
{
/// Callback manager map. Maps replay phases to callback manager objects.
typedef std::map< std::string, pearl::CallbackManager* >
    CallbackManagerMap;

/// @struct CommInfo
/// @brief  global rank and communicator information for non-communicating events.
///
/// This data structure is used to save communication relevant parameters for events
/// don't have their inherent communciation information. This is needed for
/// patterns like "Wait for Progress" when non-communication events can be synchronization
/// points.
struct CommInfo
{
    int           globalRank;
    std::uint32_t commId;
};


/// Container for communication information
/// @note The multimap is needed as for patterns like "Wait for progress" a single
/// Enter event can be synchpoint to multiple remote events.
typedef std::multimap< pearl::Event, CommInfo, const pearl::EventKeyCompare >
    CommInfoMap;

/// @struct SynchpointInfo
/// @brief  Synchronization information for MPI and OpenMP calls

struct SynchpointInfo
{
    double waitTime;
    double communicationTime;
    double completionTime;
    double totalTime;
};


/// @struct ContentionInfo
/// @brief  Information needed to reconstruct contention chain in passive target RMA
struct ContentionInfo
{
    pearl::ident_t     eventIndex;
    pearl::timestamp_t waitTime;
    std::uint32_t      syncRank;
};


typedef std::stack< ContentionInfo >         ContentionInfoStack;
typedef std::map< int, ContentionInfoStack > ContentionInfoStackMap;

/// Type definition of a cnode -> time map
typedef fixed_priority_map< pearl::Callpath::IdType, double, 128 >
    timemap_t;

/// Type definition of a SynchpointInfo -> time map
typedef std::map< pearl::Event, SynchpointInfo, pearl::EventKeyCompare >
    SynchpointInfoMap;

/// Type definition of a Event -> double map
typedef std::map< pearl::Event, double, pearl::EventKeyCompare >
    ScaleMap;

/**
 * @struct TimeMapSum
 * @brief  A combination of a timemap and the sum of its values
 */

struct TimeMapSum
{
    TimeMapSum()
        : mSum(0.0)
    {
    }

    timemap_t mMap; // CNode -> time map
    double    mSum; // Sum of times
};


/**
 * @struct CollectiveInfo
 * @brief  Comprises global information on a collective operation
 **/

struct CollectiveInfo
{
    TimeRank earliest_end;
    TimeRank latest;
    TimeRank root;
    TimeRank my;
};
}    // namespace scout


#endif    // !SCOUT_SCOUT_TYPES_H
