/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2021                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "ContextTree.h"

#include <sstream>
#include <string>
#include <utility>

#include <cxxslb/memory.h>

#include <gtest/gtest.h>

#include <pearl/CallingContext.h>
#include <pearl/Region.h>
#include <pearl/SourceLocation.h>
#include <pearl/String.h>

#include "DefinitionContainer.h"

using namespace std;
using namespace testing;
using namespace pearl;
using namespace pearl::detail;


namespace
{
// --- Helper ---------------------------------------------------------------

// Test fixture for calling context tree tests
class ContextTreeT
    : public Test
{
    public:
        const CallingContext&
        createContextTree();


    protected:
        DefinitionContainer< String >         mStrings;
        DefinitionContainer< Region >         mRegions;
        DefinitionContainer< SourceLocation > mSourceLocs;
        DefinitionContainer< CallingContext > mContexts;
};


// --- print tests (context trees) ------------------------------------------

TEST_F(ContextTreeT,
       print_undefinedContext_printsUndefined)
{
    string expected(" + (undefined)\n");

    ostringstream result;
    print(result, CallingContext::UNDEFINED, true);
    EXPECT_EQ(expected, result.str());

    expected = "# + (undefined)\n";

    result.str("");
    print(result, CallingContext::UNDEFINED, true, "#");
    EXPECT_EQ(expected, result.str());
}


TEST_F(ContextTreeT,
       print_contextTreeLastChild_printsTree)
{
    const CallingContext& tree = createContextTree();

    string expected(" + main @ test.c:24  <0>\n"
                    "    + foo @ test.c:19  <1>\n"
                    "    |  + bar @ test.c:14  <2>\n"
                    "    + foo @ test.c  <3>\n"
                    "       + bar  <4>\n");

    ostringstream result;
    print(result, tree, true);
    EXPECT_EQ(expected, result.str());
}


TEST_F(ContextTreeT,
       print_contextTree_printsTreeWithPrefix)
{
    const CallingContext& tree = createContextTree();

    string expected(" + main @ test.c:24  <0>\n"
                    " |  + foo @ test.c:19  <1>\n"
                    " |  |  + bar @ test.c:14  <2>\n"
                    " |  + foo @ test.c  <3>\n"
                    " |     + bar  <4>\n");

    ostringstream result;
    print(result, tree, false);
    EXPECT_EQ(expected, result.str());
}


// --- Helper ---------------------------------------------------------------

// Creates a calling context tree covering all possible situations w.r.t.
// the available information:
//
//    + main @ test.c:24  <0>           # Full info
//       + foo @ test.c:19  <1>         # Full info
//       |  + bar @ test.c:14  <2>      # No file info (region)
//       + foo @ test.c:19  <3>         # No source info (calling context)
//          + bar  <4>                  # Neither file info (region) nor source
//                                      # info (calling context)
//
const CallingContext&
ContextTreeT::createContextTree()
{
    // Create String definitions
    auto strEmpty  = cxxslb::make_unique< String >(0, "");
    auto strSource = cxxslb::make_unique< String >(1, "test.c");
    auto strMain   = cxxslb::make_unique< String >(2, "main");
    auto strFoo    = cxxslb::make_unique< String >(3, "foo");
    auto strBar    = cxxslb::make_unique< String >(4, "bar");

    // Create Region definitions
    auto regMain =
        cxxslb::make_unique< Region >(0,
                                      *strMain,
                                      *strMain,
                                      *strEmpty,
                                      Region::ROLE_FUNCTION,
                                      Paradigm::SAMPLING,
                                      *strSource,
                                      23, 25);
    auto regFoo =
        cxxslb::make_unique< Region >(1,
                                      *strFoo,
                                      *strFoo,
                                      *strEmpty,
                                      Region::ROLE_FUNCTION,
                                      Paradigm::SAMPLING,
                                      *strSource,
                                      18, 20);

    // No file info
    auto regBar =
        cxxslb::make_unique< Region >(2,
                                      *strBar,
                                      *strBar,
                                      *strEmpty,
                                      Region::ROLE_FUNCTION,
                                      Paradigm::SAMPLING,
                                      String::UNDEFINED,
                                      13, 15);

    // Create SourceLocation definitions
    auto sclMain = cxxslb::make_unique< SourceLocation >(0, *strSource, 24);
    auto sclFoo  = cxxslb::make_unique< SourceLocation >(1, *strSource, 19);
    auto sclBar  = cxxslb::make_unique< SourceLocation >(2, *strSource, 14);

    // Create CallingContext definitions
    auto ccMain = cxxslb::make_unique< CallingContext >(0,
                                                        *regMain,
                                                        *sclMain,
                                                        nullptr);
    auto ccMainFoo0 = cxxslb::make_unique< CallingContext >(1,
                                                            *regFoo,
                                                            *sclFoo,
                                                            ccMain.get());

    // No file info
    auto ccMainFoo0Bar = cxxslb::make_unique< CallingContext >(2,
                                                               *regBar,
                                                               *sclBar,
                                                               ccMainFoo0.get());

    // No source location info
    auto ccMainFoo1 = cxxslb::make_unique< CallingContext >(3,
                                                            *regFoo,
                                                            SourceLocation::UNDEFINED,
                                                            ccMain.get());

    // No file & no source location info
    auto ccMainFoo1Bar = cxxslb::make_unique< CallingContext >(4,
                                                               *regBar,
                                                               SourceLocation::UNDEFINED,
                                                               ccMainFoo1.get());
    mStrings.addItem(std::move(strEmpty));
    mStrings.addItem(std::move(strSource));
    mStrings.addItem(std::move(strMain));
    mStrings.addItem(std::move(strFoo));
    mStrings.addItem(std::move(strBar));

    mRegions.addItem(std::move(regMain));
    mRegions.addItem(std::move(regFoo));
    mRegions.addItem(std::move(regBar));

    mSourceLocs.addItem(std::move(sclMain));
    mSourceLocs.addItem(std::move(sclFoo));
    mSourceLocs.addItem(std::move(sclBar));

    auto* root = ccMain.get();
    mContexts.addItem(std::move(ccMain));
    mContexts.addItem(std::move(ccMainFoo0));
    mContexts.addItem(std::move(ccMainFoo0Bar));
    mContexts.addItem(std::move(ccMainFoo1));
    mContexts.addItem(std::move(ccMainFoo1Bar));

    return *root;
}
}    // unnamed namespace
