/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2017-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2021-2023                                                **
**  RWTH Aachen University, IT Center                                      **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "MutableGlobalDefs.h"

#include <utility>
#include <vector>

#include <cxxslb/memory.h>

#include <gtest/gtest.h>

#include <pearl/CallingContext.h>
#include <pearl/Callsite.h>
#include <pearl/CartDimension.h>
#include <pearl/CartTopology.h>
#include <pearl/CommLocationSet.h>
#include <pearl/CommSet.h>
#include <pearl/Communicator.h>
#include <pearl/Error.h>
#include <pearl/Location.h>
#include <pearl/LocationGroup.h>
#include <pearl/LocationSet.h>
#include <pearl/Metric.h>
#include <pearl/Region.h>
#include <pearl/RmaWindow.h>
#include <pearl/SourceLocation.h>
#include <pearl/String.h>
#include <pearl/SystemNode.h>

#include "Process.h"

using namespace std;
using namespace testing;
using namespace pearl;
using namespace pearl::detail;


namespace
{
// --- Helper ---------------------------------------------------------------

// Dummy test fixture needed for MutableGlobalDefs typed tests
template< typename DefClassT >
class MutableGlobalDefsTypedT
    : public Test
{
};


// Instantiate test cases for definition types supported by MutableGlobalDefs
typedef Types< CallingContext,
               Callsite,
               CartDimension,
               CartTopology,
               Communicator,
               GroupingSet,
               Location,
               LocationGroup,
               Metric,
               Region,
               RmaWindow,
               SourceLocation,
               String,
               SystemNode > DefinitionTypes;
TYPED_TEST_SUITE(MutableGlobalDefsTypedT, DefinitionTypes);


// createItem function template and specializations for supported definitions
template< typename DefClassT >
std::unique_ptr< DefClassT >
createItem(const typename DefClassT::IdType id)
{
    return nullptr;
}


template< >
std::unique_ptr< CallingContext >
createItem(const CallingContext::IdType id)
{
    return cxxslb::make_unique< CallingContext >(id,
                                                 Region::UNDEFINED,
                                                 SourceLocation::UNDEFINED,
                                                 nullptr);
}


template< >
std::unique_ptr< Callsite >
createItem(const Callsite::IdType id)
{
    return cxxslb::make_unique< Callsite >(id,
                                           String::UNDEFINED,
                                           0,
                                           Region::UNDEFINED);
}


template< >
std::unique_ptr< CartDimension >
createItem(const CartDimension::IdType id)
{
    return cxxslb::make_unique< CartDimension >(id,
                                                String::UNDEFINED,
                                                0,
                                                CartDimension::PERIODIC);
}


template< >
std::unique_ptr< CartTopology >
createItem(const CartTopology::IdType id)
{
    CartTopology::DimensionContainer dimensions;

    return cxxslb::make_unique< CartTopology >(id,
                                               String::UNDEFINED,
                                               Communicator::UNDEFINED,
                                               dimensions);
}


template< >
std::unique_ptr< Communicator >
createItem(const Communicator::IdType id)
{
    return cxxslb::make_unique< Communicator >(id,
                                               String::UNDEFINED,
                                               Communicator::PROPERTY_NONE,
                                               CommSet::UNDEFINED,
                                               nullptr);
}


template< >
std::unique_ptr< GroupingSet >
createItem(const GroupingSet::IdType id)
{
    vector< const Location* > members;

    return cxxslb::make_unique< LocationSet >(id,
                                              String::UNDEFINED,
                                              members);
}


template< >
std::unique_ptr< Location >
createItem(const Location::IdType id)
{
    return cxxslb::make_unique< Location >(id,
                                           String::UNDEFINED,
                                           Location::TYPE_CPU_THREAD,
                                           0,
                                           const_cast< LocationGroup* >(&LocationGroup::UNDEFINED));
}


template< >
std::unique_ptr< LocationGroup >
createItem(const LocationGroup::IdType id)
{
    return cxxslb::make_unique< Process >(id,
                                          String::UNDEFINED,
                                          const_cast< SystemNode* >(&SystemNode::UNDEFINED),
                                          nullptr);
}


template< >
std::unique_ptr< Metric >
createItem(const Metric::IdType id)
{
    return cxxslb::make_unique< Metric >(id,
                                         "",
                                         "",
                                         "INTEGER",
                                         "COUNTER",
                                         "START");
}


template< >
std::unique_ptr< Region >
createItem(const Region::IdType id)
{
    return cxxslb::make_unique< Region >(id,
                                         String::UNDEFINED,
                                         String::UNDEFINED,
                                         String::UNDEFINED,
                                         Region::ROLE_UNKNOWN,
                                         Paradigm::UNKNOWN,
                                         String::UNDEFINED,
                                         0,
                                         0);
}


template< >
std::unique_ptr< RmaWindow >
createItem(const RmaWindow::IdType id)
{
    return cxxslb::make_unique< RmaWindow >(id,
                                            Communicator::UNDEFINED);
}


template< >
std::unique_ptr< SourceLocation >
createItem(const SourceLocation::IdType id)
{
    return cxxslb::make_unique< SourceLocation >(id,
                                                 String::UNDEFINED,
                                                 0);
}


template< >
std::unique_ptr< String >
createItem(const String::IdType id)
{
    return cxxslb::make_unique< String >(id,
                                         "");
}


template< >
std::unique_ptr< SystemNode >
createItem(const SystemNode::IdType id)
{
    return cxxslb::make_unique< SystemNode >(id,
                                             String::UNDEFINED,
                                             String::UNDEFINED,
                                             nullptr);
}


// addItem function template and specializations for supported definitions
template< typename DefClassT >
void
addItem(MutableGlobalDefs&           defs,
        std::unique_ptr< DefClassT > item)
{
}


template< >
void
addItem(MutableGlobalDefs&                defs,
        std::unique_ptr< CallingContext > item)
{
    defs.addCallingContext(std::move(item));
}


template< >
void
addItem(MutableGlobalDefs&          defs,
        std::unique_ptr< Callsite > item)
{
    defs.addCallsite(std::move(item));
}


template< >
void
addItem(MutableGlobalDefs&               defs,
        std::unique_ptr< CartDimension > item)
{
    defs.addCartDimension(std::move(item));
}


template< >
void
addItem(MutableGlobalDefs&              defs,
        std::unique_ptr< CartTopology > item)
{
    defs.addCartTopology(std::move(item));
}


template< >
void
addItem(MutableGlobalDefs&              defs,
        std::unique_ptr< Communicator > item)
{
    defs.addCommunicator(std::move(item));
}


template< >
void
addItem(MutableGlobalDefs&             defs,
        std::unique_ptr< GroupingSet > item)
{
    defs.addGroupingSet(std::move(item));
}


template< >
void
addItem(MutableGlobalDefs&          defs,
        std::unique_ptr< Location > item)
{
    defs.addLocation(std::move(item));
}


template< >
void
addItem(MutableGlobalDefs&               defs,
        std::unique_ptr< LocationGroup > item)
{
    defs.addLocationGroup(std::move(item));
}


template< >
void
addItem(MutableGlobalDefs&        defs,
        std::unique_ptr< Metric > item)
{
    defs.addMetric(std::move(item));
}


template< >
void
addItem(MutableGlobalDefs&        defs,
        std::unique_ptr< Region > item)
{
    defs.addRegion(std::move(item));
}


template< >
void
addItem(MutableGlobalDefs&           defs,
        std::unique_ptr< RmaWindow > item)
{
    defs.addRmaWindow(std::move(item));
}


template< >
void
addItem(MutableGlobalDefs&                defs,
        std::unique_ptr< SourceLocation > item)
{
    defs.addSourceLocation(std::move(item));
}


template< >
void
addItem(MutableGlobalDefs&        defs,
        std::unique_ptr< String > item)
{
    defs.addString(std::move(item));
}


template< >
void
addItem(MutableGlobalDefs&            defs,
        std::unique_ptr< SystemNode > item)
{
    defs.addSystemNode(std::move(item));
}


// --- MutableGlobalDefs tests ----------------------------------------------

TEST(MutableGlobalDefsT,
     assignCartCoordinate_validCartTopoId_addsCoordinate)
{
    MutableGlobalDefs defs;

    addItem< Location >(defs, createItem< Location >(0));

    vector< const Location* > locations;
    locations.push_back(&defs.getLocation(0));
    addItem< GroupingSet >(defs,
                           cxxslb::make_unique< CommLocationSet >(0,
                                                                  String::UNDEFINED,
                                                                  Paradigm::USER,
                                                                  locations));

    vector< uint32_t > ranks(1, 0);
    addItem< GroupingSet >(defs,
                           cxxslb::make_unique< CommSet >(1,
                                                          String::UNDEFINED,
                                                          GroupingSet::PROPERTY_NONE,
                                                          defs.getCommLocationSet(Paradigm::USER),
                                                          ranks));

    const GroupingSet& groupingSet = defs.getGroupingSet(1);
    const CommSet&     commSet     = static_cast< const CommSet& >(groupingSet);
    addItem< Communicator >(defs,
                            cxxslb::make_unique< Communicator >(0,
                                                                String::UNDEFINED,
                                                                Communicator::PROPERTY_NONE,
                                                                commSet,
                                                                nullptr));

    addItem< CartDimension >(defs,
                             cxxslb::make_unique< CartDimension >(0,
                                                                  String::UNDEFINED,
                                                                  3,
                                                                  CartDimension::PERIODIC));
    addItem< CartDimension >(defs,
                             cxxslb::make_unique< CartDimension >(1,
                                                                  String::UNDEFINED,
                                                                  3,
                                                                  CartDimension::PERIODIC));

    CartTopology::DimensionContainer dimensions;
    dimensions.push_back(&defs.getCartDimension(0));
    dimensions.push_back(&defs.getCartDimension(1));
    addItem< CartTopology >(defs,
                            cxxslb::make_unique< CartTopology >(0,
                                                                String::UNDEFINED,
                                                                defs.getCommunicator(0),
                                                                dimensions));

    CartCoordinate coordinate;
    coordinate.push_back(1);
    coordinate.push_back(2);
    defs.assignCartCoordinate(0, 0, coordinate);

    const CartCoordList coords = defs.getCartTopology(0).getRankCoordinates(0);
    ASSERT_EQ(1, coords.size());
    EXPECT_EQ(coordinate, coords[0]);
}


TEST(MutableGlobalDefsT,
     assignCartCoordinate_invalidCartTopoId_throws)
{
    CartCoordinate coordinate;

    MutableGlobalDefs defs;
    EXPECT_THROW(defs.assignCartCoordinate(0, 0, coordinate), RuntimeError);
}


TEST(MutableGlobalDefsT,
     assignCartCommunicator_validCartTopoId_setsCommunicator)
{
    MutableGlobalDefs defs;

    addItem< CartTopology >(defs, createItem< CartTopology >(0));

    addItem< Communicator >(defs, createItem< Communicator >(0));
    defs.assignCartCommunicator(0, defs.getCommunicator(0));

    const CartTopology& topology = defs.getCartTopology(0);
    EXPECT_EQ(&defs.getCommunicator(0), &topology.getCommunicator());
}


TEST(MutableGlobalDefsT,
     assignCartCommunicator_invalidCartTopoId_throws)
{
    CartCoordinate coordinate;

    MutableGlobalDefs defs;
    EXPECT_THROW(defs.assignCartCommunicator(0, Communicator::UNDEFINED),
                 RuntimeError);
}


TEST(MutableGlobalDefsT,
     getCommLocationSet_validParadigm_returnsSet)
{
    vector< const Location* > locations;

    auto comm =
        cxxslb::make_unique< CommLocationSet >(0,
                                               String::UNDEFINED,
                                               Paradigm::MPI,
                                               locations);
    auto* commPtr = comm.get();

    MutableGlobalDefs defs;
    defs.addGroupingSet(std::move(comm));

    EXPECT_EQ(commPtr, &defs.getCommLocationSet(Paradigm::MPI));
}


TEST(MutableGlobalDefsT,
     getCommLocationSet_invalidParadigm_throws)
{
    MutableGlobalDefs defs;

    EXPECT_THROW(defs.getCommLocationSet(Paradigm::MPI), RuntimeError);
}


TYPED_TEST(MutableGlobalDefsTypedT,
           validate_defsWithUniqueIds_doesNotThrow)
{
    MutableGlobalDefs defs;

    addItem< TypeParam >(defs, createItem< TypeParam >(7));
    addItem< TypeParam >(defs, createItem< TypeParam >(2));
    addItem< TypeParam >(defs, createItem< TypeParam >(5));
    addItem< TypeParam >(defs, createItem< TypeParam >(3));

    EXPECT_NO_THROW(defs.validate());
}


TYPED_TEST(MutableGlobalDefsTypedT,
           validate_defsWithDuplicateIds_throws)
{
    MutableGlobalDefs defs;

    addItem< TypeParam >(defs, createItem< TypeParam >(7));
    addItem< TypeParam >(defs, createItem< TypeParam >(2));
    addItem< TypeParam >(defs, createItem< TypeParam >(5));
    addItem< TypeParam >(defs, createItem< TypeParam >(2));

    EXPECT_THROW(defs.validate(), RuntimeError);
}


TEST(MutableGlobalDefsT,
     validate_defsWithUniqueCommLocationSets_doesNotThrow)
{
    vector< const Location* > locations;

    MutableGlobalDefs defs;
    defs.addGroupingSet(
        cxxslb::make_unique< CommLocationSet >(0,
                                               String::UNDEFINED,
                                               Paradigm::MPI,
                                               locations));

    EXPECT_NO_THROW(defs.validate());
}


TEST(MutableGlobalDefsT,
     validate_defsWithDuplicateCommLocationSets_throws)
{
    vector< const Location* > locations;

    MutableGlobalDefs defs;
    defs.addGroupingSet(
        cxxslb::make_unique< CommLocationSet >(0,
                                               String::UNDEFINED,
                                               Paradigm::MPI,
                                               locations));
    defs.addGroupingSet(
        cxxslb::make_unique< CommLocationSet >(1,
                                               String::UNDEFINED,
                                               Paradigm::MPI,
                                               locations));

    EXPECT_THROW(defs.validate(), RuntimeError);
}
}    // unnamed namespace
