/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2019-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2021-2023                                                **
**  RWTH Aachen University, IT Center                                      **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include <pearl/RmaWindow.h>

#include <iostream>
#include <string>

#include <cxxslb/memory.h>

#include <gtest/gtest.h>

#include <pearl/CommLocationSet.h>
#include <pearl/CommSet.h>
#include <pearl/Communicator.h>
#include <pearl/Location.h>
#include <pearl/String.h>
#include <pearl/SystemNode.h>
#include <pearl/iomanip.h>

#include "DefinitionContainer.h"
#include "Process.h"
#include "iomanip_detail.h"

using namespace std;
using namespace testing;
using namespace pearl;
using namespace pearl::detail;


namespace
{
// --- Helper ---------------------------------------------------------------

// Test fixture for RmaWindow tests
class RmaWindowT
    : public Test
{
    public:
        RmaWindowT();


    protected:
        std::unique_ptr< SystemNode >      mSystemNode;
        std::unique_ptr< LocationGroup >   mLocationGroup;
        DefinitionContainer< String >      mStrings;
        DefinitionContainer< Location >    mLocations;
        std::unique_ptr< CommLocationSet > mCommLocationSet;
        std::unique_ptr< CommSet >         mCommSet;
        std::unique_ptr< Communicator >    mComm;
        std::unique_ptr< RmaWindow >       mWindow;
};


// --- RmaWindow tests ------------------------------------------------------

TEST_F(RmaWindowT,
       operatorLeftShift_itemAtSkipDetail_printsNothing)
{
    string expected;

    ostringstream result;
    result << setDetail(DETAIL_SKIP)
           << *mWindow;
    EXPECT_EQ(expected, result.str());
}


TEST_F(RmaWindowT,
       operatorLeftShift_itemAtAbbrevDetail_printsIdentifier)
{
    string expected("42");

    ostringstream result;
    result << setDetail(DETAIL_ABBRV)
           << *mWindow;
    EXPECT_EQ(expected, result.str());
}


TEST_F(RmaWindowT,
       operatorLeftShift_itemAtCompactDetail_printsOneliner)
{
    string expected("RmaWindow { id: 42, comm: 0 }");

    ostringstream result;
    result << *mWindow;
    EXPECT_EQ(expected, result.str());
}


TEST_F(RmaWindowT,
       operatorLeftShift_itemAtFullDetail_printsData)
{
    string expected("RmaWindow {\n"
                    "    id: 42\n"
                    "    comm: Communicator { id: 0, name: \"world_comm\", comm set: 0, parent: <undef>, flags: 0 (NONE) }\n"
                    "}");

    ostringstream result;
    result << setDetail(DETAIL_FULL)
           << *mWindow;
    EXPECT_EQ(expected, result.str());
}


// --- Helper ---------------------------------------------------------------

RmaWindowT::RmaWindowT()
{
    // Create dummy SystemNode (used as parent)
    mSystemNode =
        cxxslb::make_unique< SystemNode >(0,
                                          String::UNDEFINED,
                                          String::UNDEFINED,
                                          nullptr);

    // Create dummy LocationGroup (used as parent)
    mLocationGroup =
        cxxslb::make_unique< Process >(0,
                                       String::UNDEFINED,
                                       mSystemNode.get(),
                                       nullptr);

    // Create locations
    std::vector< const Location* > setMembers;
    setMembers.reserve(16);
    for (uint32_t id = 0; id < 16; ++id)
    {
        mStrings.addItem(
            cxxslb::make_unique< String >(id,
                                          "Rank " + std::to_string(id)));

        mLocations.addItem(
            cxxslb::make_unique< Location >(id,
                                            *mStrings.getItemById(id),
                                            Location::TYPE_CPU_THREAD,
                                            0,
                                            mLocationGroup.get()));
        setMembers.push_back(mLocations.getItemById(id));
    }

    // Create communication location set
    mCommLocationSet =
        cxxslb::make_unique< CommLocationSet >(0,
                                               String::UNDEFINED,
                                               Paradigm::MPI,
                                               setMembers);

    // Create "world" communication set
    vector< uint32_t > ranks;
    ranks.reserve(16);
    for (uint32_t rank = 0; rank < 16; ++rank)
    {
        ranks.push_back(rank);
    }
    mStrings.addItem(cxxslb::make_unique< String >(16, "world_group"));
    mCommSet =
        cxxslb::make_unique< CommSet >(0,
                                       *mStrings.getItemById(16),
                                       GroupingSet::PROPERTY_WORLD,
                                       *mCommLocationSet,
                                       ranks);

    // Create "world" communicator
    mStrings.addItem(cxxslb::make_unique< String >(17, "world_comm"));
    mComm =
        cxxslb::make_unique< Communicator >(0,
                                            *mStrings.getItemById(17),
                                            Communicator::PROPERTY_NONE,
                                            *mCommSet,
                                            nullptr);

    // Create RMA window
    mWindow =
        cxxslb::make_unique< RmaWindow >(42, *mComm);
}
}    // unnamed namespace
