/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2024                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2014                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  Copyright (c) 2019-2023                                                **
**  RWTH Aachen University, IT Center                                      **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "Patterns_gen.h"

#include <map>
#include <string>

#include <UTILS_Error.h>

#include <cubew_cube.h>

#include <pearl/Callback.h>
#include <pearl/CallbackManager.h>
#include <pearl/Event.h>

#include "AnalyzeTask.h"
#include "CbData.h"
#include "TypedPattern.h"
#include "scout_types.h"

#if defined(_MPI)
    #include "LongTermDelayPattern.h"
#endif    // _MPI

using namespace std;
using namespace pearl;


#line 17 "Common.pattern"

#include <sys/stat.h>

#include <algorithm>
#include <cstddef>
#include <cstdint>
#include <cstdio>
#include <cstring>
#include <iostream>
#include <limits>
#include <list>
#include <numeric>
#include <set>
#include <sstream>
#include <utility>

#include <cxxslb/memory.h>

#define SCALASCA_DEBUG_MODULE_NAME    SCOUT
#include <UTILS_Debug.h>

#include <pearl/AmListenerFactory.h>
#include <pearl/AmRequestFactory.h>
#include <pearl/Buffer.h>
#include <pearl/GlobalDefs.h>
#include <pearl/Leave_rep.h>
#include <pearl/Location.h>
#include <pearl/Region.h>
#include <pearl/String.h>
#include <pearl/ThreadEnd_rep.h>
#include <pearl/pearl_replay.h>

#include "Callstack.h"
#include "DelayOps.h"
#include "LockTracking.h"
#include "Quantile.h"
#include "ReportData.h"
#include "Roles.h"
#include "Settings.h"
#include "SynchpointHandler.h"
#include "TmapCacheHandler.h"
#include "WaitstateStats.h"
#include "user_events.h"

#if defined(_MPI)
    #include <cinttypes>

    #include <pearl/AmNull.h>
    #include <pearl/AmRuntime.h>
    #include <pearl/Communicator.h>
    #include <pearl/EventSet.h>
    #include <pearl/LocalData.h>
    #include <pearl/MpiAmListener.h>
    #include <pearl/MpiAmRequest.h>
    #include <pearl/MpiCollEnd_rep.h>
    #include <pearl/MpiGroup.h>
    #include <pearl/MpiHelpers.h>
    #include <pearl/MpiMessage.h>
    #include <pearl/RemoteData.h>
    #include <pearl/RemoteEventSet.h>
    #include <pearl/RmaEpoch.h>
    #include <pearl/RmaWindow.h>

    #include "AmLockContention.h"
    #include "AmSeverityExchange.h"
    #include "AmSyncpointExchange.h"
    #include "AmWaitForProgressRequest.h"
    #include "AmWaitForProgressResponse.h"
    #include "LockEpochQueue.h"
    #include "MpiDatatypes.h"
    #include "MpiDelayOps.h"
    #include "MpiOperators.h"
    #include "WaitstateCandidates.h"
#endif    // _MPI
#if defined(_OPENMP)
    #include <omp.h>

    #include "OmpDelayOps.h"
    #include "OmpEventHandler.h"
#endif    // _OPENMP


// --- Local variables --------------------------------------------------

#if defined(_OPENMP)
    namespace
    {
    int omp_nest_level = 0;
    }    // unnamed namespace
#endif   // _OPENMP


/// Maximum number of coefficients used for quantile approximation
#define NUMBER_COEFF    60

/* For statistics debugging: #define WRITE_CONTROL_VALUES */


namespace scout
{
namespace metric
{
/*-------------------------------------------------------------------------*/
/**
 *  @class Time
 *  @brief Time metric.
 *
 *  Implements the "Time" metric.
 */
/*-------------------------------------------------------------------------*/

class Time
    : public TypedPattern< double >
{
    public:
        /// @name Callback registration
        /// @{

        /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
        ///
        void
        registerCallbacks(CallbackManagerMap& cbmanagers) override;

        /// @}
        /// @name Pattern information access
        /// @{

        /// @copydoc scout::Pattern::getId()
        ///
        long
        getId() const override
        {
            return PAT_TIME;
        }

        /// @copydoc scout::Pattern::getParent()
        ///
        long
        getParent() const override
        {
            return PAT_NONE;
        }

        /// @copydoc scout::Pattern::getName()
        ///
        string
        getName() const override
        {
            return "Time";
        }

        /// @copydoc scout::Pattern::getUniqueName()
        ///
        string
        getUniqueName() const override
        {
            return "time";
        }

        /// @copydoc scout::Pattern::getDescription()
        ///
        string
        getDescription() const override
        {
            return "Total CPU allocation time (includes time allocated"
                   " for idle threads)";
        }

        /// @copydoc scout::Pattern::getUnit()
        ///
        string
        getUnit() const override
        {
            return "sec";
        }

        /// @copydoc scout::Pattern::getStorageMode()
        ///
        CubeMetricType
        getStorageMode() const override
        {
            return CUBE_METRIC_INCLUSIVE;
        }

        /// @}


    private:
        /// @name Callback functions (main)
        /// @{

        /// @brief Callback function for GROUP_LEAVE events.
        ///
        /// Callback function triggered by GROUP_LEAVE events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        group_leave_cb(const CallbackManager& cbmanager,
                       int                    user_event,
                       const Event&           event,
                       CallbackData*          cdata);

        /// @brief Callback function for THREAD_TASK_COMPLETE events.
        ///
        /// Callback function triggered by THREAD_TASK_COMPLETE events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        thread_task_complete_cb(const CallbackManager& cbmanager,
                                int                    user_event,
                                const Event&           event,
                                CallbackData*          cdata);

        /// @brief Callback function for OMP_MGMT_FORK events.
        ///
        /// Callback function triggered by OMP_MGMT_FORK events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        omp_mgmt_fork_cb(const CallbackManager& cbmanager,
                         int                    user_event,
                         const Event&           event,
                         CallbackData*          cdata);

        /// @brief Callback function for OMP_MGMT_JOIN events.
        ///
        /// Callback function triggered by OMP_MGMT_JOIN events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        omp_mgmt_join_cb(const CallbackManager& cbmanager,
                         int                    user_event,
                         const Event&           event,
                         CallbackData*          cdata);

        /// @brief Callback function for FINISHED events.
        ///
        /// Callback function triggered by FINISHED events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        finished_cb(const CallbackManager& cbmanager,
                    int                    user_event,
                    const Event&           event,
                    CallbackData*          cdata);

        /// @}
};


// --- Callback registration ------------------------------------------------

void
Time::registerCallbacks(CallbackManagerMap& cbmanagers)
{
    TypedPattern< double >::registerCallbacks(cbmanagers);

    CallbackManagerMap::iterator    it;
    CallbackManagerMap::mapped_type cbmanager;

    // Register main callbacks
    it = cbmanagers.find("");
    UTILS_ASSERT(it != cbmanagers.end());

    cbmanager = it->second;
    cbmanager->register_callback(
        GROUP_LEAVE,
        PEARL_create_callback(
            this,
            &Time::group_leave_cb));
    cbmanager->register_callback(
        THREAD_TASK_COMPLETE,
        PEARL_create_callback(
            this,
            &Time::thread_task_complete_cb));
    cbmanager->register_callback(
        OMP_MGMT_FORK,
        PEARL_create_callback(
            this,
            &Time::omp_mgmt_fork_cb));
    cbmanager->register_callback(
        OMP_MGMT_JOIN,
        PEARL_create_callback(
            this,
            &Time::omp_mgmt_join_cb));
    cbmanager->register_callback(
        FINISHED,
        PEARL_create_callback(
            this,
            &Time::finished_cb));
}


// --- Callback functions (main) --------------------------------------------

void
Time::group_leave_cb(const CallbackManager& cbmanager,
                     const int              user_event,
                     const Event&           event,
                     CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 47 "Generic.pattern"

    Event enter = data->mCallstack->top();

    INC_METRIC_VALUE(SCALASCA_DEBUG_TIME_METRIC,
                     enter.get_cnode(),
                     event->getTimestamp() - enter->getTimestamp() - data->mCallstack->getCurrentTimeOffset());
}


void
Time::thread_task_complete_cb(const CallbackManager& cbmanager,
                              const int              user_event,
                              const Event&           event,
                              CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 55 "Generic.pattern"

    // Updating stub nodes in calling tree (only for the implicit task)
    data->mIdle = event->getTimestamp() - data->mCallstack->getStartTime() - data->mCallstack->getTotalTimeOffset();

    INC_METRIC_VALUE(SCALASCA_DEBUG_TIME_METRIC,
                     data->mDefs->get_cnode(data->mCallstack->getStubNodeId()),
                     data->mIdle);

    cbmanager.notify(TASK_CORRECTION, event, data);
}


void
Time::omp_mgmt_fork_cb(const CallbackManager& cbmanager,
                       const int              user_event,
                       const Event&           event,
                       CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 66 "Generic.pattern"

    INC_METRIC_VALUE(SCALASCA_DEBUG_TIME_METRIC,
                     event.get_cnode(),
                     data->mIdle);
}


void
Time::omp_mgmt_join_cb(const CallbackManager& cbmanager,
                       const int              user_event,
                       const Event&           event,
                       CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 72 "Generic.pattern"

    INC_METRIC_VALUE(SCALASCA_DEBUG_TIME_METRIC,
                     event.get_cnode(),
                     data->mIdle);
}


void
Time::finished_cb(const CallbackManager& cbmanager,
                  const int              user_event,
                  const Event&           event,
                  CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 78 "Generic.pattern"

    // Combined loop over all callpaths to:
    // (1) Inclusify profile of non-master threads outside of parallel regions
    // (2) Fix the task root value by attributing its children

    uint32_t thread_id = data->mTrace->get_location().getThreadId();

    uint32_t count = data->mDefs->numCallpaths();
    for (uint32_t id = 0; id < count; ++id)
    {
        Callpath* callpath = data->mDefs->get_cnode(id);

        // Fix severity of artificial TASKS/THREADS callpath
        if (  (callpath->getRegion() == data->mDefs->getTaskRootRegion())
           || (callpath->getRegion() == data->mDefs->getThreadRootRegion()))
        {
            for (uint32_t i = 0; i < callpath->numChildren(); ++i)
            {
                uint32_t childId = callpath->getChild(i).getId();

                INC_METRIC_VALUE(SCALASCA_DEBUG_TIME_METRIC,
                                 callpath,
                                 m_severity[data->mDefs->get_cnode(childId)]);
            }

            // Propagate severity of artificial THREADS callpath to
            // program callpath if existent
            if (callpath->getRegion() == data->mDefs->getThreadRootRegion())
            {
                Callpath* parent = callpath->getParent();
                if (parent)
                {
                    INC_METRIC_VALUE(SCALASCA_DEBUG_TIME_METRIC,
                                     parent,
                                     m_severity[callpath]);
                }
            }
        }

        // skip in case of thread 0
        if (thread_id != 0)
        {
            // Does the callpath refer to an OpenMP parallel region?
            if (!is_omp_parallel(callpath->getRegion()))
            {
                continue;
            }

            // Propagate severity value "upwards"
            double    value  = m_severity[callpath];
            Callpath* parent = callpath->getParent();
            while (parent)
            {
                INC_METRIC_VALUE(SCALASCA_DEBUG_TIME_METRIC,
                                 parent,
                                 value);

                parent = parent->getParent();
            }
        }
    }
}


/*-------------------------------------------------------------------------*/
/**
 *  @class Visits
 *  @brief Visits metric.
 *
 *  Implements the "Visits" metric.
 */
/*-------------------------------------------------------------------------*/

class Visits
    : public TypedPattern< uint64_t >
{
    public:
        /// @name Callback registration
        /// @{

        /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
        ///
        void
        registerCallbacks(CallbackManagerMap& cbmanagers) override;

        /// @}
        /// @name Pattern information access
        /// @{

        /// @copydoc scout::Pattern::getId()
        ///
        long
        getId() const override
        {
            return PAT_VISITS;
        }

        /// @copydoc scout::Pattern::getParent()
        ///
        long
        getParent() const override
        {
            return PAT_NONE;
        }

        /// @copydoc scout::Pattern::getName()
        ///
        string
        getName() const override
        {
            return "Visits";
        }

        /// @copydoc scout::Pattern::getUniqueName()
        ///
        string
        getUniqueName() const override
        {
            return "visits";
        }

        /// @copydoc scout::Pattern::getDescription()
        ///
        string
        getDescription() const override
        {
            return "Number of visits";
        }

        /// @copydoc scout::Pattern::getUnit()
        ///
        string
        getUnit() const override
        {
            return "occ";
        }

        /// @copydoc scout::Pattern::getStorageMode()
        ///
        CubeMetricType
        getStorageMode() const override
        {
            return CUBE_METRIC_EXCLUSIVE;
        }

        /// @}


    private:
        /// @name Callback functions (main)
        /// @{

        /// @brief Callback function for GROUP_ENTER events.
        ///
        /// Callback function triggered by GROUP_ENTER events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        group_enter_cb(const CallbackManager& cbmanager,
                       int                    user_event,
                       const Event&           event,
                       CallbackData*          cdata);

        /// @brief Callback function for THREAD_TASK_COMPLETE events.
        ///
        /// Callback function triggered by THREAD_TASK_COMPLETE events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        thread_task_complete_cb(const CallbackManager& cbmanager,
                                int                    user_event,
                                const Event&           event,
                                CallbackData*          cdata);

        /// @}
};


// --- Callback registration ------------------------------------------------

void
Visits::registerCallbacks(CallbackManagerMap& cbmanagers)
{
    TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

    CallbackManagerMap::iterator    it;
    CallbackManagerMap::mapped_type cbmanager;

    // Register main callbacks
    it = cbmanagers.find("");
    UTILS_ASSERT(it != cbmanagers.end());

    cbmanager = it->second;
    cbmanager->register_callback(
        GROUP_ENTER,
        PEARL_create_callback(
            this,
            &Visits::group_enter_cb));
    cbmanager->register_callback(
        THREAD_TASK_COMPLETE,
        PEARL_create_callback(
            this,
            &Visits::thread_task_complete_cb));
}


// --- Callback functions (main) --------------------------------------------

void
Visits::group_enter_cb(const CallbackManager& cbmanager,
                       const int              user_event,
                       const Event&           event,
                       CallbackData* const    cdata)
{
    #line 181 "Generic.pattern"

    INC_METRIC_VALUE(SCALASCA_DEBUG_VISITS_METRIC,
                     event.get_cnode(),
                     1);
}


void
Visits::thread_task_complete_cb(const CallbackManager& cbmanager,
                                const int              user_event,
                                const Event&           event,
                                CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 187 "Generic.pattern"

    // updating visits for the task stub nodes on the implicit task
    // current mode: one stub node per instance and one visit per instance
    INC_METRIC_VALUE(SCALASCA_DEBUG_VISITS_METRIC,
                     data->mDefs->get_cnode(data->mCallstack->getStubNodeId()),
                     1);
}


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Async
     *  @brief Asyncronous Driver metric.
     *
     *  Implements the "Asyncronous Driver" metric.
     */
    /*---------------------------------------------------------------------*/

    class Async
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_ASYNC;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_NONE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "ASYNC";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "async";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Driver for asynchronous pattern detection";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @copydoc scout::Pattern::isVisible()
            ///
            bool
            isVisible() const override
            {
                return false;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for PREPARE events.
            ///
            /// Callback function triggered by PREPARE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            prepare_cb(const CallbackManager& cbmanager,
                       int                    user_event,
                       const Event&           event,
                       CallbackData*          cdata);

            /// @brief Callback function for FINISHED events.
            ///
            /// Callback function triggered by FINISHED events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            finished_cb(const CallbackManager& cbmanager,
                        int                    user_event,
                        const Event&           event,
                        CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Async::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            PREPARE,
            PEARL_create_callback(
                this,
                &Async::prepare_cb));
        cbmanager->register_callback(
            FINISHED,
            PEARL_create_callback(
                this,
                &Async::finished_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    Async::prepare_cb(const CallbackManager& cbmanager,
                      const int              user_event,
                      const Event&           event,
                      CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 35 "Async.pattern"

        pearl::AmRuntime::getInstance().setCbdata(data);
        pearl::AmRuntime::getInstance().setCbmanager(cbmanager);

        pearl::AmRuntime::getInstance().attach(AmNull::getInstance());
        pearl::AmRuntime::getInstance().attach(AmLockContention::getInstance());
        pearl::AmRuntime::getInstance().attach(AmSeverityExchange::getInstance());
        pearl::AmRuntime::getInstance().attach(AmSyncpointExchange::getInstance());
        pearl::AmRuntime::getInstance().attach(AmWaitForProgressRequest::getInstance());
        pearl::AmRuntime::getInstance().attach(AmWaitForProgressResponse::getInstance());

        // Register MPI specific factory callbacks
        pearl::AmRequestFactory::registerCallback(Paradigm::MPI,
                                                  &MpiAmRequest::create);
        pearl::AmListenerFactory::registerCallback(Paradigm::MPI,
                                                   &MpiAmListener::create);
    }


    void
    Async::finished_cb(const CallbackManager& cbmanager,
                       const int              user_event,
                       const Event&           event,
                       CallbackData* const    cdata)
    {
        #line 53 "Async.pattern"

        AmRuntime::getInstance().shutdown();
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_InitCompletion
     *  @brief MPI Initialization Completion Time metric.
     *
     *  Implements the "MPI Initialization Completion Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_InitCompletion
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_INIT_COMPLETION;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_INIT_EXIT;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Initialization Completion";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_init_completion";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time needed to finish MPI initialization";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for PRE_INIT events.
            ///
            /// Callback function triggered by PRE_INIT events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            pre_init_cb(const CallbackManager& cbmanager,
                        int                    user_event,
                        const Event&           event,
                        CallbackData*          cdata);

            /// @brief Callback function for INIT events.
            ///
            /// Callback function triggered by INIT events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            init_cb(const CallbackManager& cbmanager,
                    int                    user_event,
                    const Event&           event,
                    CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_InitCompletion::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            PRE_INIT,
            PEARL_create_callback(
                this,
                &MPI_InitCompletion::pre_init_cb));
        cbmanager->register_callback(
            INIT,
            PEARL_create_callback(
                this,
                &MPI_InitCompletion::init_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_InitCompletion::pre_init_cb(const CallbackManager& cbmanager,
                                    const int              user_event,
                                    const Event&           event,
                                    CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 124 "MPI.pattern"

        const CollectiveInfo& ci = data->mCollinfo;

        // Validate clock condition
        const timestamp_t localEndT = event->getTimestamp();
        if (ci.latest.mTime > localEndT)
        {
            UTILS_DLOG_LEVEL(SCALASCA_DEBUG_CLOCK_VIOLATION)
                << "Clock violation @ " << *event.get_cnode()
                << ": " << (ci.latest.mTime - localEndT);

            cbmanager.notify(CCV_COLL, event, data);
        }

        // Restrict wait & completion time to time spent in operation
        const timestamp_t latestBeginT = min(ci.latest.mTime, localEndT);
        const timestamp_t earliestEndT = max(ci.earliest_end.mTime, latestBeginT);

        // Calculate completion time, but no wait time.  Otherwise, delay
        // costs may be assigned to MPI_Init(_thread) to propagate further,
        // which are then never assigned to any call path.
        data->mCompletion = max(0.0, localEndT - earliestEndT);
    }


    void
    MPI_InitCompletion::init_cb(const CallbackManager& cbmanager,
                                const int              user_event,
                                const Event&           event,
                                CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 148 "MPI.pattern"

        if (data->mCompletion > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             data->mCompletion);
        }

        // There will always be completion time at MPI_Init; notify all
        // processes
        cbmanager.notify(INIT_COMPL, event, data);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_FinalizeWait
     *  @brief Wait at MPI Finalize Time metric.
     *
     *  Implements the "Wait at MPI Finalize Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_FinalizeWait
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_FINALIZE_WAIT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_INIT_EXIT;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Wait at Finalize";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_finalize_wait";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Waiting time in front of MPI_Finalize";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for PRE_FINALIZE events.
            ///
            /// Callback function triggered by PRE_FINALIZE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            pre_finalize_cb(const CallbackManager& cbmanager,
                            int                    user_event,
                            const Event&           event,
                            CallbackData*          cdata);

            /// @brief Callback function for FINALIZE events.
            ///
            /// Callback function triggered by FINALIZE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            finalize_cb(const CallbackManager& cbmanager,
                        int                    user_event,
                        const Event&           event,
                        CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_FinalizeWait::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            PRE_FINALIZE,
            PEARL_create_callback(
                this,
                &MPI_FinalizeWait::pre_finalize_cb));
        cbmanager->register_callback(
            FINALIZE,
            PEARL_create_callback(
                this,
                &MPI_FinalizeWait::finalize_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_FinalizeWait::pre_finalize_cb(const CallbackManager& cbmanager,
                                      const int              user_event,
                                      const Event&           event,
                                      CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 187 "MPI.pattern"

        const CollectiveInfo& ci = data->mCollinfo;

        // Validate clock condition
        const timestamp_t localEndT = event->getTimestamp();
        if (ci.latest.mTime > localEndT)
        {
            UTILS_DLOG_LEVEL(SCALASCA_DEBUG_CLOCK_VIOLATION)
                << "Clock violation @ " << *event.get_cnode()
                << ": " << (ci.latest.mTime - localEndT);

            cbmanager.notify(CCV_COLL, event, data);
        }

        // Restrict wait & completion time to time spent in operation
        const timestamp_t latestBeginT = min(ci.latest.mTime, localEndT);
        const timestamp_t earliestEndT = max(ci.earliest_end.mTime, latestBeginT);

        // Calculate wait & completion time
        data->mIdle       = max(0.0, latestBeginT - ci.my.mTime);
        data->mCompletion = max(0.0, localEndT - earliestEndT);
    }


    void
    MPI_FinalizeWait::finalize_cb(const CallbackManager& cbmanager,
                                  const int              user_event,
                                  const Event&           event,
                                  CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 210 "MPI.pattern"

        if (data->mIdle > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             data->mIdle);
        }

        // There will always be waiting time at MPI_Finalize; notify all
        // processes
        cbmanager.notify(WAIT_FINALIZE, event, data);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_WinMgmt
     *  @brief MPI Communicator Management Time metric.
     *
     *  Implements the "MPI Communicator Management Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_WinMgmt
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_MGMT_COMM;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_MANAGEMENT;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Communicator Management";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_mgmt_comm";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent in MPI Communicator management calls";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @copydoc scout::Pattern::isVisible()
            ///
            bool
            isVisible() const override
            {
                return false;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for PRE_COLL_COMM_CREATE events.
            ///
            /// Callback function triggered by PRE_COLL_COMM_CREATE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            pre_coll_comm_create_cb(const CallbackManager& cbmanager,
                                    int                    user_event,
                                    const Event&           event,
                                    CallbackData*          cdata);

            /// @brief Callback function for PRE_COLL_COMM_FREE events.
            ///
            /// Callback function triggered by PRE_COLL_COMM_FREE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            pre_coll_comm_free_cb(const CallbackManager& cbmanager,
                                  int                    user_event,
                                  const Event&           event,
                                  CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_WinMgmt::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            PRE_COLL_COMM_CREATE,
            PEARL_create_callback(
                this,
                &MPI_WinMgmt::pre_coll_comm_create_cb));
        cbmanager->register_callback(
            PRE_COLL_COMM_FREE,
            PEARL_create_callback(
                this,
                &MPI_WinMgmt::pre_coll_comm_free_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_WinMgmt::pre_coll_comm_create_cb(const CallbackManager& cbmanager,
                                         const int              user_event,
                                         const Event&           event,
                                         CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 253 "MPI.pattern"

        if (data->mIsSyncpoint)
        {
            const CollectiveInfo& ci = data->mCollinfo;

            // Validate clock condition
            const timestamp_t localEndT = event->getTimestamp();
            if (ci.latest.mTime > localEndT)
            {
                UTILS_DLOG_LEVEL(SCALASCA_DEBUG_CLOCK_VIOLATION)
                    << "Clock violation @ " << *event.get_cnode()
                    << ": " << (ci.latest.mTime - localEndT);

                cbmanager.notify(CCV_COLL, event, data);
            }

            // Restrict wait & completion time to time spent in operation
            const timestamp_t latestBeginT = min(ci.latest.mTime, localEndT);
            const timestamp_t earliestEndT = max(ci.earliest_end.mTime, latestBeginT);

            // Calculate wait time
            data->mIdle       = max(0.0, latestBeginT - ci.my.mTime);
            data->mCompletion = max(0.0, localEndT - earliestEndT);
        }
    }


    void
    MPI_WinMgmt::pre_coll_comm_free_cb(const CallbackManager& cbmanager,
                                       const int              user_event,
                                       const Event&           event,
                                       CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 279 "MPI.pattern"

        if (data->mIsSyncpoint)
        {
            const CollectiveInfo& ci = data->mCollinfo;

            // Validate clock condition
            const timestamp_t localEndT = event->getTimestamp();
            if (ci.latest.mTime > localEndT)
            {
                UTILS_DLOG_LEVEL(SCALASCA_DEBUG_CLOCK_VIOLATION)
                    << "Clock violation @ " << *event.get_cnode()
                    << ": " << (ci.latest.mTime - localEndT);

                cbmanager.notify(CCV_COLL, event, data);
            }

            // Restrict wait & completion time to time spent in operation
            const timestamp_t latestBeginT = min(ci.latest.mTime, localEndT);
            const timestamp_t earliestEndT = max(ci.earliest_end.mTime, latestBeginT);

            // Calculate wait time
            data->mIdle       = max(0.0, latestBeginT - ci.my.mTime);
            data->mCompletion = max(0.0, localEndT - earliestEndT);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_WaitCommCreate
     *  @brief MPI Wait at Communicator Create Time metric.
     *
     *  Implements the "MPI Wait at Communicator Create Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_WaitCommCreate
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_WAIT_COMM_CREATE;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_MGMT_COMM;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Wait at Communicator Create";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_wait_comm_create";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time due to inherent synchronization in MPI"
                       " communicator creation operations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for COLL_COMM_CREATE events.
            ///
            /// Callback function triggered by COLL_COMM_CREATE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            coll_comm_create_cb(const CallbackManager& cbmanager,
                                int                    user_event,
                                const Event&           event,
                                CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_WaitCommCreate::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            COLL_COMM_CREATE,
            PEARL_create_callback(
                this,
                &MPI_WaitCommCreate::coll_comm_create_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_WaitCommCreate::coll_comm_create_cb(const CallbackManager& cbmanager,
                                            const int              user_event,
                                            const Event&           event,
                                            CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 335 "MPI.pattern"

        if (data->mIdle > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             data->mIdle);
        }

        // Notify all processes to allow for collective communication (e.g.,
        // during most-severe instance tracking)
        cbmanager.notify(WAIT_COMM_CREATE, event, data);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_CommCreateCompletion
     *  @brief MPI Communicator Create Completion Time metric.
     *
     *  Implements the "MPI Communicator Create Completion Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_CommCreateCompletion
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_COMM_CREATE_COMPLETION;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_MGMT_COMM;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Communicator Create Completion";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_comm_create_completion";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time needed to finish an MPI communicator creation"
                       " operation";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for COLL_COMM_CREATE events.
            ///
            /// Callback function triggered by COLL_COMM_CREATE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            coll_comm_create_cb(const CallbackManager& cbmanager,
                                int                    user_event,
                                const Event&           event,
                                CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_CommCreateCompletion::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            COLL_COMM_CREATE,
            PEARL_create_callback(
                this,
                &MPI_CommCreateCompletion::coll_comm_create_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_CommCreateCompletion::coll_comm_create_cb(const CallbackManager& cbmanager,
                                                  const int              user_event,
                                                  const Event&           event,
                                                  CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 376 "MPI.pattern"

        if (data->mCompletion > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             data->mCompletion);
        }

        // Notify all processes to allow for collective communication (e.g.,
        // during most-severe instance tracking)
        cbmanager.notify(COMM_CREATE_COMPL, event, data);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_WaitCommFree
     *  @brief MPI Wait at Communicator Free Time metric.
     *
     *  Implements the "MPI Wait at Communicator Free Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_WaitCommFree
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_WAIT_COMM_FREE;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_MGMT_COMM;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Wait at Communicator Free";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_wait_comm_free";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time due to inherent synchronization in MPI"
                       " communicator destruction operations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for COLL_COMM_FREE events.
            ///
            /// Callback function triggered by COLL_COMM_FREE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            coll_comm_free_cb(const CallbackManager& cbmanager,
                              int                    user_event,
                              const Event&           event,
                              CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_WaitCommFree::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            COLL_COMM_FREE,
            PEARL_create_callback(
                this,
                &MPI_WaitCommFree::coll_comm_free_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_WaitCommFree::coll_comm_free_cb(const CallbackManager& cbmanager,
                                        const int              user_event,
                                        const Event&           event,
                                        CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 411 "MPI.pattern"

        if (data->mIdle > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             data->mIdle);
        }

        // Notify all processes to allow for collective communication (e.g.,
        // during most-severe instance tracking)
        cbmanager.notify(WAIT_COMM_FREE, event, data);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_SyncCollective
     *  @brief MPI Collective Synchronization Time metric.
     *
     *  Implements the "MPI Collective Synchronization Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_SyncCollective
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_SYNC_COLLECTIVE;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_SYNCHRONIZATION;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Collective Synchronization";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_sync_collective";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent in MPI barriers";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @copydoc scout::Pattern::isVisible()
            ///
            bool
            isVisible() const override
            {
                return false;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for PRE_SYNC_COLL events.
            ///
            /// Callback function triggered by PRE_SYNC_COLL events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            pre_sync_coll_cb(const CallbackManager& cbmanager,
                             int                    user_event,
                             const Event&           event,
                             CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_SyncCollective::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            PRE_SYNC_COLL,
            PEARL_create_callback(
                this,
                &MPI_SyncCollective::pre_sync_coll_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_SyncCollective::pre_sync_coll_cb(const CallbackManager& cbmanager,
                                         const int              user_event,
                                         const Event&           event,
                                         CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 521 "MPI.pattern"

        const CollectiveInfo& ci = data->mCollinfo;

        // Validate clock condition
        const timestamp_t localEndT = event->getTimestamp();
        if (ci.latest.mTime > localEndT)
        {
            UTILS_DLOG_LEVEL(SCALASCA_DEBUG_CLOCK_VIOLATION)
                << "Clock violation @ " << *event.get_cnode()
                << ": " << (ci.latest.mTime - localEndT);

            cbmanager.notify(CCV_COLL, event, data);
        }

        // Restrict wait & completion time to time spent in operation
        const timestamp_t latestBeginT = min(ci.latest.mTime, localEndT);
        const timestamp_t earliestEndT = max(ci.earliest_end.mTime, latestBeginT);

        // Calculate wait & completion time
        data->mIdle       = max(0.0, latestBeginT - ci.my.mTime);
        data->mCompletion = max(0.0, localEndT - earliestEndT);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_BarrierWait
     *  @brief Wait at MPI Barrier Time metric.
     *
     *  Implements the "Wait at MPI Barrier Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_BarrierWait
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_BARRIER_WAIT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_SYNC_COLLECTIVE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Wait at Barrier";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_barrier_wait";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Waiting time in front of MPI barriers";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for SYNC_COLL events.
            ///
            /// Callback function triggered by SYNC_COLL events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            sync_coll_cb(const CallbackManager& cbmanager,
                         int                    user_event,
                         const Event&           event,
                         CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_BarrierWait::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            SYNC_COLL,
            PEARL_create_callback(
                this,
                &MPI_BarrierWait::sync_coll_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_BarrierWait::sync_coll_cb(const CallbackManager& cbmanager,
                                  const int              user_event,
                                  const Event&           event,
                                  CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 572 "MPI.pattern"

        if (data->mIdle > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             data->mIdle);
        }

        // There will always be waiting time at barriers; notify all processes
        cbmanager.notify(WAIT_BARRIER, event, data);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_BarrierCompletion
     *  @brief MPI Barrier Completion Time metric.
     *
     *  Implements the "MPI Barrier Completion Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_BarrierCompletion
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_BARRIER_COMPLETION;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_SYNC_COLLECTIVE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Barrier Completion";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_barrier_completion";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time needed to finish MPI barriers";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for SYNC_COLL events.
            ///
            /// Callback function triggered by SYNC_COLL events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            sync_coll_cb(const CallbackManager& cbmanager,
                         int                    user_event,
                         const Event&           event,
                         CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_BarrierCompletion::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            SYNC_COLL,
            PEARL_create_callback(
                this,
                &MPI_BarrierCompletion::sync_coll_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_BarrierCompletion::sync_coll_cb(const CallbackManager& cbmanager,
                                        const int              user_event,
                                        const Event&           event,
                                        CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 611 "MPI.pattern"

        if (data->mCompletion > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             data->mCompletion);
        }

        // There will always be completion time at barriers; notify all
        // processes
        cbmanager.notify(BARRIER_COMPL, event, data);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_P2P
     *  @brief MPI Point-to-point Communication Time metric.
     *
     *  Implements the "MPI Point-to-point Communication Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_P2P
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_POINT2POINT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_COMMUNICATION;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI point-to-point communication";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_point2point";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent in MPI point-to-point communication";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @copydoc scout::Pattern::isVisible()
            ///
            bool
            isVisible() const override
            {
                return false;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for POST_RECV events.
            ///
            /// Callback function triggered by POST_RECV events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            post_recv_cb(const CallbackManager& cbmanager,
                         int                    user_event,
                         const Event&           event,
                         CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_P2P::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            POST_RECV,
            PEARL_create_callback(
                this,
                &MPI_P2P::post_recv_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_P2P::post_recv_cb(const CallbackManager& cbmanager,
                          const int              user_event,
                          const Event&           event,
                          CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 679 "MPI.pattern"

        RemoteEvent send = data->mRemote->getEvent(ROLE_SEND);

        // Validate clock condition
        const auto sendT = send->getTimestamp();
        const auto recvT = event->getTimestamp();
        if (sendT > recvT)
        {
            UTILS_DLOG_LEVEL(SCALASCA_DEBUG_CLOCK_VIOLATION)
                << "Clock violation @ " << *event.get_cnode()
                << ": " << (sendT - recvT);

            cbmanager.notify(CCV_P2P, event, data);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_LateSender
     *  @brief MPI Late Sender Time metric.
     *
     *  Implements the "MPI Late Sender Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_LateSender
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_LATESENDER;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_POINT2POINT;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Late Sender";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_latesender";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time in MPI point-to-point receive operation waiting"
                       " for a message";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            #line 728 "MPI.pattern"

            // Tracks the most severe Late Sender instance found in multi-receive
            // completion calls (e.g., `MPI_Waitall`)
            std::unique_ptr< LateSenderCandidate > mMaxCandidate;

            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for POST_RECV events.
            ///
            /// Callback function triggered by POST_RECV events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            post_recv_cb(const CallbackManager& cbmanager,
                         int                    user_event,
                         const Event&           event,
                         CallbackData*          cdata);

            /// @brief Callback function for GROUP_LEAVE events.
            ///
            /// Callback function triggered by GROUP_LEAVE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            group_leave_cb(const CallbackManager& cbmanager,
                           int                    user_event,
                           const Event&           event,
                           CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_LateSender::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            POST_RECV,
            PEARL_create_callback(
                this,
                &MPI_LateSender::post_recv_cb));
        cbmanager->register_callback(
            GROUP_LEAVE,
            PEARL_create_callback(
                this,
                &MPI_LateSender::group_leave_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_LateSender::post_recv_cb(const CallbackManager& cbmanager,
                                 const int              user_event,
                                 const Event&           event,
                                 CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 734 "MPI.pattern"

        // Receive completions in `MPI_Test*` do not synchronize
        const auto enterRecv = data->mLocal->getEvent(ROLE_ENTER_RECV);
        auto&      region    = enterRecv->getRegion();
        if (is_mpi_testx(region))
        {
            return;
        }

        // Calculate waiting time, restricted to time spent in operation
        const auto enterSend  = data->mRemote->getEvent(ROLE_ENTER_SEND);
        const auto enterSendT = enterSend->getTimestamp();
        const auto leaveRecv  = data->mLocal->getEvent(ROLE_LEAVE_RECV);
        const auto leaveRecvT = leaveRecv->getTimestamp();

        const auto minEnterSendT = std::min(enterSendT, leaveRecvT);
        const auto enterRecvT    = enterRecv->getTimestamp();

        data->mIdle = std::max(0.0, minEnterSendT - enterRecvT);

        if (data->mIdle > 0)
        {
            // Calculate completion time
            // Since a Late Sender has been found, leaveSendT > enterSendT > enterRecvT
            // holds.  Thus, the completion time as calculated below is guaranteed to
            // not exceed the time spent in the operation.
            const auto leaveSend  = data->mRemote->getEvent(ROLE_LEAVE_SEND);
            const auto leaveSendT = leaveSend->getTimestamp();

            data->mCompletion = std::max(0.0, leaveRecvT - leaveSendT);

            const LateSenderCandidate candidate(data->mIdle,
                                                data->mCompletion,
                                                data->mRemote->getEvent(ROLE_SEND),
                                                event);
            if (is_mpi_wait_multi(region))
            {
                // Track most severe instance within this call
                if (mMaxCandidate)
                {
                    *mMaxCandidate = std::max(*mMaxCandidate, candidate);
                }
                else
                {
                    mMaxCandidate = cxxslb::make_unique< LateSenderCandidate >(candidate);
                }
            }
            else
            {
                INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                                 event.get_cnode(),
                                 data->mIdle);

                candidate.confirm(cbmanager, event, data);
            }
        }
    }


    void
    MPI_LateSender::group_leave_cb(const CallbackManager& cbmanager,
                                   const int              user_event,
                                   const Event&           event,
                                   CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 792 "MPI.pattern"

        if (!mMaxCandidate)
        {
            return;
        }

        INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                         event.get_cnode(),
                         mMaxCandidate->mIdle);

        mMaxCandidate->announce(cbmanager, event, data);
        mMaxCandidate->confirm(cbmanager, mMaxCandidate->mRecv, data);
        mMaxCandidate.reset();
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_LateSenderWO
     *  @brief MPI Late Sender, Wrong Order Time metric.
     *
     *  Implements the "MPI Late Sender, Wrong Order Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_LateSenderWO
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_LATESENDER_WO;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_LATESENDER;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Late Sender, wrong order";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_latesender_wo";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Late Sender situation due to MPI messages received"
                       " in the wrong order";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            #line 838 "MPI.pattern"

            /// Maximum number of "Late Sender" instances to be considered for
            /// detecting wrong-order situations
            static constexpr uint32_t MAX_INSTANCES = 100;

            /// Confirmed "Late Sender" instances -> LSWO candidates (implicitly
            /// sorted by RECV timestamp, as instances are detected "in order"
            /// during forward replay)
            std::list< LswoCandidate > mLateSenders;

            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for LATE_SENDER events.
            ///
            /// Callback function triggered by LATE_SENDER events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            late_sender_cb(const CallbackManager& cbmanager,
                           int                    user_event,
                           const Event&           event,
                           CallbackData*          cdata);

            /// @brief Callback function for POST_RECV events.
            ///
            /// Callback function triggered by POST_RECV events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            post_recv_cb(const CallbackManager& cbmanager,
                         int                    user_event,
                         const Event&           event,
                         CallbackData*          cdata);

            /// @brief Callback function for FINISHED events.
            ///
            /// Callback function triggered by FINISHED events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            finished_cb(const CallbackManager& cbmanager,
                        int                    user_event,
                        const Event&           event,
                        CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_LateSenderWO::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            LATE_SENDER,
            PEARL_create_callback(
                this,
                &MPI_LateSenderWO::late_sender_cb));
        cbmanager->register_callback(
            POST_RECV,
            PEARL_create_callback(
                this,
                &MPI_LateSenderWO::post_recv_cb));
        cbmanager->register_callback(
            FINISHED,
            PEARL_create_callback(
                this,
                &MPI_LateSenderWO::finished_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_LateSenderWO::late_sender_cb(const CallbackManager& cbmanager,
                                     const int              user_event,
                                     const Event&           event,
                                     CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 849 "MPI.pattern"

        if (mLateSenders.size() == MAX_INSTANCES)
        {
            mLateSenders.pop_front();
        }

        mLateSenders.emplace_back(data->mIdle,
                                  data->mCompletion,
                                  data->mRemote->getEvent(ROLE_SEND_LS),
                                  data->mLocal->getEvent(ROLE_RECV_LS));
    }


    void
    MPI_LateSenderWO::post_recv_cb(const CallbackManager& cbmanager,
                                   const int              user_event,
                                   const Event&           event,
                                   CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 861 "MPI.pattern"

        const auto send = data->mRemote->getEvent(ROLE_SEND);

        auto candidate = mLateSenders.begin();
        while (candidate != mLateSenders.end())
        {
            if (candidate->mSend->getTimestamp() > send->getTimestamp())
            {
                INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                                 candidate->mRecv.get_cnode(),
                                 candidate->mIdle);

                candidate->confirm(cbmanager, event, data);

                candidate = mLateSenders.erase(candidate);
                continue;
            }

            ++candidate;
        }
    }


    void
    MPI_LateSenderWO::finished_cb(const CallbackManager& cbmanager,
                                  const int              user_event,
                                  const Event&           event,
                                  CallbackData* const    cdata)
    {
        #line 883 "MPI.pattern"

        mLateSenders.clear();
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_LswoSame
     *  @brief MPI Late Sender, Wrong Order Time / Same Source metric.
     *
     *  Implements the "MPI Late Sender, Wrong Order Time / Same Source" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_LswoSame
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_LSWO_SAME;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_LATESENDER_WO;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Late Sender, wrong order (same source)";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_lswo_same";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Late Sender, Wrong Order situation due to MPI"
                       " messages received from the same source";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for LATE_SENDER_WO events.
            ///
            /// Callback function triggered by LATE_SENDER_WO events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            late_sender_wo_cb(const CallbackManager& cbmanager,
                              int                    user_event,
                              const Event&           event,
                              CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_LswoSame::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            LATE_SENDER_WO,
            PEARL_create_callback(
                this,
                &MPI_LswoSame::late_sender_wo_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_LswoSame::late_sender_wo_cb(const CallbackManager& cbmanager,
                                    const int              user_event,
                                    const Event&           event,
                                    CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 940 "MPI.pattern"

        const auto lsSend   = data->mRemote->getEvent(ROLE_SEND_LSWO);
        const auto currSend = data->mRemote->getEvent(ROLE_SEND);

        if (lsSend.get_location() == currSend.get_location())
        {
            const auto lsRecv = data->mLocal->getEvent(ROLE_RECV_LSWO);

            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             lsRecv.get_cnode(),
                             data->mIdle);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_LateReceiver
     *  @brief MPI Late Receiver Time metric.
     *
     *  Implements the "MPI Late Receiver Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_LateReceiver
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_LATERECEIVER;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_POINT2POINT;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Late Receiver";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_latereceiver";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time in MPI point-to-point send operation waiting"
                       " for the receiver to become ready";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            #line 987 "MPI.pattern"

            struct multiwait_info_t
            {
                int    sendcount;
                Event  maxevent;
                double maxidle;
                double maxcmpl;
            };


            typedef std::set< pearl::Event, const pearl::EventKeyCompare >
                event_set_t;
            typedef std::map< Event, multiwait_info_t, pearl::EventKeyCompare >
                multiwait_map_t;

            event_set_t     m_ls_set;
            multiwait_map_t m_multiwait_map;

            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for LATE_SENDER_CANDIDATE events.
            ///
            /// Callback function triggered by LATE_SENDER_CANDIDATE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            late_sender_candidate_cb(const CallbackManager& cbmanager,
                                     int                    user_event,
                                     const Event&           event,
                                     CallbackData*          cdata);

            /// @}
            /// @name Callback functions (bws)
            /// @{

            /// @brief Callback function for POST_SEND events.
            ///
            /// Callback function triggered by POST_SEND events
            /// during the @"bws@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            post_send_bws_cb(const CallbackManager& cbmanager,
                             int                    user_event,
                             const Event&           event,
                             CallbackData*          cdata);

            /// @brief Callback function for MPI_SEND_COMPLETE events.
            ///
            /// Callback function triggered by MPI_SEND_COMPLETE events
            /// during the @"bws@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_send_complete_bws_cb(const CallbackManager& cbmanager,
                                     int                    user_event,
                                     const Event&           event,
                                     CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_LateReceiver::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            LATE_SENDER_CANDIDATE,
            PEARL_create_callback(
                this,
                &MPI_LateReceiver::late_sender_candidate_cb));

        // Register bws callbacks
        it = cbmanagers.find("bws");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            POST_SEND,
            PEARL_create_callback(
                this,
                &MPI_LateReceiver::post_send_bws_cb));
        cbmanager->register_callback(
            MPI_SEND_COMPLETE,
            PEARL_create_callback(
                this,
                &MPI_LateReceiver::mpi_send_complete_bws_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_LateReceiver::late_sender_candidate_cb(const CallbackManager& cbmanager,
                                               const int              user_event,
                                               const Event&           event,
                                               CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 1006 "MPI.pattern"

        m_ls_set.insert(data->mCallstack->top());
    }


    // --- Callback functions (bws) -----------------------------------------

    void
    MPI_LateReceiver::post_send_bws_cb(const CallbackManager& cbmanager,
                                       const int              user_event,
                                       const Event&           event,
                                       CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 1012 "MPI.pattern"

        // backward replay: msg received from destination here

        Event       enter_sendcmp = data->mLocal->getEvent(ROLE_ENTER_SENDCMP);
        Event       leave_sendcmp = data->mLocal->getEvent(ROLE_LEAVE_SENDCMP);
        RemoteEvent enter_recvreq = data->mRemote->getEvent(ROLE_ENTER_RECVREQ);
        RemoteEvent leave_recvreq = data->mRemote->getEvent(ROLE_LEAVE_RECVREQ);

        // Calculate waiting time
        data->mIdle =
            leave_sendcmp->getTimestamp() <= enter_recvreq->getTimestamp()
            ? 0
            : enter_recvreq->getTimestamp() - enter_sendcmp->getTimestamp();

        if (data->mIdle > 0)
        {
            data->mCompletion =
                leave_sendcmp->getTimestamp() > leave_recvreq->getTimestamp()
                ? 0
                : leave_sendcmp->getTimestamp() - leave_recvreq->getTimestamp();
        }

        Region region = enter_sendcmp->getRegion();

        if (  is_mpi_block_send(region)
           || is_mpi_wait_single(region))
        {
            if (data->mIdle > 0)
            {
                INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                                 enter_sendcmp.get_cnode(),
                                 data->mIdle);

                RemoteEvent recv = data->mRemote->getEvent(ROLE_RECV);
                data->mRemote->addEvent(recv, ROLE_RECV_LR);
                data->mLocal->addEvent(enter_sendcmp, ROLE_ENTER_SENDCMP_LR);

                cbmanager.notify(LATE_RECEIVER, event, data);
            }
        }
        else if (is_mpi_wait_multi(region))
        {
            multiwait_map_t::iterator it = m_multiwait_map.find(enter_sendcmp);

            if (it == m_multiwait_map.end())
            {
                return;
            }

            if (data->mIdle > it->second.maxidle)
            {
                it->second.maxidle  = data->mIdle;
                it->second.maxcmpl  = data->mCompletion;
                it->second.maxevent = event;
            }

            --it->second.sendcount;

            if (it->second.sendcount == 0)
            {
                if (it->second.maxidle > 0.0)
                {
                    INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                                     it->first.get_cnode(),
                                     it->second.maxidle);

                    data->mIdle       = it->second.maxidle;
                    data->mCompletion = it->second.maxcmpl;

                    RemoteEvent recv = data->mRemote->getEvent(ROLE_RECV);
                    data->mRemote->addEvent(recv, ROLE_RECV_LR);
                    data->mLocal->addEvent(enter_sendcmp, ROLE_ENTER_SENDCMP_LR);

                    cbmanager.notify(LATE_RECEIVER, it->second.maxevent, data);
                }

                m_multiwait_map.erase(it);
            }
        }
    }


    void
    MPI_LateReceiver::mpi_send_complete_bws_cb(const CallbackManager& cbmanager,
                                               const int              user_event,
                                               const Event&           event,
                                               CallbackData* const    cdata)
    {
        #line 1093 "MPI.pattern"

        Event enter = event.enterptr();

        if (  is_mpi_wait_multi(enter->getRegion())
           && !m_ls_set.count(enter))
        {
            multiwait_map_t::iterator it = m_multiwait_map.find(enter);

            if (it == m_multiwait_map.end())
            {
                multiwait_info_t info;

                info.sendcount = 1;
                info.maxidle   = 0.0;

                m_multiwait_map.insert(std::make_pair(enter, info));
            }
            else
            {
                it->second.sendcount++;
            }
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_Collective
     *  @brief MPI Collective Communication Time metric.
     *
     *  Implements the "MPI Collective Communication Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_Collective
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_COLLECTIVE;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_COMMUNICATION;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI collective communication";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_collective";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent in MPI collective communication";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @copydoc scout::Pattern::isVisible()
            ///
            bool
            isVisible() const override
            {
                return false;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for PRE_COLL_12N events.
            ///
            /// Callback function triggered by PRE_COLL_12N events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            pre_coll_12n_cb(const CallbackManager& cbmanager,
                            int                    user_event,
                            const Event&           event,
                            CallbackData*          cdata);

            /// @brief Callback function for PRE_COLL_N2N events.
            ///
            /// Callback function triggered by PRE_COLL_N2N events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            pre_coll_n2n_cb(const CallbackManager& cbmanager,
                            int                    user_event,
                            const Event&           event,
                            CallbackData*          cdata);

            /// @brief Callback function for PRE_COLL_N21 events.
            ///
            /// Callback function triggered by PRE_COLL_N21 events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            pre_coll_n21_cb(const CallbackManager& cbmanager,
                            int                    user_event,
                            const Event&           event,
                            CallbackData*          cdata);

            /// @brief Callback function for PRE_COLL_SCAN events.
            ///
            /// Callback function triggered by PRE_COLL_SCAN events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            pre_coll_scan_cb(const CallbackManager& cbmanager,
                             int                    user_event,
                             const Event&           event,
                             CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_Collective::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            PRE_COLL_12N,
            PEARL_create_callback(
                this,
                &MPI_Collective::pre_coll_12n_cb));
        cbmanager->register_callback(
            PRE_COLL_N2N,
            PEARL_create_callback(
                this,
                &MPI_Collective::pre_coll_n2n_cb));
        cbmanager->register_callback(
            PRE_COLL_N21,
            PEARL_create_callback(
                this,
                &MPI_Collective::pre_coll_n21_cb));
        cbmanager->register_callback(
            PRE_COLL_SCAN,
            PEARL_create_callback(
                this,
                &MPI_Collective::pre_coll_scan_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_Collective::pre_coll_12n_cb(const CallbackManager& cbmanager,
                                    const int              user_event,
                                    const Event&           event,
                                    CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 1164 "MPI.pattern"

        if (data->mIsSyncpoint)
        {
            const CollectiveInfo& ci = data->mCollinfo;

            // Validate clock condition
            const timestamp_t localEndT = event->getTimestamp();
            if (ci.root.mTime > localEndT)
            {
                UTILS_DLOG_LEVEL(SCALASCA_DEBUG_CLOCK_VIOLATION)
                    << "Clock violation @ " << *event.get_cnode()
                    << ": " << (ci.latest.mTime - localEndT);

                cbmanager.notify(CCV_COLL, event, data);
            }

            // Restrict wait & completion time to time spent in operation
            const timestamp_t rootBeginT   = min(ci.root.mTime, localEndT);
            const timestamp_t earliestEndT = max(ci.earliest_end.mTime, rootBeginT);

            // Calculate wait & completion time
            data->mIdle       = max(0.0, rootBeginT - ci.my.mTime);
            data->mCompletion = max(0.0, localEndT - earliestEndT);
        }
    }


    void
    MPI_Collective::pre_coll_n2n_cb(const CallbackManager& cbmanager,
                                    const int              user_event,
                                    const Event&           event,
                                    CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 1190 "MPI.pattern"

        if (data->mIsSyncpoint)
        {
            const CollectiveInfo& ci = data->mCollinfo;

            // Validate clock condition
            const timestamp_t localEndT = event->getTimestamp();
            if (ci.latest.mTime > localEndT)
            {
                UTILS_DLOG_LEVEL(SCALASCA_DEBUG_CLOCK_VIOLATION)
                    << "Clock violation @ " << *event.get_cnode()
                    << ": " << (ci.latest.mTime - localEndT);

                cbmanager.notify(CCV_COLL, event, data);
            }

            // Restrict wait & completion time to time spent in operation
            const timestamp_t latestBeginT = min(ci.latest.mTime, localEndT);
            const timestamp_t earliestEndT = max(ci.earliest_end.mTime, latestBeginT);

            // Calculate wait & completion time
            data->mIdle       = max(0.0, latestBeginT - ci.my.mTime);
            data->mCompletion = max(0.0, localEndT - earliestEndT);
        }
    }


    void
    MPI_Collective::pre_coll_n21_cb(const CallbackManager& cbmanager,
                                    const int              user_event,
                                    const Event&           event,
                                    CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 1216 "MPI.pattern"

        if (data->mIsSyncpoint)
        {
            const CollectiveInfo& ci = data->mCollinfo;

            // Validate clock condition
            const timestamp_t localEndT = event->getTimestamp();
            if (ci.latest.mTime > localEndT)
            {
                UTILS_DLOG_LEVEL(SCALASCA_DEBUG_CLOCK_VIOLATION)
                    << "Clock violation @ " << *event.get_cnode()
                    << ": " << (ci.latest.mTime - localEndT);

                cbmanager.notify(CCV_COLL, event, data);
            }

            // Restrict wait & completion time to time spent in operation
            const timestamp_t latestBeginT = min(ci.latest.mTime, localEndT);
            const timestamp_t earliestEndT = max(ci.earliest_end.mTime, latestBeginT);

            // Calculate wait & completion time
            data->mIdle       = max(0.0, latestBeginT - ci.my.mTime);
            data->mCompletion = max(0.0, localEndT - earliestEndT);
        }
    }


    void
    MPI_Collective::pre_coll_scan_cb(const CallbackManager& cbmanager,
                                     const int              user_event,
                                     const Event&           event,
                                     CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 1242 "MPI.pattern"

        if (data->mIsSyncpoint)
        {
            const CollectiveInfo& ci = data->mCollinfo;

            // Validate clock condition
            const timestamp_t localEndT = event->getTimestamp();
            if (ci.latest.mTime > localEndT)
            {
                UTILS_DLOG_LEVEL(SCALASCA_DEBUG_CLOCK_VIOLATION)
                    << "Clock violation @ " << *event.get_cnode()
                    << ": " << (ci.latest.mTime - localEndT);

                cbmanager.notify(CCV_COLL, event, data);
            }

            // Restrict wait time to time spent in operation
            const timestamp_t latestBeginT = min(ci.latest.mTime, localEndT);

            // Calculate wait time
            data->mIdle = max(0.0, latestBeginT - ci.my.mTime);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_EarlyReduce
     *  @brief MPI Early Reduce Time metric.
     *
     *  Implements the "MPI Early Reduce Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_EarlyReduce
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_EARLYREDUCE;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_COLLECTIVE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Early Reduce";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_earlyreduce";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Waiting time due to an early receiver in MPI n-to-1"
                       " operations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for COLL_N21 events.
            ///
            /// Callback function triggered by COLL_N21 events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            coll_n21_cb(const CallbackManager& cbmanager,
                        int                    user_event,
                        const Event&           event,
                        CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_EarlyReduce::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            COLL_N21,
            PEARL_create_callback(
                this,
                &MPI_EarlyReduce::coll_n21_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_EarlyReduce::coll_n21_cb(const CallbackManager& cbmanager,
                                 const int              user_event,
                                 const Event&           event,
                                 CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 1293 "MPI.pattern"

        if (data->mIdle > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             data->mIdle);

            // Early Reduce time only occurs on the root process, i.e.,
            // most-severe instance tracking can be performed locally
            cbmanager.notify(EARLY_REDUCE, event, data);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_EarlyScan
     *  @brief MPI Early Scan Time metric.
     *
     *  Implements the "MPI Early Scan Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_EarlyScan
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_EARLYSCAN;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_COLLECTIVE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Early Scan";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_earlyscan";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Waiting time due to an early receiver in an MPI scan"
                       " operation";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for COLL_SCAN events.
            ///
            /// Callback function triggered by COLL_SCAN events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            coll_scan_cb(const CallbackManager& cbmanager,
                         int                    user_event,
                         const Event&           event,
                         CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_EarlyScan::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            COLL_SCAN,
            PEARL_create_callback(
                this,
                &MPI_EarlyScan::coll_scan_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_EarlyScan::coll_scan_cb(const CallbackManager& cbmanager,
                                const int              user_event,
                                const Event&           event,
                                CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 1330 "MPI.pattern"

        if (data->mIdle > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             data->mIdle);
        }

        // All processes need to take part in the most-severe instance tracking
        cbmanager.notify(EARLY_SCAN, event, data);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_LateBroadcast
     *  @brief MPI Late Broadcast Time metric.
     *
     *  Implements the "MPI Late Broadcast Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_LateBroadcast
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_LATEBROADCAST;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_COLLECTIVE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Late Broadcast";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_latebroadcast";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Waiting time due to a late sender in MPI 1-to-n"
                       " operations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for COLL_12N events.
            ///
            /// Callback function triggered by COLL_12N events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            coll_12n_cb(const CallbackManager& cbmanager,
                        int                    user_event,
                        const Event&           event,
                        CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_LateBroadcast::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            COLL_12N,
            PEARL_create_callback(
                this,
                &MPI_LateBroadcast::coll_12n_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_LateBroadcast::coll_12n_cb(const CallbackManager& cbmanager,
                                   const int              user_event,
                                   const Event&           event,
                                   CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 1369 "MPI.pattern"

        if (data->mIdle > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             data->mIdle);
        }

        // Notify all processes to allow for collective communication (e.g.,
        // during most-severe instance tracking)
        cbmanager.notify(LATE_BCAST, event, data);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_WaitNxN
     *  @brief MPI Wait at N x N Time metric.
     *
     *  Implements the "MPI Wait at N x N Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_WaitNxN
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_WAIT_NXN;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_COLLECTIVE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Wait at N x N";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_wait_nxn";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time due to inherent synchronization in MPI n-to-n"
                       " operations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for COLL_N2N events.
            ///
            /// Callback function triggered by COLL_N2N events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            coll_n2n_cb(const CallbackManager& cbmanager,
                        int                    user_event,
                        const Event&           event,
                        CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_WaitNxN::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            COLL_N2N,
            PEARL_create_callback(
                this,
                &MPI_WaitNxN::coll_n2n_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_WaitNxN::coll_n2n_cb(const CallbackManager& cbmanager,
                             const int              user_event,
                             const Event&           event,
                             CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 1415 "MPI.pattern"

        if (data->mIdle > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             data->mIdle);
        }

        // Notify all processes to allow for collective communication (e.g.,
        // during most-severe instance tracking)
        cbmanager.notify(WAIT_NXN, event, data);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_NxNCompletion
     *  @brief MPI N x N Completion Time metric.
     *
     *  Implements the "MPI N x N Completion Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_NxNCompletion
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_NXN_COMPLETION;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_COLLECTIVE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI N x N Completion";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_nxn_completion";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time needed to finish an MPI n-to-n collective"
                       " operation";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for COLL_N2N events.
            ///
            /// Callback function triggered by COLL_N2N events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            coll_n2n_cb(const CallbackManager& cbmanager,
                        int                    user_event,
                        const Event&           event,
                        CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_NxNCompletion::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            COLL_N2N,
            PEARL_create_callback(
                this,
                &MPI_NxNCompletion::coll_n2n_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_NxNCompletion::coll_n2n_cb(const CallbackManager& cbmanager,
                                   const int              user_event,
                                   const Event&           event,
                                   CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 1455 "MPI.pattern"

        if (data->mCompletion > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             data->mCompletion);
        }

        // Notify all processes to allow for collective communication (e.g.,
        // during most-severe instance tracking)
        cbmanager.notify(NXN_COMPL, event, data);
    }
#endif    // _MPI


#if defined(_OPENMP)
    /*---------------------------------------------------------------------*/
    /**
     *  @class OMP_Mgmt
     *  @brief OpenMP Thread Management Time metric.
     *
     *  Implements the "OpenMP Thread Management Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class OMP_Mgmt
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_OMP_MANAGEMENT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_OMP_TIME;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "OMP Management";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "omp_management";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent in OpenMP thread management";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for THREAD_FORK events.
            ///
            /// Callback function triggered by THREAD_FORK events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            thread_fork_cb(const CallbackManager& cbmanager,
                           int                    user_event,
                           const Event&           event,
                           CallbackData*          cdata);

            /// @brief Callback function for THREAD_JOIN events.
            ///
            /// Callback function triggered by THREAD_JOIN events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            thread_join_cb(const CallbackManager& cbmanager,
                           int                    user_event,
                           const Event&           event,
                           CallbackData*          cdata);

            /// @brief Callback function for GROUP_ENTER events.
            ///
            /// Callback function triggered by GROUP_ENTER events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            group_enter_cb(const CallbackManager& cbmanager,
                           int                    user_event,
                           const Event&           event,
                           CallbackData*          cdata);

            /// @brief Callback function for GROUP_LEAVE events.
            ///
            /// Callback function triggered by GROUP_LEAVE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            group_leave_cb(const CallbackManager& cbmanager,
                           int                    user_event,
                           const Event&           event,
                           CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    OMP_Mgmt::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            THREAD_FORK,
            PEARL_create_callback(
                this,
                &OMP_Mgmt::thread_fork_cb));
        cbmanager->register_callback(
            THREAD_JOIN,
            PEARL_create_callback(
                this,
                &OMP_Mgmt::thread_join_cb));
        cbmanager->register_callback(
            GROUP_ENTER,
            PEARL_create_callback(
                this,
                &OMP_Mgmt::group_enter_cb));
        cbmanager->register_callback(
            GROUP_LEAVE,
            PEARL_create_callback(
                this,
                &OMP_Mgmt::group_leave_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    OMP_Mgmt::thread_fork_cb(const CallbackManager& cbmanager,
                             const int              user_event,
                             const Event&           event,
                             CallbackData* const    cdata)
    {
        #line 143 "OMP.pattern"

        // Increment nesting level
        omp_nest_level++;
    }


    void
    OMP_Mgmt::thread_join_cb(const CallbackManager& cbmanager,
                             const int              user_event,
                             const Event&           event,
                             CallbackData* const    cdata)
    {
        #line 148 "OMP.pattern"

        // Decrement nesting level
        omp_nest_level--;
    }


    void
    OMP_Mgmt::group_enter_cb(const CallbackManager& cbmanager,
                             const int              user_event,
                             const Event&           event,
                             CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 153 "OMP.pattern"

        // Are we entering an OpenMP parallel region?
        const Region& region = event->getRegion();
        if (!is_omp_parallel(region))
        {
            return;
        }

        // Determine timestamp of FORK event and the earliest ENTER into
        // the parallel region
        timestamp_t enterT = event->getTimestamp();

        static timestamp_t forkT;
        static timestamp_t earliestEnterT;
        #pragma omp master
        {
            Event fork = event.prev();
            while (fork->getType() != THREAD_FORK)
            {
                --fork;
            }
            forkT          = fork->getTimestamp();
            earliestEnterT = enterT;
        }
        #pragma omp barrier
        #pragma omp critical
        {
            earliestEnterT = std::min(earliestEnterT, enterT);
        }

        // Calculate thread fork and completion time
        #pragma omp barrier
        data->mIdle       = std::max(0.0, enterT - forkT);
        data->mCompletion = std::max(0.0, enterT - earliestEnterT);

        INC_METRIC_VALUE(SCALASCA_DEBUG_OVERHEAD_METRICS,
                         event.get_cnode(),
                         data->mIdle);

        cbmanager.notify(OMP_MGMT_FORK, event, data);
        #pragma omp barrier
    }


    void
    OMP_Mgmt::group_leave_cb(const CallbackManager& cbmanager,
                             const int              user_event,
                             const Event&           event,
                             CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 196 "OMP.pattern"

        // Are we leaving an OpenMP parallel region?
        Event         enter  = data->mCallstack->top();
        const Region& region = enter->getRegion();
        if (!is_omp_parallel(region))
        {
            return;
        }

        // Determine timestamp of JOIN event
        static timestamp_t joinTimestamp;
        #pragma omp master
        {
            Event join = event.next();
            while (join->getType() != THREAD_JOIN)
            {
                ++join;
            }

            joinTimestamp = join->getTimestamp();
        }

        // Calculate thread join time
        #pragma omp barrier
        data->mIdle = std::max(0.0, joinTimestamp - event->getTimestamp());

        INC_METRIC_VALUE(SCALASCA_DEBUG_OVERHEAD_METRICS,
                         enter.get_cnode(),
                         data->mIdle);

        cbmanager.notify(OMP_MGMT_JOIN, event, data);
        #pragma omp barrier
    }
#endif    // _OPENMP


#if defined(_OPENMP)
    /*---------------------------------------------------------------------*/
    /**
     *  @class OMP_Mgmt_Fork
     *  @brief OpenMP Thread Team Fork Time metric.
     *
     *  Implements the "OpenMP Thread Team Fork Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class OMP_Mgmt_Fork
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_OMP_FORK;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_OMP_MANAGEMENT;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "OMP Fork";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "omp_fork";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent in forking OpenMP thread teams";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for OMP_MGMT_FORK events.
            ///
            /// Callback function triggered by OMP_MGMT_FORK events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            omp_mgmt_fork_cb(const CallbackManager& cbmanager,
                             int                    user_event,
                             const Event&           event,
                             CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    OMP_Mgmt_Fork::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            OMP_MGMT_FORK,
            PEARL_create_callback(
                this,
                &OMP_Mgmt_Fork::omp_mgmt_fork_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    OMP_Mgmt_Fork::omp_mgmt_fork_cb(const CallbackManager& cbmanager,
                                    const int              user_event,
                                    const Event&           event,
                                    CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 248 "OMP.pattern"

        INC_METRIC_VALUE(SCALASCA_DEBUG_OVERHEAD_METRICS,
                         event.get_cnode(),
                         data->mIdle);
    }
#endif    // _OPENMP


#if defined(_OPENMP)
    /*---------------------------------------------------------------------*/
    /**
     *  @class OMP_Sync_Barrier
     *  @brief OpenMP Barrier Synchronization Time metric.
     *
     *  Implements the "OpenMP Barrier Synchronization Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class OMP_Sync_Barrier
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_OMP_BARRIER;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_OMP_SYNCHRONIZATION;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "OMP Barrier";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "omp_barrier";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent in OpenMP barrier synchronization";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @copydoc scout::Pattern::isVisible()
            ///
            bool
            isVisible() const override
            {
                return false;
            }

            /// @}


        private:
            #line 296 "OMP.pattern"

            double mTaskExecutionTime;

            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for TASK_CORRECTION events.
            ///
            /// Callback function triggered by TASK_CORRECTION events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            task_correction_cb(const CallbackManager& cbmanager,
                               int                    user_event,
                               const Event&           event,
                               CallbackData*          cdata);

            /// @brief Callback function for GROUP_ENTER events.
            ///
            /// Callback function triggered by GROUP_ENTER events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            group_enter_cb(const CallbackManager& cbmanager,
                           int                    user_event,
                           const Event&           event,
                           CallbackData*          cdata);

            /// @brief Callback function for GROUP_LEAVE events.
            ///
            /// Callback function triggered by GROUP_LEAVE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            group_leave_cb(const CallbackManager& cbmanager,
                           int                    user_event,
                           const Event&           event,
                           CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    OMP_Sync_Barrier::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            TASK_CORRECTION,
            PEARL_create_callback(
                this,
                &OMP_Sync_Barrier::task_correction_cb));
        cbmanager->register_callback(
            GROUP_ENTER,
            PEARL_create_callback(
                this,
                &OMP_Sync_Barrier::group_enter_cb));
        cbmanager->register_callback(
            GROUP_LEAVE,
            PEARL_create_callback(
                this,
                &OMP_Sync_Barrier::group_leave_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    OMP_Sync_Barrier::task_correction_cb(const CallbackManager& cbmanager,
                                         const int              user_event,
                                         const Event&           event,
                                         CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 300 "OMP.pattern"

        mTaskExecutionTime += data->mIdle;
    }


    void
    OMP_Sync_Barrier::group_enter_cb(const CallbackManager& cbmanager,
                                     const int              user_event,
                                     const Event&           event,
                                     CallbackData* const    cdata)
    {
        #line 304 "OMP.pattern"

        const Region& region = event.get_cnode()->getRegion();
        if (!is_omp_barrier(region))
        {
            return;
        }

        // Reset task execution time when entering the barrier
        mTaskExecutionTime = 0;
    }


    void
    OMP_Sync_Barrier::group_leave_cb(const CallbackManager& cbmanager,
                                     const int              user_event,
                                     const Event&           event,
                                     CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 315 "OMP.pattern"

        // Are we leaving an (implicit or explicit) OpenMP barrier?
        Event         enter  = data->mCallstack->top();
        const Region& region = enter->getRegion();
        if (!is_omp_barrier(region))
        {
            return;
        }

        // Skip OpenMP barrier waiting time calculation when running serially
        // (the barriers called in the algorithm cause ill side effects...)
        if (omp_nest_level == 0)
        {
            return;
        }

        // Determine latest enter time
        timestamp_t enterTime = enter->getTimestamp();

        // Implicitly shared timestamp for reduction
        static timestamp_t latestEnter;
        #pragma omp barrier
        #pragma omp master
        latestEnter = 0;
        #pragma omp barrier
        #pragma omp critical
        {
            if (latestEnter < enterTime)
            {
                latestEnter = enterTime;
            }
        }
        #pragma omp barrier

        // Calculate waiting time
        data->mIdle = std::max(0.0, latestEnter - enterTime - mTaskExecutionTime);

        // Notify sub-patterns
        if (is_omp_ebarrier(region))
        {
            cbmanager.notify(OMP_EBARRIER, event, data);
        }
        else if (is_omp_ibarrier(region))
        {
            cbmanager.notify(OMP_IBARRIER, event, data);
        }
    }
#endif    // _OPENMP


#if defined(_OPENMP)
    /*---------------------------------------------------------------------*/
    /**
     *  @class OMP_Sync_Ebarrier_Wait
     *  @brief Wait at Explicit OpenMP Barrier Time metric.
     *
     *  Implements the "Wait at Explicit OpenMP Barrier Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class OMP_Sync_Ebarrier_Wait
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_OMP_EBARRIER_WAIT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_OMP_EBARRIER;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "OMP Wait at Explicit Barrier";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "omp_ebarrier_wait";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Waiting time in front of explicit OpenMP barriers";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for OMP_EBARRIER events.
            ///
            /// Callback function triggered by OMP_EBARRIER events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            omp_ebarrier_cb(const CallbackManager& cbmanager,
                            int                    user_event,
                            const Event&           event,
                            CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    OMP_Sync_Ebarrier_Wait::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            OMP_EBARRIER,
            PEARL_create_callback(
                this,
                &OMP_Sync_Ebarrier_Wait::omp_ebarrier_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    OMP_Sync_Ebarrier_Wait::omp_ebarrier_cb(const CallbackManager& cbmanager,
                                            const int              user_event,
                                            const Event&           event,
                                            CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 413 "OMP.pattern"

        Event enter = data->mCallstack->top();

        if (data->mIdle > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             enter.get_cnode(),
                             data->mIdle);
        }

        // There will always be waiting time at barriers; all threads need
        // to take part in most-severe instance tracking
        cbmanager.notify(OMP_EBARRIER_WAIT, event, data);
    }
#endif    // _OPENMP


#if defined(_OPENMP)
    /*---------------------------------------------------------------------*/
    /**
     *  @class OMP_Sync_Ibarrier_Wait
     *  @brief Wait at Implicit OpenMP Barrier Time metric.
     *
     *  Implements the "Wait at Implicit OpenMP Barrier Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class OMP_Sync_Ibarrier_Wait
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_OMP_IBARRIER_WAIT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_OMP_IBARRIER;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "OMP Wait at Implicit Barrier";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "omp_ibarrier_wait";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Waiting time in front of implicit OpenMP barriers";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for OMP_IBARRIER events.
            ///
            /// Callback function triggered by OMP_IBARRIER events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            omp_ibarrier_cb(const CallbackManager& cbmanager,
                            int                    user_event,
                            const Event&           event,
                            CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    OMP_Sync_Ibarrier_Wait::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            OMP_IBARRIER,
            PEARL_create_callback(
                this,
                &OMP_Sync_Ibarrier_Wait::omp_ibarrier_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    OMP_Sync_Ibarrier_Wait::omp_ibarrier_cb(const CallbackManager& cbmanager,
                                            const int              user_event,
                                            const Event&           event,
                                            CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 478 "OMP.pattern"

        Event enter = data->mCallstack->top();

        if (data->mIdle > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             enter.get_cnode(),
                             data->mIdle);
        }

        // There will always be waiting time at barriers; all threads need
        // to take part in most-severe instance tracking
        cbmanager.notify(OMP_IBARRIER_WAIT, event, data);
    }
#endif    // _OPENMP


#if defined(_OPENMP)
    /*---------------------------------------------------------------------*/
    /**
     *  @class OMP_Lock_Contention_Critical
     *  @brief OpenMP Critical Contention Time metric.
     *
     *  Implements the "OpenMP Critical Contention Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class OMP_Lock_Contention_Critical
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_OMP_LOCK_CONTENTION_CRITICAL;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_OMP_CRITICAL;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "OMP Critical Contention";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "omp_lock_contention_critical";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time lost waiting for an OpenMP critical section to"
                       " become available";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for OMP_CONTENTION_CRITICAL events.
            ///
            /// Callback function triggered by OMP_CONTENTION_CRITICAL events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            omp_contention_critical_cb(const CallbackManager& cbmanager,
                                       int                    user_event,
                                       const Event&           event,
                                       CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    OMP_Lock_Contention_Critical::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            OMP_CONTENTION_CRITICAL,
            PEARL_create_callback(
                this,
                &OMP_Lock_Contention_Critical::omp_contention_critical_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    OMP_Lock_Contention_Critical::omp_contention_critical_cb(const CallbackManager& cbmanager,
                                                             const int              user_event,
                                                             const Event&           event,
                                                             CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 541 "OMP.pattern"

        if (data->mIdle > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             data->mIdle);

            // instance tracking
            cbmanager.notify(THREAD_LOCK_CONTENTION_OMPCRITICAL_WAIT, event, data);
        }
    }
#endif    // _OPENMP


#if defined(_OPENMP)
    /*---------------------------------------------------------------------*/
    /**
     *  @class OMP_Lock_Contention_API
     *  @brief OpenMP Lock API Contention Time metric.
     *
     *  Implements the "OpenMP Lock API Contention Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class OMP_Lock_Contention_API
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_OMP_LOCK_CONTENTION_API;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_OMP_LOCK_API;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "OMP Lock API Contention";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "omp_lock_contention_api";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time lost waiting for an OpenMP lock to be acquired";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for OMP_CONTENTION_API events.
            ///
            /// Callback function triggered by OMP_CONTENTION_API events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            omp_contention_api_cb(const CallbackManager& cbmanager,
                                  int                    user_event,
                                  const Event&           event,
                                  CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    OMP_Lock_Contention_API::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            OMP_CONTENTION_API,
            PEARL_create_callback(
                this,
                &OMP_Lock_Contention_API::omp_contention_api_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    OMP_Lock_Contention_API::omp_contention_api_cb(const CallbackManager& cbmanager,
                                                   const int              user_event,
                                                   const Event&           event,
                                                   CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 600 "OMP.pattern"

        if (data->mIdle > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             data->mIdle);

            // instance tracking
            cbmanager.notify(THREAD_LOCK_CONTENTION_OMPLOCK_WAIT, event, data);
        }
    }
#endif    // _OPENMP


#if defined(_OPENMP)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Pthread_Lock_Contention_Mutex_Lock
     *  @brief POSIX Threads Mutex Contention Time metric.
     *
     *  Implements the "POSIX Threads Mutex Contention Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class Pthread_Lock_Contention_Mutex_Lock
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_PTHREAD_LOCK_CONTENTION_MUTEX_LOCK;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_PTHREAD_LOCK_API;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "Pthread Mutex Lock Contention";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "pthread_lock_contention_mutex_lock";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time lost waiting for a POSIX threads mutex to be"
                       " acquired";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for PTHREAD_CONTENTION_MUTEX_LOCK events.
            ///
            /// Callback function triggered by PTHREAD_CONTENTION_MUTEX_LOCK events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            pthread_contention_mutex_lock_cb(const CallbackManager& cbmanager,
                                             int                    user_event,
                                             const Event&           event,
                                             CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Pthread_Lock_Contention_Mutex_Lock::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            PTHREAD_CONTENTION_MUTEX_LOCK,
            PEARL_create_callback(
                this,
                &Pthread_Lock_Contention_Mutex_Lock::pthread_contention_mutex_lock_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    Pthread_Lock_Contention_Mutex_Lock::pthread_contention_mutex_lock_cb(const CallbackManager& cbmanager,
                                                                         const int              user_event,
                                                                         const Event&           event,
                                                                         CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 134 "Pthread.pattern"

        if (data->mIdle > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             data->mIdle);

            // instance tracking
            cbmanager.notify(THREAD_LOCK_CONTENTION_PTHREADMUTEX_WAIT, event, data);
        }
    }
#endif    // _OPENMP


#if defined(_OPENMP)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Pthread_Lock_Contention_Conditional
     *  @brief POSIX Threads Condition Contention Time metric.
     *
     *  Implements the "POSIX Threads Condition Contention Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class Pthread_Lock_Contention_Conditional
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_PTHREAD_LOCK_CONTENTION_CONDITIONAL;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_PTHREAD_CONDITIONAL;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "Pthread Condition Lock Contention";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "pthread_lock_contention_conditional";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time lost waiting for a POSIX threads mutex to be"
                       " acquired in a condition operation";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for PTHREAD_CONTENTION_CONDITIONAL events.
            ///
            /// Callback function triggered by PTHREAD_CONTENTION_CONDITIONAL events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            pthread_contention_conditional_cb(const CallbackManager& cbmanager,
                                              int                    user_event,
                                              const Event&           event,
                                              CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Pthread_Lock_Contention_Conditional::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            PTHREAD_CONTENTION_CONDITIONAL,
            PEARL_create_callback(
                this,
                &Pthread_Lock_Contention_Conditional::pthread_contention_conditional_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    Pthread_Lock_Contention_Conditional::pthread_contention_conditional_cb(const CallbackManager& cbmanager,
                                                                           const int              user_event,
                                                                           const Event&           event,
                                                                           CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 192 "Pthread.pattern"

        if (data->mIdle > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             data->mIdle);

            // instance tracking
            cbmanager.notify(THREAD_LOCK_CONTENTION_PTHREADCONDITION_WAIT, event, data);
        }
    }
#endif    // _OPENMP


#if defined(_OPENMP)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Thread_LockContention
     *  @brief Thread Lock Contention Time metric.
     *
     *  Implements the "Thread Lock Contention Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class Thread_LockContention
        : public TypedPattern< double >
    {
        public:
            /// @name Constructors & destructor
            /// @{

            /// @copydoc scout::Pattern::~Pattern()
            ///
            ~Thread_LockContention() override
            {
                #line 61 "Thread.pattern"

                PEARL_Barrier();
                if (PEARL_GetThreadNumber() == 0)
                {
                    delete mOpenmpLockTracking;
                    delete mPthreadLockTracking;
                }
            }

            /// @}
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_THREAD_LOCK_CONTENTION;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_NONE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "Thread Lock Contention";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "thread_lock_contention";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent in thread lock contention";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @copydoc scout::Pattern::isVisible()
            ///
            bool
            isVisible() const override
            {
                return false;
            }

            /// @}


        protected:
            /// @name Pattern data management
            /// @{

            /// @brief Initialize local data.
            ///
            /// This method is implicitly called from registerCallbacks() and
            /// initializes pattern-local data.
            ///
            void
            init() override
            {
                TypedPattern< double >::init();

                #line 53 "Thread.pattern"

                if (PEARL_GetThreadNumber() == 0)
                {
                    mOpenmpLockTracking  = new scout::LockTracking();
                    mPthreadLockTracking = new scout::LockTracking();
                }
                PEARL_Barrier();
            }

            /// @}


        private:
            #line 29 "Thread.pattern"

            LockTracking*
            getLockTrackingVariant(Paradigm paradigm)
            {
                switch (paradigm)
                {
                    case Paradigm::OPENMP:
                        return mOpenmpLockTracking;

                    case Paradigm::PTHREAD:
                        return mPthreadLockTracking;

                    default:
                        return nullptr;
                }
            }

            static LockTracking* mOpenmpLockTracking;
            static LockTracking* mPthreadLockTracking;

            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for THREAD_ACQUIRE_LOCK events.
            ///
            /// Callback function triggered by THREAD_ACQUIRE_LOCK events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            thread_acquire_lock_cb(const CallbackManager& cbmanager,
                                   int                    user_event,
                                   const Event&           event,
                                   CallbackData*          cdata);

            /// @brief Callback function for THREAD_RELEASE_LOCK events.
            ///
            /// Callback function triggered by THREAD_RELEASE_LOCK events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            thread_release_lock_cb(const CallbackManager& cbmanager,
                                   int                    user_event,
                                   const Event&           event,
                                   CallbackData*          cdata);

            /// @}
    };


    // --- Static class data ------------------------------------------------

    #line 49 "Thread.pattern"

    LockTracking* Thread_LockContention::mOpenmpLockTracking;
    LockTracking* Thread_LockContention::mPthreadLockTracking;


    // --- Callback registration --------------------------------------------

    void
    Thread_LockContention::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            THREAD_ACQUIRE_LOCK,
            PEARL_create_callback(
                this,
                &Thread_LockContention::thread_acquire_lock_cb));
        cbmanager->register_callback(
            THREAD_RELEASE_LOCK,
            PEARL_create_callback(
                this,
                &Thread_LockContention::thread_release_lock_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    Thread_LockContention::thread_acquire_lock_cb(const CallbackManager& cbmanager,
                                                  const int              user_event,
                                                  const Event&           event,
                                                  CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 70 "Thread.pattern"

        Paradigm paradigm = getLockParadigm(event);

        LockTracking* lockMap = getLockTrackingVariant(paradigm);
        if (lockMap == nullptr)
        {
            // unsupported paradigm
            return;
        }

        // wait time=last release time  - respective Enter timestamp
        data->mIdle = lockMap->acquireLock(event)
                      - event.enterptr()->getTimestamp();

        LockType lockType = getLockType(event);

        // distribute time dependent on lock type
        if (lockType == LockType::OMP_LOCK_API)
        {
            cbmanager.notify(OMP_CONTENTION_API, event, data);
        }
        else if (lockType == LockType::OMP_CRITICAL)
        {
            cbmanager.notify(OMP_CONTENTION_CRITICAL, event, data);
        }
        else if (  (lockType == LockType::PTHREAD_MUTEX_LOCK)
                || (lockType == LockType::PTHREAD_MUTEX_TRYLOCK))
        {
            cbmanager.notify(PTHREAD_CONTENTION_MUTEX_LOCK, event, data);
        }
        else if (lockType == LockType::PTHREAD_CONDITION_WAIT)
        {
            cbmanager.notify(PTHREAD_CONTENTION_CONDITIONAL, event, data);
        }
    }


    void
    Thread_LockContention::thread_release_lock_cb(const CallbackManager& cbmanager,
                                                  const int              user_event,
                                                  const Event&           event,
                                                  CallbackData* const    cdata)
    {
        #line 106 "Thread.pattern"

        Paradigm paradigm = getLockParadigm(event);

        LockTracking* lockMap = getLockTrackingVariant(paradigm);
        if (lockMap == nullptr)
        {
            // unsupported paradigm
            return;
        }
        lockMap->releaseLock(event);
    }
#endif    // _OPENMP


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_SyncsSend
     *  @brief MPI Point-to-point Send Synchronization Operations metric.
     *
     *  Implements the "MPI Point-to-point Send Synchronization Operations" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_SyncsSend
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_SYNCS_SEND;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_SYNCS_P2P;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI P2P send synchronizations";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "syncs_send";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of MPI point-to-point send synchronization"
                       " operations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "occ";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for GROUP_SEND events.
            ///
            /// Callback function triggered by GROUP_SEND events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            group_send_cb(const CallbackManager& cbmanager,
                          int                    user_event,
                          const Event&           event,
                          CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_SyncsSend::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            GROUP_SEND,
            PEARL_create_callback(
                this,
                &MPI_SyncsSend::group_send_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_SyncsSend::group_send_cb(const CallbackManager& cbmanager,
                                 const int              user_event,
                                 const Event&           event,
                                 CallbackData* const    cdata)
    {
        #line 73 "Message.pattern"

        if (event->getBytesSent() == 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_INSTANCE_METRICS,
                             event.get_cnode(),
                             1);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_SyncsRecv
     *  @brief MPI Point-to-point Receive Synchronization Operations metric.
     *
     *  Implements the "MPI Point-to-point Receive Synchronization Operations" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_SyncsRecv
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_SYNCS_RECV;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_SYNCS_P2P;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI P2P recv synchronizations";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "syncs_recv";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of MPI point-to-point receive synchronization"
                       " operations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "occ";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for POST_RECV events.
            ///
            /// Callback function triggered by POST_RECV events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            post_recv_cb(const CallbackManager& cbmanager,
                         int                    user_event,
                         const Event&           event,
                         CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_SyncsRecv::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            POST_RECV,
            PEARL_create_callback(
                this,
                &MPI_SyncsRecv::post_recv_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_SyncsRecv::post_recv_cb(const CallbackManager& cbmanager,
                                const int              user_event,
                                const Event&           event,
                                CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 99 "Message.pattern"

        RemoteEvent send = data->mRemote->getEvent(ROLE_SEND);

        if (send->getBytesSent() == 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_INSTANCE_METRICS,
                             event.get_cnode(),
                             1);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_SyncsColl
     *  @brief MPI Collective Synchronizations metric.
     *
     *  Implements the "MPI Collective Synchronizations" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_SyncsColl
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_SYNCS_COLL;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_SYNCS;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI collective synchronizations";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "syncs_coll";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of MPI collective synchronization operations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "occ";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_COLLECTIVE_END events.
            ///
            /// Callback function triggered by MPI_COLLECTIVE_END events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_collective_end_cb(const CallbackManager& cbmanager,
                                  int                    user_event,
                                  const Event&           event,
                                  CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_SyncsColl::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_COLLECTIVE_END,
            PEARL_create_callback(
                this,
                &MPI_SyncsColl::mpi_collective_end_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_SyncsColl::mpi_collective_end_cb(const CallbackManager& cbmanager,
                                         const int              user_event,
                                         const Event&           event,
                                         CallbackData* const    cdata)
    {
        #line 138 "Message.pattern"

        if (  (event->getBytesSent() == 0)
           && (event->getBytesReceived() == 0))
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_INSTANCE_METRICS,
                             event.get_cnode(),
                             1);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_CommsSend
     *  @brief MPI Point-to-point Send Communication Operations metric.
     *
     *  Implements the "MPI Point-to-point Send Communication Operations" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_CommsSend
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_COMMS_SEND;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_COMMS_P2P;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI P2P send communications";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "comms_send";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of MPI point-to-point send communication"
                       " operations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "occ";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for GROUP_SEND events.
            ///
            /// Callback function triggered by GROUP_SEND events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            group_send_cb(const CallbackManager& cbmanager,
                          int                    user_event,
                          const Event&           event,
                          CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_CommsSend::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            GROUP_SEND,
            PEARL_create_callback(
                this,
                &MPI_CommsSend::group_send_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_CommsSend::group_send_cb(const CallbackManager& cbmanager,
                                 const int              user_event,
                                 const Event&           event,
                                 CallbackData* const    cdata)
    {
        #line 246 "Message.pattern"

        if (event->getBytesSent() != 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_INSTANCE_METRICS,
                             event.get_cnode(),
                             1);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_CommsRecv
     *  @brief MPI Point-to-point Receive Communication Operations metric.
     *
     *  Implements the "MPI Point-to-point Receive Communication Operations" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_CommsRecv
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_COMMS_RECV;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_COMMS_P2P;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI P2P recv communications";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "comms_recv";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of MPI point-to-point receive communication"
                       " operations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "occ";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for POST_RECV events.
            ///
            /// Callback function triggered by POST_RECV events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            post_recv_cb(const CallbackManager& cbmanager,
                         int                    user_event,
                         const Event&           event,
                         CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_CommsRecv::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            POST_RECV,
            PEARL_create_callback(
                this,
                &MPI_CommsRecv::post_recv_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_CommsRecv::post_recv_cb(const CallbackManager& cbmanager,
                                const int              user_event,
                                const Event&           event,
                                CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 272 "Message.pattern"

        RemoteEvent send = data->mRemote->getEvent(ROLE_SEND);

        if (send->getBytesSent() != 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_INSTANCE_METRICS,
                             event.get_cnode(),
                             1);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_CommsExch
     *  @brief MPI Collective Exchange Communications metric.
     *
     *  Implements the "MPI Collective Exchange Communications" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_CommsExch
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_COMMS_CXCH;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_COMMS_COLL;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI collective exchange communications";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "comms_cxch";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of MPI collective communications as source"
                       " and destination";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "occ";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_COLLECTIVE_END events.
            ///
            /// Callback function triggered by MPI_COLLECTIVE_END events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_collective_end_cb(const CallbackManager& cbmanager,
                                  int                    user_event,
                                  const Event&           event,
                                  CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_CommsExch::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_COLLECTIVE_END,
            PEARL_create_callback(
                this,
                &MPI_CommsExch::mpi_collective_end_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_CommsExch::mpi_collective_end_cb(const CallbackManager& cbmanager,
                                         const int              user_event,
                                         const Event&           event,
                                         CallbackData* const    cdata)
    {
        #line 326 "Message.pattern"

        if (  (event->getBytesSent() != 0)
           && (event->getBytesReceived() != 0))
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_INSTANCE_METRICS,
                             event.get_cnode(),
                             1);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_CommsSrc
     *  @brief MPI Collective Communications as Source metric.
     *
     *  Implements the "MPI Collective Communications as Source" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_CommsSrc
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_COMMS_CSRC;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_COMMS_COLL;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI collective communications as source";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "comms_csrc";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of MPI collective communications as source";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "occ";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_COLLECTIVE_END events.
            ///
            /// Callback function triggered by MPI_COLLECTIVE_END events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_collective_end_cb(const CallbackManager& cbmanager,
                                  int                    user_event,
                                  const Event&           event,
                                  CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_CommsSrc::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_COLLECTIVE_END,
            PEARL_create_callback(
                this,
                &MPI_CommsSrc::mpi_collective_end_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_CommsSrc::mpi_collective_end_cb(const CallbackManager& cbmanager,
                                        const int              user_event,
                                        const Event&           event,
                                        CallbackData* const    cdata)
    {
        #line 354 "Message.pattern"

        if (  (event->getBytesSent() != 0)
           && (event->getBytesReceived() == 0))
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_INSTANCE_METRICS,
                             event.get_cnode(),
                             1);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_CommsDst
     *  @brief MPI Collective Communications as Destination metric.
     *
     *  Implements the "MPI Collective Communications as Destination" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_CommsDst
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_COMMS_CDST;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_COMMS_COLL;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI collective communications as destination";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "comms_cdst";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of MPI collective communications as"
                       " destination";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "occ";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_COLLECTIVE_END events.
            ///
            /// Callback function triggered by MPI_COLLECTIVE_END events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_collective_end_cb(const CallbackManager& cbmanager,
                                  int                    user_event,
                                  const Event&           event,
                                  CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_CommsDst::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_COLLECTIVE_END,
            PEARL_create_callback(
                this,
                &MPI_CommsDst::mpi_collective_end_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_CommsDst::mpi_collective_end_cb(const CallbackManager& cbmanager,
                                        const int              user_event,
                                        const Event&           event,
                                        CallbackData* const    cdata)
    {
        #line 382 "Message.pattern"

        if (  (event->getBytesSent() == 0)
           && (event->getBytesReceived() != 0))
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_INSTANCE_METRICS,
                             event.get_cnode(),
                             1);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_BytesSent
     *  @brief MPI Point-to-point Bytes Sent metric.
     *
     *  Implements the "MPI Point-to-point Bytes Sent" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_BytesSent
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_BYTES_SENT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_BYTES_P2P;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI P2P bytes sent";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "bytes_sent";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of bytes sent in MPI point-to-point"
                       " communication operations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "bytes";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for GROUP_SEND events.
            ///
            /// Callback function triggered by GROUP_SEND events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            group_send_cb(const CallbackManager& cbmanager,
                          int                    user_event,
                          const Event&           event,
                          CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_BytesSent::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            GROUP_SEND,
            PEARL_create_callback(
                this,
                &MPI_BytesSent::group_send_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_BytesSent::group_send_cb(const CallbackManager& cbmanager,
                                 const int              user_event,
                                 const Event&           event,
                                 CallbackData* const    cdata)
    {
        #line 553 "Message.pattern"

        if (event->getBytesSent() > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_TRANSFER_METRICS,
                             event.get_cnode(),
                             event->getBytesSent());
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_BytesRcvd
     *  @brief MPI Point-to-point Bytes Received metric.
     *
     *  Implements the "MPI Point-to-point Bytes Received" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_BytesRcvd
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_BYTES_RCVD;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_BYTES_P2P;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI P2P bytes received";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "bytes_rcvd";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of bytes received in MPI point-to-point"
                       " communication operations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "bytes";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for POST_RECV events.
            ///
            /// Callback function triggered by POST_RECV events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            post_recv_cb(const CallbackManager& cbmanager,
                         int                    user_event,
                         const Event&           event,
                         CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_BytesRcvd::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            POST_RECV,
            PEARL_create_callback(
                this,
                &MPI_BytesRcvd::post_recv_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_BytesRcvd::post_recv_cb(const CallbackManager& cbmanager,
                                const int              user_event,
                                const Event&           event,
                                CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 597 "Message.pattern"

        RemoteEvent send = data->mRemote->getEvent(ROLE_SEND);

        if (send->getBytesSent() > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_TRANSFER_METRICS,
                             event.get_cnode(),
                             send->getBytesSent());
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_BytesCout
     *  @brief MPI Collective Bytes Outgoing metric.
     *
     *  Implements the "MPI Collective Bytes Outgoing" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_BytesCout
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_BYTES_COUT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_BYTES_COLL;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI collective bytes outgoing";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "bytes_cout";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of bytes outgoing in MPI collective"
                       " communication operations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "bytes";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_COLLECTIVE_END events.
            ///
            /// Callback function triggered by MPI_COLLECTIVE_END events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_collective_end_cb(const CallbackManager& cbmanager,
                                  int                    user_event,
                                  const Event&           event,
                                  CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_BytesCout::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_COLLECTIVE_END,
            PEARL_create_callback(
                this,
                &MPI_BytesCout::mpi_collective_end_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_BytesCout::mpi_collective_end_cb(const CallbackManager& cbmanager,
                                         const int              user_event,
                                         const Event&           event,
                                         CallbackData* const    cdata)
    {
        #line 659 "Message.pattern"

        uint64_t bytesSent = event->getBytesSent();
        if (bytesSent > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_TRANSFER_METRICS,
                             event.get_cnode(),
                             bytesSent);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_BytesCin
     *  @brief MPI Collective Bytes Incoming metric.
     *
     *  Implements the "MPI Collective Bytes Incoming" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_BytesCin
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_BYTES_CIN;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_BYTES_COLL;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI collective bytes incoming";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "bytes_cin";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of bytes incoming in MPI collective"
                       " communication operations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "bytes";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_COLLECTIVE_END events.
            ///
            /// Callback function triggered by MPI_COLLECTIVE_END events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_collective_end_cb(const CallbackManager& cbmanager,
                                  int                    user_event,
                                  const Event&           event,
                                  CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_BytesCin::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_COLLECTIVE_END,
            PEARL_create_callback(
                this,
                &MPI_BytesCin::mpi_collective_end_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_BytesCin::mpi_collective_end_cb(const CallbackManager& cbmanager,
                                        const int              user_event,
                                        const Event&           event,
                                        CallbackData* const    cdata)
    {
        #line 691 "Message.pattern"

        uint64_t bytesReceived = event->getBytesReceived();
        if (bytesReceived > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_TRANSFER_METRICS,
                             event.get_cnode(),
                             bytesReceived);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_BytesPut
     *  @brief MPI One-sided Bytes Sent metric.
     *
     *  Implements the "MPI One-sided Bytes Sent" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_BytesPut
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_BYTES_PUT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_BYTES_RMA;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI one-sided bytes sent";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "bytes_put";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of bytes sent in MPI one-sided communication"
                       " operations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "bytes";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_RMA_PUT_START events.
            ///
            /// Callback function triggered by MPI_RMA_PUT_START events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_put_start_cb(const CallbackManager& cbmanager,
                                 int                    user_event,
                                 const Event&           event,
                                 CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_BytesPut::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_RMA_PUT_START,
            PEARL_create_callback(
                this,
                &MPI_BytesPut::mpi_rma_put_start_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_BytesPut::mpi_rma_put_start_cb(const CallbackManager& cbmanager,
                                       const int              user_event,
                                       const Event&           event,
                                       CallbackData* const    cdata)
    {
        #line 746 "Message.pattern"

        INC_METRIC_VALUE(SCALASCA_DEBUG_TRANSFER_METRICS,
                         event.get_cnode(),
                         event->getBytesSent());
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_BytesGet
     *  @brief MPI One-sided Bytes Received metric.
     *
     *  Implements the "MPI One-sided Bytes Received" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_BytesGet
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_BYTES_GET;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_BYTES_RMA;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI one-sided bytes received";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "bytes_get";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of bytes received in MPI one-sided"
                       " communication operations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "bytes";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_RMA_GET_START events.
            ///
            /// Callback function triggered by MPI_RMA_GET_START events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_get_start_cb(const CallbackManager& cbmanager,
                                 int                    user_event,
                                 const Event&           event,
                                 CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_BytesGet::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_RMA_GET_START,
            PEARL_create_callback(
                this,
                &MPI_BytesGet::mpi_rma_get_start_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_BytesGet::mpi_rma_get_start_cb(const CallbackManager& cbmanager,
                                       const int              user_event,
                                       const Event&           event,
                                       CallbackData* const    cdata)
    {
        #line 774 "Message.pattern"

        INC_METRIC_VALUE(SCALASCA_DEBUG_TRANSFER_METRICS,
                         event.get_cnode(),
                         event->getBytesReceived());
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_SLR_Count
     *  @brief MPI Late Receiver Instances (Synchronizations) metric.
     *
     *  Implements the "MPI Late Receiver Instances (Synchronizations)" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_SLR_Count
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_SLR_COUNT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_SYNCS_SEND;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Late Receiver Instances (Synchronizations)";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_slr_count";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of Late Receiver instances in MPI"
                       " synchronizations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "occ";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (bws)
            /// @{

            /// @brief Callback function for LATE_RECEIVER events.
            ///
            /// Callback function triggered by LATE_RECEIVER events
            /// during the @"bws@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            late_receiver_bws_cb(const CallbackManager& cbmanager,
                                 int                    user_event,
                                 const Event&           event,
                                 CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_SLR_Count::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register bws callbacks
        it = cbmanagers.find("bws");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            LATE_RECEIVER,
            PEARL_create_callback(
                this,
                &MPI_SLR_Count::late_receiver_bws_cb));
    }


    // --- Callback functions (bws) -----------------------------------------

    void
    MPI_SLR_Count::late_receiver_bws_cb(const CallbackManager& cbmanager,
                                        const int              user_event,
                                        const Event&           event,
                                        CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 31 "Counts.pattern"

        Event completion = data->mLocal->getEvent(ROLE_ENTER_SENDCMP_LR);

        if (event->getBytesSent() == 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_INSTANCE_METRICS,
                             completion.get_cnode(),
                             1);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_SLS_Count
     *  @brief MPI Late Sender Instances (Synchronizations) metric.
     *
     *  Implements the "MPI Late Sender Instances (Synchronizations)" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_SLS_Count
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_SLS_COUNT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_SYNCS_RECV;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Late Sender Instances (Synchronizations)";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_sls_count";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of Late Sender instances in MPI"
                       " synchronizations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "occ";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for LATE_SENDER events.
            ///
            /// Callback function triggered by LATE_SENDER events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            late_sender_cb(const CallbackManager& cbmanager,
                           int                    user_event,
                           const Event&           event,
                           CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_SLS_Count::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            LATE_SENDER,
            PEARL_create_callback(
                this,
                &MPI_SLS_Count::late_sender_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_SLS_Count::late_sender_cb(const CallbackManager& cbmanager,
                                  const int              user_event,
                                  const Event&           event,
                                  CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 59 "Counts.pattern"

        RemoteEvent send = data->mRemote->getEvent(ROLE_SEND_LS);

        if (send->getBytesSent() == 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_INSTANCE_METRICS,
                             event.get_cnode(),
                             1);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_SLSWO_Count
     *  @brief MPI Late Sender, Wrong Order Instances (Synchronizations) metric.
     *
     *  Implements the "MPI Late Sender, Wrong Order Instances (Synchronizations)" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_SLSWO_Count
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_SLSWO_COUNT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_SLS_COUNT;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Late Sender, Wrong Order Instances"
                       " (Synchronizations)";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_slswo_count";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of Late Sender instances in MPI"
                       " synchronizations where messages are received in"
                       " wrong order";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "occ";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for LATE_SENDER_WO events.
            ///
            /// Callback function triggered by LATE_SENDER_WO events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            late_sender_wo_cb(const CallbackManager& cbmanager,
                              int                    user_event,
                              const Event&           event,
                              CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_SLSWO_Count::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            LATE_SENDER_WO,
            PEARL_create_callback(
                this,
                &MPI_SLSWO_Count::late_sender_wo_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_SLSWO_Count::late_sender_wo_cb(const CallbackManager& cbmanager,
                                       const int              user_event,
                                       const Event&           event,
                                       CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 88 "Counts.pattern"

        RemoteEvent send = data->mRemote->getEvent(ROLE_SEND_LSWO);
        Event       recv = data->mLocal->getEvent(ROLE_RECV_LSWO);

        if (send->getBytesSent() == 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_INSTANCE_METRICS,
                             recv.get_cnode(),
                             1);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_CLR_Count
     *  @brief MPI Late Receiver Instances (Communications) metric.
     *
     *  Implements the "MPI Late Receiver Instances (Communications)" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_CLR_Count
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_CLR_COUNT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_COMMS_SEND;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Late Receiver Instances (Communications)";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_clr_count";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of Late Receiver instances in MPI"
                       " communications";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "occ";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (bws)
            /// @{

            /// @brief Callback function for LATE_RECEIVER events.
            ///
            /// Callback function triggered by LATE_RECEIVER events
            /// during the @"bws@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            late_receiver_bws_cb(const CallbackManager& cbmanager,
                                 int                    user_event,
                                 const Event&           event,
                                 CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_CLR_Count::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register bws callbacks
        it = cbmanagers.find("bws");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            LATE_RECEIVER,
            PEARL_create_callback(
                this,
                &MPI_CLR_Count::late_receiver_bws_cb));
    }


    // --- Callback functions (bws) -----------------------------------------

    void
    MPI_CLR_Count::late_receiver_bws_cb(const CallbackManager& cbmanager,
                                        const int              user_event,
                                        const Event&           event,
                                        CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 117 "Counts.pattern"

        Event completion = data->mLocal->getEvent(ROLE_ENTER_SENDCMP_LR);

        if (event->getBytesSent() > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_INSTANCE_METRICS,
                             completion.get_cnode(),
                             1);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_CLS_Count
     *  @brief MPI Late Sender Instances (Communications) metric.
     *
     *  Implements the "MPI Late Sender Instances (Communications)" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_CLS_Count
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_CLS_COUNT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_COMMS_RECV;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Late Sender Instances (Communications)";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_cls_count";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of Late Sender instances in MPI"
                       " communications";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "occ";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for LATE_SENDER events.
            ///
            /// Callback function triggered by LATE_SENDER events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            late_sender_cb(const CallbackManager& cbmanager,
                           int                    user_event,
                           const Event&           event,
                           CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_CLS_Count::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            LATE_SENDER,
            PEARL_create_callback(
                this,
                &MPI_CLS_Count::late_sender_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_CLS_Count::late_sender_cb(const CallbackManager& cbmanager,
                                  const int              user_event,
                                  const Event&           event,
                                  CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 145 "Counts.pattern"

        RemoteEvent send = data->mRemote->getEvent(ROLE_SEND_LS);

        if (send->getBytesSent() != 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_INSTANCE_METRICS,
                             event.get_cnode(),
                             1);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_CLSWO_Count
     *  @brief MPI Late Sender, Wrong Order Instances (Communications) metric.
     *
     *  Implements the "MPI Late Sender, Wrong Order Instances (Communications)" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_CLSWO_Count
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_CLSWO_COUNT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_CLS_COUNT;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Late Sender, Wrong Order Instances"
                       " (Communications)";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_clswo_count";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of Late Sender instances in MPI"
                       " communications where messages are received in wrong"
                       " order";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "occ";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for LATE_SENDER_WO events.
            ///
            /// Callback function triggered by LATE_SENDER_WO events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            late_sender_wo_cb(const CallbackManager& cbmanager,
                              int                    user_event,
                              const Event&           event,
                              CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_CLSWO_Count::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            LATE_SENDER_WO,
            PEARL_create_callback(
                this,
                &MPI_CLSWO_Count::late_sender_wo_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_CLSWO_Count::late_sender_wo_cb(const CallbackManager& cbmanager,
                                       const int              user_event,
                                       const Event&           event,
                                       CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 173 "Counts.pattern"

        RemoteEvent send = data->mRemote->getEvent(ROLE_SEND_LSWO);
        Event       recv = data->mLocal->getEvent(ROLE_RECV_LSWO);

        if (send->getBytesSent() != 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_INSTANCE_METRICS,
                             recv.get_cnode(),
                             1);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_RmaWaitAtCreate
     *  @brief MPI Wait at Window Create Time metric.
     *
     *  Implements the "MPI Wait at Window Create Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_RmaWaitAtCreate
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_RMA_WAIT_AT_CREATE;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_MGMT_WIN;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Wait at Window Create";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_rma_wait_at_create";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent in MPI_Win_create, waiting for the last"
                       " process to enter the call";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_RMA_WAIT_AT_CREATE events.
            ///
            /// Callback function triggered by MPI_RMA_WAIT_AT_CREATE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_wait_at_create_cb(const CallbackManager& cbmanager,
                                      int                    user_event,
                                      const Event&           event,
                                      CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_RmaWaitAtCreate::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_RMA_WAIT_AT_CREATE,
            PEARL_create_callback(
                this,
                &MPI_RmaWaitAtCreate::mpi_rma_wait_at_create_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_RmaWaitAtCreate::mpi_rma_wait_at_create_cb(const CallbackManager& cbmanager,
                                                   const int              user_event,
                                                   const Event&           event,
                                                   CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 37 "MPI_RMA.pattern"

        INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                         event.get_cnode(),
                         data->mIdle);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_RmaWaitAtFree
     *  @brief MPI Wait at Window Free Time metric.
     *
     *  Implements the "MPI Wait at Window Free Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_RmaWaitAtFree
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_RMA_WAIT_AT_FREE;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_MGMT_WIN;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Wait at Window Free";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_rma_wait_at_free";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent in MPI_Win_free, waiting for the last"
                       " process to enter the call";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_RMA_WAIT_AT_FREE events.
            ///
            /// Callback function triggered by MPI_RMA_WAIT_AT_FREE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_wait_at_free_cb(const CallbackManager& cbmanager,
                                    int                    user_event,
                                    const Event&           event,
                                    CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_RmaWaitAtFree::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_RMA_WAIT_AT_FREE,
            PEARL_create_callback(
                this,
                &MPI_RmaWaitAtFree::mpi_rma_wait_at_free_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_RmaWaitAtFree::mpi_rma_wait_at_free_cb(const CallbackManager& cbmanager,
                                               const int              user_event,
                                               const Event&           event,
                                               CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 60 "MPI_RMA.pattern"

        INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                         event.get_cnode(),
                         data->mIdle);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_RmaSync
     *  @brief MPI One-sided Synchronization Time metric.
     *
     *  Implements the "MPI One-sided Synchronization Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_RmaSync
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_RMA_SYNCHRONIZATION;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_SYNCHRONIZATION;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI One-sided Synchronization";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_rma_synchronization";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent in MPI one-sided synchronization calls";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @copydoc scout::Pattern::isVisible()
            ///
            bool
            isVisible() const override
            {
                return false;
            }

            /// @}


        private:
            #line 82 "MPI_RMA.pattern"

            typedef struct
            {
                LocalData                 events;              ///< Local event set for caching GATS start events
                map< uint32_t, RmaEpoch > lockEpoch;           ///< Map to store different
                LockEpochQueue            epochQueue;          ///< Queue to save epoch events of remote processes
                ContentionInfoStackMap    syncPointInfoStacks; ///< Map to store syncpoint exchange data
            } win_t;

            map< const RmaWindow*, win_t > m_windows;    ///< Map of window-related cache data

            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_RMA_LOCK events.
            ///
            /// Callback function triggered by MPI_RMA_LOCK events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_lock_cb(const CallbackManager& cbmanager,
                            int                    user_event,
                            const Event&           event,
                            CallbackData*          cdata);

            /// @brief Callback function for MPI_RMA_UNLOCK events.
            ///
            /// Callback function triggered by MPI_RMA_UNLOCK events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_unlock_cb(const CallbackManager& cbmanager,
                              int                    user_event,
                              const Event&           event,
                              CallbackData*          cdata);

            /// @brief Callback function for MPI_RMA_GATS events.
            ///
            /// Callback function triggered by MPI_RMA_GATS events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_gats_cb(const CallbackManager& cbmanager,
                            int                    user_event,
                            const Event&           event,
                            CallbackData*          cdata);

            /// @brief Callback function for MPI_RMA_COLLECTIVE_END events.
            ///
            /// Callback function triggered by MPI_RMA_COLLECTIVE_END events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_collective_end_cb(const CallbackManager& cbmanager,
                                      int                    user_event,
                                      const Event&           event,
                                      CallbackData*          cdata);

            /// @brief Callback function for MPI_RMA_LC_COMPUTE events.
            ///
            /// Callback function triggered by MPI_RMA_LC_COMPUTE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_lc_compute_cb(const CallbackManager& cbmanager,
                                  int                    user_event,
                                  const Event&           event,
                                  CallbackData*          cdata);

            /// @brief Callback function for MPI_RMA_POST_COMPLETE events.
            ///
            /// Callback function triggered by MPI_RMA_POST_COMPLETE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_post_complete_cb(const CallbackManager& cbmanager,
                                     int                    user_event,
                                     const Event&           event,
                                     CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_RmaSync::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_RMA_LOCK,
            PEARL_create_callback(
                this,
                &MPI_RmaSync::mpi_rma_lock_cb));
        cbmanager->register_callback(
            MPI_RMA_UNLOCK,
            PEARL_create_callback(
                this,
                &MPI_RmaSync::mpi_rma_unlock_cb));
        cbmanager->register_callback(
            MPI_RMA_GATS,
            PEARL_create_callback(
                this,
                &MPI_RmaSync::mpi_rma_gats_cb));
        cbmanager->register_callback(
            MPI_RMA_COLLECTIVE_END,
            PEARL_create_callback(
                this,
                &MPI_RmaSync::mpi_rma_collective_end_cb));
        cbmanager->register_callback(
            MPI_RMA_LC_COMPUTE,
            PEARL_create_callback(
                this,
                &MPI_RmaSync::mpi_rma_lc_compute_cb));
        cbmanager->register_callback(
            MPI_RMA_POST_COMPLETE,
            PEARL_create_callback(
                this,
                &MPI_RmaSync::mpi_rma_post_complete_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_RmaSync::mpi_rma_lock_cb(const CallbackManager& cbmanager,
                                 const int              user_event,
                                 const Event&           event,
                                 CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 94 "MPI_RMA.pattern"

        // get reference to window data cache
        auto& win   = event->getWindow();
        auto& entry = m_windows[&win];

        // save complete lock region to epoch store
        entry.lockEpoch[event->getRemote()].push_back(event.enterptr());
        entry.lockEpoch[event->getRemote()].push_back(event);
        entry.lockEpoch[event->getRemote()].push_back(event.leaveptr());

        // notify direct RMA synchronization
        cbmanager.notify(RMA_SYNC_DIRECT, event, data);
    }


    void
    MPI_RmaSync::mpi_rma_unlock_cb(const CallbackManager& cbmanager,
                                   const int              user_event,
                                   const Event&           event,
                                   CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 108 "MPI_RMA.pattern"

        // get reference to window data cache
        auto& win   = event->getWindow();
        auto& entry = m_windows[&win];

        // save complete unlock region to epoch store
        entry.lockEpoch[event->getRemote()].push_back(event.enterptr());
        entry.lockEpoch[event->getRemote()].push_back(event);
        entry.lockEpoch[event->getRemote()].push_back(event.leaveptr());

        // create active message request
        AmRequest::Ptr request = AmRequestFactory::create(event->getRemote(),
                                                          win.getCommunicator());

        // pack lock epoch data to check for lock competition
        AmLockContention::pack(win.getId(),
                               entry.lockEpoch[event->getRemote()],
                               request.get());

        // enqueue request
        AmRuntime::getInstance().enqueue(std::move(request));

        // clear lock epoch cache
        entry.lockEpoch[event->getRemote()].clear();
    }


    void
    MPI_RmaSync::mpi_rma_gats_cb(const CallbackManager& cbmanager,
                                 const int              user_event,
                                 const Event&           event,
                                 CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 134 "MPI_RMA.pattern"

        auto  enter  = event.enterptr();
        auto& win    = event->getWindow();
        auto& region = enter->getRegion();
        auto& group  = event->getGroup();
        auto& entry  = m_windows[&win];

        // check which GATS region we are in
        if (is_mpi_rma_wait(region))
        {
            // analyze pairwise RMA sync pattern on target
            data->mCount = group.numRanks();
            cbmanager.notify(MPI_RMA_PWS_COUNT, event, data);

            data->mCount -= data->mCountOrigin;
            cbmanager.notify(MPI_RMA_PWS_UNNEEDED_COUNT, event, data);

            // notify callback for full GATS pattern analysis on target
            cbmanager.notify(MPI_RMA_POST_WAIT, event, data);
        }
        else if (is_mpi_rma_start(region))
        {
            // cache start events for later analysis
            entry.events.addEvent(enter, ROLE_ENTER_RMA_START);
            entry.events.addEvent(event, ROLE_LEAVE_RMA_START);
        }
        else if (is_mpi_rma_complete(region))
        {
            // analyze pairwise RMA sync pattern on origin
            data->mCount = group.numRanks();
            cbmanager.notify(MPI_RMA_PWS_COUNT, event, data);

            data->mCount -= data->mCountTarget;
            cbmanager.notify(MPI_RMA_PWS_UNNEEDED_COUNT, event, data);

            // set cached data as current
            *data->mLocal = entry.events;

            cbmanager.notify(MPI_RMA_POST_COMPLETE, event, data);

            // Clean up local data event cache
            entry.events.clear();
        }
    }


    void
    MPI_RmaSync::mpi_rma_collective_end_cb(const CallbackManager& cbmanager,
                                           const int              user_event,
                                           const Event&           event,
                                           CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 179 "MPI_RMA.pattern"

        // NOTE: The heuristic for collective calls in RMA is to check
        //       whether all the recorded timespans of all participating
        //       processes overlap. If not all timespans overlap, no
        //       collective synchronization is assumed.
        Event         enter  = event.enterptr();
        const Region& region = enter->getRegion();

        auto& win  = event->getWindow();
        auto& comm = win.getCommunicator();

        // Check valid overlap & possible idle time
        const CollectiveInfo& coll(data->mCollinfo);
        if (  (coll.earliest_end.mTime > coll.latest.mTime)
           && (coll.my.mTime < coll.latest.mTime))
        {
            data->mIdle = coll.latest.mTime - coll.my.mTime;
        }
        else
        {
            data->mIdle = 0;
        }

        if (is_mpi_rma_fence(region))
        {
            if (data->mIdle > 0)
            {
                cbmanager.notify(MPI_RMA_WAIT_AT_FENCE, event, data);

                if (  (data->mLastRmaOp.mTime > enter->getTimestamp())
                   && (data->mLastRmaOp.mTime <= event->getTimestamp()))
                {
                    data->mIdle = data->mLastRmaOp.mTime - enter->getTimestamp();
                    cbmanager.notify(MPI_RMA_EARLY_FENCE, event, data);
                }
            }

            // analyze pairwise sync pattern
            int comm_size = 0;
            MPI_Comm_size(getMpiHandle(comm), &comm_size);

            // count of remote partners is stored in data->mCount
            data->mCount = comm_size - data->mCount;
            cbmanager.notify(MPI_RMA_PWS_UNNEEDED_COUNT, event, data);

            data->mCount = comm_size;
            cbmanager.notify(MPI_RMA_PWS_COUNT, event, data);

            // compute lock competition
            data->mWindow = &win;
            cbmanager.notify(MPI_RMA_LC_COMPUTE, event, data);
        }
        else if (is_mpi_rma_create(region))
        {
            auto& win   = event->getWindow();
            auto& entry = m_windows[&win];

            if (data->mIdle > 0)
            {
                cbmanager.notify(MPI_RMA_WAIT_AT_CREATE, event, data);
            }

            // register lock epoch queue for new window
            AmLockContention::getInstance().registerQueue(win.getId(), &entry.epochQueue);
        }
        else if (is_mpi_rma_free(region))
        {
            if (data->mIdle > 0)
            {
                cbmanager.notify(MPI_RMA_WAIT_AT_FREE, event, data);

                if (data->mLastRmaOp.mTime > enter->getTimestamp())
                {
                    data->mIdle = data->mLastRmaOp.mTime - enter->getTimestamp();
                    cbmanager.notify(MPI_RMA_EARLY_FREE, event, data);
                }
            }

            // compute lock competition
            data->mWindow = &win;
            cbmanager.notify(MPI_RMA_LC_COMPUTE, event, data);
        }
        data->mLastRmaOp.mTime = std::numeric_limits< double >::lowest();
    }


    void
    MPI_RmaSync::mpi_rma_lc_compute_cb(const CallbackManager& cbmanager,
                                       const int              user_event,
                                       const Event&           event,
                                       CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 264 "MPI_RMA.pattern"

        auto& win   = *(data->mWindow);
        auto& entry = m_windows[&win];
        bool  current_exclusive_epoch;

        while (entry.epochQueue.size() >= 2)
        {
            // Priority queue of epochs to be processed
            LockEpochQueue scheduled_epochs;

            // Remove next epoch from queue and schedule it
            RemoteRmaEpochPtr current_epoch(entry.epochQueue.top());
            entry.epochQueue.pop();

            // check if current_epoch is exclusive
            RemoteRmaEpoch::const_iterator lock_it = current_epoch->begin();
            while (  (lock_it != current_epoch->end())
                  && !(*lock_it)->isOfType(MPI_RMA_LOCK))
            {
                ++lock_it;
            }

            current_exclusive_epoch = (*lock_it)->isExclusive();

            scheduled_epochs.push(current_epoch);

            RemoteRmaEpochPtr reference_epoch(entry.epochQueue.top());
            bool              exclusive_epoch_found = false;

            if (!current_exclusive_epoch)
            {
                do
                {
                    reference_epoch = entry.epochQueue.top();

                    // get lock event
                    RemoteRmaEpoch::const_iterator lock_it = reference_epoch->begin();
                    while (  (lock_it != reference_epoch->end())
                          && !(*lock_it)->isOfType(MPI_RMA_LOCK))
                    {
                        ++lock_it;
                    }

                    // if lock is shared, schedule epoch and get next
                    if (!(*lock_it)->isExclusive())
                    {
                        scheduled_epochs.push(reference_epoch);
                        entry.epochQueue.pop();
                    }
                    else
                    {
                        exclusive_epoch_found = true;
                    }
                }
                while (!exclusive_epoch_found && !entry.epochQueue.empty());
            }

            if (current_exclusive_epoch || exclusive_epoch_found)
            {
                while (!scheduled_epochs.empty())
                {
                    current_epoch = scheduled_epochs.top();
                    scheduled_epochs.pop();
                    RemoteRmaEpoch::reverse_iterator ce_it = current_epoch->rbegin();

                    // get previous exclusive lock owner's unlock event
                    RemoteRmaEpoch::reverse_iterator ref_it = reference_epoch->rbegin();
                    while (!(*ref_it)->isOfType(MPI_RMA_UNLOCK))
                    {
                        ++ref_it;
                    }

                    // Search event of current lock owner preceeding previous owners unlock
                    while (  (ce_it != current_epoch->rend())
                          && ((*ce_it)->getTimestamp() > (*ref_it)->getTimestamp()))
                    {
                        ++ce_it;
                    }

                    if (  (ce_it != current_epoch->rend())
                       && (  (*ce_it)->isOfType(GROUP_ENTER)
                          || (*ce_it)->isOfType(MPI_RMA_LOCK)
                          || (*ce_it)->isOfType(MPI_RMA_PUT_START)    // FIXME: this may have to be PUT_END!
                          || (*ce_it)->isOfType(MPI_RMA_GET_START)))  // FIXME: this may have to be GET_END!
                    {
                        timestamp_t idle_time = (*ref_it)->getTimestamp() - (*ce_it)->getTimestamp();

                        if (idle_time > 0)
                        {
                            RemoteEvent& wait_event = *ce_it;

                            // determine next Event type of MPI_RMA_LOCK, MPI_RMA_PUT_START or MPI_RMA_GET_START
                            if ((*ce_it)->isOfType(GROUP_ENTER))
                            {
                                // set iterator to ce_it
                                RemoteRmaEpoch::iterator it = current_epoch->begin();
                                while (  (it != current_epoch->end())
                                      && ((*it)->getTimestamp() <= (*ce_it)->getTimestamp()))
                                {
                                    ++it;
                                }

                                // set it to MPI_RMA_LOCK, MPI_RMA_PUT_START or MPI_RMA_GET_START or MPI_RMA_UNLOCK
                                while (  (it != current_epoch->end())
                                      && !(  (*it)->isOfType(MPI_RMA_LOCK)
                                          || (*it)->isOfType(MPI_RMA_PUT_START)
                                          || (*it)->isOfType(MPI_RMA_GET_START)
                                          || (*it)->isOfType(MPI_RMA_UNLOCK)))
                                {
                                    ++it;
                                }

                                wait_event = *it;
                            }

                            // prepare info to push on stacks for both events
                            RemoteEvent first_event     = *ref_it;
                            ident_t     wait_id         = wait_event.get_remote_index();
                            ident_t     first_id        = first_event.get_remote_index();
                            uint32_t    first_rank      = ref_it->get_location().getId();
                            uint32_t    wait_rank       = ce_it->get_location().getId();
                            uint32_t    first_sync_rank = wait_rank;
                            uint32_t    wait_sync_rank  = first_rank;

                            // build info-structs - negative idletime means other processes have waited
                            ContentionInfo wait_info  = { wait_id, idle_time, wait_sync_rank };
                            ContentionInfo first_info = { first_id, -idle_time, first_sync_rank };

                            // push wait_info on stack of correspondent map
                            if (entry.syncPointInfoStacks.find(wait_rank) == entry.syncPointInfoStacks.end())
                            {
                                ContentionInfoStack wait_stack;
                                wait_stack.push(wait_info);
                                entry.syncPointInfoStacks[wait_rank] = wait_stack;
                            }
                            else
                            {
                                entry.syncPointInfoStacks[wait_rank].push(wait_info);
                            }

                            // push first_info on stack of correspondent map
                            if (entry.syncPointInfoStacks.find(first_rank) == entry.syncPointInfoStacks.end())
                            {
                                ContentionInfoStack first_stack;
                                first_stack.push(first_info);
                                entry.syncPointInfoStacks[first_rank] = first_stack;
                            }
                            else
                            {
                                entry.syncPointInfoStacks[first_rank].push(first_info);
                            }
                        }    // if (idle_time > 0)

                        // delete current_epoch;
                    } // if ce_it type_of...
                }     // while (scheduled_epochs.size() > 0)
            }         // if (current_exclusive_epoch || exclusive_epoch_found)
        }             // while (entry.epochQueue.size() >= 2)

        // pop last entry in queue
        if (!entry.epochQueue.empty())
        {
            entry.epochQueue.pop();
        }

        // send synchpoint information for lock-epochs to origins
        ContentionInfoStackMap::iterator mit = entry.syncPointInfoStacks.begin();
        while (mit != entry.syncPointInfoStacks.end())
        {
            ContentionInfoStack infoStack = mit->second;

            auto& comm = win.getCommunicator();

            while (!infoStack.empty())
            {
                ContentionInfo& info    = infoStack.top();
                AmRequest::Ptr  request = AmRequestFactory::create(comm.getCommSet().getLocalRank(mit->first),
                                                                   comm);
                AmSyncpointExchange::pack(MPI_RMA_LOCK_COMPETITION, 0, info, request.get());

                AmRuntime::getInstance().enqueue(std::move(request));
                infoStack.pop();
            }

            ++mit;
        }
        entry.syncPointInfoStacks.clear();
    }


    void
    MPI_RmaSync::mpi_rma_post_complete_cb(const CallbackManager& cbmanager,
                                          const int              user_event,
                                          const Event&           event,
                                          CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 453 "MPI_RMA.pattern"

        Event       enter       = event.enterptr();
        timestamp_t enter_time  = enter->getTimestamp();
        Event       start_enter = data->mLocal->getEvent(ROLE_ENTER_RMA_START);
        Event       start_exit  = data->mLocal->getEvent(ROLE_LEAVE_RMA_START);

        if (  (data->mLastPost.mTime > enter_time)
           && (data->mLastPost.mTime < event->getTimestamp()))
        {
            /* non-blocking start and post concurrent to complete */
            data->mIdle       = data->mLastPost.mTime - enter_time;
            data->mCallpathId = enter.get_cnode()->getId();

            cbmanager.notify(MPI_RMA_LATE_POST, event, data);
        }
        else if (  (data->mLastPost.mTime < start_exit->getTimestamp())
                && (data->mLastPost.mTime > start_enter->getTimestamp()))
        {
            /* blocking start */
            data->mIdle       = data->mLastPost.mTime - start_enter->getTimestamp();
            data->mCallpathId = start_enter.get_cnode()->getId();

            cbmanager.notify(MPI_RMA_LATE_POST, start_exit, data);
        }
        else
        {
            // check all RMA operations between start and complete
            auto& win         = event->getWindow();
            auto  evt_it      = event.prev().prev();
            bool  start_found = false;

            while (!start_found)
            {
                if (  (  evt_it->isOfType(MPI_RMA_PUT_START)
                      || evt_it->isOfType(MPI_RMA_GET_START))
                   && (&(evt_it->getWindow()) == &win))
                {
                    if (  (data->mLastPost.mTime > evt_it->getTimestamp())
                       && (data->mLastPost.mTime < evt_it.leaveptr()->getTimestamp()))
                    {
                        data->mIdle       = data->mLastPost.mTime - evt_it->getTimestamp();
                        data->mCallpathId = evt_it.enterptr().get_cnode()->getId();

                        cbmanager.notify(MPI_RMA_LATE_POST, evt_it, data);
                    }
                }

                // FIXME: The 'leaveptr().prev()' construct is only guaranteed to
                //        work for EPIK traces and needs to be revised once OTF2
                //        RMA records are supported!
                if (  evt_it->isOfType(GROUP_ENTER)
                   && is_mpi_rma_start(evt_it->getRegion())
                   && (&(evt_it.leaveptr().prev()->getWindow()) == &win))
                {
                    start_found = true;
                }

                /* check previous event */
                --evt_it;
            }
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_RmaSyncLatePost
     *  @brief MPI Late Post Time in Synchronizations metric.
     *
     *  Implements the "MPI Late Post Time in Synchronizations" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_RmaSyncLatePost
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_RMA_SYNC_LATE_POST;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_RMA_SYNC_ACTIVE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Late Post (Synchronization)";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_rma_sync_late_post";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent in MPI one-sided active target"
                       " synchronization waiting for exposure epoch to start";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_RMA_LATE_POST events.
            ///
            /// Callback function triggered by MPI_RMA_LATE_POST events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_late_post_cb(const CallbackManager& cbmanager,
                                 int                    user_event,
                                 const Event&           event,
                                 CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_RmaSyncLatePost::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_RMA_LATE_POST,
            PEARL_create_callback(
                this,
                &MPI_RmaSyncLatePost::mpi_rma_late_post_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_RmaSyncLatePost::mpi_rma_late_post_cb(const CallbackManager& cbmanager,
                                              const int              user_event,
                                              const Event&           event,
                                              CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 551 "MPI_RMA.pattern"

        Callpath* callpath = data->mDefs->get_cnode(data->mCallpathId);

        if (is_mpi_rma_sync(callpath->getRegion()))
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             callpath,
                             data->mIdle);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_RmaEarlyWait
     *  @brief MPI Early Wait Time metric.
     *
     *  Implements the "MPI Early Wait Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_RmaEarlyWait
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_RMA_EARLY_WAIT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_RMA_SYNC_ACTIVE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Early Wait";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_rma_early_wait";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent in MPI_Win_wait waiting for last exposure"
                       " epoch to finish";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_RMA_POST_WAIT events.
            ///
            /// Callback function triggered by MPI_RMA_POST_WAIT events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_post_wait_cb(const CallbackManager& cbmanager,
                                 int                    user_event,
                                 const Event&           event,
                                 CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_RmaEarlyWait::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_RMA_POST_WAIT,
            PEARL_create_callback(
                this,
                &MPI_RmaEarlyWait::mpi_rma_post_wait_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_RmaEarlyWait::mpi_rma_post_wait_cb(const CallbackManager& cbmanager,
                                           const int              user_event,
                                           const Event&           event,
                                           CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 581 "MPI_RMA.pattern"

        Event       enter      = event.enterptr();
        timestamp_t enter_time = enter->getTimestamp();

        if (data->mLastComplete.mTime > enter_time)
        {
            // compute early wait
            data->mIdle = data->mLastComplete.mTime - enter_time;

            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             enter.get_cnode(),
                             data->mIdle);

            cbmanager.notify(MPI_RMA_EARLY_WAIT, event, data);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_RmaLateComplete
     *  @brief MPI Late Complete Time metric.
     *
     *  Implements the "MPI Late Complete Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_RmaLateComplete
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_RMA_LATE_COMPLETE;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_RMA_EARLY_WAIT;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Late Complete";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_rma_late_complete";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent waiting in MPI_Win_wait between last RMA"
                       " access and last MPI_Win_complete";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_RMA_EARLY_WAIT events.
            ///
            /// Callback function triggered by MPI_RMA_EARLY_WAIT events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_early_wait_cb(const CallbackManager& cbmanager,
                                  int                    user_event,
                                  const Event&           event,
                                  CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_RmaLateComplete::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_RMA_EARLY_WAIT,
            PEARL_create_callback(
                this,
                &MPI_RmaLateComplete::mpi_rma_early_wait_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_RmaLateComplete::mpi_rma_early_wait_cb(const CallbackManager& cbmanager,
                                               const int              user_event,
                                               const Event&           event,
                                               CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 619 "MPI_RMA.pattern"

        Event       enter      = event.enterptr();
        timestamp_t enter_time = enter->getTimestamp();
        timestamp_t max_ts     = max(data->mLastRmaOp.mTime, enter_time);

        // save severity (it does not have to be checked as this is only
        // triggered if an Early Wait is found.)
        INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                         enter.get_cnode(),
                         data->mLastComplete.mTime - max_ts);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_RmaWaitAtFence
     *  @brief MPI Wait at Fence Time metric.
     *
     *  Implements the "MPI Wait at Fence Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_RmaWaitAtFence
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_RMA_WAIT_AT_FENCE;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_RMA_SYNC_ACTIVE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Wait at Fence";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_rma_wait_at_fence";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent in front of MPI_Win_fence, waiting for"
                       " other processes";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_RMA_WAIT_AT_FENCE events.
            ///
            /// Callback function triggered by MPI_RMA_WAIT_AT_FENCE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_wait_at_fence_cb(const CallbackManager& cbmanager,
                                     int                    user_event,
                                     const Event&           event,
                                     CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_RmaWaitAtFence::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_RMA_WAIT_AT_FENCE,
            PEARL_create_callback(
                this,
                &MPI_RmaWaitAtFence::mpi_rma_wait_at_fence_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_RmaWaitAtFence::mpi_rma_wait_at_fence_cb(const CallbackManager& cbmanager,
                                                 const int              user_event,
                                                 const Event&           event,
                                                 CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 650 "MPI_RMA.pattern"

        INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                         event.get_cnode(),
                         data->mIdle);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_RmaEarlyFence
     *  @brief MPI Early Fence Time metric.
     *
     *  Implements the "MPI Early Fence Time" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_RmaEarlyFence
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_RMA_EARLY_FENCE;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_RMA_WAIT_AT_FENCE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Early Fence";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_rma_early_fence";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Waiting time in MPI_Win_fence due to pending RMA"
                       " operations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_RMA_EARLY_FENCE events.
            ///
            /// Callback function triggered by MPI_RMA_EARLY_FENCE events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_early_fence_cb(const CallbackManager& cbmanager,
                                   int                    user_event,
                                   const Event&           event,
                                   CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_RmaEarlyFence::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_RMA_EARLY_FENCE,
            PEARL_create_callback(
                this,
                &MPI_RmaEarlyFence::mpi_rma_early_fence_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_RmaEarlyFence::mpi_rma_early_fence_cb(const CallbackManager& cbmanager,
                                              const int              user_event,
                                              const Event&           event,
                                              CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 675 "MPI_RMA.pattern"

        INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                         event.get_cnode(),
                         data->mIdle);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Mpi_RmaSyncLockContention
     *  @brief MPI Lock Contention Time in Synchronizations metric.
     *
     *  Implements the "MPI Lock Contention Time in Synchronizations" metric.
     */
    /*---------------------------------------------------------------------*/

    class Mpi_RmaSyncLockContention
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_RMA_SYNC_LOCK_COMPETITION;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_RMA_SYNC_PASSIVE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Lock Contention (Synchronization)";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_rma_sync_lock_competition";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent waiting in MPI_Win_lock/MPI_Win_unlock"
                       " acquiring a lock for a window";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_RMA_LOCK_COMPETITION events.
            ///
            /// Callback function triggered by MPI_RMA_LOCK_COMPETITION events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_lock_competition_cb(const CallbackManager& cbmanager,
                                        int                    user_event,
                                        const Event&           event,
                                        CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Mpi_RmaSyncLockContention::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_RMA_LOCK_COMPETITION,
            PEARL_create_callback(
                this,
                &Mpi_RmaSyncLockContention::mpi_rma_lock_competition_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    Mpi_RmaSyncLockContention::mpi_rma_lock_competition_cb(const CallbackManager& cbmanager,
                                                           const int              user_event,
                                                           const Event&           event,
                                                           CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 714 "MPI_RMA.pattern"

        Callpath* callpath = data->mDefs->get_cnode(data->mCallpathId);

        if (  (data->mIdle > 0)
           && is_mpi_rma_passive(callpath->getRegion()))
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             callpath,
                             data->mIdle);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Mpi_RmaSyncWaitForProgress
     *  @brief MPI Wait for Progress Time in Synchronizations metric.
     *
     *  Implements the "MPI Wait for Progress Time in Synchronizations" metric.
     */
    /*---------------------------------------------------------------------*/

    class Mpi_RmaSyncWaitForProgress
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_RMA_SYNC_WAIT_FOR_PROGRESS;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_RMA_SYNC_PASSIVE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Wait for Progress (Synchronization)";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_rma_sync_wait_for_progress";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent waiting in MPI_Win_lock/MPI_Win_unlock"
                       " waiting for progress on the target";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_RMA_WAIT_FOR_PROGRESS events.
            ///
            /// Callback function triggered by MPI_RMA_WAIT_FOR_PROGRESS events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_wait_for_progress_cb(const CallbackManager& cbmanager,
                                         int                    user_event,
                                         const Event&           event,
                                         CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Mpi_RmaSyncWaitForProgress::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_RMA_WAIT_FOR_PROGRESS,
            PEARL_create_callback(
                this,
                &Mpi_RmaSyncWaitForProgress::mpi_rma_wait_for_progress_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    Mpi_RmaSyncWaitForProgress::mpi_rma_wait_for_progress_cb(const CallbackManager& cbmanager,
                                                             const int              user_event,
                                                             const Event&           event,
                                                             CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 744 "MPI_RMA.pattern"

        Callpath* callpath = data->mDefs->get_cnode(data->mCallpathId);

        if (is_mpi_rma_passive(callpath->getRegion()))
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             callpath,
                             data->mIdle);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_RmaCommLatePost
     *  @brief MPI Late Post Time in Communications metric.
     *
     *  Implements the "MPI Late Post Time in Communications" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_RmaCommLatePost
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_RMA_COMM_LATE_POST;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_RMA_COMMUNICATION;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Late Post (Communication)";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_rma_comm_late_post";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent in MPI one-sided communication operations"
                       " waiting for exposure epoch to start";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_RMA_LATE_POST events.
            ///
            /// Callback function triggered by MPI_RMA_LATE_POST events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_late_post_cb(const CallbackManager& cbmanager,
                                 int                    user_event,
                                 const Event&           event,
                                 CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_RmaCommLatePost::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_RMA_LATE_POST,
            PEARL_create_callback(
                this,
                &MPI_RmaCommLatePost::mpi_rma_late_post_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_RmaCommLatePost::mpi_rma_late_post_cb(const CallbackManager& cbmanager,
                                              const int              user_event,
                                              const Event&           event,
                                              CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 791 "MPI_RMA.pattern"

        Callpath* callpath = data->mDefs->get_cnode(data->mCallpathId);

        if (is_mpi_rma_comm(callpath->getRegion()))
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             callpath,
                             data->mIdle);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Mpi_RmaCommLockContention
     *  @brief MPI Lock Contention Time in Communications metric.
     *
     *  Implements the "MPI Lock Contention Time in Communications" metric.
     */
    /*---------------------------------------------------------------------*/

    class Mpi_RmaCommLockContention
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_RMA_COMM_LOCK_COMPETITION;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_RMA_COMMUNICATION;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Lock Contention (Communication)";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_rma_comm_lock_competition";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent waiting in MPI_Win_lock/MPI_Win_unlock"
                       " acquiring a lock for a window";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_RMA_LOCK_COMPETITION events.
            ///
            /// Callback function triggered by MPI_RMA_LOCK_COMPETITION events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_lock_competition_cb(const CallbackManager& cbmanager,
                                        int                    user_event,
                                        const Event&           event,
                                        CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Mpi_RmaCommLockContention::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_RMA_LOCK_COMPETITION,
            PEARL_create_callback(
                this,
                &Mpi_RmaCommLockContention::mpi_rma_lock_competition_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    Mpi_RmaCommLockContention::mpi_rma_lock_competition_cb(const CallbackManager& cbmanager,
                                                           const int              user_event,
                                                           const Event&           event,
                                                           CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 819 "MPI_RMA.pattern"

        Callpath* callpath = data->mDefs->get_cnode(data->mCallpathId);

        if (  (data->mIdle > 0)
           && is_mpi_rma_comm(callpath->getRegion()))
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             callpath,
                             data->mIdle);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Mpi_RmaCommWaitForProgress
     *  @brief MPI Wait for Progress Time in Communications metric.
     *
     *  Implements the "MPI Wait for Progress Time in Communications" metric.
     */
    /*---------------------------------------------------------------------*/

    class Mpi_RmaCommWaitForProgress
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_RMA_COMM_WAIT_FOR_PROGRESS;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_RMA_COMMUNICATION;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI Wait for Progress (Communication)";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_rma_comm_wait_for_progress";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Time spent waiting in MPI_Win_lock/MPI_Win_unlock"
                       " waiting for progress on the target";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_RMA_WAIT_FOR_PROGRESS events.
            ///
            /// Callback function triggered by MPI_RMA_WAIT_FOR_PROGRESS events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_wait_for_progress_cb(const CallbackManager& cbmanager,
                                         int                    user_event,
                                         const Event&           event,
                                         CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Mpi_RmaCommWaitForProgress::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_RMA_WAIT_FOR_PROGRESS,
            PEARL_create_callback(
                this,
                &Mpi_RmaCommWaitForProgress::mpi_rma_wait_for_progress_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    Mpi_RmaCommWaitForProgress::mpi_rma_wait_for_progress_cb(const CallbackManager& cbmanager,
                                                             const int              user_event,
                                                             const Event&           event,
                                                             CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 849 "MPI_RMA.pattern"

        Callpath* callpath = data->mDefs->get_cnode(data->mCallpathId);

        if (is_mpi_rma_comm(callpath->getRegion()))
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             callpath,
                             data->mIdle);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_RmaPairsyncCount
     *  @brief Pair-wise MPI One-sided Synchronizations metric.
     *
     *  Implements the "Pair-wise MPI One-sided Synchronizations" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_RmaPairsyncCount
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_RMA_PAIRSYNC_COUNT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_NONE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI pair-wise one-sided synchronizations";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_rma_pairsync_count";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of pair-wise MPI one-sided synchronizations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "occ";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_RMA_PWS_COUNT events.
            ///
            /// Callback function triggered by MPI_RMA_PWS_COUNT events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_pws_count_cb(const CallbackManager& cbmanager,
                                 int                    user_event,
                                 const Event&           event,
                                 CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_RmaPairsyncCount::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_RMA_PWS_COUNT,
            PEARL_create_callback(
                this,
                &MPI_RmaPairsyncCount::mpi_rma_pws_count_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_RmaPairsyncCount::mpi_rma_pws_count_cb(const CallbackManager& cbmanager,
                                               const int              user_event,
                                               const Event&           event,
                                               CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 888 "MPI_RMA.pattern"

        INC_METRIC_VALUE(SCALASCA_DEBUG_INSTANCE_METRICS,
                         event.get_cnode(),
                         data->mCount);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_RmaPairsyncUnneededCount
     *  @brief Unneeded Pair-wise MPI One-sided Synchronizations metric.
     *
     *  Implements the "Unneeded Pair-wise MPI One-sided Synchronizations" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_RmaPairsyncUnneededCount
        : public TypedPattern< uint64_t >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_RMA_PAIRSYNC_UNNEEDED_COUNT;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_RMA_PAIRSYNC_COUNT;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI unneeded pair-wise one-sided synchronizations";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_rma_pairsync_unneeded_count";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Number of unneeded pair-wise MPI one-sided"
                       " synchronizations";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "occ";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for MPI_RMA_PWS_UNNEEDED_COUNT events.
            ///
            /// Callback function triggered by MPI_RMA_PWS_UNNEEDED_COUNT events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_rma_pws_unneeded_count_cb(const CallbackManager& cbmanager,
                                          int                    user_event,
                                          const Event&           event,
                                          CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_RmaPairsyncUnneededCount::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< uint64_t >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            MPI_RMA_PWS_UNNEEDED_COUNT,
            PEARL_create_callback(
                this,
                &MPI_RmaPairsyncUnneededCount::mpi_rma_pws_unneeded_count_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_RmaPairsyncUnneededCount::mpi_rma_pws_unneeded_count_cb(const CallbackManager& cbmanager,
                                                                const int              user_event,
                                                                const Event&           event,
                                                                CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 921 "MPI_RMA.pattern"

        INC_METRIC_VALUE(SCALASCA_DEBUG_INSTANCE_METRICS,
                         event.get_cnode(),
                         data->mCount);
    }
#endif    // _MPI


/*-------------------------------------------------------------------------*/
/**
 *  @class Statistics
 *  @brief STATISTICS metric.
 *
 *  Implements the "STATISTICS" metric.
 */
/*-------------------------------------------------------------------------*/

class Statistics
    : public TypedPattern< double >
{
    public:
        /// @name Constructors & destructor
        /// @{

        /// @copydoc scout::Pattern::~Pattern()
        ///
        ~Statistics() override
        {
            #line 1236 "Statistics.pattern"

            // Release statistics objects
            #if defined(_MPI)
                delete ls_quant;
                delete lsw_quant;
                delete lr_quant;
                delete wnxn_quant;
                delete wb_quant;
                delete er_quant;
                delete es_quant;
                delete lb_quant;
                delete bc_quant;
                delete nxnc_quant;
            #endif

            #if defined(_OPENMP)
                delete omp_eb_quant;
                delete omp_ib_quant;
                delete thread_lc_ol_quant;
                delete thread_lc_oc_quant;
                delete thread_lc_pm_quant;
                delete thread_lc_pc_quant;
            #endif    // _OPENMP
        }

        /// @}
        /// @name Callback registration
        /// @{

        /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
        ///
        void
        registerCallbacks(CallbackManagerMap& cbmanagers) override;

        /// @}
        /// @name Pattern information access
        /// @{

        /// @copydoc scout::Pattern::getId()
        ///
        long
        getId() const override
        {
            return PAT_STATISTICS;
        }

        /// @copydoc scout::Pattern::getParent()
        ///
        long
        getParent() const override
        {
            return PAT_NONE;
        }

        /// @copydoc scout::Pattern::getName()
        ///
        string
        getName() const override
        {
            return "STATISTICS";
        }

        /// @copydoc scout::Pattern::getUniqueName()
        ///
        string
        getUniqueName() const override
        {
            return "statistics";
        }

        /// @copydoc scout::Pattern::getDescription()
        ///
        string
        getDescription() const override
        {
            return "Statistics for waiting time distribution";
        }

        /// @copydoc scout::Pattern::getUnit()
        ///
        string
        getUnit() const override
        {
            return "sec";
        }

        /// @copydoc scout::Pattern::getStorageMode()
        ///
        CubeMetricType
        getStorageMode() const override
        {
            return CUBE_METRIC_EXCLUSIVE;
        }

        /// @copydoc scout::Pattern::isVisible()
        ///
        bool
        isVisible() const override
        {
            return false;
        }

        /// @}


    protected:
        /// @name Pattern data management
        /// @{

        /// @brief Initialize local data.
        ///
        /// This method is implicitly called from registerCallbacks() and
        /// initializes pattern-local data.
        ///
        void
        init() override
        {
            TypedPattern< double >::init();

            #line 1211 "Statistics.pattern"

            // Create statistics objects for MPI metrics
            #if defined(_MPI)
                ls_quant   = new Quantile(NUMBER_COEFF);
                lsw_quant  = new Quantile(NUMBER_COEFF);
                lr_quant   = new Quantile(NUMBER_COEFF);
                wnxn_quant = new Quantile(NUMBER_COEFF);
                wb_quant   = new Quantile(NUMBER_COEFF);
                er_quant   = new Quantile(NUMBER_COEFF);
                es_quant   = new Quantile(NUMBER_COEFF);
                lb_quant   = new Quantile(NUMBER_COEFF);
                bc_quant   = new Quantile(NUMBER_COEFF);
                nxnc_quant = new Quantile(NUMBER_COEFF);
            #endif

            #if defined(_OPENMP)
                omp_eb_quant       = new Quantile(NUMBER_COEFF);
                omp_ib_quant       = new Quantile(NUMBER_COEFF);
                thread_lc_ol_quant = new Quantile(NUMBER_COEFF);
                thread_lc_oc_quant = new Quantile(NUMBER_COEFF);
                thread_lc_pm_quant = new Quantile(NUMBER_COEFF);
                thread_lc_pc_quant = new Quantile(NUMBER_COEFF);
            #endif    // _OPENMP
        }

        /// @}


    private:
        #line 37 "Statistics.pattern"

        /// Symbolic names for entries in arrays storing upper bounds of metric
        /// durations
        enum duration_t
        {
            LS_MAX_DURATION = 0,
            LR_MAX_DURATION,
            WNXN_MAX_DURATION,
            WB_MAX_DURATION,
            ER_MAX_DURATION,
            ES_MAX_DURATION,
            LB_MAX_DURATION,
            BC_MAX_DURATION,
            NXNC_MAX_DURATION,
            OMP_EB_MAX_DURATION,
            OMP_IB_MAX_DURATION,
            THREAD_LC_OL_MAX_DURATION,
            THREAD_LC_OC_MAX_DURATION,
            THREAD_LC_PM_MAX_DURATION,
            THREAD_LC_PC_MAX_DURATION,
            MAX_DURATION_ENTRIES
        };


        /// CallbackData struct used for preparation replay
        struct MaxDurations
            : public pearl::CallbackData
        {
            double m_max_duration[MAX_DURATION_ENTRIES];

            MaxDurations()
            {
                memset(m_max_duration, 0, MAX_DURATION_ENTRIES * sizeof(double));
            }

            void
            update_duration(duration_t idx,
                            double     value)
            {
                if (value > m_max_duration[idx])
                {
                    m_max_duration[idx] = value;
                }
            }
        };


        void
        mpicollective_cb(const pearl::CallbackManager& cbmanager,
                         int                           user_event,
                         const pearl::Event&           event,
                         pearl::CallbackData*          cdata)
        {
            Event         begin    = event.beginptr();
            double        duration = event->getTimestamp() - begin->getTimestamp();
            MaxDurations* data     = static_cast< MaxDurations* >(cdata);

            const Region& region = event.get_cnode()->getRegion();
            if (is_mpi_barrier(region))
            {
                // Wait at Barrier
                data->update_duration(WB_MAX_DURATION, duration);

                // Barrier Completion
                data->update_duration(BC_MAX_DURATION, duration);
            }
            else if (is_mpi_12n(region))
            {
                // Late Broadcast
                data->update_duration(LB_MAX_DURATION, duration);
            }
            else if (is_mpi_n21(region))
            {
                // Early Reduce
                data->update_duration(ER_MAX_DURATION, duration);
            }
            else if (is_mpi_scan(region))
            {
                // Early Scan
                data->update_duration(ES_MAX_DURATION, duration);
            }
            else if (is_mpi_n2n(region))
            {
                // Wait at NxN
                data->update_duration(WNXN_MAX_DURATION, duration);

                // NxN Completion
                data->update_duration(NXNC_MAX_DURATION, duration);
            }
        }

        void
        send_cb(const pearl::CallbackManager& cbmanager,
                int                           user_event,
                const pearl::Event&           event,
                pearl::CallbackData*          cdata)
        {
            Event         enter    = event.enterptr();
            Event         leave    = event.leaveptr();
            double        duration = leave->getTimestamp() - enter->getTimestamp();
            MaxDurations* data     = static_cast< MaxDurations* >(cdata);

            // Late Receiver
            data->update_duration(LR_MAX_DURATION, duration);
        }

        void
        recv_cb(const pearl::CallbackManager& cbmanager,
                int                           user_event,
                const pearl::Event&           event,
                pearl::CallbackData*          cdata)
        {
            Event         enter    = event.enterptr();
            Event         leave    = event.leaveptr();
            double        duration = leave->getTimestamp() - enter->getTimestamp();
            MaxDurations* data     = static_cast< MaxDurations* >(cdata);

            // Late Sender
            data->update_duration(LS_MAX_DURATION, duration);
        }

        void
        leave_cb(const pearl::CallbackManager& cbmanager,
                 int                           user_event,
                 const pearl::Event&           event,
                 pearl::CallbackData*          cdata)
        {
            const Region& region = event.get_cnode()->getRegion();
            if (!is_omp_barrier(region))
            {
                return;
            }

            Event         enter    = event.enterptr();
            double        duration = event->getTimestamp() - enter->getTimestamp();
            MaxDurations* data     = static_cast< MaxDurations* >(cdata);

            if (is_omp_ebarrier(region))
            {
                // OpenMP explicit barrier
                data->update_duration(OMP_EB_MAX_DURATION, duration);
            }
            else if (is_omp_ibarrier(region))
            {
                // OpenMP implicit barrier
                data->update_duration(OMP_IB_MAX_DURATION, duration);
            }
        }

        void
        acquire_cb(const pearl::CallbackManager& cbmanager,
                   int                           user_event,
                   const pearl::Event&           event,
                   pearl::CallbackData*          cdata)
        {
            Event         enter    = event.enterptr();
            double        duration = event->getTimestamp() - enter->getTimestamp();
            MaxDurations* data     = static_cast< MaxDurations* >(cdata);

            LockType lockType = getLockType(event);
            if (lockType == LockType::OMP_LOCK_API)
            {
                data->update_duration(THREAD_LC_OL_MAX_DURATION, duration);
            }
            else if (lockType == LockType::OMP_CRITICAL)
            {
                data->update_duration(THREAD_LC_OC_MAX_DURATION, duration);
            }
            else if (  (lockType == LockType::PTHREAD_MUTEX_LOCK)
                    || (lockType == LockType::PTHREAD_MUTEX_TRYLOCK))
            {
                data->update_duration(THREAD_LC_PM_MAX_DURATION, duration);
            }
            else if (lockType == LockType::PTHREAD_CONDITION_WAIT)
            {
                data->update_duration(THREAD_LC_PC_MAX_DURATION, duration);
            }
        }

        // Statistics collector objects for individual patterns
        #if defined(_MPI)
            Quantile* ls_quant;
            Quantile* lsw_quant;
            Quantile* lr_quant;
            Quantile* wnxn_quant;
            Quantile* wb_quant;
            Quantile* er_quant;
            Quantile* es_quant;
            Quantile* lb_quant;
            Quantile* bc_quant;
            Quantile* nxnc_quant;
        #endif    // _MPI

        #if defined(_OPENMP)
            Quantile* omp_eb_quant;
            Quantile* omp_ib_quant;
            Quantile* thread_lc_ol_quant;
            Quantile* thread_lc_oc_quant;
            Quantile* thread_lc_pm_quant;
            Quantile* thread_lc_pc_quant;
        #endif    // (_OPENMP)

        // define deques for most severe instances
        #if defined(_MPI)
            vector< TopMostSevere > LateSender, LateSenderWO, LateReceiver, EarlyReduce;
            vector< TopMostSevere > BarrierNxN, Barrier, NxnCompl, BarrierCompl, LateBcast, EarlyScan;
        #endif    // _MPI

        static MaxDurations sharedDurations;

        #if defined(_OPENMP)
            vector< TopMostSevere > OmpEBarrier, OmpIBarrier;
            vector< TopMostSevere > ThreadLockContentionOMPLockAPI;
            vector< TopMostSevere > ThreadLockContentionOMPCritical;
            vector< TopMostSevere > ThreadLockContentionPthreadMutex;
            vector< TopMostSevere > ThreadLockContentionPthreadCondition;

            static vector< TopMostSevere > sharedLC_OL;
            static vector< TopMostSevere > sharedLC_OC;
            static vector< TopMostSevere > sharedLC_PM;
            static vector< TopMostSevere > sharedLC_PC;


            // merges thread local TopMostSevere vectors into a shared one using the stat_collector method
            void
            result_merge_thread_topmost(vector< TopMostSevere >& sharedInst,
                                        vector< TopMostSevere >  inst)
            {
                #pragma omp critical
                {
                    for (size_t i = 0; i < inst.size(); i++)
                    {
                        stat_collector(sharedInst,
                                       inst[i].idletime,
                                       inst[i].entertime,
                                       inst[i].exittime,
                                       inst[i].cnode,
                                       inst[i].rank);
                    }
                }
            }
        #endif    // _OPENMP

        // merge results of found patterns
        // from all processes and threads
        // static variables used here as an implicit shared variables
        // to perform reduction for OpenMP threads
        void
        result_merge(Quantile* quant)
        {
            // Determine global number of instances
            // (The 'static' variable 'shared_n' is implicitly shared!)
            double        n, global_n;
            static double shared_n;
            #pragma omp master
            {
                shared_n = 0.0;
            }

            // Determine global coefficients for collecting statistics
            // (The 'static' array 'shared_coeff' is implicitly shared!)
            double        coeff[NUMBER_COEFF], global_coeff[NUMBER_COEFF];
            static double shared_coeff[NUMBER_COEFF];

            #pragma omp master
            {
                for (int i = 0; i < NUMBER_COEFF; i++)
                {
                    shared_coeff[i] = 0.0;
                }
            }

            // Determine global coefficients for control values
            // (The 'static' array 'shared_control_val' is implicitly shared!)
            double        control_val[NUMBER_COEFF], global_control_val[NUMBER_COEFF];
            static double shared_control_val[NUMBER_COEFF];

            #pragma omp master
            {
                for (int i = 0; i < NUMBER_COEFF; i++)
                {
                    shared_control_val[i] = 0.0;
                }
            }

            // Determine global max
            // (The 'static' variable 'shared_max' is implicitly shared!)
            double        max_val, global_max;
            static double shared_max;
            #pragma omp master
            {
                shared_max = std::numeric_limits< double >::lowest();
            }

            // Determine global min
            // (The 'static' variable 'shared_min' is implicitly shared!)
            double        min_val, global_min;
            static double shared_min;
            #pragma omp master
            {
                shared_min = std::numeric_limits< double >::max();
            }

            // Determine global sum
            // (The 'static' variable 'shared_sum' is implicitly shared!)
            double        sum, global_sum;
            static double shared_sum;
            #pragma omp master
            {
                shared_sum = 0.0;
            }

            // Determine global squared sum
            // (The 'static' variable 'shared_squared_sum' is implicitly shared!)
            double        squared_sum, global_squared_sum;
            static double shared_squared_sum;
            #pragma omp master
            {
                shared_squared_sum = 0.0;
            }

            n = quant->get_n();

            #pragma omp barrier
            #pragma omp critical
            {
                shared_n += n;
            }
            #pragma omp barrier

            #if defined(_MPI)
                #pragma omp master
                {
                    MPI_Allreduce(&shared_n, &global_n, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
                    shared_n = global_n;
                }
            #endif    // _MPI

            #pragma omp barrier
            global_n = shared_n;
            #pragma omp barrier

            for (int i = 0; i < NUMBER_COEFF; i++)
            {
                coeff[i] = (n / global_n) * quant->get_coeff(i);
            }

            #pragma omp barrier
            #pragma omp critical
            {
                for (int i = 0; i < NUMBER_COEFF; i++)
                {
                    shared_coeff[i] += coeff[i];
                }
            }
            #pragma omp barrier

            #if defined(_MPI)
                #pragma omp master
                {
                    MPI_Allreduce(shared_coeff, global_coeff, NUMBER_COEFF, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
                    for (int i = 0; i < NUMBER_COEFF; i++)
                    {
                        shared_coeff[i] = global_coeff[i];
                    }
                }
            #endif    // _MPI

            #pragma omp barrier
            for (int i = 0; i < NUMBER_COEFF; i++)
            {
                global_coeff[i] = shared_coeff[i];
                control_val[i]  = (n / global_n) * quant->get_control_val(i);
            }

            #pragma omp barrier
            #pragma omp critical
            {
                for (int i = 0; i < NUMBER_COEFF; i++)
                {
                    shared_control_val[i] += control_val[i];
                }
            }
            #pragma omp barrier
            #if defined(_MPI)
                #pragma omp master
                {
                    MPI_Allreduce(shared_control_val, global_control_val, NUMBER_COEFF, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
                    for (int i = 0; i < NUMBER_COEFF; i++)
                    {
                        shared_control_val[i] = global_control_val[i];
                    }
                }
            #endif    // _MPI

            #pragma omp barrier
            for (int i = 0; i < NUMBER_COEFF; i++)
            {
                global_control_val[i] = shared_control_val[i];
            }

            max_val     = quant->get_max_val();
            min_val     = quant->get_min_val();
            sum         = quant->get_sum();
            squared_sum = quant->get_squared_sum();

            #pragma omp barrier

            #pragma omp critical
            {
                if (shared_max < max_val)
                {
                    shared_max = max_val;
                }

                if (shared_min > min_val)
                {
                    shared_min = min_val;
                }

                shared_sum         += sum;
                shared_squared_sum += squared_sum;
            }
            #pragma omp barrier

            #if defined(_MPI)
                #pragma omp master
                {
                    MPI_Allreduce(&shared_max, &global_max, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
                    MPI_Allreduce(&shared_min, &global_min, 1, MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD);
                    MPI_Allreduce(&shared_sum, &global_sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
                    MPI_Allreduce(&shared_squared_sum, &global_squared_sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
                    shared_max         = global_max;
                    shared_min         = global_min;
                    shared_sum         = global_sum;
                    shared_squared_sum = global_squared_sum;
                }
            #endif    // _MPI

            #pragma omp barrier
            global_max         = shared_max;
            global_min         = shared_min;
            global_sum         = shared_sum;
            global_squared_sum = shared_squared_sum;
            #pragma omp barrier

            quant->set_global_values(global_n, global_sum, global_squared_sum, global_min, global_max, global_coeff, global_control_val);
        }

        void
        write_cube_file(Quantile*                quant,
                        vector< TopMostSevere >& instance,
                        ReportData&              data,
                        FILE*                    cube_fp)
        {
            string patternName     = quant->get_metric();
            double number_obs      = quant->get_n();
            double sum_obs         = quant->get_sum();
            double squared_sum_obs = quant->get_squared_sum();
            double min_val         = quant->get_min_val();
            double max_val         = quant->get_max_val();
            double variance        = (squared_sum_obs / number_obs) - (sum_obs * sum_obs) / (number_obs * number_obs);
            double mean            = sum_obs / number_obs;
            double lower_quant     = quant->get_lower_quant();
            double median          = quant->get_median();
            double upper_quant     = quant->get_upper_quant();

            if (number_obs > 0)
            {
                fprintf(cube_fp, "%-26s %9.0f %1.7f %1.7f %1.10f %1.10f %1.10f", patternName.c_str(), number_obs, mean, median, min_val, max_val, sum_obs);
                if (number_obs >= 2)
                {
                    fprintf(cube_fp, " %1.10f", variance);
                }
                if (number_obs >= 5)
                {
                    fprintf(cube_fp, " %1.10f %1.10f", lower_quant, upper_quant);
                }

                unsigned int i = 0;
                if (instance[i].idletime > 0.0)
                {
                    fprintf(cube_fp, "\n");
                }
                while (  (i < instance.size())
                      && (instance[i].idletime > 0.0))
                {
                    fprintf(cube_fp, "- cnode %d enter: %1.10f exit: %1.10f duration: %1.10f rank: %d\n", data.cnodes[instance[i].cnode]->id, instance[i].entertime, instance[i].exittime, instance[i].idletime, instance[i].rank);
                    i++;
                }

                fprintf(cube_fp, "\n");
            }
        }

        #ifdef WRITE_CONTROL_VALUES
            #if defined(_MPI)
                vector< double > md_ls, md_lsw, md_lr, md_lrw, md_wnxn, md_wb, md_er, md_es, md_lb, md_bc, md_nxnc;
            #endif    // _MPI

            #if defined(_OPENMP)
                vector< double > md_omp_eb, md_omp_ib;
                vector< double > md_thread_lc_ol, md_thread_lc_oc, md_thread_lc_pm, md_thread_lc_pc;
            #endif    // _OPENMP

            void
            write_values(Quantile* quant)
            {
                string filename = "stats/evaluation/" + quant->get_metric() + "_values.txt";
                FILE*  quant_fp = fopen(filename.c_str(), "w");

                int    opt_number_coeff = quant->get_number_coeff();
                double number_obs       = quant->get_n();
                double sum_obs          = quant->get_sum();
                double squared_sum_obs  = quant->get_squared_sum();
                double min_val          = quant->get_min_val();
                double max_val          = quant->get_max_val();
                double factor           = quant->get_factor();
                double coeff[NUMBER_COEFF];

                for (int i = 0; i < NUMBER_COEFF; i++)
                {
                    coeff[i] = quant->get_coeff(i);
                }

                double mean     = sum_obs / number_obs;
                double variance = (squared_sum_obs / number_obs) - (sum_obs * sum_obs) / (number_obs * number_obs);

                fprintf(quant_fp, "Number of all coefficients and all observations n : %d and %10.0f\n", NUMBER_COEFF, number_obs);
                fprintf(quant_fp, "Optimal number of coefficients: %d\n", opt_number_coeff);
                fprintf(quant_fp, "sum : %2.10f, squared sum: %2.10f\n", sum_obs, squared_sum_obs);
                fprintf(quant_fp, "mean : %2.10f, variance: %2.10f\n", mean, variance);

                double lower_quant = quant->get_lower_quant();
                double median      = quant->get_median();
                double upper_quant = quant->get_upper_quant();

                fprintf(quant_fp, "upper bound   : %2.15f\n\n", (1 / factor));
                fprintf(quant_fp, "minimum       : %2.15f\n", min_val);
                fprintf(quant_fp, "lower quartile: %2.15f\n", (lower_quant));
                fprintf(quant_fp, "median        : %2.15f\n", (median));
                fprintf(quant_fp, "upper quartile: %2.15f\n", (upper_quant));
                fprintf(quant_fp, "maximum       : %2.15f\n\n", max_val);

                for (int i = 0; i < NUMBER_COEFF; i++)
                {
                    fprintf(quant_fp, "coeff nr %2d : %2.15f\n", (i + 1), coeff[i]);
                }

                fclose(quant_fp);
            }
        #endif

        #if defined(_MPI)
            /// @brief Collate most severe waitstate instances on rank 0.
            ///
            /// Performs a reduction of the process-local waitstate instance
            /// tracking vector @p instance to collate the globally most
            /// severe instances on rank 0.
            ///
            /// @param instance
            ///     Process-local waitstate instance tracking vector.
            ///     Contains the globally most severe instances on rank 0
            ///     after the call.
            /// @param rank
            ///     Global process rank
            ///
            void
            result_top_most(vector< TopMostSevere >& instance,
                            const int                rank)
            {
                // Skip reduction if there are no instances
                int maxCount = instance.size();
                MPI_Allreduce(MPI_IN_PLACE, &maxCount, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
                if (maxCount == 0)
                {
                    return;
                }

                // Since the reduction operator assumes fixed-sized arrays,
                // fill up the vector with dummy entries if necessary
                instance.resize(TOP_SEVERE_INSTANCES);

                // Collate most severe instances on rank 0
                MPI_Reduce((rank == 0) ? MPI_IN_PLACE : instance.data(),
                           instance.data(), 1, TOPSEVEREARRAY,
                           MAX_TOPSEVEREARRAY, 0, MPI_COMM_WORLD);
            }
        #endif    // _MPI

        /// @brief Track a local waitstate instance.
        ///
        /// Compare a detected local waitstate instance with the given
        /// properties with the entries in the @p instance vector tracking
        /// the @$N@$ most severe instances at different callpaths, and
        /// update @p instance if necessary.  The entries in the vector
        /// are always kept sorted.
        ///
        /// @param instance
        ///     Vector tracking the @$N@$ most severe waitstate instances
        /// @param idle
        ///     Local waitstate idle time
        /// @param enter
        ///     Local waitstate begin time
        /// @param exit
        ///     Local waitstate end time
        /// @param cnode
        ///     Local waitstate callpath ID
        /// @param rank
        ///     Global process rank
        ///
        void
        stat_collector(vector< TopMostSevere >& instance,
                       const pearl::timestamp_t idle,
                       const pearl::timestamp_t enter,
                       const pearl::timestamp_t exit,
                       const pearl::ident_t     cnode,
                       const pearl::ident_t     rank)
        {
            TopMostSevere item(idle, enter, exit, cnode, rank);

            // Search for an existing entry with identical callpath ID
            auto it = std::find_if(instance.begin(), instance.end(),
                                   [&item](const TopMostSevere& entry)
                                   {
                                       return (entry.cnode == item.cnode);
                                   });

            // Found => Update entry and "bubble" it to the correct position
            if (it != instance.end())
            {
                *it = std::max(*it, item);

                auto prev = std::prev(it);
                while (  (it != instance.begin())
                      && (*prev < *it))
                {
                    std::swap(*prev, *it);

                    it = prev--;
                }

                return;
            }

            // Not found => Insert item at the correct position
            it = std::find_if(instance.begin(), instance.end(),
                              [&item](const TopMostSevere& entry)
                              {
                                  return (entry < item);
                              });
            if (instance.size() < TOP_SEVERE_INSTANCES)
            {
                // Case 1: Limit not yet reached => insert item
                instance.insert(it, item);

                return;
            }

            if (it == instance.end())
            {
                // Case 2: Limit reached and item less than any existing entry
                //         => no update
                return;
            }

            // Case 3: Limit reached and item larger than any existing entry
            //         => insert, dropping smallest entry
            std::move_backward(it, std::prev(instance.end()), instance.end());
            *it = item;
        }

        #if defined(_MPI)
            /// @brief Reduction of collective MPI waitstate information.
            ///
            /// Merges partial waitstate information for collective MPI
            /// operations, and passes the result to the local waitstate
            /// instance tracking.
            ///
            /// @param inst
            ///     Vector tracking the @$N@$ most severe waitstate instances
            /// @param idle
            ///     Local waitstate idle time
            /// @param enter
            ///     Local waitstate begin timestamp
            /// @param exit
            ///     Local waitstate timestamp
            /// @param cnode
            ///     Local waitstate callpath ID
            /// @param rank
            ///     Global process rank
            /// @param comm
            ///     Communicator used in the MPI collective operation
            ///
            void
            stat_collective(vector< TopMostSevere >& inst,
                            const pearl::timestamp_t idle,
                            const pearl::timestamp_t enter,
                            const pearl::timestamp_t exit,
                            const Callpath::IdType   cnode,
                            const int                rank,
                            const Communicator&      comm)
            {
                const bool isRoot = (comm.getCommSet().getGlobalRank(0)
                                     == static_cast< std::uint32_t >(rank));

                TopMostSevereMaxSum buffer(idle, enter, exit, cnode, rank);
                MPI_Reduce(isRoot ? MPI_IN_PLACE : &buffer,
                           &buffer, 1, TOPSEVERECOLL,
                           MAXSUM_TOPSEVERECOLL, 0, getMpiHandle(comm));

                if (  isRoot
                   && (buffer.idletime > 0))
                {
                    stat_collector(inst,
                                   buffer.idletime,
                                   buffer.entertime,
                                   buffer.exittime,
                                   buffer.cnode,
                                   buffer.rank);
                }
            }
        #endif    // _MPI

        #if defined(_OPENMP)
            /// @brief Reduction of collective OpenMP waitstate information.
            ///
            /// Merges partial waitstate information for OpenMP collective
            /// constructs (i.e., barriers), and passes the result to the
            /// local waitstate instance tracking.
            ///
            /// @param inst
            ///     Vector tracking the @$N@$ most severe waitstate instances
            /// @param idle
            ///     Local waitstate idle time
            /// @param enter
            ///     Local waitstate begin timestamp
            /// @param exit
            ///     Local waitstate timestamp
            /// @param cnode
            ///     Local waitstate callpath ID
            /// @param rank
            ///     Global process rank
            ///
            void
            stat_collective(vector< TopMostSevere >& inst,
                            const pearl::timestamp_t idle,
                            const pearl::timestamp_t enter,
                            const pearl::timestamp_t exit,
                            const Callpath::IdType   cnode,
                            const int                rank)
            {
                // This variable is implicitly shared
                static TopMostSevereMaxSum sharedMaxSum;

                #pragma omp barrier
                #pragma omp master
                {
                    sharedMaxSum = TopMostSevereMaxSum(0.0,
                                                       std::numeric_limits< double >::max(),
                                                       std::numeric_limits< double >::lowest(),
                                                       Callpath::NO_ID,
                                                       -1);
                }
                #pragma omp barrier
                #pragma omp critical
                {
                    sharedMaxSum = merge(sharedMaxSum, TopMostSevereMaxSum(idle, enter, exit, cnode, rank));
                }
                #pragma omp barrier
                #pragma omp master
                {
                    if (sharedMaxSum.idletime > 0)
                    {
                        stat_collector(inst,
                                       sharedMaxSum.idletime,
                                       sharedMaxSum.entertime,
                                       sharedMaxSum.exittime,
                                       sharedMaxSum.cnode,
                                       sharedMaxSum.rank);
                    }
                }
            }
        #endif    // _OPENMP


    public:
        /// @brief Collate and write wait-state statistics report.
        ///
        /// Generates the wait-state statistics report.  Therefore, it first
        /// collates the instance statistics from each process/thread
        /// on a @"master node@" (details depend on the parallel programming
        /// paradigm used) and then writes the gathered data to file.
        ///
        /// @param data
        ///     Temporary data used for report writing
        /// @param rank
        ///     Global process rank (MPI rank or 0)
        /// @param trace
        ///     Local trace-data object
        ///
        void
        collateSeverities(ReportData&              data,
                          int                      rank,
                          const pearl::LocalTrace& trace) override
        {
            #if defined(_MPI)
                result_merge(ls_quant);
                result_merge(lsw_quant);
                result_merge(lr_quant);
                result_merge(wnxn_quant);
                result_merge(wb_quant);
                result_merge(er_quant);
                result_merge(es_quant);
                result_merge(lb_quant);
                result_merge(bc_quant);
                result_merge(nxnc_quant);
            #endif    // _MPI
            #if defined(_OPENMP)
                result_merge(omp_eb_quant);
                result_merge(omp_ib_quant);
                result_merge(thread_lc_ol_quant);
                result_merge(thread_lc_oc_quant);
                result_merge(thread_lc_pm_quant);
                result_merge(thread_lc_pc_quant);
            #endif

            #if defined(_OPENMP)
                #pragma omp barrier
                result_merge_thread_topmost(sharedLC_OL, ThreadLockContentionOMPLockAPI);
                result_merge_thread_topmost(sharedLC_OC, ThreadLockContentionOMPCritical);
                result_merge_thread_topmost(sharedLC_PM, ThreadLockContentionPthreadMutex);
                result_merge_thread_topmost(sharedLC_PC, ThreadLockContentionPthreadCondition);

                #pragma omp barrier
            #endif    // _OPENMP

            #ifdef WRITE_CONTROL_VALUES
                #pragma omp master
                {
                    if (rank == 0)
                    {
                        mkdir("stats", 0700);
                        mkdir("stats/evaluation", 0700);
                        #if defined(_MPI)
                            mkdir("stats/ls_stats", 0700);
                            mkdir("stats/lsw_stats", 0700);
                            mkdir("stats/lr_stats", 0700);
                            mkdir("stats/wnxn_stats", 0700);
                            mkdir("stats/wb_stats", 0700);
                            mkdir("stats/er_stats", 0700);
                            mkdir("stats/es_stats", 0700);
                            mkdir("stats/lb_stats", 0700);
                            mkdir("stats/bc_stats", 0700);
                            mkdir("stats/nxnc_stats", 0700);
                        #endif    // _MPI

                        #if defined(_OPENMP)
                            mkdir("stats/omp_eb_stats", 0700);
                            mkdir("stats/omp_ib_stats", 0700);
                            mkdir("stats/thread_lc_ol_stats", 0700);
                            mkdir("stats/thread_lc_oc_stats", 0700);
                            mkdir("stats/thread_lc_pm_stats", 0700);
                            mkdir("stats/thread_lc_pc_stats", 0700);
                        #endif    // _OPENMP
                    }
                    #if defined(_MPI)
                        MPI_Barrier(MPI_COMM_WORLD);
                    #endif    // _MPI
                }

                // write values for OpenMp patterns
                #if defined(_OPENMP)
                    char         name_omp[50];
                    FILE*        stats_omp_fp;
                    unsigned int thread_id;
                    thread_id = omp_get_thread_num();
                    #pragma omp barrier

                    // OpenMP explicit barrier
                    snprintf(name_omp, 50, "stats/omp_eb_stats/omp_eb_stats.%06dx%06u", rank, thread_id);
                    stats_omp_fp = fopen(name_omp, "w");
                    for (size_t i = 0; i < md_omp_eb.size(); i++)
                    {
                        fprintf(stats_omp_fp, "%2.14f\n", md_omp_eb[i]);
                    }
                    fclose(stats_omp_fp);

                    // OpenMP implicit barrier
                    snprintf(name_omp, 50, "stats/omp_ib_stats/omp_ib_stats.%06dx%06u", rank, thread_id);
                    stats_omp_fp = fopen(name_omp, "w");
                    for (size_t i = 0; i < md_omp_ib.size(); i++)
                    {
                        fprintf(stats_omp_fp, "%2.14f\n", md_omp_ib[i]);
                    }
                    fclose(stats_omp_fp);

                    // Thread LockContention OMP Lock API
                    snprintf(name_omp, 50, "stats/thread_lc_ol_stats/thread_lc_ol_stats.%06dx%06u", rank, thread_id);
                    stats_omp_fp = fopen(name_omp, "w");
                    for (size_t i = 0; i < md_thread_lc_ol.size(); i++)
                    {
                        fprintf(stats_omp_fp, "%2.14f\n", md_thread_lc_ol[i]);
                    }
                    fclose(stats_omp_fp);

                    // Thread LockContention OMP Critical
                    snprintf(name_omp, 50, "stats/thread_lc_oc_stats/thread_lc_oc_stats.%06dx%06u", rank, thread_id);
                    stats_omp_fp = fopen(name_omp, "w");
                    for (size_t i = 0; i < md_thread_lc_oc.size(); i++)
                    {
                        fprintf(stats_omp_fp, "%2.14f\n", md_thread_lc_oc[i]);
                    }
                    fclose(stats_omp_fp);

                    // Thread LockContention Pthread Mutex
                    snprintf(name_omp, 50, "stats/thread_lc_pm_stats/thread_lc_pm_stats.%06dx%06u", rank, thread_id);
                    stats_omp_fp = fopen(name_omp, "w");
                    for (size_t i = 0; i < md_thread_lc_pm.size(); i++)
                    {
                        fprintf(stats_omp_fp, "%2.14f\n", md_thread_lc_pm[i]);
                    }
                    fclose(stats_omp_fp);

                    // Thread LockContention Pthread Condition
                    snprintf(name_omp, 50, "stats/thread_lc_pc_stats/thread_lc_pc_stats.%06dx%06u", rank, thread_id);
                    stats_omp_fp = fopen(name_omp, "w");
                    for (size_t i = 0; i < md_thread_lc_pc.size(); i++)
                    {
                        fprintf(stats_omp_fp, "%2.14f\n", md_thread_lc_pc[i]);
                    }
                    fclose(stats_omp_fp);
                #endif    // _OPENMP

                #if defined(_MPI)
                    #pragma omp master
                    {
                        char  name_mpi[40];
                        FILE* stats_mpi_fp;

                        // Late Sender values
                        snprintf(name_mpi, 40, "stats/ls_stats/ls_stats.%06d", rank);
                        stats_mpi_fp = fopen(name_mpi, "w");
                        for (size_t i = 0; i < md_ls.size(); i++)
                        {
                            fprintf(stats_mpi_fp, "%2.14f\n", md_ls[i]);
                        }
                        fclose(stats_mpi_fp);

                        // Late Sender, Wrong Order values
                        snprintf(name_mpi, 40, "stats/lsw_stats/lsw_stats.%06d", rank);
                        stats_mpi_fp = fopen(name_mpi, "w");
                        for (size_t i = 0; i < md_lsw.size(); i++)
                        {
                            fprintf(stats_mpi_fp, "%2.14f\n", md_lsw[i]);
                        }
                        fclose(stats_mpi_fp);

                        // Late Receiver values
                        snprintf(name_mpi, 40, "stats/lr_stats/lr_stats.%06d", rank);
                        stats_mpi_fp = fopen(name_mpi, "w");
                        for (size_t i = 0; i < md_lr.size(); i++)
                        {
                            fprintf(stats_mpi_fp, "%2.14f\n", md_lr[i]);
                        }
                        fclose(stats_mpi_fp);

                        // Wait at NxN values
                        snprintf(name_mpi, 40, "stats/wnxn_stats/wnxn_stats.%06d", rank);
                        stats_mpi_fp = fopen(name_mpi, "w");
                        for (size_t i = 0; i < md_wnxn.size(); i++)
                        {
                            fprintf(stats_mpi_fp, "%2.14f\n", md_wnxn[i]);
                        }
                        fclose(stats_mpi_fp);

                        // Wait at Barrier values
                        snprintf(name_mpi, 40, "stats/wb_stats/wb_stats.%06d", rank);
                        stats_mpi_fp = fopen(name_mpi, "w");
                        for (size_t i = 0; i < md_wb.size(); i++)
                        {
                            fprintf(stats_mpi_fp, "%2.14f\n", md_wb[i]);
                        }
                        fclose(stats_mpi_fp);

                        // Early Reduce values
                        snprintf(name_mpi, 40, "stats/er_stats/er_stats.%06d", rank);
                        stats_mpi_fp = fopen(name_mpi, "w");
                        for (size_t i = 0; i < md_er.size(); i++)
                        {
                            fprintf(stats_mpi_fp, "%2.14f\n", md_er[i]);
                        }
                        fclose(stats_mpi_fp);

                        // Early Scan values
                        snprintf(name_mpi, 40, "stats/es_stats/es_stats.%06d", rank);
                        stats_mpi_fp = fopen(name_mpi, "w");
                        for (size_t i = 0; i < md_es.size(); i++)
                        {
                            fprintf(stats_mpi_fp, "%2.14f\n", md_es[i]);
                        }
                        fclose(stats_mpi_fp);

                        // Late Broadcast values
                        snprintf(name_mpi, 40, "stats/lb_stats/lb_stats.%06d", rank);
                        stats_mpi_fp = fopen(name_mpi, "w");
                        for (size_t i = 0; i < md_lb.size(); i++)
                        {
                            fprintf(stats_mpi_fp, "%2.14f\n", md_lb[i]);
                        }
                        fclose(stats_mpi_fp);

                        // Barrier Completion values
                        snprintf(name_mpi, 40, "stats/bc_stats/bc_stats.%06d", rank);
                        stats_mpi_fp = fopen(name_mpi, "w");
                        for (size_t i = 0; i < md_bc.size(); i++)
                        {
                            fprintf(stats_mpi_fp, "%2.14f\n", md_bc[i]);
                        }
                        fclose(stats_mpi_fp);

                        // NxN Completion values
                        snprintf(name_mpi, 40, "stats/nxnc_stats/nxnc_stats.%06d", rank);
                        stats_mpi_fp = fopen(name_mpi, "w");
                        for (size_t i = 0; i < md_nxnc.size(); i++)
                        {
                            fprintf(stats_mpi_fp, "%2.14f\n", md_nxnc[i]);
                        }
                        fclose(stats_mpi_fp);
                    }
                #endif // _MPI
            #endif     // WRITE_CONTROL_VALUES

            #pragma omp master
            {
                // Merge results for each pattern
                #if defined(_MPI)
                    result_top_most(LateSender, rank);
                    result_top_most(LateSenderWO, rank);
                    result_top_most(LateReceiver, rank);
                    result_top_most(EarlyReduce, rank);
                    result_top_most(BarrierNxN, rank);
                    result_top_most(Barrier, rank);
                    result_top_most(NxnCompl, rank);
                    result_top_most(BarrierCompl, rank);
                    result_top_most(LateBcast, rank);
                    result_top_most(EarlyScan, rank);
                    #if defined(_OPENMP)
                        result_top_most(OmpEBarrier, rank);
                        result_top_most(OmpIBarrier, rank);
                        result_top_most(sharedLC_OL, rank);
                        result_top_most(sharedLC_OC, rank);
                        result_top_most(sharedLC_PM, rank);
                        result_top_most(sharedLC_PC, rank);
                    #endif // _OPENMP
                #endif     // _MPI

                if (rank == 0)
                {
                    // Open statistics file
                    string filename = archiveDirectory + "/trace.stat";
                    FILE*  cube_fp  = fopen(filename.c_str(), "w");
                    if (cube_fp == nullptr)
                    {
                        // Only warn (instead of abort) since we still might get a
                        // valid CUBE file
                        UTILS_WARNING("Could not create statistics file!");
                    }
                    else
                    {
                        // Write statistics
                        fprintf(cube_fp, "PatternName               Count      Mean    Median      Minimum      Maximum      Sum     Variance    Quartil25    Quartil75\n");

                        #if defined(_MPI)
                            ls_quant->set_metric("mpi_latesender");
                            ls_quant->calc_quantiles();
                            write_cube_file(ls_quant, LateSender, data, cube_fp);

                            lsw_quant->set_metric("mpi_latesender_wo");
                            lsw_quant->calc_quantiles();
                            write_cube_file(lsw_quant, LateSenderWO, data, cube_fp);

                            lr_quant->set_metric("mpi_latereceiver");
                            lr_quant->calc_quantiles();
                            write_cube_file(lr_quant, LateReceiver, data, cube_fp);

                            wnxn_quant->set_metric("mpi_wait_nxn");
                            wnxn_quant->calc_quantiles();
                            write_cube_file(wnxn_quant, BarrierNxN, data, cube_fp);

                            wb_quant->set_metric("mpi_barrier_wait");
                            wb_quant->calc_quantiles();
                            write_cube_file(wb_quant, Barrier, data, cube_fp);

                            er_quant->set_metric("mpi_earlyreduce");
                            er_quant->calc_quantiles();
                            write_cube_file(er_quant, EarlyReduce, data, cube_fp);

                            es_quant->set_metric("mpi_earlyscan");
                            es_quant->calc_quantiles();
                            write_cube_file(es_quant, EarlyScan, data, cube_fp);

                            lb_quant->set_metric("mpi_latebroadcast");
                            lb_quant->calc_quantiles();
                            write_cube_file(lb_quant, LateBcast, data, cube_fp);

                            bc_quant->set_metric("mpi_barrier_completion");
                            bc_quant->calc_quantiles();
                            write_cube_file(bc_quant, BarrierCompl, data, cube_fp);

                            nxnc_quant->set_metric("mpi_nxn_completion");
                            nxnc_quant->calc_quantiles();
                            write_cube_file(nxnc_quant, NxnCompl, data, cube_fp);
                        #endif    // _MPI

                        #if defined(_OPENMP)
                            omp_eb_quant->set_metric("omp_ebarrier_wait");
                            omp_eb_quant->calc_quantiles();
                            write_cube_file(omp_eb_quant, OmpEBarrier, data, cube_fp);

                            omp_ib_quant->set_metric("omp_ibarrier_wait");
                            omp_ib_quant->calc_quantiles();
                            write_cube_file(omp_ib_quant, OmpIBarrier, data, cube_fp);

                            thread_lc_ol_quant->set_metric("omp_lock_contention_api");
                            thread_lc_ol_quant->calc_quantiles();
                            write_cube_file(thread_lc_ol_quant, sharedLC_OL, data, cube_fp);

                            thread_lc_oc_quant->set_metric("omp_lock_contention_critical");
                            thread_lc_oc_quant->calc_quantiles();
                            write_cube_file(thread_lc_oc_quant, sharedLC_OC, data, cube_fp);

                            thread_lc_pm_quant->set_metric("pthread_lock_contention_mutex_lock");
                            thread_lc_pm_quant->calc_quantiles();
                            write_cube_file(thread_lc_pm_quant, sharedLC_PM, data, cube_fp);

                            thread_lc_pc_quant->set_metric("pthread_lock_contention_conditional");
                            thread_lc_pc_quant->calc_quantiles();
                            write_cube_file(thread_lc_pc_quant, sharedLC_PC, data, cube_fp);
                        #endif    // _OPENMP

                        // Close statistics file
                        fclose(cube_fp);

                        #ifdef WRITE_CONTROL_VALUES
                            // Write coefficients per pattern for debugging
                            #if defined(_MPI)
                                write_values(ls_quant);
                                write_values(lsw_quant);
                                write_values(lr_quant);
                                write_values(wnxn_quant);
                                write_values(wb_quant);
                                write_values(er_quant);
                                write_values(es_quant);
                                write_values(lb_quant);
                                write_values(bc_quant);
                                write_values(nxnc_quant);
                            #endif    // _MPI

                            #if defined(_OPENMP)
                                write_values(omp_eb_quant);
                                write_values(omp_ib_quant);
                                write_values(thread_lc_ol_quant);
                                write_values(thread_lc_oc_quant);
                                write_values(thread_lc_pm_quant);
                                write_values(thread_lc_pc_quant);
                            #endif // _OPENMP
                        #endif     // WRITE_CONTROL_VALUES
                    }
                }
            }
        }

        /// @name Callback functions (main)
        /// @{

        /// @brief Callback function for PREPARE events.
        ///
        /// Callback function triggered by PREPARE events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        prepare_cb(const CallbackManager& cbmanager,
                   int                    user_event,
                   const Event&           event,
                   CallbackData*          cdata);

        /// @brief Callback function for LATE_SENDER events.
        ///
        /// Callback function triggered by LATE_SENDER events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        late_sender_cb(const CallbackManager& cbmanager,
                       int                    user_event,
                       const Event&           event,
                       CallbackData*          cdata);

        /// @brief Callback function for LATE_SENDER_WO events.
        ///
        /// Callback function triggered by LATE_SENDER_WO events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        late_sender_wo_cb(const CallbackManager& cbmanager,
                          int                    user_event,
                          const Event&           event,
                          CallbackData*          cdata);

        /// @brief Callback function for WAIT_NXN events.
        ///
        /// Callback function triggered by WAIT_NXN events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        wait_nxn_cb(const CallbackManager& cbmanager,
                    int                    user_event,
                    const Event&           event,
                    CallbackData*          cdata);

        /// @brief Callback function for WAIT_BARRIER events.
        ///
        /// Callback function triggered by WAIT_BARRIER events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        wait_barrier_cb(const CallbackManager& cbmanager,
                        int                    user_event,
                        const Event&           event,
                        CallbackData*          cdata);

        /// @brief Callback function for EARLY_REDUCE events.
        ///
        /// Callback function triggered by EARLY_REDUCE events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        early_reduce_cb(const CallbackManager& cbmanager,
                        int                    user_event,
                        const Event&           event,
                        CallbackData*          cdata);

        /// @brief Callback function for EARLY_SCAN events.
        ///
        /// Callback function triggered by EARLY_SCAN events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        early_scan_cb(const CallbackManager& cbmanager,
                      int                    user_event,
                      const Event&           event,
                      CallbackData*          cdata);

        /// @brief Callback function for LATE_BCAST events.
        ///
        /// Callback function triggered by LATE_BCAST events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        late_bcast_cb(const CallbackManager& cbmanager,
                      int                    user_event,
                      const Event&           event,
                      CallbackData*          cdata);

        /// @brief Callback function for NXN_COMPL events.
        ///
        /// Callback function triggered by NXN_COMPL events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        nxn_compl_cb(const CallbackManager& cbmanager,
                     int                    user_event,
                     const Event&           event,
                     CallbackData*          cdata);

        /// @brief Callback function for BARRIER_COMPL events.
        ///
        /// Callback function triggered by BARRIER_COMPL events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        barrier_compl_cb(const CallbackManager& cbmanager,
                         int                    user_event,
                         const Event&           event,
                         CallbackData*          cdata);

        /// @brief Callback function for OMP_EBARRIER_WAIT events.
        ///
        /// Callback function triggered by OMP_EBARRIER_WAIT events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        omp_ebarrier_wait_cb(const CallbackManager& cbmanager,
                             int                    user_event,
                             const Event&           event,
                             CallbackData*          cdata);

        /// @brief Callback function for OMP_IBARRIER_WAIT events.
        ///
        /// Callback function triggered by OMP_IBARRIER_WAIT events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        omp_ibarrier_wait_cb(const CallbackManager& cbmanager,
                             int                    user_event,
                             const Event&           event,
                             CallbackData*          cdata);

        /// @brief Callback function for THREAD_LOCK_CONTENTION_OMPLOCK_WAIT events.
        ///
        /// Callback function triggered by THREAD_LOCK_CONTENTION_OMPLOCK_WAIT events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        thread_lock_contention_omplock_wait_cb(const CallbackManager& cbmanager,
                                               int                    user_event,
                                               const Event&           event,
                                               CallbackData*          cdata);

        /// @brief Callback function for THREAD_LOCK_CONTENTION_OMPCRITICAL_WAIT events.
        ///
        /// Callback function triggered by THREAD_LOCK_CONTENTION_OMPCRITICAL_WAIT events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        thread_lock_contention_ompcritical_wait_cb(const CallbackManager& cbmanager,
                                                   int                    user_event,
                                                   const Event&           event,
                                                   CallbackData*          cdata);

        /// @brief Callback function for THREAD_LOCK_CONTENTION_PTHREADMUTEX_WAIT events.
        ///
        /// Callback function triggered by THREAD_LOCK_CONTENTION_PTHREADMUTEX_WAIT events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        thread_lock_contention_pthreadmutex_wait_cb(const CallbackManager& cbmanager,
                                                    int                    user_event,
                                                    const Event&           event,
                                                    CallbackData*          cdata);

        /// @brief Callback function for THREAD_LOCK_CONTENTION_PTHREADCONDITION_WAIT events.
        ///
        /// Callback function triggered by THREAD_LOCK_CONTENTION_PTHREADCONDITION_WAIT events
        /// during the @"main@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        thread_lock_contention_pthreadcondition_wait_cb(const CallbackManager& cbmanager,
                                                        int                    user_event,
                                                        const Event&           event,
                                                        CallbackData*          cdata);

        /// @}
        /// @name Callback functions (bws)
        /// @{

        /// @brief Callback function for LATE_RECEIVER events.
        ///
        /// Callback function triggered by LATE_RECEIVER events
        /// during the @"bws@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        late_receiver_bws_cb(const CallbackManager& cbmanager,
                             int                    user_event,
                             const Event&           event,
                             CallbackData*          cdata);

        /// @}
};


// --- Static class data ----------------------------------------------------

#line 27 "Statistics.pattern"

#if defined(_OPENMP)
    vector< TopMostSevere > Statistics::sharedLC_OL;
    vector< TopMostSevere > Statistics::sharedLC_OC;
    vector< TopMostSevere > Statistics::sharedLC_PM;
    vector< TopMostSevere > Statistics::sharedLC_PC;
#endif    // (_OPENMP)

Statistics::MaxDurations Statistics::sharedDurations;


// --- Callback registration ------------------------------------------------

void
Statistics::registerCallbacks(CallbackManagerMap& cbmanagers)
{
    TypedPattern< double >::registerCallbacks(cbmanagers);

    CallbackManagerMap::iterator    it;
    CallbackManagerMap::mapped_type cbmanager;

    // Register main callbacks
    it = cbmanagers.find("");
    UTILS_ASSERT(it != cbmanagers.end());

    cbmanager = it->second;
    cbmanager->register_callback(
        PREPARE,
        PEARL_create_callback(
            this,
            &Statistics::prepare_cb));
    cbmanager->register_callback(
        LATE_SENDER,
        PEARL_create_callback(
            this,
            &Statistics::late_sender_cb));
    cbmanager->register_callback(
        LATE_SENDER_WO,
        PEARL_create_callback(
            this,
            &Statistics::late_sender_wo_cb));
    cbmanager->register_callback(
        WAIT_NXN,
        PEARL_create_callback(
            this,
            &Statistics::wait_nxn_cb));
    cbmanager->register_callback(
        WAIT_BARRIER,
        PEARL_create_callback(
            this,
            &Statistics::wait_barrier_cb));
    cbmanager->register_callback(
        EARLY_REDUCE,
        PEARL_create_callback(
            this,
            &Statistics::early_reduce_cb));
    cbmanager->register_callback(
        EARLY_SCAN,
        PEARL_create_callback(
            this,
            &Statistics::early_scan_cb));
    cbmanager->register_callback(
        LATE_BCAST,
        PEARL_create_callback(
            this,
            &Statistics::late_bcast_cb));
    cbmanager->register_callback(
        NXN_COMPL,
        PEARL_create_callback(
            this,
            &Statistics::nxn_compl_cb));
    cbmanager->register_callback(
        BARRIER_COMPL,
        PEARL_create_callback(
            this,
            &Statistics::barrier_compl_cb));
    cbmanager->register_callback(
        OMP_EBARRIER_WAIT,
        PEARL_create_callback(
            this,
            &Statistics::omp_ebarrier_wait_cb));
    cbmanager->register_callback(
        OMP_IBARRIER_WAIT,
        PEARL_create_callback(
            this,
            &Statistics::omp_ibarrier_wait_cb));
    cbmanager->register_callback(
        THREAD_LOCK_CONTENTION_OMPLOCK_WAIT,
        PEARL_create_callback(
            this,
            &Statistics::thread_lock_contention_omplock_wait_cb));
    cbmanager->register_callback(
        THREAD_LOCK_CONTENTION_OMPCRITICAL_WAIT,
        PEARL_create_callback(
            this,
            &Statistics::thread_lock_contention_ompcritical_wait_cb));
    cbmanager->register_callback(
        THREAD_LOCK_CONTENTION_PTHREADMUTEX_WAIT,
        PEARL_create_callback(
            this,
            &Statistics::thread_lock_contention_pthreadmutex_wait_cb));
    cbmanager->register_callback(
        THREAD_LOCK_CONTENTION_PTHREADCONDITION_WAIT,
        PEARL_create_callback(
            this,
            &Statistics::thread_lock_contention_pthreadcondition_wait_cb));

    // Register bws callbacks
    it = cbmanagers.find("bws");
    UTILS_ASSERT(it != cbmanagers.end());

    cbmanager = it->second;
    cbmanager->register_callback(
        LATE_RECEIVER,
        PEARL_create_callback(
            this,
            &Statistics::late_receiver_bws_cb));
}


// --- Callback functions (main) --------------------------------------------

void
Statistics::prepare_cb(const CallbackManager& cbmanager,
                       const int              user_event,
                       const Event&           event,
                       CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 1262 "Statistics.pattern"

    // Determine local upper bounds for pattern durations
    CallbackManager cb;
    MaxDurations    durations;

    #if defined(_MPI)
        cb.register_callback(MPI_COLLECTIVE_END, PEARL_create_callback(this, &Statistics::mpicollective_cb));
        cb.register_callback(GROUP_SEND, PEARL_create_callback(this, &Statistics::send_cb));
        cb.register_callback(GROUP_RECV, PEARL_create_callback(this, &Statistics::recv_cb));
    #endif    // _MPI

    #if defined(_OPENMP)
        cb.register_callback(GROUP_LEAVE, PEARL_create_callback(this, &Statistics::leave_cb));
        cb.register_callback(THREAD_ACQUIRE_LOCK, PEARL_create_callback(this, &Statistics::acquire_cb));
    #endif    // _OPENMP

    PEARL_forward_replay(*(data->mTrace), cb, &durations);

    // Determine global upper bounds for pattern durations
    #pragma omp master
    {
        sharedDurations.m_max_duration[OMP_EB_MAX_DURATION]       = std::numeric_limits< double >::lowest();
        sharedDurations.m_max_duration[OMP_IB_MAX_DURATION]       = std::numeric_limits< double >::lowest();
        sharedDurations.m_max_duration[THREAD_LC_OL_MAX_DURATION] = std::numeric_limits< double >::lowest();
        sharedDurations.m_max_duration[THREAD_LC_OC_MAX_DURATION] = std::numeric_limits< double >::lowest();
        sharedDurations.m_max_duration[THREAD_LC_PM_MAX_DURATION] = std::numeric_limits< double >::lowest();
        sharedDurations.m_max_duration[THREAD_LC_PC_MAX_DURATION] = std::numeric_limits< double >::lowest();
    }

    #pragma omp barrier
    #pragma omp critical
    {
        if (sharedDurations.m_max_duration[OMP_EB_MAX_DURATION] < durations.m_max_duration[OMP_EB_MAX_DURATION])
        {
            sharedDurations.m_max_duration[OMP_EB_MAX_DURATION] = durations.m_max_duration[OMP_EB_MAX_DURATION];
        }
        if (sharedDurations.m_max_duration[OMP_IB_MAX_DURATION] < durations.m_max_duration[OMP_IB_MAX_DURATION])
        {
            sharedDurations.m_max_duration[OMP_IB_MAX_DURATION] = durations.m_max_duration[OMP_IB_MAX_DURATION];
        }
        if (sharedDurations.m_max_duration[THREAD_LC_OL_MAX_DURATION] < durations.m_max_duration[THREAD_LC_OL_MAX_DURATION])
        {
            sharedDurations.m_max_duration[THREAD_LC_OL_MAX_DURATION] = durations.m_max_duration[THREAD_LC_OL_MAX_DURATION];
        }
        if (sharedDurations.m_max_duration[THREAD_LC_OC_MAX_DURATION] < durations.m_max_duration[THREAD_LC_OC_MAX_DURATION])
        {
            sharedDurations.m_max_duration[THREAD_LC_OC_MAX_DURATION] = durations.m_max_duration[THREAD_LC_OC_MAX_DURATION];
        }
        if (sharedDurations.m_max_duration[THREAD_LC_PM_MAX_DURATION] < durations.m_max_duration[THREAD_LC_PM_MAX_DURATION])
        {
            sharedDurations.m_max_duration[THREAD_LC_PM_MAX_DURATION] = durations.m_max_duration[THREAD_LC_PM_MAX_DURATION];
        }
        if (sharedDurations.m_max_duration[THREAD_LC_PC_MAX_DURATION] < durations.m_max_duration[THREAD_LC_PC_MAX_DURATION])
        {
            sharedDurations.m_max_duration[THREAD_LC_PC_MAX_DURATION] = durations.m_max_duration[THREAD_LC_PC_MAX_DURATION];
        }
    }
    #pragma omp barrier

    #if defined(_MPI)
        #pragma omp master
        {
            durations.m_max_duration[OMP_EB_MAX_DURATION]       = sharedDurations.m_max_duration[OMP_EB_MAX_DURATION];
            durations.m_max_duration[OMP_IB_MAX_DURATION]       = sharedDurations.m_max_duration[OMP_IB_MAX_DURATION];
            durations.m_max_duration[THREAD_LC_OL_MAX_DURATION] = sharedDurations.m_max_duration[THREAD_LC_OL_MAX_DURATION];
            durations.m_max_duration[THREAD_LC_OC_MAX_DURATION] = sharedDurations.m_max_duration[THREAD_LC_OC_MAX_DURATION];
            durations.m_max_duration[THREAD_LC_PM_MAX_DURATION] = sharedDurations.m_max_duration[THREAD_LC_PM_MAX_DURATION];
            durations.m_max_duration[THREAD_LC_PC_MAX_DURATION] = sharedDurations.m_max_duration[THREAD_LC_PC_MAX_DURATION];
            MPI_Allreduce(durations.m_max_duration,
                          sharedDurations.m_max_duration,
                          MAX_DURATION_ENTRIES, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
        }
    #endif    // _MPI

    #pragma omp barrier

    #if defined(_MPI)
        ls_quant->set_factor_by_upperbound(sharedDurations.m_max_duration[LS_MAX_DURATION]);
        lsw_quant->set_factor_by_upperbound(sharedDurations.m_max_duration[LS_MAX_DURATION]);
        lr_quant->set_factor_by_upperbound(sharedDurations.m_max_duration[LR_MAX_DURATION]);
        wnxn_quant->set_factor_by_upperbound(sharedDurations.m_max_duration[WNXN_MAX_DURATION]);
        wb_quant->set_factor_by_upperbound(sharedDurations.m_max_duration[WB_MAX_DURATION]);
        er_quant->set_factor_by_upperbound(sharedDurations.m_max_duration[ER_MAX_DURATION]);
        es_quant->set_factor_by_upperbound(sharedDurations.m_max_duration[ES_MAX_DURATION]);
        lb_quant->set_factor_by_upperbound(sharedDurations.m_max_duration[LB_MAX_DURATION]);
        bc_quant->set_factor_by_upperbound(sharedDurations.m_max_duration[BC_MAX_DURATION]);
        nxnc_quant->set_factor_by_upperbound(sharedDurations.m_max_duration[NXNC_MAX_DURATION]);
    #endif    // _MPI

    #if defined(_OPENMP)
        omp_eb_quant->set_factor_by_upperbound(sharedDurations.m_max_duration[OMP_EB_MAX_DURATION]);
        omp_ib_quant->set_factor_by_upperbound(sharedDurations.m_max_duration[OMP_IB_MAX_DURATION]);
        thread_lc_ol_quant->set_factor_by_upperbound(sharedDurations.m_max_duration[THREAD_LC_OL_MAX_DURATION]);
        thread_lc_oc_quant->set_factor_by_upperbound(sharedDurations.m_max_duration[THREAD_LC_OC_MAX_DURATION]);
        thread_lc_pm_quant->set_factor_by_upperbound(sharedDurations.m_max_duration[THREAD_LC_PM_MAX_DURATION]);
        thread_lc_pc_quant->set_factor_by_upperbound(sharedDurations.m_max_duration[THREAD_LC_PC_MAX_DURATION]);
    #endif    // _OPENMP
}


void
Statistics::late_sender_cb(const CallbackManager& cbmanager,
                           const int              user_event,
                           const Event&           event,
                           CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 1362 "Statistics.pattern"

    #if defined(_MPI)
        UTILS_ASSERT(data->mIdle > 0);

        ls_quant->add_value(data->mIdle);
        #ifdef WRITE_CONTROL_VALUES
            md_ls.push_back(data->mIdle);
        #endif

        const Event enter = data->mLocal->getEvent(ROLE_RECV_LS);

        stat_collector(LateSender,
                       data->mIdle,
                       enter.enterptr()->getTimestamp(),
                       event->getTimestamp(),
                       enter.get_cnode()->getId(),
                       event.get_location().getRank());
    #endif    // _MPI
}


void
Statistics::late_sender_wo_cb(const CallbackManager& cbmanager,
                              const int              user_event,
                              const Event&           event,
                              CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 1383 "Statistics.pattern"

    #if defined(_MPI)
        UTILS_ASSERT(data->mIdle > 0);

        lsw_quant->add_value(data->mIdle);
        #ifdef WRITE_CONTROL_VALUES
            md_lsw.push_back(data->mIdle);
        #endif

        const RemoteEvent enter = data->mRemote->getEvent(ROLE_SEND);
        const Event       exit  = data->mLocal->getEvent(ROLE_RECV_LSWO);

        stat_collector(LateSenderWO,
                       data->mIdle,
                       enter->getTimestamp(),
                       event->getTimestamp(),
                       exit.get_cnode()->getId(),
                       event.get_location().getRank());
    #endif    // _MPI
}


void
Statistics::wait_nxn_cb(const CallbackManager& cbmanager,
                        const int              user_event,
                        const Event&           event,
                        CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 1405 "Statistics.pattern"

    #if defined(_MPI)
        UTILS_ASSERT(data->mIdle >= 0);

        if (data->mIdle > 0)
        {
            wnxn_quant->add_value(data->mIdle);
            #ifdef WRITE_CONTROL_VALUES
                md_wnxn.push_back(data->mIdle);
            #endif
        }

        // FIXME: This should use ROLE_BEGIN_COLL instead to match WAIT_BARRIER!
        const Event begin = data->mLocal->getEvent(ROLE_ENTER_COLL);

        stat_collective(BarrierNxN,
                        data->mIdle,
                        begin->getTimestamp(),
                        event->getTimestamp(),
                        event.get_cnode()->getId(),
                        event.get_location().getRank(),
                        event->getComm());
    #endif    // _MPI
}


void
Statistics::wait_barrier_cb(const CallbackManager& cbmanager,
                            const int              user_event,
                            const Event&           event,
                            CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 1431 "Statistics.pattern"

    #if defined(_MPI)
        UTILS_ASSERT(data->mIdle >= 0);

        if (data->mIdle > 0)
        {
            wb_quant->add_value(data->mIdle);
            #ifdef WRITE_CONTROL_VALUES
                md_wb.push_back(data->mIdle);
            #endif
        }

        const Event begin = data->mLocal->getEvent(ROLE_BEGIN_COLL);

        stat_collective(Barrier,
                        data->mIdle,
                        begin->getTimestamp(),
                        event->getTimestamp(),
                        event.get_cnode()->getId(),
                        event.get_location().getRank(),
                        event->getComm());
    #endif    // _MPI
}


void
Statistics::early_reduce_cb(const CallbackManager& cbmanager,
                            const int              user_event,
                            const Event&           event,
                            CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 1456 "Statistics.pattern"

    #if defined(_MPI)
        UTILS_ASSERT(data->mIdle > 0);

        er_quant->add_value(data->mIdle);
        #ifdef WRITE_CONTROL_VALUES
            md_er.push_back(data->mIdle);
        #endif

        const Event begin = data->mLocal->getEvent(ROLE_BEGIN_COLL);

        stat_collector(EarlyReduce,
                       data->mIdle,
                       begin->getTimestamp(),
                       event->getTimestamp(),
                       event.get_cnode()->getId(),
                       event.get_location().getRank());
    #endif    // _MPI
}


void
Statistics::early_scan_cb(const CallbackManager& cbmanager,
                          const int              user_event,
                          const Event&           event,
                          CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 1477 "Statistics.pattern"

    #if defined(_MPI)
        UTILS_ASSERT(data->mIdle >= 0);

        if (data->mIdle > 0)
        {
            es_quant->add_value(data->mIdle);
            #ifdef WRITE_CONTROL_VALUES
                md_es.push_back(data->mIdle);
            #endif
        }

        const Event begin = data->mLocal->getEvent(ROLE_BEGIN_COLL);

        stat_collective(EarlyScan,
                        data->mIdle,
                        begin->getTimestamp(),
                        event->getTimestamp(),
                        event.get_cnode()->getId(),
                        event.get_location().getRank(),
                        event->getComm());
    #endif    // _MPI
}


void
Statistics::late_bcast_cb(const CallbackManager& cbmanager,
                          const int              user_event,
                          const Event&           event,
                          CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 1502 "Statistics.pattern"

    #if defined(_MPI)
        UTILS_ASSERT(data->mIdle >= 0);

        if (data->mIdle > 0)
        {
            lb_quant->add_value(data->mIdle);
            #ifdef WRITE_CONTROL_VALUES
                md_lb.push_back(data->mIdle);
            #endif
        }

        const Event begin = data->mLocal->getEvent(ROLE_BEGIN_COLL);

        stat_collective(LateBcast, data->mIdle,
                        begin->getTimestamp(),
                        event->getTimestamp(),
                        event.get_cnode()->getId(),
                        event.get_location().getRank(),
                        event->getComm());
    #endif    // _MPI
}


void
Statistics::nxn_compl_cb(const CallbackManager& cbmanager,
                         const int              user_event,
                         const Event&           event,
                         CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 1526 "Statistics.pattern"

    #if defined(_MPI)
        UTILS_ASSERT(data->mIdle >= 0);

        if (data->mCompletion > 0)
        {
            nxnc_quant->add_value(data->mCompletion);
            #ifdef WRITE_CONTROL_VALUES
                md_nxnc.push_back(data->mCompletion);
            #endif
        }

        const Event begin = data->mLocal->getEvent(ROLE_BEGIN_COLL);

        stat_collective(NxnCompl,
                        data->mCompletion,
                        begin->getTimestamp(),
                        event->getTimestamp(),
                        event.get_cnode()->getId(),
                        event.get_location().getRank(),
                        event->getComm());
    #endif    // _MPI
}


void
Statistics::barrier_compl_cb(const CallbackManager& cbmanager,
                             const int              user_event,
                             const Event&           event,
                             CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 1551 "Statistics.pattern"

    #if defined(_MPI)
        UTILS_ASSERT(data->mIdle >= 0);

        if (data->mCompletion > 0)
        {
            bc_quant->add_value(data->mCompletion);
            #ifdef WRITE_CONTROL_VALUES
                md_bc.push_back(data->mCompletion);
            #endif
        }

        const Event begin = data->mLocal->getEvent(ROLE_BEGIN_COLL);

        stat_collective(BarrierCompl,
                        data->mCompletion,
                        begin->getTimestamp(),
                        event->getTimestamp(),
                        event.get_cnode()->getId(),
                        event.get_location().getRank(),
                        event->getComm());
    #endif    // _MPI
}


void
Statistics::omp_ebarrier_wait_cb(const CallbackManager& cbmanager,
                                 const int              user_event,
                                 const Event&           event,
                                 CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 1576 "Statistics.pattern"

    #if defined(_OPENMP)
        UTILS_ASSERT(data->mIdle >= 0);

        if (data->mIdle > 0)
        {
            omp_eb_quant->add_value(data->mIdle);
            #ifdef WRITE_CONTROL_VALUES
                md_omp_eb.push_back(data->mIdle);
            #endif
        }

        // FIXME: For non-MPI, Location::getRank() is -1 => map to 0
        const int rank = std::max(0, event.get_location().getRank());

        const Event begin = data->mCallstack->top();

        stat_collective(OmpEBarrier,
                        data->mIdle,
                        begin->getTimestamp(),
                        event->getTimestamp(),
                        event.get_cnode()->getId(),
                        rank);
    #endif    // _OPENMP
}


void
Statistics::omp_ibarrier_wait_cb(const CallbackManager& cbmanager,
                                 const int              user_event,
                                 const Event&           event,
                                 CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 1603 "Statistics.pattern"

    #if defined(_OPENMP)
        UTILS_ASSERT(data->mIdle >= 0);

        if (data->mIdle > 0)
        {
            omp_ib_quant->add_value(data->mIdle);
            #ifdef WRITE_CONTROL_VALUES
                md_omp_ib.push_back(data->mIdle);
            #endif
        }

        // FIXME: For non-MPI, Location::getRank() is -1 => map to 0
        const int rank = std::max(0, event.get_location().getRank());

        const Event begin = data->mCallstack->top();

        stat_collective(OmpIBarrier,
                        data->mIdle,
                        begin->getTimestamp(),
                        event->getTimestamp(),
                        event.get_cnode()->getId(),
                        rank);
    #endif    // _OPENMP
}


void
Statistics::thread_lock_contention_omplock_wait_cb(const CallbackManager& cbmanager,
                                                   const int              user_event,
                                                   const Event&           event,
                                                   CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 1630 "Statistics.pattern"

    #if defined(_OPENMP)
        UTILS_ASSERT(data->mIdle > 0);

        thread_lc_ol_quant->add_value(data->mIdle);
        #ifdef WRITE_CONTROL_VALUES
            md_thread_lc_ol.push_back(data->mIdle);
        #endif

        // FIXME: For non-MPI, Location::getRank() is -1 => map to 0
        const int rank = std::max(0, event.get_location().getRank());

        const Event begin = data->mCallstack->top();

        stat_collector(ThreadLockContentionOMPLockAPI,
                       data->mIdle,
                       begin->getTimestamp(),
                       event->getTimestamp(),
                       event.get_cnode()->getId(),
                       rank);
    #endif    // _OPENMP
}


void
Statistics::thread_lock_contention_ompcritical_wait_cb(const CallbackManager& cbmanager,
                                                       const int              user_event,
                                                       const Event&           event,
                                                       CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 1654 "Statistics.pattern"

    #if defined(_OPENMP)
        UTILS_ASSERT(data->mIdle > 0);

        thread_lc_oc_quant->add_value(data->mIdle);
        #ifdef WRITE_CONTROL_VALUES
            md_thread_lc_oc.push_back(data->mIdle);
        #endif

        // FIXME: For non-MPI, Location::getRank() is -1 => map to 0
        const int rank = std::max(0, event.get_location().getRank());

        const Event begin = data->mCallstack->top();

        stat_collector(ThreadLockContentionOMPCritical,
                       data->mIdle,
                       begin->getTimestamp(),
                       event->getTimestamp(),
                       event.get_cnode()->getId(),
                       rank);
    #endif    // _OPENMP
}


void
Statistics::thread_lock_contention_pthreadmutex_wait_cb(const CallbackManager& cbmanager,
                                                        const int              user_event,
                                                        const Event&           event,
                                                        CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 1678 "Statistics.pattern"

    #if defined(_OPENMP)
        UTILS_ASSERT(data->mIdle > 0);

        thread_lc_pm_quant->add_value(data->mIdle);
        #ifdef WRITE_CONTROL_VALUES
            md_thread_lc_pm.push_back(data->mIdle);
        #endif

        // FIXME: For non-MPI, Location::getRank() is -1 => map to 0
        const int rank = std::max(0, event.get_location().getRank());

        const Event begin = data->mCallstack->top();

        stat_collector(ThreadLockContentionPthreadMutex,
                       data->mIdle,
                       begin->getTimestamp(),
                       event->getTimestamp(),
                       event.get_cnode()->getId(),
                       rank);
    #endif    // _OPENMP
}


void
Statistics::thread_lock_contention_pthreadcondition_wait_cb(const CallbackManager& cbmanager,
                                                            const int              user_event,
                                                            const Event&           event,
                                                            CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 1702 "Statistics.pattern"

    #if defined(_OPENMP)
        UTILS_ASSERT(data->mIdle > 0);

        thread_lc_pc_quant->add_value(data->mIdle);
        #ifdef WRITE_CONTROL_VALUES
            md_thread_lc_pc.push_back(data->mIdle);
        #endif

        // FIXME: For non-MPI, Location::getRank() is -1 => map to 0
        const int rank = std::max(0, event.get_location().getRank());

        const Event begin = data->mCallstack->top();

        stat_collector(ThreadLockContentionPthreadCondition,
                       data->mIdle,
                       begin->getTimestamp(),
                       event->getTimestamp(),
                       event.get_cnode()->getId(),
                       rank);
    #endif    // _OPENMP
}


// --- Callback functions (bws) ---------------------------------------------

void
Statistics::late_receiver_bws_cb(const CallbackManager& cbmanager,
                                 const int              user_event,
                                 const Event&           event,
                                 CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 1727 "Statistics.pattern"

    #if defined(_MPI)
        UTILS_ASSERT(data->mIdle > 0);

        lr_quant->add_value(data->mIdle);
        #ifdef WRITE_CONTROL_VALUES
            md_lr.push_back(data->mIdle);
        #endif

        const Event       enter = data->mLocal->getEvent(ROLE_ENTER_SENDCMP_LR);
        const RemoteEvent recv  = data->mRemote->getEvent(ROLE_RECV_LR);

        stat_collector(LateReceiver,
                       data->mIdle,
                       enter->getTimestamp(),
                       recv->getTimestamp(),
                       enter.get_cnode()->getId(),
                       enter.get_location().getRank());
    #endif    // _MPI
}


/*-------------------------------------------------------------------------*/
/**
 *  @class CriticalPath
 *  @brief Critical Path Profile metric.
 *
 *  Implements the "Critical Path Profile" metric.
 */
/*-------------------------------------------------------------------------*/

class CriticalPath
    : public TypedPattern< double >
{
    public:
        /// @name Callback registration
        /// @{

        /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
        ///
        void
        registerCallbacks(CallbackManagerMap& cbmanagers) override;

        /// @}
        /// @name Pattern information access
        /// @{

        /// @copydoc scout::Pattern::getId()
        ///
        long
        getId() const override
        {
            return PAT_CRITICAL_PATH;
        }

        /// @copydoc scout::Pattern::getParent()
        ///
        long
        getParent() const override
        {
            return PAT_NONE;
        }

        /// @copydoc scout::Pattern::getName()
        ///
        string
        getName() const override
        {
            return "Critical path profile";
        }

        /// @copydoc scout::Pattern::getUniqueName()
        ///
        string
        getUniqueName() const override
        {
            return "critical_path";
        }

        /// @copydoc scout::Pattern::getDescription()
        ///
        string
        getDescription() const override
        {
            return "Profile of the application's critical path";
        }

        /// @copydoc scout::Pattern::getUnit()
        ///
        string
        getUnit() const override
        {
            return "sec";
        }

        /// @copydoc scout::Pattern::getStorageMode()
        ///
        CubeMetricType
        getStorageMode() const override
        {
            return CUBE_METRIC_EXCLUSIVE;
        }

        /// @}


    private:
        #line 67 "CriticalPath.pattern"

        // *INDENT-OFF*    Uncrustify issue #2789
        bool  m_on_critical_path;
        Event m_end_event;
        // *INDENT-ON*

        static std::vector< double > m_tmpvec;


        /// @todo Once all Syncpoints are adjusted, `spi.waitTime` should be &le;
        ///       the time spent in the comm/sync operation (i.e., `min` may not
        ///       be necessary unless there are rounding errors)
        void
        update_path_severities(const Event         from,
                               const CbData* const data)
        {
            timemap_t tmap = data->mTmapCacheHandler->getRuntimes(from, m_end_event);

            for (timemap_t::const_iterator it = tmap.begin();
                 it != tmap.end();
                 ++it)
            {
                INC_METRIC_VALUE(SCALASCA_DEBUG_CRITICAL_PATH_METRICS,
                                 data->mDefs->get_cnode(it->first),
                                 it->second);
            }

            const SynchpointInfo spi = data->mSynchpointHandler->getSynchpointInfo(from);

            Callpath* const callpath = from.get_cnode();
            const double    waitT    = min(spi.waitTime, tmap[callpath->getId()]);

            DEC_METRIC_VALUE(SCALASCA_DEBUG_CRITICAL_PATH_METRICS,
                             callpath,
                             waitT);
        }

        void
        update_first_path_severities(const CbData* const data)
        {
            timemap_t tmap = data->mTmapCacheHandler->getRuntimes(Event(), m_end_event);

            for (timemap_t::const_iterator it = tmap.begin();
                 it != tmap.end();
                 ++it)
            {
                INC_METRIC_VALUE(SCALASCA_DEBUG_CRITICAL_PATH_METRICS,
                                 data->mDefs->get_cnode(it->first),
                                 it->second);
            }
        }

        /// @todo Once all Syncpoints are adjusted, `wtsm` entries should be &le;
        ///       the time spent in the comm/sync operation (i.e., `min` may not
        ///       be necessary unless there are rounding errors)
        void
        collate_global_critical_path(CbData* const data)
        {
            const uint32_t numCallpaths = data->mDefs->numCallpaths();

            // Collate global critical-path profile
            data->mGlobalCriticalPath.resize(numCallpaths, 0.0);
            data->mLocalCriticalPath.resize(numCallpaths, 0.0);

            for (map< Callpath*, double >::const_iterator it = m_severity.begin();
                 it != m_severity.end();
                 ++it)
            {
                data->mLocalCriticalPath[(it->first)->getId()] = it->second;
            }

            #pragma omp single
            {
                m_tmpvec.assign(numCallpaths, 0.0);
            }
            #pragma omp critical
            {
                transform(m_tmpvec.begin(), m_tmpvec.end(),
                          data->mLocalCriticalPath.begin(),
                          m_tmpvec.begin(),
                          plus< double >());
            }
            #pragma omp barrier

            #ifdef _MPI
                #pragma omp master
                {
                    MPI_Allreduce(MPI_IN_PLACE, &m_tmpvec[0], numCallpaths,
                                  MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
                }
                #pragma omp barrier
            #endif

            data->mGlobalCriticalPath.assign(m_tmpvec.begin(), m_tmpvec.end());

            // Collate local non-waiting time profile
            data->mLocalTimeProfile.resize(numCallpaths, 0.0);

            timemap_t  ltmap = data->mTmapCacheHandler->getRuntimes(Event(), Event());
            TimeMapSum wtsm  = CalculateWaitMap(data->mSynchpointHandler->getSynchpoints());

            for (timemap_t::const_iterator it = ltmap.begin();
                 it != ltmap.end();
                 ++it)
            {
                data->mLocalTimeProfile[it->first] += it->second;

                timemap_t::const_iterator wit = wtsm.mMap.find(it->first);
                if (wit != wtsm.mMap.end())
                {
                    data->mLocalTimeProfile[it->first] -= min(it->second, wit->second);
                }
            }
        }

        #ifdef _MPI
            void
            handle_mpi_collective(const Event&        event,
                                  const CbData* const data,
                                  const int           destRank)
            {
                int inFlag = 0;
                if (  m_on_critical_path
                   && data->mSynchpointHandler->isWaitstate(event))
                {
                    update_path_severities(event, data);

                    inFlag             = 1;
                    m_on_critical_path = false;

                    UTILS_DLOG_LEVEL(SCALASCA_DEBUG_CRITICAL_PATH)
                        << "Critical Path: Moves to rank "
                        << event->getComm().getCommSet().getGlobalRank(destRank);
                }

                int outFlag = 0;
                MPI_Reduce(&inFlag, &outFlag, 1, MPI_INT, MPI_MAX, destRank,
                           getMpiHandle(event->getComm()));

                if (outFlag)
                {
                    m_on_critical_path = true;
                    m_end_event        = event;

                    UTILS_DLOG_LEVEL(SCALASCA_DEBUG_CRITICAL_PATH)
                        << "Critical Path: Continues on " << event.get_location();
                }
            }
        #endif    // _MPI

        /// @name Callback functions (bwc)
        /// @{

        /// @brief Callback function for PREPARE events.
        ///
        /// Callback function triggered by PREPARE events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        prepare_bwc_cb(const CallbackManager& cbmanager,
                       int                    user_event,
                       const Event&           event,
                       CallbackData*          cdata);

        /// @brief Callback function for FINISHED events.
        ///
        /// Callback function triggered by FINISHED events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        finished_bwc_cb(const CallbackManager& cbmanager,
                        int                    user_event,
                        const Event&           event,
                        CallbackData*          cdata);

        /// @brief Callback function for OMP_MGMT_FORK events.
        ///
        /// Callback function triggered by OMP_MGMT_FORK events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        omp_mgmt_fork_bwc_cb(const CallbackManager& cbmanager,
                             int                    user_event,
                             const Event&           event,
                             CallbackData*          cdata);

        /// @brief Callback function for OMP_BARRIER events.
        ///
        /// Callback function triggered by OMP_BARRIER events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        omp_barrier_bwc_cb(const CallbackManager& cbmanager,
                           int                    user_event,
                           const Event&           event,
                           CallbackData*          cdata);

        /// @brief Callback function for FINALIZE events.
        ///
        /// Callback function triggered by FINALIZE events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        finalize_bwc_cb(const CallbackManager& cbmanager,
                        int                    user_event,
                        const Event&           event,
                        CallbackData*          cdata);

        /// @brief Callback function for INIT events.
        ///
        /// Callback function triggered by INIT events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        init_bwc_cb(const CallbackManager& cbmanager,
                    int                    user_event,
                    const Event&           event,
                    CallbackData*          cdata);

        /// @brief Callback function for COLL_12N events.
        ///
        /// Callback function triggered by COLL_12N events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        coll_12n_bwc_cb(const CallbackManager& cbmanager,
                        int                    user_event,
                        const Event&           event,
                        CallbackData*          cdata);

        /// @brief Callback function for COLL_N2N events.
        ///
        /// Callback function triggered by COLL_N2N events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        coll_n2n_bwc_cb(const CallbackManager& cbmanager,
                        int                    user_event,
                        const Event&           event,
                        CallbackData*          cdata);

        /// @brief Callback function for COLL_COMM_CREATE events.
        ///
        /// Callback function triggered by COLL_COMM_CREATE events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        coll_comm_create_bwc_cb(const CallbackManager& cbmanager,
                                int                    user_event,
                                const Event&           event,
                                CallbackData*          cdata);

        /// @brief Callback function for COLL_COMM_FREE events.
        ///
        /// Callback function triggered by COLL_COMM_FREE events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        coll_comm_free_bwc_cb(const CallbackManager& cbmanager,
                              int                    user_event,
                              const Event&           event,
                              CallbackData*          cdata);

        /// @brief Callback function for SYNC_COLL events.
        ///
        /// Callback function triggered by SYNC_COLL events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        sync_coll_bwc_cb(const CallbackManager& cbmanager,
                         int                    user_event,
                         const Event&           event,
                         CallbackData*          cdata);

        /// @brief Callback function for COLL_N21 events.
        ///
        /// Callback function triggered by COLL_N21 events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        coll_n21_bwc_cb(const CallbackManager& cbmanager,
                        int                    user_event,
                        const Event&           event,
                        CallbackData*          cdata);

        /// @brief Callback function for PRE_RECV events.
        ///
        /// Callback function triggered by PRE_RECV events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        pre_recv_bwc_cb(const CallbackManager& cbmanager,
                        int                    user_event,
                        const Event&           event,
                        CallbackData*          cdata);

        /// @brief Callback function for POST_SEND events.
        ///
        /// Callback function triggered by POST_SEND events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        post_send_bwc_cb(const CallbackManager& cbmanager,
                         int                    user_event,
                         const Event&           event,
                         CallbackData*          cdata);

        /// @brief Callback function for POST_INV_RECVREQ events.
        ///
        /// Callback function triggered by POST_INV_RECVREQ events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        post_inv_recvreq_bwc_cb(const CallbackManager& cbmanager,
                                int                    user_event,
                                const Event&           event,
                                CallbackData*          cdata);

        /// @brief Callback function for PRE_INV_SENDCMP events.
        ///
        /// Callback function triggered by PRE_INV_SENDCMP events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        pre_inv_sendcmp_bwc_cb(const CallbackManager& cbmanager,
                               int                    user_event,
                               const Event&           event,
                               CallbackData*          cdata);

        /// @}
};


// --- Static class data ----------------------------------------------------

#line 64 "CriticalPath.pattern"

std::vector< double > CriticalPath::m_tmpvec;


// --- Callback registration ------------------------------------------------

void
CriticalPath::registerCallbacks(CallbackManagerMap& cbmanagers)
{
    TypedPattern< double >::registerCallbacks(cbmanagers);

    CallbackManagerMap::iterator    it;
    CallbackManagerMap::mapped_type cbmanager;

    // Register bwc callbacks
    it = cbmanagers.find("bwc");
    UTILS_ASSERT(it != cbmanagers.end());

    cbmanager = it->second;
    cbmanager->register_callback(
        PREPARE,
        PEARL_create_callback(
            this,
            &CriticalPath::prepare_bwc_cb));
    cbmanager->register_callback(
        FINISHED,
        PEARL_create_callback(
            this,
            &CriticalPath::finished_bwc_cb));
    cbmanager->register_callback(
        OMP_MGMT_FORK,
        PEARL_create_callback(
            this,
            &CriticalPath::omp_mgmt_fork_bwc_cb));
    cbmanager->register_callback(
        OMP_BARRIER,
        PEARL_create_callback(
            this,
            &CriticalPath::omp_barrier_bwc_cb));
    cbmanager->register_callback(
        FINALIZE,
        PEARL_create_callback(
            this,
            &CriticalPath::finalize_bwc_cb));
    cbmanager->register_callback(
        INIT,
        PEARL_create_callback(
            this,
            &CriticalPath::init_bwc_cb));
    cbmanager->register_callback(
        COLL_12N,
        PEARL_create_callback(
            this,
            &CriticalPath::coll_12n_bwc_cb));
    cbmanager->register_callback(
        COLL_N2N,
        PEARL_create_callback(
            this,
            &CriticalPath::coll_n2n_bwc_cb));
    cbmanager->register_callback(
        COLL_COMM_CREATE,
        PEARL_create_callback(
            this,
            &CriticalPath::coll_comm_create_bwc_cb));
    cbmanager->register_callback(
        COLL_COMM_FREE,
        PEARL_create_callback(
            this,
            &CriticalPath::coll_comm_free_bwc_cb));
    cbmanager->register_callback(
        SYNC_COLL,
        PEARL_create_callback(
            this,
            &CriticalPath::sync_coll_bwc_cb));
    cbmanager->register_callback(
        COLL_N21,
        PEARL_create_callback(
            this,
            &CriticalPath::coll_n21_bwc_cb));
    cbmanager->register_callback(
        PRE_RECV,
        PEARL_create_callback(
            this,
            &CriticalPath::pre_recv_bwc_cb));
    cbmanager->register_callback(
        POST_SEND,
        PEARL_create_callback(
            this,
            &CriticalPath::post_send_bwc_cb));
    cbmanager->register_callback(
        POST_INV_RECVREQ,
        PEARL_create_callback(
            this,
            &CriticalPath::post_inv_recvreq_bwc_cb));
    cbmanager->register_callback(
        PRE_INV_SENDCMP,
        PEARL_create_callback(
            this,
            &CriticalPath::pre_inv_sendcmp_bwc_cb));
}


// --- Callback functions (bwc) ---------------------------------------------

void
CriticalPath::prepare_bwc_cb(const CallbackManager& cbmanager,
                             const int              user_event,
                             const Event&           event,
                             CallbackData* const    cdata)
{
    #line 219 "CriticalPath.pattern"

    m_on_critical_path = false;

    #ifndef _MPI
        #pragma omp master
        {
            m_on_critical_path = true;

            UTILS_DLOG_LEVEL(SCALASCA_DEBUG_CRITICAL_PATH)
                << "Critical Path: Starts on " << event.get_location();
        }
    #endif
}


void
CriticalPath::finished_bwc_cb(const CallbackManager& cbmanager,
                              const int              user_event,
                              const Event&           event,
                              CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 233 "CriticalPath.pattern"

    #ifndef _MPI
        #pragma omp master
        {
            update_first_path_severities(data);

            UTILS_DLOG_LEVEL(SCALASCA_DEBUG_CRITICAL_PATH)
                << "Critical Path: Finishes on " << event.get_location();
        }
    #endif

    collate_global_critical_path(data);

    cbmanager.notify(CRITICALPATH_FINISHED, event, data);
}


void
CriticalPath::omp_mgmt_fork_bwc_cb(const CallbackManager& cbmanager,
                                   const int              user_event,
                                   const Event&           event,
                                   CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 249 "CriticalPath.pattern"

    #ifdef _OPENMP
        static bool flag;

        #pragma omp single
        {
            flag = false;
        }
        #pragma omp critical
        {
            if (m_on_critical_path)
            {
                // --- set severities
                update_path_severities(event, data);

                m_on_critical_path = false;
                flag               = true;
            }
        }

        #pragma omp barrier
        { }

        #pragma omp master
        {
            if (flag)
            {
                m_on_critical_path = true;
                m_end_event        = event;
            }
        }
        #pragma omp barrier
        { }
    #endif
}


void
CriticalPath::omp_barrier_bwc_cb(const CallbackManager& cbmanager,
                                 const int              user_event,
                                 const Event&           event,
                                 CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 285 "CriticalPath.pattern"

    #ifdef _OPENMP
        static bool flag;

        #pragma omp single
        {
            flag = false;
        }
        #pragma omp critical
        {
            if (m_on_critical_path)
            {
                // --- set severities
                update_path_severities(event, data);

                m_on_critical_path = false;
                flag               = true;
            }
        }

        #pragma omp barrier
        { }

        if (  flag
           && !data->mSynchpointHandler->isWaitstate(event))
        {
            // FIXME: bit of a problem here ... might have two critical paths
            m_on_critical_path = true;
            m_end_event        = event;
        }
        #pragma omp barrier
        { }
    #endif
}


void
CriticalPath::finalize_bwc_cb(const CallbackManager& cbmanager,
                              const int              user_event,
                              const Event&           event,
                              CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 321 "CriticalPath.pattern"

    const CollectiveInfo& ci = data->mCollinfo;

    if (ci.my.mRank == ci.latest.mRank)
    {
        m_on_critical_path = true;
        m_end_event        = event;

        UTILS_DLOG_LEVEL(SCALASCA_DEBUG_CRITICAL_PATH)
            << "Critical Path: Starts on " << event.get_location();
    }
}


void
CriticalPath::init_bwc_cb(const CallbackManager& cbmanager,
                          const int              user_event,
                          const Event&           event,
                          CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 335 "CriticalPath.pattern"

    if (m_on_critical_path)
    {
        update_path_severities(event, data);

        UTILS_DLOG_LEVEL(SCALASCA_DEBUG_CRITICAL_PATH)
            << "Critical Path: Finishes on " << event.get_location();
    }
}


void
CriticalPath::coll_12n_bwc_cb(const CallbackManager& cbmanager,
                              const int              user_event,
                              const Event&           event,
                              CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 345 "CriticalPath.pattern"

    #ifdef _MPI
        const CollectiveInfo& ci = data->mCollinfo;

        handle_mpi_collective(event, data, ci.root.mRank);
    #endif
}


void
CriticalPath::coll_n2n_bwc_cb(const CallbackManager& cbmanager,
                              const int              user_event,
                              const Event&           event,
                              CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 353 "CriticalPath.pattern"

    #ifdef _MPI
        const CollectiveInfo& ci = data->mCollinfo;

        handle_mpi_collective(event, data, ci.latest.mRank);
    #endif
}


void
CriticalPath::coll_comm_create_bwc_cb(const CallbackManager& cbmanager,
                                      const int              user_event,
                                      const Event&           event,
                                      CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 361 "CriticalPath.pattern"

    #ifdef _MPI
        const CollectiveInfo& ci = data->mCollinfo;

        handle_mpi_collective(event, data, ci.latest.mRank);
    #endif
}


void
CriticalPath::coll_comm_free_bwc_cb(const CallbackManager& cbmanager,
                                    const int              user_event,
                                    const Event&           event,
                                    CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 369 "CriticalPath.pattern"

    #ifdef _MPI
        const CollectiveInfo& ci = data->mCollinfo;

        handle_mpi_collective(event, data, ci.latest.mRank);
    #endif
}


void
CriticalPath::sync_coll_bwc_cb(const CallbackManager& cbmanager,
                               const int              user_event,
                               const Event&           event,
                               CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 377 "CriticalPath.pattern"

    #ifdef _MPI
        const CollectiveInfo& ci = data->mCollinfo;

        handle_mpi_collective(event, data, ci.latest.mRank);
    #endif
}


void
CriticalPath::coll_n21_bwc_cb(const CallbackManager& cbmanager,
                              const int              user_event,
                              const Event&           event,
                              CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 385 "CriticalPath.pattern"

    #ifdef _MPI
        const CollectiveInfo& ci = data->mCollinfo;

        handle_mpi_collective(event, data, ci.latest.mRank);
    #endif
}


void
CriticalPath::pre_recv_bwc_cb(const CallbackManager& cbmanager,
                              const int              user_event,
                              const Event&           event,
                              CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 393 "CriticalPath.pattern"

    #ifdef _MPI
        Buffer*  buffer = new Buffer(16);
        uint32_t flag   = 0;

        if (  m_on_critical_path
           && data->mSynchpointHandler->isWaitstate(event))
        {
            // --- set severities
            update_path_severities(event, data);

            // --- critical path continues on wait-state causing process

            flag               = 1;
            m_on_critical_path = false;
        }

        buffer->put_uint32(flag);
        data->mLocal->addBuffer(buffer, BUFFER_CRITICALPATH);
    #endif
}


void
CriticalPath::post_send_bwc_cb(const CallbackManager& cbmanager,
                               const int              user_event,
                               const Event&           event,
                               CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 415 "CriticalPath.pattern"

    #ifdef _MPI
        if (!data->mSynchpointHandler->isSynchpoint(event))
        {
            return;
        }

        // --- get critical path info flag

        uint32_t flag = data->mRemote->getBuffer(BUFFER_CRITICALPATH)->get_uint32();

        if (flag)
        {
            m_on_critical_path = true;
            m_end_event        = event;
        }
    #endif
}


void
CriticalPath::post_inv_recvreq_bwc_cb(const CallbackManager& cbmanager,
                                      const int              user_event,
                                      const Event&           event,
                                      CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 434 "CriticalPath.pattern"

    #ifdef _MPI
        // --- get critical path info flag

        uint32_t flag = data->mInvRemote->getBuffer(BUFFER_CRITICALPATH)->get_uint32();

        if (flag)
        {
            m_on_critical_path = true;
            m_end_event        = event;
        }
    #endif
}


void
CriticalPath::pre_inv_sendcmp_bwc_cb(const CallbackManager& cbmanager,
                                     const int              user_event,
                                     const Event&           event,
                                     CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 448 "CriticalPath.pattern"

    #ifdef _MPI
        Buffer*  buffer = new Buffer(16);
        uint32_t flag   = 0;

        if (  m_on_critical_path
           && data->mSynchpointHandler->isWaitstate(event))
        {
            // --- set severities
            update_path_severities(event, data);

            // --- critical path continues on wait-state causing process

            flag               = 1;
            m_on_critical_path = false;
        }

        buffer->put_uint32(flag);
        data->mInvLocal->addBuffer(buffer, BUFFER_CRITICALPATH);
    #endif
}


/*-------------------------------------------------------------------------*/
/**
 *  @class CriticalPathActivities
 *  @brief Activity Impact metric.
 *
 *  Implements the "Activity Impact" metric.
 */
/*-------------------------------------------------------------------------*/

class CriticalPathActivities
    : public TypedPattern< double >
{
    public:
        /// @name Callback registration
        /// @{

        /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
        ///
        void
        registerCallbacks(CallbackManagerMap& cbmanagers) override;

        /// @}
        /// @name Pattern information access
        /// @{

        /// @copydoc scout::Pattern::getId()
        ///
        long
        getId() const override
        {
            return PAT_CRITICAL_PATH_ACTIVITIES;
        }

        /// @copydoc scout::Pattern::getParent()
        ///
        long
        getParent() const override
        {
            return PAT_PERFORMANCE_IMPACT_CRITICALPATH;
        }

        /// @copydoc scout::Pattern::getName()
        ///
        string
        getName() const override
        {
            return "Activity impact";
        }

        /// @copydoc scout::Pattern::getUniqueName()
        ///
        string
        getUniqueName() const override
        {
            return "critical_path_activities";
        }

        /// @copydoc scout::Pattern::getDescription()
        ///
        string
        getDescription() const override
        {
            return "Resource consumption of activities on the critical"
                   " path";
        }

        /// @copydoc scout::Pattern::getUnit()
        ///
        string
        getUnit() const override
        {
            return "sec";
        }

        /// @copydoc scout::Pattern::getStorageMode()
        ///
        CubeMetricType
        getStorageMode() const override
        {
            return CUBE_METRIC_EXCLUSIVE;
        }

        /// @}


    private:
        /// @name Callback functions (bwc)
        /// @{

        /// @brief Callback function for CRITICALPATH_FINISHED events.
        ///
        /// Callback function triggered by CRITICALPATH_FINISHED events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        criticalpath_finished_bwc_cb(const CallbackManager& cbmanager,
                                     int                    user_event,
                                     const Event&           event,
                                     CallbackData*          cdata);

        /// @}
};


// --- Callback registration ------------------------------------------------

void
CriticalPathActivities::registerCallbacks(CallbackManagerMap& cbmanagers)
{
    TypedPattern< double >::registerCallbacks(cbmanagers);

    CallbackManagerMap::iterator    it;
    CallbackManagerMap::mapped_type cbmanager;

    // Register bwc callbacks
    it = cbmanagers.find("bwc");
    UTILS_ASSERT(it != cbmanagers.end());

    cbmanager = it->second;
    cbmanager->register_callback(
        CRITICALPATH_FINISHED,
        PEARL_create_callback(
            this,
            &CriticalPathActivities::criticalpath_finished_bwc_cb));
}


// --- Callback functions (bwc) ---------------------------------------------

void
CriticalPathActivities::criticalpath_finished_bwc_cb(const CallbackManager& cbmanager,
                                                     const int              user_event,
                                                     const Event&           event,
                                                     CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 581 "CriticalPath.pattern"

    // Shortcuts
    const vector< double >& gcpath = data->mGlobalCriticalPath;
    const vector< double >& ltimes = data->mLocalTimeProfile;

    const size_t numCallpaths = data->mDefs->numCallpaths();
    for (size_t i = 0; i < numCallpaths; ++i)
    {
        if (gcpath[i] > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_CRITICAL_PATH_METRICS,
                             data->mDefs->get_cnode(i),
                             min(gcpath[i], ltimes[i]));
        }
    }
}


/*-------------------------------------------------------------------------*/
/**
 *  @class CriticalImbalanceImpact
 *  @brief Critical Imbalance Impact metric.
 *
 *  Implements the "Critical Imbalance Impact" metric.
 */
/*-------------------------------------------------------------------------*/

class CriticalImbalanceImpact
    : public TypedPattern< double >
{
    public:
        /// @name Callback registration
        /// @{

        /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
        ///
        void
        registerCallbacks(CallbackManagerMap& cbmanagers) override;

        /// @}
        /// @name Pattern information access
        /// @{

        /// @copydoc scout::Pattern::getId()
        ///
        long
        getId() const override
        {
            return PAT_CRITICAL_IMBALANCE_IMPACT;
        }

        /// @copydoc scout::Pattern::getParent()
        ///
        long
        getParent() const override
        {
            return PAT_PERFORMANCE_IMPACT_CRITICALPATH;
        }

        /// @copydoc scout::Pattern::getName()
        ///
        string
        getName() const override
        {
            return "Critical-path imbalance impact";
        }

        /// @copydoc scout::Pattern::getUniqueName()
        ///
        string
        getUniqueName() const override
        {
            return "critical_imbalance_impact";
        }

        /// @copydoc scout::Pattern::getDescription()
        ///
        string
        getDescription() const override
        {
            return "Performance impact of load imbalance on the critical"
                   " path";
        }

        /// @copydoc scout::Pattern::getUnit()
        ///
        string
        getUnit() const override
        {
            return "sec";
        }

        /// @copydoc scout::Pattern::getStorageMode()
        ///
        CubeMetricType
        getStorageMode() const override
        {
            return CUBE_METRIC_EXCLUSIVE;
        }

        /// @}


    private:
        #line 628 "CriticalPath.pattern"

        static vector< double > globalImbalanceCosts;

        /// @name Callback functions (bwc)
        /// @{

        /// @brief Callback function for CRITICALPATH_FINISHED events.
        ///
        /// Callback function triggered by CRITICALPATH_FINISHED events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        criticalpath_finished_bwc_cb(const CallbackManager& cbmanager,
                                     int                    user_event,
                                     const Event&           event,
                                     CallbackData*          cdata);

        /// @}
};


// --- Static class data ----------------------------------------------------

#line 631 "CriticalPath.pattern"

vector< double > CriticalImbalanceImpact::globalImbalanceCosts;


// --- Callback registration ------------------------------------------------

void
CriticalImbalanceImpact::registerCallbacks(CallbackManagerMap& cbmanagers)
{
    TypedPattern< double >::registerCallbacks(cbmanagers);

    CallbackManagerMap::iterator    it;
    CallbackManagerMap::mapped_type cbmanager;

    // Register bwc callbacks
    it = cbmanagers.find("bwc");
    UTILS_ASSERT(it != cbmanagers.end());

    cbmanager = it->second;
    cbmanager->register_callback(
        CRITICALPATH_FINISHED,
        PEARL_create_callback(
            this,
            &CriticalImbalanceImpact::criticalpath_finished_bwc_cb));
}


// --- Callback functions (bwc) ---------------------------------------------

void
CriticalImbalanceImpact::criticalpath_finished_bwc_cb(const CallbackManager& cbmanager,
                                                      const int              user_event,
                                                      const Event&           event,
                                                      CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 635 "CriticalPath.pattern"

    // Shortcuts
    const size_t numCallpaths = data->mDefs->numCallpaths();

    const vector< double >& gcpath = data->mGlobalCriticalPath;
    const vector< double >& lcpath = data->mLocalCriticalPath;
    const vector< double >& ltimes = data->mLocalTimeProfile;

    // Calculate aggregate local wait time
    // FIXME:
    // In theory, the wait time can be calculated as the totel time on the
    // global critical path minus the local time profile (which excludes
    // wait time).  However, in practice this difference may be negative.
    // This needs to be investigated in more depth...
    const double gcplen    = accumulate(gcpath.begin(), gcpath.end(), 0.0);
    const double lplen     = accumulate(ltimes.begin(), ltimes.end(), 0.0);
    const double aggrWaitT = gcplen - lplen;

    // Calculate local per-callpath critical-path excess time
    // (see thesis D. Boehme, p. 83, Eq. 1)
    vector< double > localImbalanceCosts;
    localImbalanceCosts.reserve(numCallpaths);
    for (size_t i = 0; i < numCallpaths; ++i)
    {
        localImbalanceCosts.push_back(max(gcpath[i] - ltimes[i], 0.0));
    }

    // Calculate aggregate local critical-path excess time (p. 84, Eq. 1);
    // excess time for call paths not on critical path is 0
    const double aggrExcessT = accumulate(localImbalanceCosts.begin(),
                                          localImbalanceCosts.end(), 0.0);

    // Calculate local critical-path imbalance costs (p. 83, Eq. 2)
    const double scalingFactor = aggrWaitT / aggrExcessT;
    transform(localImbalanceCosts.begin(), localImbalanceCosts.end(),
              localImbalanceCosts.begin(),
              [scalingFactor](const double value)
              {
                  return (value * scalingFactor);
              });

    // Accumulate global critical-path imbalance costs
    #pragma omp single
    {
        globalImbalanceCosts.assign(numCallpaths, 0.0);
    }
    #pragma omp critical
    {
        transform(globalImbalanceCosts.begin(), globalImbalanceCosts.end(),
                  localImbalanceCosts.begin(),
                  globalImbalanceCosts.begin(),
                  plus< double >());
    }
    #pragma omp barrier

    #ifdef _MPI
        #pragma omp master
        {
            MPI_Allreduce(MPI_IN_PLACE, &globalImbalanceCosts[0], numCallpaths,
                          MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
        }
        #pragma omp barrier
    #endif

    // Proportionally distribute global critical-path imbalance costs to
    // call paths on the critical path
    for (size_t i = 0; i < numCallpaths; ++i)
    {
        if (lcpath[i] > 0)    // implies 'gcpath[i] > 0'
        {
            // FIXME: calculated costs should never be negative
            const double costs = (lcpath[i] / gcpath[i]) * globalImbalanceCosts[i];
            if (costs > 0)
            {
                INC_METRIC_VALUE(SCALASCA_DEBUG_CRITICAL_PATH_METRICS,
                                 data->mDefs->get_cnode(i),
                                 costs);
            }
        }
    }
}


/*-------------------------------------------------------------------------*/
/**
 *  @class InterPartitionImbalance
 *  @brief Inter-partition Imbalance metric.
 *
 *  Implements the "Inter-partition Imbalance" metric.
 */
/*-------------------------------------------------------------------------*/

class InterPartitionImbalance
    : public TypedPattern< double >
{
    public:
        /// @name Callback registration
        /// @{

        /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
        ///
        void
        registerCallbacks(CallbackManagerMap& cbmanagers) override;

        /// @}
        /// @name Pattern information access
        /// @{

        /// @copydoc scout::Pattern::getId()
        ///
        long
        getId() const override
        {
            return PAT_INTER_PARTITION_IMBALANCE;
        }

        /// @copydoc scout::Pattern::getParent()
        ///
        long
        getParent() const override
        {
            return PAT_CRITICAL_IMBALANCE_IMPACT;
        }

        /// @copydoc scout::Pattern::getName()
        ///
        string
        getName() const override
        {
            return "Inter-partition imbalance";
        }

        /// @copydoc scout::Pattern::getUniqueName()
        ///
        string
        getUniqueName() const override
        {
            return "inter_partition_imbalance";
        }

        /// @copydoc scout::Pattern::getDescription()
        ///
        string
        getDescription() const override
        {
            return "Performance impact of load imbalance between process"
                   " partitions that perform different activities";
        }

        /// @copydoc scout::Pattern::getUnit()
        ///
        string
        getUnit() const override
        {
            return "sec";
        }

        /// @copydoc scout::Pattern::getStorageMode()
        ///
        CubeMetricType
        getStorageMode() const override
        {
            return CUBE_METRIC_EXCLUSIVE;
        }

        /// @}


    private:
        #line 765 "CriticalPath.pattern"

        static vector< double > phr;

        /// @name Callback functions (bwc)
        /// @{

        /// @brief Callback function for CRITICALPATH_FINISHED events.
        ///
        /// Callback function triggered by CRITICALPATH_FINISHED events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        criticalpath_finished_bwc_cb(const CallbackManager& cbmanager,
                                     int                    user_event,
                                     const Event&           event,
                                     CallbackData*          cdata);

        /// @}
};


// --- Static class data ----------------------------------------------------

#line 768 "CriticalPath.pattern"

vector< double > InterPartitionImbalance::phr;


// --- Callback registration ------------------------------------------------

void
InterPartitionImbalance::registerCallbacks(CallbackManagerMap& cbmanagers)
{
    TypedPattern< double >::registerCallbacks(cbmanagers);

    CallbackManagerMap::iterator    it;
    CallbackManagerMap::mapped_type cbmanager;

    // Register bwc callbacks
    it = cbmanagers.find("bwc");
    UTILS_ASSERT(it != cbmanagers.end());

    cbmanager = it->second;
    cbmanager->register_callback(
        CRITICALPATH_FINISHED,
        PEARL_create_callback(
            this,
            &InterPartitionImbalance::criticalpath_finished_bwc_cb));
}


// --- Callback functions (bwc) ---------------------------------------------

void
InterPartitionImbalance::criticalpath_finished_bwc_cb(const CallbackManager& cbmanager,
                                                      const int              user_event,
                                                      const Event&           event,
                                                      CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 772 "CriticalPath.pattern"

    //
    // --- calculate local headroom

    // shortcuts
    const size_t            ncnodes = data->mDefs->numCallpaths();
    const vector< double >& gcpath(data->mGlobalCriticalPath);
    const vector< double >& lcpath(data->mLocalCriticalPath);
    const vector< double >& ltimes(data->mLocalTimeProfile);

    double gcplen = std::accumulate(gcpath.begin(), gcpath.end(), 0.0);
    double lplen  = 0.0;

    vector< double > hr(gcpath);

    for (size_t i = 0; i < ncnodes; ++i)
    {
        if (ltimes[i] > 0.0)
        {
            lplen += ltimes[i] + (hr[i] - std::min(ltimes[i], hr[i]));
            hr[i]  = 0.0;
        }
    }

    double lhrlen = std::accumulate(hr.begin(), hr.end(), 0.0);
    double hrf    = (gcplen - lplen) / lhrlen;

    for (size_t i = 0; i < ncnodes; ++i)
    {
        hr[i] *= hrf;
    }

    //
    // --- exchange headrooms and calculate severity

    #pragma omp master
    {
        phr.assign(ncnodes, 0.0);
    }
    #pragma omp barrier
    { }
    #pragma omp critical
    {
        std::transform(phr.begin(), phr.end(), hr.begin(),
                       phr.begin(), std::plus< double >());
    }
    #pragma omp barrier
    { }

    #ifdef _MPI
        #pragma omp master
        {
            MPI_Allreduce(MPI_IN_PLACE, &phr[0], ncnodes, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
        }
    #endif

    #pragma omp barrier
    { }
    hr.assign(phr.begin(), phr.end());

    for (uint32_t i = 0; i < ncnodes; ++i)
    {
        if (!(lcpath[i] > 0.0))
        {
            continue;
        }

        double sev = (lcpath[i] / gcpath[i]) * hr[i];

        if (sev > 0.0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_CRITICAL_PATH_METRICS,
                             data->mDefs->get_cnode(i),
                             sev);
        }
    }
}


/*-------------------------------------------------------------------------*/
/**
 *  @class NonCriticalPathActivities
 *  @brief Non-critical-path Activities metric.
 *
 *  Implements the "Non-critical-path Activities" metric.
 */
/*-------------------------------------------------------------------------*/

class NonCriticalPathActivities
    : public TypedPattern< double >
{
    public:
        /// @name Callback registration
        /// @{

        /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
        ///
        void
        registerCallbacks(CallbackManagerMap& cbmanagers) override;

        /// @}
        /// @name Pattern information access
        /// @{

        /// @copydoc scout::Pattern::getId()
        ///
        long
        getId() const override
        {
            return PAT_NON_CRITICAL_PATH_ACTIVITIES;
        }

        /// @copydoc scout::Pattern::getParent()
        ///
        long
        getParent() const override
        {
            return PAT_PERFORMANCE_IMPACT;
        }

        /// @copydoc scout::Pattern::getName()
        ///
        string
        getName() const override
        {
            return "Non-critical-path activities";
        }

        /// @copydoc scout::Pattern::getUniqueName()
        ///
        string
        getUniqueName() const override
        {
            return "non_critical_path_activities";
        }

        /// @copydoc scout::Pattern::getDescription()
        ///
        string
        getDescription() const override
        {
            return "Overall resource consumption of activities that are"
                   " not on the critical path";
        }

        /// @copydoc scout::Pattern::getUnit()
        ///
        string
        getUnit() const override
        {
            return "sec";
        }

        /// @copydoc scout::Pattern::getStorageMode()
        ///
        CubeMetricType
        getStorageMode() const override
        {
            return CUBE_METRIC_EXCLUSIVE;
        }

        /// @}


    private:
        /// @name Callback functions (bwc)
        /// @{

        /// @brief Callback function for CRITICALPATH_FINISHED events.
        ///
        /// Callback function triggered by CRITICALPATH_FINISHED events
        /// during the @"bwc@" replay pass.
        ///
        /// @param cbmanager
        ///     Callback manager instance
        /// @param user_event
        ///     User event triggering this callback (if applicable)
        /// @param event
        ///     Currently processed event
        /// @param cdata
        ///     Callback data object
        ///
        void
        criticalpath_finished_bwc_cb(const CallbackManager& cbmanager,
                                     int                    user_event,
                                     const Event&           event,
                                     CallbackData*          cdata);

        /// @}
};


// --- Callback registration ------------------------------------------------

void
NonCriticalPathActivities::registerCallbacks(CallbackManagerMap& cbmanagers)
{
    TypedPattern< double >::registerCallbacks(cbmanagers);

    CallbackManagerMap::iterator    it;
    CallbackManagerMap::mapped_type cbmanager;

    // Register bwc callbacks
    it = cbmanagers.find("bwc");
    UTILS_ASSERT(it != cbmanagers.end());

    cbmanager = it->second;
    cbmanager->register_callback(
        CRITICALPATH_FINISHED,
        PEARL_create_callback(
            this,
            &NonCriticalPathActivities::criticalpath_finished_bwc_cb));
}


// --- Callback functions (bwc) ---------------------------------------------

void
NonCriticalPathActivities::criticalpath_finished_bwc_cb(const CallbackManager& cbmanager,
                                                        const int              user_event,
                                                        const Event&           event,
                                                        CallbackData* const    cdata)
{
    SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

    #line 868 "CriticalPath.pattern"

    // Shortcuts
    const vector< double >& gcpath = data->mGlobalCriticalPath;
    const vector< double >& ltimes = data->mLocalTimeProfile;

    const size_t numCallpaths = data->mDefs->numCallpaths();
    for (uint32_t i = 0; i < numCallpaths; ++i)
    {
        const double impact = ltimes[i] - gcpath[i];
        if (impact > 0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_CRITICAL_PATH_METRICS,
                             data->mDefs->get_cnode(i),
                             impact);
        }
    }
}


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_Finalize
     *  @brief Short-term MPI Finalize Delay Costs metric.
     *
     *  Implements the "Short-term MPI Finalize Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_Finalize
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_FINALIZE;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_FINALIZE_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI short-term finalize delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_finalize";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that directly cause wait states in front of"
                       " MPI_Finalize";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (bwc)
            /// @{

            /// @brief Callback function for FINALIZE events.
            ///
            /// Callback function triggered by FINALIZE events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            finalize_bwc_cb(const CallbackManager& cbmanager,
                            int                    user_event,
                            const Event&           event,
                            CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Delay_Finalize::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register bwc callbacks
        it = cbmanagers.find("bwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            FINALIZE,
            PEARL_create_callback(
                this,
                &Delay_Finalize::finalize_bwc_cb));
    }


    // --- Callback functions (bwc) -----------------------------------------

    void
    Delay_Finalize::finalize_bwc_cb(const CallbackManager& cbmanager,
                                    const int              user_event,
                                    const Event&           event,
                                    CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 177 "Delay.MPI.pattern"

        const CollectiveInfo& ci = data->mCollinfo;

        data->mInputScales.clear();

        cbmanager.notify(PRE_DELAY_MPI_FINALIZE, event, data);

        CalculateCollectiveDelay(ci.my, ci.latest, event, data);

        if (ci.my.mRank == ci.latest.mRank)
        {
            // algorithm 1, line 24
            map< uint32_t, TimeMapSum >::const_iterator dWit = data->mCollDelayInfo.find(KEY_WAITTIME);
            if (  (dWit != data->mCollDelayInfo.end())
               && (dWit->second.mSum > 0.0))
            {
                for (timemap_t::const_iterator it = dWit->second.mMap.begin();
                     it != dWit->second.mMap.end();
                     ++it)
                {
                    INC_METRIC_VALUE(SCALASCA_DEBUG_DELAY_METRICS,
                                     data->mDefs->get_cnode(it->first),
                                     it->second);
                }
            }

            cbmanager.notify(DELAY_MPI_FINALIZE, event, data);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_Finalize_LT
     *  @brief Long-term MPI Finalize Delay Costs metric.
     *
     *  Implements the "Long-term MPI Finalize Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_Finalize_LT
        : public LongTermDelayPattern
    {
        public:
            /// @name Constructors & destructor
            /// @{

            /// @brief Constructor.
            ///
            /// Creates a new instance.
            ///
            Delay_Finalize_LT()
                : LongTermDelayPattern(BUFFER_RCOST_MPI_FINALIZE,
                                       KEY_RCOST_MPI_FINALIZE,
                                       DELAY_MPI_FINALIZE)
            {
            }

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_FINALIZE_LONGTERM;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_FINALIZE_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI long-term finalize delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_finalize_longterm";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that indirectly cause wait states in in front"
                       " of MPI_Finalize through wait-state propagation";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}
    };
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_CommCreate
     *  @brief Short-term MPI Communicator Create Delay Costs metric.
     *
     *  Implements the "Short-term MPI Communicator Create Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_CommCreate
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_COMM_CREATE;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_COMM_CREATE_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI short-term communicator create delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_comm_create";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that directly cause wait states in MPI"
                       " communicator creation routines";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (bwc)
            /// @{

            /// @brief Callback function for COLL_COMM_CREATE events.
            ///
            /// Callback function triggered by COLL_COMM_CREATE events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            coll_comm_create_bwc_cb(const CallbackManager& cbmanager,
                                    int                    user_event,
                                    const Event&           event,
                                    CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Delay_CommCreate::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register bwc callbacks
        it = cbmanagers.find("bwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            COLL_COMM_CREATE,
            PEARL_create_callback(
                this,
                &Delay_CommCreate::coll_comm_create_bwc_cb));
    }


    // --- Callback functions (bwc) -----------------------------------------

    void
    Delay_CommCreate::coll_comm_create_bwc_cb(const CallbackManager& cbmanager,
                                              const int              user_event,
                                              const Event&           event,
                                              CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 300 "Delay.MPI.pattern"

        const CollectiveInfo& ci = data->mCollinfo;

        data->mInputScales.clear();

        cbmanager.notify(PRE_DELAY_MPI_COLL_COMM_CREATE, event, data);

        CalculateCollectiveDelay(ci.my, ci.latest, event, data);

        if (ci.my.mRank == ci.latest.mRank)
        {
            // algorithm 1, line 24
            map< uint32_t, TimeMapSum >::const_iterator dWit = data->mCollDelayInfo.find(KEY_WAITTIME);
            if (  (dWit != data->mCollDelayInfo.end())
               && (dWit->second.mSum > 0.0))
            {
                for (timemap_t::const_iterator it = dWit->second.mMap.begin();
                     it != dWit->second.mMap.end();
                     ++it)
                {
                    INC_METRIC_VALUE(SCALASCA_DEBUG_DELAY_METRICS,
                                     data->mDefs->get_cnode(it->first),
                                     it->second);
                }
            }

            cbmanager.notify(DELAY_MPI_COLL_COMM_CREATE, event, data);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_CommCreate_LT
     *  @brief Long-term MPI Communicator Create Delay Costs metric.
     *
     *  Implements the "Long-term MPI Communicator Create Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_CommCreate_LT
        : public LongTermDelayPattern
    {
        public:
            /// @name Constructors & destructor
            /// @{

            /// @brief Constructor.
            ///
            /// Creates a new instance.
            ///
            Delay_CommCreate_LT()
                : LongTermDelayPattern(BUFFER_RCOST_MPI_COLL_COMM_CREATE,
                                       KEY_RCOST_MPI_COLL_COMM_CREATE,
                                       DELAY_MPI_COLL_COMM_CREATE)
            {
            }

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_COMM_CREATE_LONGTERM;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_COMM_CREATE_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI long-term communicator create delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_comm_create_longterm";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that indirectly cause wait states in MPI"
                       " communicator creation routines through wait-state"
                       " propagation";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}
    };
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_CommFree
     *  @brief Short-term MPI Communicator Free Delay Costs metric.
     *
     *  Implements the "Short-term MPI Communicator Free Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_CommFree
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_COMM_FREE;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_COMM_FREE_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI short-term communicator free delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_comm_free";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that directly cause wait states in MPI"
                       " communicator destruction routines";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (bwc)
            /// @{

            /// @brief Callback function for COLL_COMM_FREE events.
            ///
            /// Callback function triggered by COLL_COMM_FREE events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            coll_comm_free_bwc_cb(const CallbackManager& cbmanager,
                                  int                    user_event,
                                  const Event&           event,
                                  CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Delay_CommFree::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register bwc callbacks
        it = cbmanagers.find("bwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            COLL_COMM_FREE,
            PEARL_create_callback(
                this,
                &Delay_CommFree::coll_comm_free_bwc_cb));
    }


    // --- Callback functions (bwc) -----------------------------------------

    void
    Delay_CommFree::coll_comm_free_bwc_cb(const CallbackManager& cbmanager,
                                          const int              user_event,
                                          const Event&           event,
                                          CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 402 "Delay.MPI.pattern"

        const CollectiveInfo& ci = data->mCollinfo;

        data->mInputScales.clear();

        cbmanager.notify(PRE_DELAY_MPI_COLL_COMM_FREE, event, data);

        CalculateCollectiveDelay(ci.my, ci.latest, event, data);

        if (ci.my.mRank == ci.latest.mRank)
        {
            // algorithm 1, line 24
            map< uint32_t, TimeMapSum >::const_iterator dWit = data->mCollDelayInfo.find(KEY_WAITTIME);
            if (  (dWit != data->mCollDelayInfo.end())
               && (dWit->second.mSum > 0.0))
            {
                for (timemap_t::const_iterator it = dWit->second.mMap.begin();
                     it != dWit->second.mMap.end();
                     ++it)
                {
                    INC_METRIC_VALUE(SCALASCA_DEBUG_DELAY_METRICS,
                                     data->mDefs->get_cnode(it->first),
                                     it->second);
                }
            }

            cbmanager.notify(DELAY_MPI_COLL_COMM_FREE, event, data);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_CommFree_LT
     *  @brief Long-term MPI Communicator Free Delay Costs metric.
     *
     *  Implements the "Long-term MPI Communicator Free Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_CommFree_LT
        : public LongTermDelayPattern
    {
        public:
            /// @name Constructors & destructor
            /// @{

            /// @brief Constructor.
            ///
            /// Creates a new instance.
            ///
            Delay_CommFree_LT()
                : LongTermDelayPattern(BUFFER_RCOST_MPI_COLL_COMM_FREE,
                                       KEY_RCOST_MPI_COLL_COMM_FREE,
                                       DELAY_MPI_COLL_COMM_FREE)
            {
            }

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_COMM_FREE_LONGTERM;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_COMM_FREE_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI long-term communicator free delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_comm_free_longterm";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that indirectly cause wait states in MPI"
                       " communicator destruction routines through"
                       " wait-state propagation";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}
    };
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_LateSender
     *  @brief Short-term MPI Late Sender Delay Costs metric.
     *
     *  Implements the "Short-term MPI Late Sender Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_LateSender
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_LATESENDER;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_LATESENDER_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI short-term Late Sender delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_latesender";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that directly cause MPI Late Sender wait"
                       " states";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (bwc)
            /// @{

            /// @brief Callback function for PRE_RECV events.
            ///
            /// Callback function triggered by PRE_RECV events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            pre_recv_bwc_cb(const CallbackManager& cbmanager,
                            int                    user_event,
                            const Event&           event,
                            CallbackData*          cdata);

            /// @brief Callback function for POST_SEND events.
            ///
            /// Callback function triggered by POST_SEND events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            post_send_bwc_cb(const CallbackManager& cbmanager,
                             int                    user_event,
                             const Event&           event,
                             CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Delay_LateSender::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register bwc callbacks
        it = cbmanagers.find("bwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            PRE_RECV,
            PEARL_create_callback(
                this,
                &Delay_LateSender::pre_recv_bwc_cb));
        cbmanager->register_callback(
            POST_SEND,
            PEARL_create_callback(
                this,
                &Delay_LateSender::post_send_bwc_cb));
    }


    // --- Callback functions (bwc) -----------------------------------------

    void
    Delay_LateSender::pre_recv_bwc_cb(const CallbackManager& cbmanager,
                                      const int              user_event,
                                      const Event&           event,
                                      CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 532 "Delay.MPI.pattern"

        // Shortcuts
        const TmapCacheHandler*  TmH = data->mTmapCacheHandler;
        const SynchpointHandler* SpH = data->mSynchpointHandler;

        // cost-accounting backward replay

        Buffer*        buffer = new Buffer(256);
        SynchpointInfo spi    = SpH->getSynchpointInfo(event);

        buffer->put_double(spi.waitTime);

        timemap_t tmap;

        if (spi.waitTime > 0)
        {
            int rank = event->getComm().getCommSet().getGlobalRank(event->getSource());
            tmap = GetMpiDelayTimemap(event, rank, SpH, TmH);
        }

        PackTimemap(buffer, tmap);
        data->mLocal->addBuffer(buffer, BUFFER_DELAY);
    }


    void
    Delay_LateSender::post_send_bwc_cb(const CallbackManager& cbmanager,
                                       const int              user_event,
                                       const Event&           event,
                                       CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 556 "Delay.MPI.pattern"

        // Shortcuts
        const TmapCacheHandler*  TmH = data->mTmapCacheHandler;
        const SynchpointHandler* SpH = data->mSynchpointHandler;

        // is this a synchpoint?
        if (!SpH->isSynchpoint(event))
        {
            return;
        }

        // --- get delay info buffer

        Buffer* buffer = data->mRemote->getBuffer(BUFFER_DELAY);
        data->mRwait = buffer->get_double();

        // was there a late-sender wait state?
        if (!(data->mRwait > 0))
        {
            return;
        }

        // --- find previous synchronization point

        int rank = event->getComm().getCommSet().getGlobalRank(event->getDestination());
        data->mPrevSp = SpH->findPreviousMpiSynchpoint(event, rank);

        // --- get local and remote time maps and calculate delay

        data->mPropSpMap = SpH->getSynchpointsBetween(data->mPrevSp, event);
        data->mWaitInfo  = CalculateWaitMap(data->mPropSpMap);
        data->mDelayInfo =
            CalculateDiff(PruneDelayTimemap(TmH->getRuntimes(data->mPrevSp, event),
                                            data->mWaitInfo.mMap,
                                            data->mPrevSp,
                                            SpH->getSynchpointInfo(data->mPrevSp)),
                          UnpackTimemap(buffer));

        // shortcuts
        const TimeMapSum& d(data->mDelayInfo);
        const TimeMapSum& w(data->mWaitInfo);

        // --- calculate delay costs

        double scale = 0.0;

        if ((d.mSum + w.mSum) > 0)
        {
            scale = std::min(data->mRwait / (d.mSum + w.mSum), 1.0);
        }

        // --- set severities

        if (d.mSum > 0)
        {
            for (timemap_t::const_iterator it = d.mMap.begin();
                 it != d.mMap.end();
                 ++it)
            {
                INC_METRIC_VALUE(SCALASCA_DEBUG_DELAY_METRICS,
                                 data->mDefs->get_cnode(it->first),
                                 scale * it->second);
            }
        }

        // --- invoke subsequent handlers

        cbmanager.notify(DELAY_LATE_SENDER, event, data);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_LateSender_LT
     *  @brief Long-term MPI Late Sender Delay Costs metric.
     *
     *  Implements the "Long-term MPI Late Sender Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_LateSender_LT
        : public LongTermDelayPattern
    {
        public:
            /// @name Constructors & destructor
            /// @{

            /// @brief Constructor.
            ///
            /// Creates a new instance.
            ///
            Delay_LateSender_LT()
                : LongTermDelayPattern(BUFFER_RCOST_LATE_SENDER,
                                       KEY_RCOST_LATE_SENDER,
                                       DELAY_LATE_SENDER)
            {
            }

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_LATESENDER_LONGTERM;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_LATESENDER_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI long-term Late Sender delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_latesender_longterm";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that indirectly cause MPI Late Sender wait"
                       " states through wait-state propagation";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}
    };
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_LateReceiver
     *  @brief Short-term MPI Late Receiver Delay Costs metric.
     *
     *  Implements the "Short-term MPI Late Receiver Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_LateReceiver
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_LATERECEIVER;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_LATERECEIVER_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI short-term Late Receiver delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_latereceiver";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that directly cause MPI Late Receiver wait"
                       " states";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (bwc)
            /// @{

            /// @brief Callback function for PRE_INV_SENDCMP events.
            ///
            /// Callback function triggered by PRE_INV_SENDCMP events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            pre_inv_sendcmp_bwc_cb(const CallbackManager& cbmanager,
                                   int                    user_event,
                                   const Event&           event,
                                   CallbackData*          cdata);

            /// @brief Callback function for POST_INV_RECVREQ events.
            ///
            /// Callback function triggered by POST_INV_RECVREQ events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            post_inv_recvreq_bwc_cb(const CallbackManager& cbmanager,
                                    int                    user_event,
                                    const Event&           event,
                                    CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Delay_LateReceiver::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register bwc callbacks
        it = cbmanagers.find("bwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            PRE_INV_SENDCMP,
            PEARL_create_callback(
                this,
                &Delay_LateReceiver::pre_inv_sendcmp_bwc_cb));
        cbmanager->register_callback(
            POST_INV_RECVREQ,
            PEARL_create_callback(
                this,
                &Delay_LateReceiver::post_inv_recvreq_bwc_cb));
    }


    // --- Callback functions (bwc) -----------------------------------------

    void
    Delay_LateReceiver::pre_inv_sendcmp_bwc_cb(const CallbackManager& cbmanager,
                                               const int              user_event,
                                               const Event&           event,
                                               CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 786 "Delay.MPI.pattern"

        // Shortcuts
        const TmapCacheHandler*  TmH = data->mTmapCacheHandler;
        const SynchpointHandler* SpH = data->mSynchpointHandler;

        // cost-accounting backward replay

        Buffer*        buffer = new Buffer(256);
        SynchpointInfo spi    = SpH->getSynchpointInfo(event);

        buffer->put_double(spi.waitTime);

        timemap_t tmap;

        if (spi.waitTime > 0)
        {
            Event sendreq = event.request();
            int   rank    = sendreq->getComm().getCommSet().getGlobalRank(sendreq->getDestination());
            tmap = GetMpiDelayTimemap(event, rank, SpH, TmH);
        }

        PackTimemap(buffer, tmap);
        data->mInvLocal->addBuffer(buffer, BUFFER_DELAY);
    }


    void
    Delay_LateReceiver::post_inv_recvreq_bwc_cb(const CallbackManager& cbmanager,
                                                const int              user_event,
                                                const Event&           event,
                                                CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 811 "Delay.MPI.pattern"

        // Shortcuts
        const TmapCacheHandler*  TmH = data->mTmapCacheHandler;
        const SynchpointHandler* SpH = data->mSynchpointHandler;

        // is this a synchpoint?
        if (!SpH->isSynchpoint(event))
        {
            return;
        }

        // --- get delay info buffer

        Buffer* buffer = data->mInvRemote->getBuffer(BUFFER_DELAY);
        data->mRwait = buffer->get_double();

        // was there a late-receiver wait state?
        if (!(data->mRwait > 0))
        {
            return;
        }

        // --- find previous synchronization point

        Event recvcmp = event.completion();
        int   rank    = recvcmp->getComm().getCommSet().getGlobalRank(recvcmp->getSource());
        data->mPrevSp = SpH->findPreviousMpiSynchpoint(event, rank);

        // --- get local and remote time maps and calculate delay

        data->mPropSpMap = SpH->getSynchpointsBetween(data->mPrevSp, event);
        data->mWaitInfo  = CalculateWaitMap(data->mPropSpMap);
        data->mDelayInfo =
            CalculateDiff(PruneDelayTimemap(TmH->getRuntimes(data->mPrevSp, event),
                                            data->mWaitInfo.mMap,
                                            data->mPrevSp,
                                            SpH->getSynchpointInfo(data->mPrevSp)),
                          UnpackTimemap(buffer));

        // shortcuts
        const TimeMapSum& d(data->mDelayInfo);
        const TimeMapSum& w(data->mWaitInfo);

        // --- calculate delay costs

        double scale = 0.0;

        if ((d.mSum + w.mSum) > 0)
        {
            scale = std::min(data->mRwait / (d.mSum + w.mSum), 1.0);
        }

        // --- set severities

        if (d.mSum > 0)
        {
            for (timemap_t::const_iterator it = d.mMap.begin();
                 it != d.mMap.end();
                 ++it)
            {
                INC_METRIC_VALUE(SCALASCA_DEBUG_DELAY_METRICS,
                                 data->mDefs->get_cnode(it->first),
                                 scale * it->second);
            }
        }

        // --- invoke subsequent handlers

        cbmanager.notify(DELAY_LATE_RECEIVER, event, data);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_LateReceiver_LT
     *  @brief Long-term MPI Late Receiver Delay Costs metric.
     *
     *  Implements the "Long-term MPI Late Receiver Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_LateReceiver_LT
        : public LongTermDelayPattern
    {
        public:
            /// @name Constructors & destructor
            /// @{

            /// @brief Constructor.
            ///
            /// Creates a new instance.
            ///
            Delay_LateReceiver_LT()
                : LongTermDelayPattern(BUFFER_RCOST_LATE_RECEIVER,
                                       KEY_RCOST_LATE_RECEIVER,
                                       DELAY_LATE_RECEIVER)
            {
            }

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_LATERECEIVER_LONGTERM;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_LATERECEIVER_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI long-term Late Receiver delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_latereceiver_longterm";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that indirectly cause MPI Late Receiver wait"
                       " states through wait-state propagation";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}
    };
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_Barrier
     *  @brief Short-term MPI Barrier Delay Costs metric.
     *
     *  Implements the "Short-term MPI Barrier Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_Barrier
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_BARRIER;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_BARRIER_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI short-term barrier delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_barrier";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that directly cause wait states in MPI"
                       " barriers";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (bwc)
            /// @{

            /// @brief Callback function for SYNC_COLL events.
            ///
            /// Callback function triggered by SYNC_COLL events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            sync_coll_bwc_cb(const CallbackManager& cbmanager,
                             int                    user_event,
                             const Event&           event,
                             CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Delay_Barrier::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register bwc callbacks
        it = cbmanagers.find("bwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            SYNC_COLL,
            PEARL_create_callback(
                this,
                &Delay_Barrier::sync_coll_bwc_cb));
    }


    // --- Callback functions (bwc) -----------------------------------------

    void
    Delay_Barrier::sync_coll_bwc_cb(const CallbackManager& cbmanager,
                                    const int              user_event,
                                    const Event&           event,
                                    CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 1057 "Delay.MPI.pattern"

        const CollectiveInfo& ci = data->mCollinfo;

        data->mInputScales.clear();

        cbmanager.notify(PRE_DELAY_MPI_BARRIER, event, data);

        CalculateCollectiveDelay(ci.my, ci.latest, event, data);

        if (ci.my.mRank == ci.latest.mRank)
        {
            // algorithm 1, line 24
            map< uint32_t, TimeMapSum >::const_iterator dWit = data->mCollDelayInfo.find(KEY_WAITTIME);
            if (  (dWit != data->mCollDelayInfo.end())
               && (dWit->second.mSum > 0.0))
            {
                for (timemap_t::const_iterator it = dWit->second.mMap.begin();
                     it != dWit->second.mMap.end();
                     ++it)
                {
                    INC_METRIC_VALUE(SCALASCA_DEBUG_DELAY_METRICS,
                                     data->mDefs->get_cnode(it->first),
                                     it->second);
                }
            }

            cbmanager.notify(DELAY_MPI_BARRIER, event, data);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_Barrier_LT
     *  @brief Long-term MPI Barrier Delay Costs metric.
     *
     *  Implements the "Long-term MPI Barrier Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_Barrier_LT
        : public LongTermDelayPattern
    {
        public:
            /// @name Constructors & destructor
            /// @{

            /// @brief Constructor.
            ///
            /// Creates a new instance.
            ///
            Delay_Barrier_LT()
                : LongTermDelayPattern(BUFFER_RCOST_MPI_BARRIER,
                                       KEY_RCOST_MPI_BARRIER,
                                       DELAY_MPI_BARRIER)
            {
            }

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_BARRIER_LONGTERM;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_BARRIER_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI long-term barrier delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_barrier_longterm";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that indirectly cause wait states in MPI"
                       " barriers through wait-state propagation";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}
    };
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_NxN
     *  @brief Short-term MPI N x N Collectives Delay Costs metric.
     *
     *  Implements the "Short-term MPI N x N Collectives Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_NxN
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_N2N;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_N2N_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI short-term N x N delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_n2n";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that directly cause wait states in MPI n-to-n"
                       " collective communications";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (bwc)
            /// @{

            /// @brief Callback function for COLL_N2N events.
            ///
            /// Callback function triggered by COLL_N2N events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            coll_n2n_bwc_cb(const CallbackManager& cbmanager,
                            int                    user_event,
                            const Event&           event,
                            CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Delay_NxN::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register bwc callbacks
        it = cbmanagers.find("bwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            COLL_N2N,
            PEARL_create_callback(
                this,
                &Delay_NxN::coll_n2n_bwc_cb));
    }


    // --- Callback functions (bwc) -----------------------------------------

    void
    Delay_NxN::coll_n2n_bwc_cb(const CallbackManager& cbmanager,
                               const int              user_event,
                               const Event&           event,
                               CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 1159 "Delay.MPI.pattern"

        const CollectiveInfo& ci = data->mCollinfo;

        data->mInputScales.clear();

        cbmanager.notify(PRE_DELAY_MPI_COLL_N2N, event, data);

        CalculateCollectiveDelay(ci.my, ci.latest, event, data);

        if (ci.my.mRank == ci.latest.mRank)
        {
            // algorithm 1, line 24
            map< uint32_t, TimeMapSum >::const_iterator dWit = data->mCollDelayInfo.find(KEY_WAITTIME);
            if (  (dWit != data->mCollDelayInfo.end())
               && (dWit->second.mSum > 0.0))
            {
                for (timemap_t::const_iterator it = dWit->second.mMap.begin();
                     it != dWit->second.mMap.end();
                     ++it)
                {
                    INC_METRIC_VALUE(SCALASCA_DEBUG_DELAY_METRICS,
                                     data->mDefs->get_cnode(it->first),
                                     it->second);
                }
            }

            cbmanager.notify(DELAY_MPI_COLL_N2N, event, data);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_NxN_LT
     *  @brief Long-term MPI N x N Collectives Delay Costs metric.
     *
     *  Implements the "Long-term MPI N x N Collectives Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_NxN_LT
        : public LongTermDelayPattern
    {
        public:
            /// @name Constructors & destructor
            /// @{

            /// @brief Constructor.
            ///
            /// Creates a new instance.
            ///
            Delay_NxN_LT()
                : LongTermDelayPattern(BUFFER_RCOST_MPI_COLL_N2N,
                                       KEY_RCOST_MPI_COLL_N2N,
                                       DELAY_MPI_COLL_N2N)
            {
            }

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_N2N_LONGTERM;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_N2N_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI long-term N x N delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_n2n_longterm";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that indirectly cause wait states in MPI"
                       " n-to-n collective communications through wait-state"
                       " propagation";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}
    };
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_12N
     *  @brief Short-term MPI 1-to-N Collectives Delay Costs metric.
     *
     *  Implements the "Short-term MPI 1-to-N Collectives Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_12N
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_12N;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_12N_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI short-term 1-to-N delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_12n";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that directly cause wait states in MPI 1-to-n"
                       " collective communications";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (bwc)
            /// @{

            /// @brief Callback function for COLL_12N events.
            ///
            /// Callback function triggered by COLL_12N events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            coll_12n_bwc_cb(const CallbackManager& cbmanager,
                            int                    user_event,
                            const Event&           event,
                            CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Delay_12N::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register bwc callbacks
        it = cbmanagers.find("bwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            COLL_12N,
            PEARL_create_callback(
                this,
                &Delay_12N::coll_12n_bwc_cb));
    }


    // --- Callback functions (bwc) -----------------------------------------

    void
    Delay_12N::coll_12n_bwc_cb(const CallbackManager& cbmanager,
                               const int              user_event,
                               const Event&           event,
                               CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 1261 "Delay.MPI.pattern"

        const CollectiveInfo& ci = data->mCollinfo;

        data->mInputScales.clear();

        cbmanager.notify(PRE_DELAY_MPI_COLL_12N, event, data);

        CalculateCollectiveDelay(ci.my, ci.root, event, data);

        if (ci.my.mRank == ci.root.mRank)
        {
            // algorithm 1, line 24
            map< uint32_t, TimeMapSum >::const_iterator dWit = data->mCollDelayInfo.find(KEY_WAITTIME);
            if (  (dWit != data->mCollDelayInfo.end())
               && (dWit->second.mSum > 0.0))
            {
                for (timemap_t::const_iterator it = dWit->second.mMap.begin();
                     it != dWit->second.mMap.end();
                     ++it)
                {
                    INC_METRIC_VALUE(SCALASCA_DEBUG_DELAY_METRICS,
                                     data->mDefs->get_cnode(it->first),
                                     it->second);
                }
            }

            cbmanager.notify(DELAY_MPI_COLL_12N, event, data);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_12N_LT
     *  @brief Long-term MPI 1-to-N Collectives Delay Costs metric.
     *
     *  Implements the "Long-term MPI 1-to-N Collectives Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_12N_LT
        : public LongTermDelayPattern
    {
        public:
            /// @name Constructors & destructor
            /// @{

            /// @brief Constructor.
            ///
            /// Creates a new instance.
            ///
            Delay_12N_LT()
                : LongTermDelayPattern(BUFFER_RCOST_MPI_COLL_12N,
                                       KEY_RCOST_MPI_COLL_12N,
                                       DELAY_MPI_COLL_12N)
            {
            }

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_12N_LONGTERM;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_12N_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI long-term 1-to-N delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_12n_longterm";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that indirectly cause wait states in MPI"
                       " 1-to-n collective communications through wait-state"
                       " propagation";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}
    };
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_N21
     *  @brief Short-term MPI N-to-1 Collectives Delay Costs metric.
     *
     *  Implements the "Short-term MPI N-to-1 Collectives Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_N21
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_N21;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_N21_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI short-term N-to-1 delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_n21";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that directly cause wait states in MPI n-to-1"
                       " collective communications";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (bwc)
            /// @{

            /// @brief Callback function for COLL_N21 events.
            ///
            /// Callback function triggered by COLL_N21 events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            coll_n21_bwc_cb(const CallbackManager& cbmanager,
                            int                    user_event,
                            const Event&           event,
                            CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Delay_N21::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register bwc callbacks
        it = cbmanagers.find("bwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            COLL_N21,
            PEARL_create_callback(
                this,
                &Delay_N21::coll_n21_bwc_cb));
    }


    // --- Callback functions (bwc) -----------------------------------------

    void
    Delay_N21::coll_n21_bwc_cb(const CallbackManager& cbmanager,
                               const int              user_event,
                               const Event&           event,
                               CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 1363 "Delay.MPI.pattern"

        const CollectiveInfo& ci = data->mCollinfo;

        data->mInputScales.clear();

        cbmanager.notify(PRE_DELAY_MPI_COLL_N21, event, data);

        CalculateCollectiveDelay(ci.my, ci.latest, event, data);

        if (ci.my.mRank == ci.latest.mRank)
        {
            // algorithm 1, line 24
            map< uint32_t, TimeMapSum >::const_iterator dWit = data->mCollDelayInfo.find(KEY_WAITTIME);
            if (  (dWit != data->mCollDelayInfo.end())
               && (dWit->second.mSum > 0.0))
            {
                for (timemap_t::const_iterator it = dWit->second.mMap.begin();
                     it != dWit->second.mMap.end();
                     ++it)
                {
                    INC_METRIC_VALUE(SCALASCA_DEBUG_DELAY_METRICS,
                                     data->mDefs->get_cnode(it->first),
                                     it->second);
                }
            }

            cbmanager.notify(DELAY_MPI_COLL_N21, event, data);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_N21_LT
     *  @brief Long-term MPI N-to-1 Collectives Delay Costs metric.
     *
     *  Implements the "Long-term MPI N-to-1 Collectives Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_N21_LT
        : public LongTermDelayPattern
    {
        public:
            /// @name Constructors & destructor
            /// @{

            /// @brief Constructor.
            ///
            /// Creates a new instance.
            ///
            Delay_N21_LT()
                : LongTermDelayPattern(BUFFER_RCOST_MPI_COLL_N21,
                                       KEY_RCOST_MPI_COLL_N21,
                                       DELAY_MPI_COLL_N21)
            {
            }

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_N21_LONGTERM;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_N21_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "MPI long-term N-to-1 delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_n21_longterm";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that indirectly cause wait states in MPI"
                       " n-to-1 collective communications through wait-state"
                       " propagation";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}
    };
#endif    // _MPI


#if defined(_OPENMP)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_OmpBarrier
     *  @brief Short-term OpenMP Barrier Delay Costs metric.
     *
     *  Implements the "Short-term OpenMP Barrier Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_OmpBarrier
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_OMPBARRIER;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_OMPBARRIER_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "OMP short-term barrier delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_ompbarrier";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that directly cause wait states in OpenMP"
                       " barriers";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (bwc)
            /// @{

            /// @brief Callback function for OMP_BARRIER events.
            ///
            /// Callback function triggered by OMP_BARRIER events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            omp_barrier_bwc_cb(const CallbackManager& cbmanager,
                               int                    user_event,
                               const Event&           event,
                               CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Delay_OmpBarrier::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register bwc callbacks
        it = cbmanagers.find("bwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            OMP_BARRIER,
            PEARL_create_callback(
                this,
                &Delay_OmpBarrier::omp_barrier_bwc_cb));
    }


    // --- Callback functions (bwc) -----------------------------------------

    void
    Delay_OmpBarrier::omp_barrier_bwc_cb(const CallbackManager& cbmanager,
                                         const int              user_event,
                                         const Event&           event,
                                         CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 79 "Delay.OMP.pattern"

        const CollectiveInfo& ci(data->mCollinfo);

        data->mInputScales.clear();

        cbmanager.notify(PRE_DELAY_OMP_BARRIER, event, data);

        CalculateOmpBarrierDelay(ci.my, ci.latest, event, data);

        if (ci.my.mRank == ci.latest.mRank)
        {
            // Set severities
            map< uint32_t, TimeMapSum >::const_iterator dWit = data->mCollDelayInfo.find(KEY_WAITTIME);

            if (  (dWit != data->mCollDelayInfo.end())
               && (dWit->second.mSum > 0.0))
            {
                for (timemap_t::const_iterator it = dWit->second.mMap.begin();
                     it != dWit->second.mMap.end();
                     ++it)
                {
                    INC_METRIC_VALUE(SCALASCA_DEBUG_DELAY_METRICS,
                                     data->mDefs->get_cnode(it->first),
                                     it->second);
                }
            }

            cbmanager.notify(DELAY_OMP_BARRIER, event, data);
        }
    }
#endif    // _OPENMP


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_OmpBarrier_LT
     *  @brief Long-term OpenMP Barrier Delay Costs metric.
     *
     *  Implements the "Long-term OpenMP Barrier Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_OmpBarrier_LT
        : public LongTermDelayPattern
    {
        public:
            /// @name Constructors & destructor
            /// @{

            /// @brief Constructor.
            ///
            /// Creates a new instance.
            ///
            Delay_OmpBarrier_LT()
                : LongTermDelayPattern(BUFFER_RCOST_OMP_BARRIER,
                                       KEY_RCOST_OMP_BARRIER,
                                       DELAY_OMP_BARRIER)
            {
            }

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_OMPBARRIER_LONGTERM;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_OMPBARRIER_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "OMP long-term barrier delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_ompbarrier_longterm";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that indirectly cause wait states in OpenMP"
                       " barriers through wait-state propagation";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}
    };
#endif    // _MPI


#if defined(_OPENMP)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_OmpIdle
     *  @brief Short-term OpenMP Thread Idleness Delay Costs metric.
     *
     *  Implements the "Short-term OpenMP Thread Idleness Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_OmpIdle
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_OMPIDLE;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_OMPIDLE_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "OMP short-term idleness delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_ompidle";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays on the OpenMP master thread that leave OpenMP"
                       " worker threads idle";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            #line 181 "Delay.OMP.pattern"

            /// @todo Implicit assumption: @#analysis threads == @#OpenMP threads
            void
            calculateDelayCosts(const CallbackManager& cbmanager,
                                const Event&           event,
                                CbData* const          data)
            {
                #pragma omp master
                {
                    // Shortcuts
                    const TmapCacheHandler*  TmH = data->mTmapCacheHandler;
                    const SynchpointHandler* SpH = data->mSynchpointHandler;

                    // Normally, the delay info mini-profile would be calculated
                    // between the current event and the previous syncpoint, incl.
                    // its completion time.  In this case, however, the previous
                    // syncpoint stored (if it exists) is the LEAVE event of the
                    // implicit barrier of the preceding parallel region, i.e., the
                    // interval based on the syncpoint is too long.  Thus, search
                    // for the respective THREAD_JOIN event starting from the
                    // syncpoint, and use it to calculate the correct mini-profile.
                    //
                    // Note that the wait map subtracted from the mini-profile can
                    // still be determined based on the syncpoint since there can
                    // be no wait states between the LEAVE of the implicit barrier
                    // and the THREAD_JOIN event of the parallel region.
                    data->mPrevSp = SpH->findPreviousOmpSynchpoint(event);
                    Event joinEvt = data->mPrevSp;
                    while (joinEvt.is_valid() && !joinEvt->isOfType(THREAD_JOIN))
                    {
                        ++joinEvt;
                    }
                    data->mPropSpMap = SpH->getSynchpointsBetween(data->mPrevSp, event);
                    data->mWaitInfo  = CalculateWaitMap(data->mPropSpMap);
                    data->mDelayInfo = CalculateDiff(TmH->getRuntimes(joinEvt, event),
                                                     data->mWaitInfo.mMap);

                    // shortcuts
                    const TimeMapSum& d(data->mDelayInfo);
                    const TimeMapSum& w(data->mWaitInfo);

                    // --- calculate delay costs

                    double scale = 0.0;

                    if ((d.mSum + w.mSum) > 0)
                    {
                        scale = (omp_get_num_threads() - 1);
                    }

                    // --- set severities

                    if (d.mSum > 0)
                    {
                        for (timemap_t::const_iterator it = d.mMap.begin();
                             it != d.mMap.end();
                             ++it)
                        {
                            INC_METRIC_VALUE(SCALASCA_DEBUG_DELAY_METRICS,
                                             data->mDefs->get_cnode(it->first),
                                             scale * it->second);
                        }
                    }

                    data->mOmpIdleScale = scale;

                    cbmanager.notify(DELAY_OMP_IDLE, event, data);
                }
            }

            /// @name Callback functions (bwc)
            /// @{

            /// @brief Callback function for PREPARE events.
            ///
            /// Callback function triggered by PREPARE events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            prepare_bwc_cb(const CallbackManager& cbmanager,
                           int                    user_event,
                           const Event&           event,
                           CallbackData*          cdata);

            /// @brief Callback function for OMP_MGMT_FORK events.
            ///
            /// Callback function triggered by OMP_MGMT_FORK events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            omp_mgmt_fork_bwc_cb(const CallbackManager& cbmanager,
                                 int                    user_event,
                                 const Event&           event,
                                 CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Delay_OmpIdle::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register bwc callbacks
        it = cbmanagers.find("bwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            PREPARE,
            PEARL_create_callback(
                this,
                &Delay_OmpIdle::prepare_bwc_cb));
        cbmanager->register_callback(
            OMP_MGMT_FORK,
            PEARL_create_callback(
                this,
                &Delay_OmpIdle::omp_mgmt_fork_bwc_cb));
    }


    // --- Callback functions (bwc) -----------------------------------------

    void
    Delay_OmpIdle::prepare_bwc_cb(const CallbackManager& cbmanager,
                                  const int              user_event,
                                  const Event&           event,
                                  CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 252 "Delay.OMP.pattern"

        // Calculate delay costs for last serial interval
        // (end of last parallel region to program end)
        calculateDelayCosts(cbmanager, event, data);
    }


    void
    Delay_OmpIdle::omp_mgmt_fork_bwc_cb(const CallbackManager& cbmanager,
                                        const int              user_event,
                                        const Event&           event,
                                        CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 258 "Delay.OMP.pattern"

        // Calculate delay costs for all serial intervals except last
        // (end of last parallel region to program end)
        calculateDelayCosts(cbmanager, event, data);
    }
#endif    // _OPENMP


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Delay_OmpIdle_LT
     *  @brief Long-term OpenMP Thread Idleness Delay Costs metric.
     *
     *  Implements the "Long-term OpenMP Thread Idleness Delay Costs" metric.
     */
    /*---------------------------------------------------------------------*/

    class Delay_OmpIdle_LT
        : public LongTermDelayPattern
    {
        public:
            /// @name Constructors & destructor
            /// @{

            /// @brief Constructor.
            ///
            /// Creates a new instance.
            ///
            Delay_OmpIdle_LT()
                : LongTermDelayPattern(BUFFER_RCOST_OMP_IDLE,
                                       KEY_RCOST_OMP_IDLE,
                                       DELAY_OMP_IDLE)
            {
            }

            /// @}
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_DELAY_OMPIDLE_LONGTERM;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_DELAY_OMPIDLE_AGGREGATE;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "OMP long-term idleness delay costs";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "delay_ompidle_longterm";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Delays that indirectly leave OpenMP worker threads"
                       " idle";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (bwc)
            /// @{

            /// @brief Callback function for DELAY_OMP_IDLE events.
            ///
            /// Callback function triggered by DELAY_OMP_IDLE events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            delay_omp_idle_bwc_cb(const CallbackManager& cbmanager,
                                  int                    user_event,
                                  const Event&           event,
                                  CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Delay_OmpIdle_LT::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        LongTermDelayPattern::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register bwc callbacks
        it = cbmanagers.find("bwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            DELAY_OMP_IDLE,
            PEARL_create_callback(
                this,
                &Delay_OmpIdle_LT::delay_omp_idle_bwc_cb));
    }


    // --- Callback functions (bwc) -----------------------------------------

    void
    Delay_OmpIdle_LT::delay_omp_idle_bwc_cb(const CallbackManager& cbmanager,
                                            const int              user_event,
                                            const Event&           event,
                                            CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 304 "Delay.OMP.pattern"

        if (data->mWaitInfo.mSum > 0.0)
        {
            mCost.addCosts(data->mPropSpMap, data->mOmpIdleScale);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_Wait_Indirect_LS
     *  @brief Indirect MPI Late Sender Wait States metric.
     *
     *  Implements the "Indirect MPI Late Sender Wait States" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_Wait_Indirect_LS
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_WAIT_INDIRECT_LATESENDER;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_WAIT_INDIRECT;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "Indirect MPI Late Sender wait states";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_wait_indirect_latesender";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Indirect waiting time in MPI Late Sender wait states"
                       " that results from propagation";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            #line 120 "Waitstates.MPI.pattern"

            typedef std::map< Event, double, EventKeyCompare > eventmap_t;

            eventmap_t m_pw;

            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for RPL_RESUME_FWD events.
            ///
            /// Callback function triggered by RPL_RESUME_FWD events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            rpl_resume_fwd_cb(const CallbackManager& cbmanager,
                              int                    user_event,
                              const Event&           event,
                              CallbackData*          cdata);

            /// @}
            /// @name Callback functions (bwc)
            /// @{

            /// @brief Callback function for DELAY_LATE_SENDER events.
            ///
            /// Callback function triggered by DELAY_LATE_SENDER events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            delay_late_sender_bwc_cb(const CallbackManager& cbmanager,
                                     int                    user_event,
                                     const Event&           event,
                                     CallbackData*          cdata);

            /// @}
            /// @name Callback functions (fwc)
            /// @{

            /// @brief Callback function for PRE_SEND events.
            ///
            /// Callback function triggered by PRE_SEND events
            /// during the @"fwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            pre_send_fwc_cb(const CallbackManager& cbmanager,
                            int                    user_event,
                            const Event&           event,
                            CallbackData*          cdata);

            /// @brief Callback function for POST_RECV events.
            ///
            /// Callback function triggered by POST_RECV events
            /// during the @"fwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            post_recv_fwc_cb(const CallbackManager& cbmanager,
                             int                    user_event,
                             const Event&           event,
                             CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_Wait_Indirect_LS::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            RPL_RESUME_FWD,
            PEARL_create_callback(
                this,
                &MPI_Wait_Indirect_LS::rpl_resume_fwd_cb));

        // Register bwc callbacks
        it = cbmanagers.find("bwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            DELAY_LATE_SENDER,
            PEARL_create_callback(
                this,
                &MPI_Wait_Indirect_LS::delay_late_sender_bwc_cb));

        // Register fwc callbacks
        it = cbmanagers.find("fwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            PRE_SEND,
            PEARL_create_callback(
                this,
                &MPI_Wait_Indirect_LS::pre_send_fwc_cb));
        cbmanager->register_callback(
            POST_RECV,
            PEARL_create_callback(
                this,
                &MPI_Wait_Indirect_LS::post_recv_fwc_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_Wait_Indirect_LS::rpl_resume_fwd_cb(const CallbackManager& cbmanager,
                                            const int              user_event,
                                            const Event&           event,
                                            CallbackData* const    cdata)
    {
        #line 126 "Waitstates.MPI.pattern"

        m_pw.clear();
    }


    // --- Callback functions (bwc) -----------------------------------------

    void
    MPI_Wait_Indirect_LS::delay_late_sender_bwc_cb(const CallbackManager& cbmanager,
                                                   const int              user_event,
                                                   const Event&           event,
                                                   CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 131 "Waitstates.MPI.pattern"

        // Shortcuts
        const TimeMapSum& d(data->mDelayInfo);
        const TimeMapSum& w(data->mWaitInfo);

        if (w.mSum > 0)
        {
            double scale = std::min(data->mRwait / (d.mSum + w.mSum), 1.0);
            m_pw[event] += scale * w.mSum;
        }
    }


    // --- Callback functions (fwc) -----------------------------------------

    void
    MPI_Wait_Indirect_LS::pre_send_fwc_cb(const CallbackManager& cbmanager,
                                          const int              user_event,
                                          const Event&           event,
                                          CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 144 "Waitstates.MPI.pattern"

        eventmap_t::iterator it = m_pw.find(event);

        Buffer* buf = new Buffer(16);
        buf->put_double(it == m_pw.end() ? 0.0 : it->second);

        data->mLocal->addBuffer(buf, BUFFER_PROPWAIT);
    }


    void
    MPI_Wait_Indirect_LS::post_recv_fwc_cb(const CallbackManager& cbmanager,
                                           const int              user_event,
                                           const Event&           event,
                                           CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 152 "Waitstates.MPI.pattern"

        double pw = data->mRemote->getBuffer(BUFFER_PROPWAIT)->get_double();

        if (pw > 0.0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             pw);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_Wait_Indirect_LR
     *  @brief Indirect MPI Late Receiver Wait States metric.
     *
     *  Implements the "Indirect MPI Late Receiver Wait States" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_Wait_Indirect_LR
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_WAIT_INDIRECT_LATERECEIVER;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_WAIT_INDIRECT;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "Indirect MPI Late Receiver wait states";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_wait_indirect_latereceiver";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Indirect waiting time in MPI Late Receiver wait"
                       " states that results from propagation";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            #line 180 "Waitstates.MPI.pattern"

            typedef std::map< Event, double, EventKeyCompare > eventmap_t;

            eventmap_t m_pw;

            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for RPL_RESUME_FWD events.
            ///
            /// Callback function triggered by RPL_RESUME_FWD events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            rpl_resume_fwd_cb(const CallbackManager& cbmanager,
                              int                    user_event,
                              const Event&           event,
                              CallbackData*          cdata);

            /// @}
            /// @name Callback functions (bwc)
            /// @{

            /// @brief Callback function for DELAY_LATE_RECEIVER events.
            ///
            /// Callback function triggered by DELAY_LATE_RECEIVER events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            delay_late_receiver_bwc_cb(const CallbackManager& cbmanager,
                                       int                    user_event,
                                       const Event&           event,
                                       CallbackData*          cdata);

            /// @}
            /// @name Callback functions (fwc)
            /// @{

            /// @brief Callback function for PRE_INV_RECVREQ events.
            ///
            /// Callback function triggered by PRE_INV_RECVREQ events
            /// during the @"fwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            pre_inv_recvreq_fwc_cb(const CallbackManager& cbmanager,
                                   int                    user_event,
                                   const Event&           event,
                                   CallbackData*          cdata);

            /// @brief Callback function for POST_INV_SENDCMP events.
            ///
            /// Callback function triggered by POST_INV_SENDCMP events
            /// during the @"fwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            post_inv_sendcmp_fwc_cb(const CallbackManager& cbmanager,
                                    int                    user_event,
                                    const Event&           event,
                                    CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_Wait_Indirect_LR::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            RPL_RESUME_FWD,
            PEARL_create_callback(
                this,
                &MPI_Wait_Indirect_LR::rpl_resume_fwd_cb));

        // Register bwc callbacks
        it = cbmanagers.find("bwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            DELAY_LATE_RECEIVER,
            PEARL_create_callback(
                this,
                &MPI_Wait_Indirect_LR::delay_late_receiver_bwc_cb));

        // Register fwc callbacks
        it = cbmanagers.find("fwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            PRE_INV_RECVREQ,
            PEARL_create_callback(
                this,
                &MPI_Wait_Indirect_LR::pre_inv_recvreq_fwc_cb));
        cbmanager->register_callback(
            POST_INV_SENDCMP,
            PEARL_create_callback(
                this,
                &MPI_Wait_Indirect_LR::post_inv_sendcmp_fwc_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_Wait_Indirect_LR::rpl_resume_fwd_cb(const CallbackManager& cbmanager,
                                            const int              user_event,
                                            const Event&           event,
                                            CallbackData* const    cdata)
    {
        #line 186 "Waitstates.MPI.pattern"

        m_pw.clear();
    }


    // --- Callback functions (bwc) -----------------------------------------

    void
    MPI_Wait_Indirect_LR::delay_late_receiver_bwc_cb(const CallbackManager& cbmanager,
                                                     const int              user_event,
                                                     const Event&           event,
                                                     CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 191 "Waitstates.MPI.pattern"

        // Shortcuts
        const TimeMapSum& d(data->mDelayInfo);
        const TimeMapSum& w(data->mWaitInfo);

        if (w.mSum > 0)
        {
            double scale = std::min(data->mRwait / (d.mSum + w.mSum), 1.0);
            m_pw[event] += scale * w.mSum;
        }
    }


    // --- Callback functions (fwc) -----------------------------------------

    void
    MPI_Wait_Indirect_LR::pre_inv_recvreq_fwc_cb(const CallbackManager& cbmanager,
                                                 const int              user_event,
                                                 const Event&           event,
                                                 CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 204 "Waitstates.MPI.pattern"

        eventmap_t::iterator it = m_pw.find(event);

        Buffer* buf = new Buffer(16);
        buf->put_double(it == m_pw.end() ? 0.0 : it->second);

        data->mInvLocal->addBuffer(buf, BUFFER_PROPWAIT);
    }


    void
    MPI_Wait_Indirect_LR::post_inv_sendcmp_fwc_cb(const CallbackManager& cbmanager,
                                                  const int              user_event,
                                                  const Event&           event,
                                                  CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 212 "Waitstates.MPI.pattern"

        double pw = data->mInvRemote->getBuffer(BUFFER_PROPWAIT)->get_double();

        if (pw > 0.0)
        {
            INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                             event.get_cnode(),
                             pw);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class MPI_Wait_PT
     *  @brief Propagating MPI Point-to-point Wait States metric.
     *
     *  Implements the "Propagating MPI Point-to-point Wait States" metric.
     */
    /*---------------------------------------------------------------------*/

    class MPI_Wait_PT
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_WAIT_PROPAGATING;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_WAITSTATES_PROPAGATING_VS_TERMINAL;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "Propagating MPI point-to-point wait states";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_wait_propagating";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "Aggregate amount of MPI point-to-point waiting time"
                       " that propagates further";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @copydoc scout::Pattern::isVisible()
            ///
            bool
            isVisible() const override
            {
                return false;
            }

            /// @}


        private:
            #line 380 "Waitstates.MPI.pattern"

            typedef std::map< Event, double, EventKeyCompare > eventmap_t;

            eventmap_t m_pw;

            /// @name Callback functions (main)
            /// @{

            /// @brief Callback function for RPL_RESUME_FWD events.
            ///
            /// Callback function triggered by RPL_RESUME_FWD events
            /// during the @"main@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            rpl_resume_fwd_cb(const CallbackManager& cbmanager,
                              int                    user_event,
                              const Event&           event,
                              CallbackData*          cdata);

            /// @}
            /// @name Callback functions (bwc)
            /// @{

            /// @brief Callback function for DELAY_LATE_SENDER events.
            ///
            /// Callback function triggered by DELAY_LATE_SENDER events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            delay_late_sender_bwc_cb(const CallbackManager& cbmanager,
                                     int                    user_event,
                                     const Event&           event,
                                     CallbackData*          cdata);

            /// @brief Callback function for DELAY_LATE_RECEIVER events.
            ///
            /// Callback function triggered by DELAY_LATE_RECEIVER events
            /// during the @"bwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            delay_late_receiver_bwc_cb(const CallbackManager& cbmanager,
                                       int                    user_event,
                                       const Event&           event,
                                       CallbackData*          cdata);

            /// @}
            /// @name Callback functions (fwc)
            /// @{

            /// @brief Callback function for GROUP_RECV events.
            ///
            /// Callback function triggered by GROUP_RECV events
            /// during the @"fwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            group_recv_fwc_cb(const CallbackManager& cbmanager,
                              int                    user_event,
                              const Event&           event,
                              CallbackData*          cdata);

            /// @brief Callback function for MPI_SEND events.
            ///
            /// Callback function triggered by MPI_SEND events
            /// during the @"fwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_send_fwc_cb(const CallbackManager& cbmanager,
                            int                    user_event,
                            const Event&           event,
                            CallbackData*          cdata);

            /// @brief Callback function for MPI_SEND_COMPLETE events.
            ///
            /// Callback function triggered by MPI_SEND_COMPLETE events
            /// during the @"fwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            mpi_send_complete_fwc_cb(const CallbackManager& cbmanager,
                                     int                    user_event,
                                     const Event&           event,
                                     CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    MPI_Wait_PT::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register main callbacks
        it = cbmanagers.find("");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            RPL_RESUME_FWD,
            PEARL_create_callback(
                this,
                &MPI_Wait_PT::rpl_resume_fwd_cb));

        // Register bwc callbacks
        it = cbmanagers.find("bwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            DELAY_LATE_SENDER,
            PEARL_create_callback(
                this,
                &MPI_Wait_PT::delay_late_sender_bwc_cb));
        cbmanager->register_callback(
            DELAY_LATE_RECEIVER,
            PEARL_create_callback(
                this,
                &MPI_Wait_PT::delay_late_receiver_bwc_cb));

        // Register fwc callbacks
        it = cbmanagers.find("fwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            GROUP_RECV,
            PEARL_create_callback(
                this,
                &MPI_Wait_PT::group_recv_fwc_cb));
        cbmanager->register_callback(
            MPI_SEND,
            PEARL_create_callback(
                this,
                &MPI_Wait_PT::mpi_send_fwc_cb));
        cbmanager->register_callback(
            MPI_SEND_COMPLETE,
            PEARL_create_callback(
                this,
                &MPI_Wait_PT::mpi_send_complete_fwc_cb));
    }


    // --- Callback functions (main) ----------------------------------------

    void
    MPI_Wait_PT::rpl_resume_fwd_cb(const CallbackManager& cbmanager,
                                   const int              user_event,
                                   const Event&           event,
                                   CallbackData* const    cdata)
    {
        #line 386 "Waitstates.MPI.pattern"

        m_pw.clear();
    }


    // --- Callback functions (bwc) -----------------------------------------

    void
    MPI_Wait_PT::delay_late_sender_bwc_cb(const CallbackManager& cbmanager,
                                          const int              user_event,
                                          const Event&           event,
                                          CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 391 "Waitstates.MPI.pattern"

        // Shortcuts
        const TimeMapSum& d(data->mDelayInfo);
        const TimeMapSum& w(data->mWaitInfo);

        if (!(w.mSum > 0))
        {
            return;
        }

        double scale = std::min(data->mRwait / (d.mSum + w.mSum), 1.0);

        if (!(scale > 0))
        {
            return;
        }

        // --- update maximum propagating wait time for wait states in current sync interval

        for (SynchpointInfoMap::const_iterator it = data->mPropSpMap.begin();
             it != data->mPropSpMap.end();
             ++it)
        {
            if (it->second.waitTime > 0)
            {
                eventmap_t::iterator pwit = m_pw.find(it->first);

                if (pwit == m_pw.end())
                {
                    m_pw.insert(std::make_pair(it->first, scale * it->second.waitTime));
                }
                else
                {
                    pwit->second = std::max(scale * it->second.waitTime, pwit->second);
                }
            }
        }
    }


    void
    MPI_Wait_PT::delay_late_receiver_bwc_cb(const CallbackManager& cbmanager,
                                            const int              user_event,
                                            const Event&           event,
                                            CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 429 "Waitstates.MPI.pattern"

        // Shortcuts
        const TimeMapSum& d(data->mDelayInfo);
        const TimeMapSum& w(data->mWaitInfo);

        if (!(w.mSum > 0))
        {
            return;
        }

        double scale = std::min(data->mRwait / (d.mSum + w.mSum), 1.0);

        if (!(scale > 0))
        {
            return;
        }

        // --- update maximum propagating wait time for wait states in current sync interval

        for (SynchpointInfoMap::const_iterator it = data->mPropSpMap.begin();
             it != data->mPropSpMap.end();
             ++it)
        {
            if (it->second.waitTime > 0)
            {
                eventmap_t::iterator pwit = m_pw.find(it->first);

                if (pwit == m_pw.end())
                {
                    m_pw.insert(std::make_pair(it->first, scale * it->second.waitTime));
                }
                else
                {
                    pwit->second = std::max(scale * it->second.waitTime, pwit->second);
                }
            }
        }
    }


    // --- Callback functions (fwc) -----------------------------------------

    void
    MPI_Wait_PT::group_recv_fwc_cb(const CallbackManager& cbmanager,
                                   const int              user_event,
                                   const Event&           event,
                                   CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 756 "Waitstates.MPI.pattern"

        eventmap_t::iterator pwmit = m_pw.find(event);

        if (pwmit != m_pw.end())
        {
            data->mIdle = pwmit->second;
            cbmanager.notify(WAITSTATE_PROP_LS, event, data);
        }
    }


    void
    MPI_Wait_PT::mpi_send_fwc_cb(const CallbackManager& cbmanager,
                                 const int              user_event,
                                 const Event&           event,
                                 CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 765 "Waitstates.MPI.pattern"

        eventmap_t::iterator pwmit = m_pw.find(event);

        if (pwmit != m_pw.end())
        {
            data->mIdle = pwmit->second;
            cbmanager.notify(WAITSTATE_PROP_LR, event, data);
        }
    }


    void
    MPI_Wait_PT::mpi_send_complete_fwc_cb(const CallbackManager& cbmanager,
                                          const int              user_event,
                                          const Event&           event,
                                          CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 774 "Waitstates.MPI.pattern"

        eventmap_t::iterator pwmit = m_pw.find(event);

        if (pwmit != m_pw.end())
        {
            data->mIdle = pwmit->second;
            cbmanager.notify(WAITSTATE_PROP_LR, event, data);
        }
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Wait_PropagatingLS
     *  @brief Propagating MPI Late Sender Wait States metric.
     *
     *  Implements the "Propagating MPI Late Sender Wait States" metric.
     */
    /*---------------------------------------------------------------------*/

    class Wait_PropagatingLS
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_WAIT_PROPAGATING_LS;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_WAIT_PROPAGATING;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "Propagating MPI Late Sender wait states";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_wait_propagating_ls";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "MPI Late Sender wait states which cause further"
                       " waiting time on other processes";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (fwc)
            /// @{

            /// @brief Callback function for WAITSTATE_PROP_LS events.
            ///
            /// Callback function triggered by WAITSTATE_PROP_LS events
            /// during the @"fwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            waitstate_prop_ls_fwc_cb(const CallbackManager& cbmanager,
                                     int                    user_event,
                                     const Event&           event,
                                     CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Wait_PropagatingLS::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register fwc callbacks
        it = cbmanagers.find("fwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            WAITSTATE_PROP_LS,
            PEARL_create_callback(
                this,
                &Wait_PropagatingLS::waitstate_prop_ls_fwc_cb));
    }


    // --- Callback functions (fwc) -----------------------------------------

    void
    Wait_PropagatingLS::waitstate_prop_ls_fwc_cb(const CallbackManager& cbmanager,
                                                 const int              user_event,
                                                 const Event&           event,
                                                 CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 823 "Waitstates.MPI.pattern"

        INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                         event.get_cnode(),
                         data->mIdle);
    }
#endif    // _MPI


#if defined(_MPI)
    /*---------------------------------------------------------------------*/
    /**
     *  @class Wait_PropagatingLR
     *  @brief Propagating MPI Late Receiver Wait States metric.
     *
     *  Implements the "Propagating MPI Late Receiver Wait States" metric.
     */
    /*---------------------------------------------------------------------*/

    class Wait_PropagatingLR
        : public TypedPattern< double >
    {
        public:
            /// @name Callback registration
            /// @{

            /// @copydoc scout::Pattern::registerCallbacks(CallbackManagerMap&)
            ///
            void
            registerCallbacks(CallbackManagerMap& cbmanagers) override;

            /// @}
            /// @name Pattern information access
            /// @{

            /// @copydoc scout::Pattern::getId()
            ///
            long
            getId() const override
            {
                return PAT_MPI_WAIT_PROPAGATING_LR;
            }

            /// @copydoc scout::Pattern::getParent()
            ///
            long
            getParent() const override
            {
                return PAT_MPI_WAIT_PROPAGATING;
            }

            /// @copydoc scout::Pattern::getName()
            ///
            string
            getName() const override
            {
                return "Propagating MPI Late Receiver wait states";
            }

            /// @copydoc scout::Pattern::getUniqueName()
            ///
            string
            getUniqueName() const override
            {
                return "mpi_wait_propagating_lr";
            }

            /// @copydoc scout::Pattern::getDescription()
            ///
            string
            getDescription() const override
            {
                return "MPI Late Receiver wait states which cause further"
                       " waiting time on other processes";
            }

            /// @copydoc scout::Pattern::getUnit()
            ///
            string
            getUnit() const override
            {
                return "sec";
            }

            /// @copydoc scout::Pattern::getStorageMode()
            ///
            CubeMetricType
            getStorageMode() const override
            {
                return CUBE_METRIC_EXCLUSIVE;
            }

            /// @}


        private:
            /// @name Callback functions (fwc)
            /// @{

            /// @brief Callback function for WAITSTATE_PROP_LR events.
            ///
            /// Callback function triggered by WAITSTATE_PROP_LR events
            /// during the @"fwc@" replay pass.
            ///
            /// @param cbmanager
            ///     Callback manager instance
            /// @param user_event
            ///     User event triggering this callback (if applicable)
            /// @param event
            ///     Currently processed event
            /// @param cdata
            ///     Callback data object
            ///
            void
            waitstate_prop_lr_fwc_cb(const CallbackManager& cbmanager,
                                     int                    user_event,
                                     const Event&           event,
                                     CallbackData*          cdata);

            /// @}
    };


    // --- Callback registration --------------------------------------------

    void
    Wait_PropagatingLR::registerCallbacks(CallbackManagerMap& cbmanagers)
    {
        TypedPattern< double >::registerCallbacks(cbmanagers);

        CallbackManagerMap::iterator    it;
        CallbackManagerMap::mapped_type cbmanager;

        // Register fwc callbacks
        it = cbmanagers.find("fwc");
        UTILS_ASSERT(it != cbmanagers.end());

        cbmanager = it->second;
        cbmanager->register_callback(
            WAITSTATE_PROP_LR,
            PEARL_create_callback(
                this,
                &Wait_PropagatingLR::waitstate_prop_lr_fwc_cb));
    }


    // --- Callback functions (fwc) -----------------------------------------

    void
    Wait_PropagatingLR::waitstate_prop_lr_fwc_cb(const CallbackManager& cbmanager,
                                                 const int              user_event,
                                                 const Event&           event,
                                                 CallbackData* const    cdata)
    {
        SCALASCA_MAYBE_UNUSED auto* const data = static_cast< CbData* >(cdata);

        #line 847 "Waitstates.MPI.pattern"

        INC_METRIC_VALUE(SCALASCA_DEBUG_WAITSTATE_METRICS,
                         event.get_cnode(),
                         data->mIdle);
    }
#endif // _MPI
}      // namespace scout::metric


// --- Pattern registration -------------------------------------------------

void
create_patterns(AnalyzeTask* const analyzer)
{
    analyzer->addPattern(new metric::Time());
    analyzer->addPattern(new metric::Visits());
    #if defined(_MPI)
        if (enableAsynchronous)
        {
            analyzer->addPattern(new metric::Async());
        }
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_InitCompletion());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_FinalizeWait());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_WinMgmt());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_WaitCommCreate());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_CommCreateCompletion());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_WaitCommFree());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_SyncCollective());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_BarrierWait());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_BarrierCompletion());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_P2P());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_LateSender());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_LateSenderWO());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_LswoSame());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_LateReceiver());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_Collective());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_EarlyReduce());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_EarlyScan());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_LateBroadcast());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_WaitNxN());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_NxNCompletion());
    #endif    // _MPI
    #if defined(_OPENMP)
        analyzer->addPattern(new metric::OMP_Mgmt());
    #endif    // _OPENMP
    #if defined(_OPENMP)
        analyzer->addPattern(new metric::OMP_Mgmt_Fork());
    #endif    // _OPENMP
    #if defined(_OPENMP)
        analyzer->addPattern(new metric::OMP_Sync_Barrier());
    #endif    // _OPENMP
    #if defined(_OPENMP)
        analyzer->addPattern(new metric::OMP_Sync_Ebarrier_Wait());
    #endif    // _OPENMP
    #if defined(_OPENMP)
        analyzer->addPattern(new metric::OMP_Sync_Ibarrier_Wait());
    #endif    // _OPENMP
    #if defined(_OPENMP)
        analyzer->addPattern(new metric::OMP_Lock_Contention_Critical());
    #endif    // _OPENMP
    #if defined(_OPENMP)
        analyzer->addPattern(new metric::OMP_Lock_Contention_API());
    #endif    // _OPENMP
    #if defined(_OPENMP)
        analyzer->addPattern(new metric::Pthread_Lock_Contention_Mutex_Lock());
    #endif    // _OPENMP
    #if defined(_OPENMP)
        analyzer->addPattern(new metric::Pthread_Lock_Contention_Conditional());
    #endif    // _OPENMP
    #if defined(_OPENMP)
        analyzer->addPattern(new metric::Thread_LockContention());
    #endif    // _OPENMP
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_SyncsSend());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_SyncsRecv());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_SyncsColl());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_CommsSend());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_CommsRecv());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_CommsExch());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_CommsSrc());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_CommsDst());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_BytesSent());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_BytesRcvd());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_BytesCout());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_BytesCin());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_BytesPut());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_BytesGet());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_SLR_Count());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_SLS_Count());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_SLSWO_Count());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_CLR_Count());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_CLS_Count());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_CLSWO_Count());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_RmaWaitAtCreate());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_RmaWaitAtFree());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_RmaSync());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_RmaSyncLatePost());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_RmaEarlyWait());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_RmaLateComplete());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_RmaWaitAtFence());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_RmaEarlyFence());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::Mpi_RmaSyncLockContention());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::Mpi_RmaSyncWaitForProgress());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_RmaCommLatePost());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::Mpi_RmaCommLockContention());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::Mpi_RmaCommWaitForProgress());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_RmaPairsyncCount());
    #endif    // _MPI
    #if defined(_MPI)
        analyzer->addPattern(new metric::MPI_RmaPairsyncUnneededCount());
    #endif    // _MPI
    if (enableStatistics)
    {
        analyzer->addPattern(new metric::Statistics());
    }
    if (enableCriticalPath)
    {
        analyzer->addPattern(new metric::CriticalPath());
    }
    if (enableCriticalPath)
    {
        analyzer->addPattern(new metric::CriticalPathActivities());
    }
    if (enableCriticalPath)
    {
        analyzer->addPattern(new metric::CriticalImbalanceImpact());
    }
    if (enableCriticalPath)
    {
        analyzer->addPattern(new metric::InterPartitionImbalance());
    }
    if (enableCriticalPath)
    {
        analyzer->addPattern(new metric::NonCriticalPathActivities());
    }
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_Finalize());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_Finalize_LT());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_CommCreate());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_CommCreate_LT());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_CommFree());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_CommFree_LT());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_LateSender());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_LateSender_LT());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_LateReceiver());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_LateReceiver_LT());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_Barrier());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_Barrier_LT());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_NxN());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_NxN_LT());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_12N());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_12N_LT());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_N21());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_N21_LT());
        }
    #endif    // _MPI
    #if defined(_OPENMP)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_OmpBarrier());
        }
    #endif    // _OPENMP
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_OmpBarrier_LT());
        }
    #endif    // _MPI
    #if defined(_OPENMP)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_OmpIdle());
        }
    #endif    // _OPENMP
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Delay_OmpIdle_LT());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::MPI_Wait_Indirect_LS());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::MPI_Wait_Indirect_LR());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::MPI_Wait_PT());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Wait_PropagatingLS());
        }
    #endif    // _MPI
    #if defined(_MPI)
        if (enableDelayAnalysis)
        {
            analyzer->addPattern(new metric::Wait_PropagatingLR());
        }
    #endif    // _MPI
}
}    // namespace scout
